/************* BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  he author accepts no responsibility or          **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#define MAX_fpbuf			65536
#define MAX_cgibuf 			4096
#define MAX_npairs 			32
#define MAX_statbuf			(1024*1024)

#include "!lib-general.c"
#include "!custom.c"
/*#include "!lib-cgi.c"*/
#include "!lib-uinfo-misc.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

static timmy thetim;
int send_doc(int w, char* p);

/***************************************************************************************************************************
** Main code. Doesn't make use of command line arguments.                                                                 **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	char *playerid = "", *run = 0, *p;
	int i, rc = 0;
	struct tm *the_tm;
	time_t thetime;
	FILE* fp = 0;


	/* What time is it? */
	time(&thetime);
	the_tm = localtime(&thetime);
	thetim.day  = the_tm->tm_yday - StartingYDay;
	thetim.hour = the_tm->tm_hour;
	thetim.min  = the_tm->tm_min;

	/* File creation umask */
	umask(0117);

	setbuf(stdout, 0);

	/* Make note in log and redirect stderr there */
	if (rc = seterrlog("Front.c", ctime(&thetime))) exit(send_doc(rc, 0));

	/* Parse CGI input */
	if ((i = parsecgi()) > 0) erf(i, 10);

	/* No input given */
	if ((i == 0) || !(*(playerid = findval("id", 1))))
	{
		playerid = PUB_ID;
		erf(ERF_Void, 1);
	}

	/* Prepare temporary file for chained stdin */
	if (!(fp = prepforstdin())) erf(ERF_tmpfileFailed, 11);

	urlencode(fp, "authname", playerid, 1);

	if (findval("sub_stats", 0)) run = CGI_ROOT "stats.cgi";

	else if (findval("sub_passw", 0))
	{
		fputs("&pl_1", fp);			
		urlencode(fp, "id", playerid, 1);
		urlencode(fp, "act", "1", 0);
		urlencode(fp, "sub_act", "auto", 0);
		run = CGI_ROOT "pled.cgi";
	}

	else if (findval("sub_plinfo", 0))
	{
		if (!(p = findval("tarpid", 0))) p = playerid;
		urlencode(fp, "id", p, 0);
		run = CGI_ROOT "plinfo.cgi";
	}
	
	else if ((p = findval("sub_addev", 0)) || findval("sub_edev", 0))
	{
		if ((!p) && (p = findval("ev_ref", 0))) urlencode(fp, "ev_ref", p, 0);
		run = CGI_ROOT "event.cgi";
	}

#if OPT_ChangePseud
	else if (findval("sub_pseud", 0))
	{
		fputs("&id_", fp);			
		urlencode(fp, playerid, "auto", 1);
		urlencode(fp, "action", "6", 0);
		urlencode(fp, "sub_act", "auto", 0);
		run = CGI_ROOT "stats.cgi";
	}
#endif

	if (run) urlencode(fp, "authpass", findval("pass", 1), 0);

	/* Execute the new binary, taking input from temporary file */
	if (run)
	{
		fprintf(stderr, "%li:   fronting %s to %s\n", uniqid, playerid, run);

		erf(launchwithstdin(run, fp), 2);
	}


_erf: /***** Error handling *****/

	/* Debug */
	if (rc > ERF_Void) fprintf(stderr, "%li:   main(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Tidy up */
	if (fp) fclose(fp);

	/* Only retain low bits of normal rc's */
	return (send_doc(ERF_LOWBITS(rc), playerid));
}

/***************************************************************************************************************************
** Produce and output HTML document data for the various situations indicated by (w).                                     **
***************************************************************************************************************************/
int send_doc(int w, char* plid)
{
	int rc = 0, i, lockfd = 0;
	FILE* fp = 0;

	/* Output HTTP headers */
    printf("Content-type: text/html" LF LF);

	/* Claim lockfile */
	if (!(lockfd = setflock("front.c"))) w = ERF_flockFailed;

	/* Not handling it here */
	if ((w) && (w != ERF_Void))
	{
		error_doc(w);
		return 0;
	}

	puts(LF "<HTML><HEAD><TITLE>Michaelmas Massacre Assassins' Game: Player frontend</TITLE></HEAD>"
		 LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		 LF "<h1>Player frontend</h1><form action=\"" CGI_URL "front.cgi\" method=\"POST\">");

	if (fp = fopen(FILE_MOTD, "rb"))
	{
		while ((i = getc(fp)) != EOF) if (EOF == putchar(i)) break;
		fclose(fp);
		fp = 0;
	}

	puts("<p>Player: &nbsp; <select size=1 name=\"id\">");

	if (rc = makeidmenu(plid, PF_Login, PF_Login, OPT_IDMenuMode)) erf(rc, 13);

	printf("</select> &nbsp; password: &nbsp;"
		   LF "<input name=\"pass\" type=\"password\" value=\"%s\" size=24 maxlength="STRING(MAX_PasswordLen)"></p><ul>"
		   LF "<!--<li><p><input value=\"Get full statistics\" name=\"sub_plinfo\" type=\"submit\">"
		   LF " for player <select size=1 name=\"tarpid\">", findval("pass", 1));

	/*if (rc = makeidmenu(findval("tarpid", 0), PF_Playing|PF_Special, PF_Playing, OPT_IDMenuMode)) erf(rc, 14);*/

	printf("</select></p></li>-->"
		   LF "<li><p><input value=\"Change password\" type=\"submit\" name=\"sub_passw\"></p></li>"
		   LF "<li><p><input value=\"Add a new event\" type=\"submit\" name=\"sub_addev\">"
		   LF " - If it hasn't yet been reported.</p></li>"
		   LF "<li><p><input value=\"Add to or edit an existing event\" type=\"submit\" name=\"sub_edev\">"
		   LF " - event ref: &nbsp; <input type=\"text\" name=\"ev_ref\" value=\"%s\" size=8 maxlength=8></p></li>"
		   LF "<li><p><input value=\"View player list\" name=\"sub_stats\" type=\"submit\">"
		   LF " - you have to authenticate as a player to see names and addresses.</p></li>"
		   LF "</ul></form>"
		   LF "<hr><p>Back to <a href=" NEWS_INDEX ">Index</a></p></BODY></HTML>",
		   findval("ev_ref", 1));


_erf: /***** Error handling *****/

	if (rc) fprintf(stderr, "%li:   send_doc(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	return 0;
}

/*** The end ***/
