/* Prepare player target/assassin topology & files according to seeding system. */

/************* BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  Permission to use this code has been granted    **
**  to the Pip Gregenbach (and associates) for the  **
**  period April to July 2002; modifications may    **
**  be made as needed to suit the St. Olaf College  **
**  assassin-like game, so long as this message is  **
**  retained and the changes clearly marked as      **
**  such. The author accepts no responsibility or   **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
**  This code will almost certainly be released     **
**  under a much less restrictive licence when      **
**  finished; please maintain contact with author.  **
*****************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#define MAX_PLAYERS 4000
#define MAX_P_CNT 800
#define MAX_S_CNT 400
#define MAX_T_CNT 400
#define MAX_TRIES 20
#define FILENAME "largelist"
#define GILENAME "testuinfo"
#define QILENAME "testevents"

enum pflg { dead, alive, seed, prim, sec, tert };
enum prest { none, xass, xtar };

void toeol(FILE* fp);

typedef struct player_type
{
	int num;
	int seed;
	enum prest rest;
	enum pflg flg;
	int ass[3];
	int tar[3];
	char id[256];
	char nid[256];
	char name[256];
	char coll[256];
	char email[256];
	char pseu[512];
	char addr[1024];
	char wwww[256];
} player_type;


static player_type player[MAX_PLAYERS+1];
static int playcnt = 0, max_p = MAX_PLAYERS;
static int seeds = 0;
static int seedlist = 0;
char buf[2560] = "Muahahahahahahahaha!";

int randint(int lim)
{
	return(1 + (rand() % lim));
}


int id2num(char *id)
{
	int i;

	for (i = 1; i <= max_p; i++)
	{
		if (!strcmp(id, player[i].nid)) return (i);
	}

	return (0);
}


int pprep(char *from)
{
	int i, j, c;
	
	char tmpc[128];

	player_type *p_ptr, *last_ptr;

	FILE *fp = stdin;

	if (from) fp = fopen(from, "rb");

	/* Scan through for valid lines */
	while ((c = fgetc(fp)) != EOF)
	{
		/* Ignore lines not beginning with "*" character */
		switch (c)
		{
			default   : break;
			case '\n' : continue;
		}
		
		/* acquire pointer */
		p_ptr = &player[++playcnt];

		p_ptr->email[0]=c;

		if (i = tread(fp, p_ptr->email+1,p_ptr->email+256, 0)) return 12341;
		if (i = tread(fp, p_ptr->name, p_ptr->name +256, 0)) return 1235;
		if (i = tread(fp, p_ptr->addr, p_ptr->addr +1024, 0)) return 1236;
		if (i = tread(fp, p_ptr->coll, p_ptr->coll +256, 0)) return 1238;
		/*if (i = tread(fp, p_ptr->nid,  p_ptr->nid  +256, 0)) return 1234;*/
		if (i = tread(fp, p_ptr->pseu, p_ptr->pseu +512, 0)) return 1237;
		if (i = tread(fp, p_ptr->wwww, p_ptr->wwww +256, 0) > 1) return 1236;
		/*p_ptr->wwww[0] = '\0';*/

		strcpy(p_ptr->nid, p_ptr->email);

		sprintf(p_ptr->id, "p%d", 77+playcnt*2);
		/*strcpy(p_ptr->id, p_ptr->nid);*/

		printf("(%s)", p_ptr->nid);

		/* initialise player */
		p_ptr->num = playcnt;
		p_ptr->flg = alive;
		p_ptr->seed = 0;
		p_ptr->rest = none;

		/* no assassins or targets allocated yet */
		for (j = 0; j < 3; j++)
		{
			p_ptr->ass[j] = 0;
			p_ptr->tar[j] = 0;
		}
	}

	/* count number of players */
	max_p = playcnt;

	/* report */
	printf("\n\nOkay. %d players entered.\n\n", max_p);

	if (from) (void) fclose(fp);

	/* no seeds yet */
	seeds = 0;

	/* no previous */
	last_ptr = 0;

	/* ask for seeds */
	printf("Specify seed number %d: ", (seeds + 1));

	/* setup seeds */
	while (gets(tmpc)[0] != '\0')
	{
		/* search for seed id */
		if (!(i = id2num(tmpc)))
		{
			printf("No such player. Try again: ");
			continue;
		}

		/* acquire new pointer */
		p_ptr = &player[i];

		/* check for duplication */
		if (p_ptr->flg != alive)
		{
			printf("Already a seed. Try again: ");
			continue;
		}

		/* start seed list */
		if (!seeds) seedlist = i;

		/* maintain seed list */
		if (last_ptr) last_ptr->seed = i;

		/* save last seed pointer */
		last_ptr = p_ptr;

		/* force seed */
		p_ptr->flg = seed;

		/* no next seed yet */
		p_ptr->seed = 0;

		/* count seeds */
		seeds++;

		/* ask for next seed */
		printf("Specify seed number %d: ", (seeds + 1));
	}

	/* okay */
	if (seeds)
	{
		printf("\n\nOkay. %d seeds entered.\n\n", seeds);
		return(0);
		
	}
	else
	{
		printf("\nNo seeds entered!");
		return (1);
	}
}


int alloc_seed(void)
{
	int thisseed, j, thisprim, c, nextseed = 0;


	/* allocate primary targets to seeds */
	for (thisseed = seedlist; thisseed; thisseed = nextseed)
	{
		/* acquire pointer */
		player_type *p_ptr = &player[thisseed];

		/* progress */
		printf("seeding %s:\n", p_ptr->id);

		/* should be a seed */
		assert(p_ptr->flg == seed);

		/* acquire next seed in list */
		nextseed = p_ptr->seed;

		/* select random unused players as target/assassin */
		for (j = 0; j < 6; j++)
		{
			player_type *q_ptr;

			/* try to find unused players */
			for (c = 0; c < MAX_P_CNT; c++)
			{
				/* choose */
				thisprim = randint(max_p - 1);

				/* don't select self */
				if (thisprim > thisseed) thisprim++;
	
				/* acquire pointer */
				q_ptr = &player[thisprim];

				/* check for pre-allocated */
				if (q_ptr->flg != alive) continue;

				/* now a primary target/assassin */
				q_ptr->flg = prim;

				/* store relation to seed */
				q_ptr->seed = thisseed;

				/* assign seed as assassin */
				if (j < 3)
				{
					q_ptr->ass[0] = thisseed;
					p_ptr->tar[j] = thisprim;
				}

				/* else assign seed as target */
				else
				{
				    q_ptr->tar[0] = thisseed;
					p_ptr->ass[j-3] = thisprim;
				}

				break;
			}

			/* failure */
			if (c == MAX_P_CNT)
			{
				printf("\nFAILURE: Could not allocated enough primary targets to seed after %d iterations.\n\n", c);
				return(0);
			}
		}
	}

	/* message */
	printf("\nPrimary targets sucessfully allocated to %d seeds.", seeds);

	/* sucess */
	return(1);
}


int alloc_prim(void)
{
	int i, c, thisseed, nextseed = 0;

	int tmp_set[8];
	

	/* finish primary allocations */
	for (thisseed = seedlist; thisseed; thisseed = nextseed)
	{
		/* acquire pointer */
		player_type *p_ptr = &player[thisseed];

		/* one link between each pair of primaries (fixed topology) */
		player[p_ptr->tar[0]].tar[0] = p_ptr->ass[2];
		player[p_ptr->ass[2]].ass[0] = p_ptr->tar[0];
		player[p_ptr->tar[1]].tar[0] = p_ptr->ass[1];
		player[p_ptr->ass[1]].ass[0] = p_ptr->tar[1];
		player[p_ptr->tar[2]].tar[0] = p_ptr->ass[0];
		player[p_ptr->ass[0]].ass[0] = p_ptr->tar[2];

		/* do next seed */
		nextseed = p_ptr->seed;

		/* allocate first (local) tranche of secondaries */
		for (i = 0; i < 8; i++)
		{
			int pk;

			player_type *q_ptr;

			for (c = 0; c < MAX_S_CNT; c++)
			{
				/* choose a random player, acquire pointer */
				q_ptr = &player[pk = randint(max_p)];

				/* need an unallocated player */
				if (q_ptr->flg != alive) continue;

				/* sucess */
				break;
			}

			/* allocation failed */
			if (c == MAX_S_CNT)
			{
				printf("\nFAILURE: Could not allocate enough secondary targets around seed after %d iterations.\n", c);
				return(0);
			}

			/* make it secondary */
			q_ptr->flg = sec;

			/* record the owning seed */
			q_ptr->seed = thisseed;

			/* store the new secondary for linking later */
			tmp_set[i] = pk;
		}

		/* link secondaries into primaries (fixed topology) */
		player[ p_ptr->tar[0] ].tar[1] = tmp_set[2];
		player[ p_ptr->tar[0] ].tar[2] = tmp_set[7];
		player[ p_ptr->tar[0] ].ass[1] = tmp_set[0];
		player[ p_ptr->tar[0] ].ass[2] = tmp_set[5];
		player[ p_ptr->tar[1] ].tar[1] = tmp_set[1];
		player[ p_ptr->tar[1] ].tar[2] = tmp_set[4];
		player[ p_ptr->tar[1] ].ass[1] = tmp_set[3];
		player[ p_ptr->tar[1] ].ass[2] = tmp_set[6];
		player[ p_ptr->tar[2] ].tar[1] = tmp_set[0];
		player[ p_ptr->tar[2] ].tar[2] = tmp_set[6];
		player[ p_ptr->tar[2] ].ass[1] = tmp_set[2];
		player[ p_ptr->tar[2] ].ass[2] = tmp_set[5];
		player[ p_ptr->ass[0] ].tar[1] = tmp_set[1];
		player[ p_ptr->ass[0] ].tar[2] = tmp_set[4];
		player[ p_ptr->ass[0] ].ass[1] = tmp_set[3];
		player[ p_ptr->ass[0] ].ass[2] = tmp_set[7];
		player[ p_ptr->ass[1] ].tar[1] = tmp_set[3];
		player[ p_ptr->ass[1] ].tar[2] = tmp_set[5];
		player[ p_ptr->ass[1] ].ass[1] = tmp_set[0];
		player[ p_ptr->ass[1] ].ass[2] = tmp_set[7];
		player[ p_ptr->ass[2] ].tar[1] = tmp_set[2];
		player[ p_ptr->ass[2] ].tar[2] = tmp_set[6];
		player[ p_ptr->ass[2] ].ass[1] = tmp_set[1];
		player[ p_ptr->ass[2] ].ass[2] = tmp_set[4];

		/* some secondaries have two assassins and one target */
		player[ tmp_set[1] ].ass[0] = p_ptr->tar[1];
		player[ tmp_set[1] ].ass[1] = p_ptr->ass[0];
		player[ tmp_set[1] ].tar[0] = p_ptr->ass[2];

		player[ tmp_set[2] ].ass[0] = p_ptr->tar[0];
		player[ tmp_set[2] ].ass[1] = p_ptr->ass[2];
		player[ tmp_set[2] ].tar[0] = p_ptr->tar[2];

		player[ tmp_set[4] ].ass[0] = p_ptr->tar[1];
		player[ tmp_set[4] ].ass[1] = p_ptr->ass[0];
		player[ tmp_set[4] ].tar[0] = p_ptr->ass[2];

		player[ tmp_set[6] ].ass[0] = p_ptr->tar[2];
		player[ tmp_set[6] ].ass[1] = p_ptr->ass[2];
		player[ tmp_set[6] ].tar[0] = p_ptr->tar[1];

		/* others have one assassin and two targets */
		player[ tmp_set[0] ].ass[0] = p_ptr->tar[2];
		player[ tmp_set[0] ].tar[0] = p_ptr->tar[0];
		player[ tmp_set[0] ].tar[1] = p_ptr->ass[1];
		player[ tmp_set[3] ].ass[0] = p_ptr->ass[1];
		player[ tmp_set[3] ].tar[0] = p_ptr->tar[1];
		player[ tmp_set[3] ].tar[1] = p_ptr->ass[0];
		player[ tmp_set[5] ].ass[0] = p_ptr->ass[1];
		player[ tmp_set[5] ].tar[0] = p_ptr->tar[0];
		player[ tmp_set[5] ].tar[1] = p_ptr->tar[2];
		player[ tmp_set[7] ].ass[0] = p_ptr->tar[0];
		player[ tmp_set[7] ].tar[0] = p_ptr->ass[0];
		player[ tmp_set[7] ].tar[1] = p_ptr->ass[1];

		/* record nature of linkage for removal later */
		player[ tmp_set[0] ].rest = xtar;
		player[ tmp_set[1] ].rest = xass;
		player[ tmp_set[2] ].rest = xass;
		player[ tmp_set[3] ].rest = xtar;
		player[ tmp_set[4] ].rest = xass;
		player[ tmp_set[5] ].rest = xtar;
		player[ tmp_set[7] ].rest = xtar;
		player[ tmp_set[6] ].rest = xass;
	}

	/* message */
	printf("\nSecondary targets sucessfully allocated to primaries.\n");

	/* success */
	return(1);
}


int alloc_sec(void)
{
	int pi, j, c;
	int ucnt = 0, scnt = 0, acnt = 0, tcnt = 0;


	/* count unallocated players or secondary players */
	for (pi = 1; pi <= max_p; pi++)
	{
		if (player[pi].flg == alive) ucnt++;
		else if (player[pi].flg == sec) scnt++;
	}

	/* report */
	printf("Found %d partially allocated (secondary) players and %d unallocated (tertiary) players.\n", scnt, ucnt);

	/* extract number of links that need to be made */
	acnt = (6 * ucnt) + (3 * scnt);

	/* report again */
	printf("Trying to arrange %d links to complete initial game topology.\n", acnt);

	/* assertion */
	assert((acnt % 2) == 0);

	/* cheat */
	tcnt = acnt = acnt / 2;

	/* try to link remaining players */
	for (pi = 1; pi <= max_p; pi++)
	{
		/* acquire pointer */
		player_type *p_ptr = &player[pi];

		/* skip players already allocated */
		if ((p_ptr->flg != alive) && (p_ptr->flg != sec)) continue;
		
		/* check for any uncreated links */
		for (j = 0; j < 6; j++)
		{
    		int l, pm, n;

			int *a_ptr;

			player_type *m_ptr;	

			/* find a target */
			if (j < 3) a_ptr = &p_ptr->tar[j];

			/* else find an assassin */
			else a_ptr = &p_ptr->ass[j-3];

			/* test link */
			if (*a_ptr) continue;

			for (c = 0; c < MAX_T_CNT; c++)
			{
				/* find somewhere for it to go (pick a random free link) */
				l = randint((j < 3) ? acnt : tcnt);

				/* search for this link */
				for (pm = 1; (pm <= max_p) && (l > 0); pm++)
				{
					/* acquire pointer */
					m_ptr = &player[pm];

					/* test */
					if ((m_ptr->flg != alive) && (m_ptr->flg != sec)) continue;

					/* check if has a free assassin */
					if (j < 3) { for (n = 0; (n < 3) && (l > 0); n++) if (!m_ptr->ass[n]) l--; }

					/* or else if has a free target */
					else { for (n = 0; (n < 3) && (l > 0); n++) if (!m_ptr->tar[n]) l--; }
				}

				/* correct things */
				n--; pm--;

				/* paranoia */
				if (l != 0)
				{
					printf("This shouldn't happen.");
					continue;
				}

				/* don't allocate to self */
				if (pm == pi)
				{
					putchar('!');
					continue;
				}

				/* don't duplicate existing links */
				if ((m_ptr->tar[0] == pi) || (m_ptr->tar[1] == pi) || (m_ptr->tar[2] == pi) ||
					(m_ptr->ass[0] == pi) || (m_ptr->ass[1] == pi) || (m_ptr->ass[2] == pi))
				{
					putchar('@');
					continue;
				}

				/* success */
				break;
			}

			/* failure in link allocation function? */
			if (c == MAX_T_CNT)
			{
				printf("\n\nFAILURE: Could not allocate secondary/tertiary link after %d iterations.\n", c);
				return(0);
			}

			/* allocate pm as target of pi, and pi as assassin of pm */
			if (j < 3)
			{
				m_ptr->ass[n] = pi;
				*a_ptr = pm;
			}

			/* otherwise allocate pm as assassin of pi, and pi as target of pm */
			else
			{
				m_ptr->tar[n] = pi;
				*a_ptr = pm;
			}

			/* decrease count of links left */
			tcnt--;
			acnt--;
		}
	}

	/* message */
	printf("\nSucessfully allocated all player links.\n\n");

	/* success */
	return(1);
}


int alloc_check(void)
{
	return(1);
}


void repair_sec(void)
{
	int pi, a = 0, b = 0;

	/* remove secondary/teriary links */
	for (pi = 1; pi <= max_p; pi++)
	{
		/* acquire pointer */
		player_type *p_ptr = &player[pi];

		/* remove all links from tertiary players */
		if (p_ptr->flg == alive)
		{
			int j;
			for (j = 0; j < 3; j++) p_ptr->tar[j] = p_ptr->ass[j] = 0;
			a++;
		}

		/* remove some links from secondary players */
		else if (p_ptr->flg == sec)
		{
			/* some automatically have two targets, one assassin */
			if (p_ptr->rest == xtar)
			{
				p_ptr->tar[2] = p_ptr->ass[1] = p_ptr->ass[2] = 0;
				b++;
			}

			/* whilst others have one automatic target, two such assassins */
			else if (p_ptr->rest == xass)
			{
				p_ptr->ass[2] = p_ptr->tar[1] = p_ptr->tar[2] = 0;
				b++;
			}
			else printf("!!!");
		}
		else if ((p_ptr->flg != prim) && (p_ptr->flg != seed)) printf("???");
	}
	printf("(%d,%d)", a, b);
}


void saveit(void)
{
	FILE *f, *g, *q;
	int i;
	char *cp, *cq;


	f = fopen(FILENAME, "wb");
	g = fopen(GILENAME, "wb");
	q = fopen(QILENAME, "wb");

#if 0
	fprintf(g, "*umpire\tPPPPPPP\tThe Umpire\tGonville and Caius\tThe Illumnati\tP9\trho21@cam.ac.uk\t%lu\t00:00:00"
			   "\t0\t0\t0\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tWanted for corruption\tThe Umpire\n"
			   "*npleg\tGGGGGGG\tThe Legal Non-Player\tNone\tNone\tNowhere\tXXX\t%lu\t00:00:00"
			   "\t0\t0\t0\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tUnwanted\tlegal non-player\n"
			   "*npill\tGGGGGGG\tThe Illegal Non-Player\tNone\tNone\tNowhere\tXXX\t%lu\t00:00:00"
			   "\t0\t0\t0\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tZZZ\tUnwanted\tillegal non-player\n",
			   0UL+(1<<0)+(1<<2)+(1<<4)+(1<<7)+(1<<10)+(1<<21),
			   0UL+(1<<0)+(1<<2)+(1<<8)+(1<<9)+(1<<11)+(1<<21),
			   0UL+(1<<0)+(1<<2)+(1<<8)+(1<<9)+(1<<12)+(1<<21));

	fprintf(q, "00000400 mark\n" 
			   "############################################\n"
			   "## Toy with the above line at your peril! ##\n"
			   "############################################\n\n");
#endif

	for (i = 1; i <= max_p; i++)
	{
		/* acquire pointer to player i */
		player_type *p_ptr = &player[i];

		switch (p_ptr->flg)
		{
			case dead:  cp = "dead"; break;
			case alive: cp = "live"; break;
			case seed:  cp = "seed"; break;
			case prim:  cp = "prim"; break;
			case sec:	cp = "secd"; break;
			default:    cp = "????";
		}
		switch (p_ptr->rest)
		{
			case none:  cq = "none"; break;
			case xtar:  cq = "xTAR"; break;
			case xass:  cq = "xASS"; break;
			default:    cq = "????";
		}
		fprintf(f, "Player %3d: (%s) %8s... (tar) %3d %3d %3d (ass) %3d %3d %3d (%s)\n",
				i, cp, p_ptr->id, p_ptr->tar[0], p_ptr->tar[1], p_ptr->tar[2],
				p_ptr->ass[0], p_ptr->ass[1], p_ptr->ass[2], cq);

		fprintf(g, "*%s\t(NoPassword)\t%s\t%s\t%s\t%s\t%s\t%lX\t13:19:00\t0\t0\t0\t0\t-\tNo comment\t%s\n",
			p_ptr->id, p_ptr->name, p_ptr->coll, p_ptr->wwww, p_ptr->addr, 
			p_ptr->email, 0x4104UL, p_ptr->pseu);

		fprintf(q, "*s%d\t00:19:00\t%lX\t%s\t6:20:00\t%s\tT%s:%s\tT%s:%s\tT%s:%s\t-\n",
			i, (1UL<<5)+(1UL<<12)+(1UL<<13)+(1UL<<20),
			p_ptr->id,
			p_ptr->pseu,
			p_ptr->id,
			player[p_ptr->tar[0]].id,
			p_ptr->id, 
			player[p_ptr->tar[1]].id,
			p_ptr->id,
			player[p_ptr->tar[2]].id
			);
	}
	(void) fclose(f);
	(void) fclose(g);
	(void) fclose(q);
}


int main(int argc, char *argv[])
{
	int c;

	umask(0x117);

	/* seed random number generator */
	srand((unsigned int)(time(0)));

	/* Initialise array, ask for seeds */
	if (c = pprep((argc > 1) ? argv[1] : 0))
	{
		printf("\nInvalid information (%d). Terminating.\n", c);
		return (1);
	}

	/* Allocate things */
	printf("Building topology...");

	/* Allocate targets and check for valid shape */
	if (alloc_seed() && alloc_prim())
	{
		for (c = 0; c < MAX_TRIES; c++)
		{
			printf("Final allocation procedure: iteration %d\n", c+1);
			printf("****************************************\n");
			if (alloc_sec()) break;
			else repair_sec();
		}
		if (c != MAX_TRIES)
		{
			printf("\nValid topology found. Saving...\n");
			saveit();
			printf("Finished.\n");
			return(0);
		}
	}

	/* total failure */
	printf("\nNo valid topology found after %d iterations. Terminating.\n", c);
	return (1);
}


/***************************************************************************************************************************
** Low-level file analysis/input routines. The toeol() function skips all characters until EOL or EOF is reported.        **
** The tread() function skips initial tabs, and reads subsequent data into the buffer delimited by (to),(end) until a     **
** tab, EOL or EOF is encountered. (*tonext) is pointed to the character following the terminator of the buffered data.   **
** It returns (0) if data read up to tab, (1) if data was read up to EOL/EOF, (2) if no data was read, (3) on error.      **
** thopp() is as tread, but does not store the data; if non-zero, (*howmuch) is set to required buffer size for data.     **
***************************************************************************************************************************/
int tread(FILE* fp, char* to, char* end, char** tonext)
{
	int c;
	char* tom = to;

	/* Skip initial tabs */
	while ((c = fgetc(fp)) == '\t');

	/* No data to read */
	switch (c)
	{
		case '\n' : ungetc(c, fp);
		case EOF  : return 2;
	}

	/* Read data into buffer */
	while ((c != EOF) && (c != '\n') && (c != '\t') && (to < end))
	{
		*(to++) = c;
		c = fgetc(fp);
	}

	/* Buffer exhaustion */
	if (to >= end) return 3;

	/* Terminate */
	*to = '\0';

	/* Remaining free space in buffer */
	if (tonext) *tonext = to + 1;

	/* Tidy up */
	switch (c)
	{
		case '\t' : return 0;
		case '\n' : ungetc(c, fp);
		case EOF  : return 1;
	}

	/* This shouldn't ever happen */
	return 3;
}
int thopp(FILE* fp, int* howmuch)
{
	int c, n = 0;

	/* Skip initial tabs */
	while ((c = fgetc(fp)) == '\t');

	/* No data to read */
	switch (c)
	{
		case '\n' : ungetc(c, fp);
		case EOF  : return 2;
	}

	/* Read data into buffer */
	while ((c != EOF) && (c != '\n') && (c != '\t')) { n++; c = fgetc(fp); }

	/* How large a buffer would be needed? */
	if (howmuch) *howmuch = n + 1;

	/* Tidy up */
	switch (c)
	{
		case '\t' : return 0;
		case '\n' : ungetc(c, fp);
		case EOF  : return 1;
	}

	/* This shouldn't ever happen */
	return 3;
}
void toeol(FILE* fp)
{
	int c;
	while (( (c = fgetc(fp)) != '\n') && (c != EOF));
}

/* The end */
