/*****************************************************
**  typedefs.h                           (library)  **
**  ----------------------------------------------  **
**  Fundamental structure, type, and preprecessor   **
**  constant definitions, also some useful macros.  **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef included__lib_typedefs
 #error "** lib_typedefs multiply #included **"
#else
 #define included__lib_typedefs fromfile


/* TWINE turns its parameter tokens into a string. */
#define TWINE(x) #x
/* STRING expands its parameter before turning it into a string. */
#define STRING(x) TWINE(x)

/* Basic macros and definitions */
#define TMTV(x,y,z) (((x)<<16)+((y)<<8)+(z))
#define TVAL(x) ((((ulong)(x).day)<<16)+(((ulong)(x).hour)<<8)+((ulong)(x).min))
#define TMAX ULONG_MAX
#define ZERO_TIMER(x) do (x).day = (x).hour = (x).min = 0; while (0)
#define TVTM(x,y) do { (x).day = (y) >> 16; (x).hour = ((y) >> 8) & 0xFF; (x).min = (y) & 0xFF; } while (0)
#define STR(x) ((x) ? (x) : "???")
#define NSTR(x) ((x) ? (x) : "")
#define ThisOrThat(x,y) ((x) ? (x) : (y))
#define streq(x,y) (strcmp((x),(y)) == 0)
#define LF "\x0A"
#define CR "\x0D"

/* Special player ids */
#define NON_ID				"none"
#define UMP_ID				"umpire"
#define LEG_ID				"npleg"
#define ILL_ID				"npill"
#define PUB_ID				"public"

/* Todo --- Remove at least some of these limits */
#define MAX_id				32	/* excluding terminator */
#define MAX_eref			32	/* excluding terminator */
#define MAX_rid				32	/* excluding terminator */
#define MAX_PSEUD			256
#define MAX_WANTED			4096
#define MAX_players			256
#define MAX_timestrfbuf 	64
#define MAX_CheckAct		200

/* General constants */
#define NO_COMMENT			"No comment"
#define SCORE_SHIFT			16
#define DANGER_SHIFT		16
#define SCORE_ROUND			(1<<15)

/* Fundamental types */
typedef struct { unsigned char day, hour, min; } timmy;
typedef unsigned long report_flags;

/* Game specific configuration */
#include "config.h"

/* HTML fragments and error documents */
#define FILE_MOTD			NEWS_ROOT "motd.html"
#define FILE_UnknownRef		DOC_ROOT "UnknownRef.html"
#define FILE_AddEvGuide		DOC_ROOT "AddEvGuide.html"
#define FILE_CantPubAddEv	DOC_ROOT "PubCantAddEv.html"
#define FILE_CantAddEv		DOC_ROOT "PlyCantAddEv.html"
#define FILE_CantPubEditEv	DOC_ROOT "PubCantEditEv.html"
#define FILE_CantEditEv		DOC_ROOT "PlyCantEditEv.html"
#define FILE_BadAuth		DOC_ROOT "BadAuth.html"
#define FILE_ReportGuide	DOC_ROOT "repguide.html"
#define FILE_PlinfoGuide	DOC_ROOT "PlinfoGuide.html"
#define FILE_NoMoreEvents	DOC_ROOT "nomore.html"
#define FILE_CantPubDoReps	DOC_ROOT "NoPubReps.html"
#define FILE_NotRepOwner	DOC_ROOT "NotRepOwner.html"

/* File pathnames */
#define FILE_DefList		DATA_ROOT "baz-plist.txt"
#define FILE_DefTargetList	DATA_ROOT "quz-tlist.txt"
#define FILE_DefMail		DATA_ROOT "email.txt"
#define FILE_PassMail		DATA_ROOT "email-pass.txt"
#define FILE_TargMail		DATA_ROOT "email-targ.txt"
#define FILE_Lock			DATA_ROOT "lock.tmp"
#define FILE_Log			DATA_ROOT "log.txt"
#define FILE_NewsLog		DATA_ROOT "log-mknews.txt"
#define FILE_ScoreLog		DATA_ROOT "log-scores.txt"
#define FILE_CronTab		DATA_ROOT "cron.txt"
#define FILE_Reports		DATA_ROOT "reports.txt"
#define FILE_NewReports		DATA_ROOT "reports-new.txt"
#define FILE_OldReports		DATA_ROOT "reports-old.txt"
#define FILE_Events			DATA_ROOT "events.txt"
#define FILE_NewEvents		DATA_ROOT "events-new.txt"
#define FILE_OldEvents		DATA_ROOT "events-old.txt"
#define FILE_UserInfo		DATA_ROOT "uinfo.txt"
#define FILE_NewUserInfo	DATA_ROOT "uinfo-new.txt"
#define FILE_OldUserInfo	DATA_ROOT "uinfo-old.txt"
#define FILE_Pseuds			DATA_ROOT "pseuds.txt"
#define FILE_NewPseuds		DATA_ROOT "pseuds-new.txt"
#define FILE_OldPseuds		DATA_ROOT "pseuds-old.txt"

/* Command strings to be passed to system() to run non-CGI executables */
#define SYS_MkNews			BIN_ROOT "mknews "
#define SYS_Scores			BIN_ROOT "scoring +e "

/* Use different files for test builds */
#ifdef TEST
 #undef FILE_UserInfo
 #undef FILE_NewUserInfo
 #undef FILE_OldUserInfo
 #define FILE_UserInfo		DATA_ROOT "uinfoT.txt"
 #define FILE_NewUserInfo	DATA_ROOT "uinfoT-new.txt"
 #define FILE_OldUserInfo	DATA_ROOT "uinfoT-old.txt"
#endif

/* Player flags */
#define PF_Hidden			(1UL<<0)	/* doesn't show up on stats */
#define PF_Retired			(1UL<<1)
#define PF_Playing			(1UL<<2)	/* can make attempts */
#define PF_Dangerous		(1UL<<3)
#define PF_Login			(1UL<<4)	/* can login */
#define PF_Dead				(1UL<<5)	/* currently dead */
#define PF_ResTimer			(1UL<<6)
#define PF_Admin			(1UL<<7)
#define PF_Victim			(1UL<<8)	/* targetable */
#define PF_PartialFix		(1UL<<9)	/* protect some flags (and possibily the timer) in uinfo entry */
#define PF_FullFix			(1UL<<10)	/* never alter this uinfo entry */
#define PF_LegalNP			(1UL<<11)	/* is the legal np */
#define PF_IllegalNP		(1UL<<12)	/* is the illegal np */
#define PF_Naughty			(1UL<<13)
#define PF_Mailable			(1UL<<14)	/* can be mailed */
#define PF_InactTimer		(1UL<<15)	/* incompetency timer running */
#define PF_InactWarning		(1UL<<16)	/* activity timer close to expiry */
#define PF_Inactive			(1UL<<17)	/* activity timer has expired */
#define PF_Incompetent		(1UL<<18)
#define PF_Wanted			(1UL<<19)
#define PF_Police			(1UL<<20)
#define PF_Special			(1UL<<21)
#define PF_Outdated			(1UL<<22)
#define PF_SendMail			(1UL<<23)
#define PF_Redeemed			(1UL<<24)
#define PF_Public			(1UL<<25)
#define PF_Observer			(1UL<<26)
#define PF_LongComment		(1UL<<27)

/* Macros and definitions for player flag manipulation */
#define PF_MASK_Flags		(0x0FFFFFFF)
#define PF_MASK_Type		(PF_Wanted|PF_Police|PF_Incompetent)
#define PF_MASK_Sorting		(PF_MASK_Type|PF_Inactive|PF_InactWarning|PF_Dead)
#define PF_MASK_Scoring		(PF_MASK_Sorting|PF_ResTimer|PF_InactTimer|PF_Naughty|PF_Redeemed)
#define PF_MASK_NoEdit		(PF_PartialFix|PF_FullFix|PF_LegalNP|PF_IllegalNP|PF_Special)
#define PF_MASK_Timers		(PF_InactTimer|PF_ResTimer)
#define PF_MASK_Fixed		(PF_MASK_Flags)
#define PF_type(x)			(((x) & (PF_MASK_Type|PF_Special)) >> 18)
#define PF_group(x)			((((x) & PF_MASK_Type) >> 18) | (((x) & PF_Dead) >> 2))
#define PF_genus(x)			(((x) & PF_MASK_Type) >> 18)
#define INITIAL_FLAGS		(0UL)

/* Event flags */
#define EF_Kill				(1UL<<0)	/* fatal (EF_Assassination) */
#define EF_Invisible		(1UL<<1)	/* does not show on news */
#define EF_Comment			(1UL<<2)	/* comment event */
#define EF_Ignore			(1UL<<3)	/* pretend this event does not exist */
#define EF_AddPseudonym		(1UL<<4)	/* give new pseudonym to player 1, and use it for this event */
#define EF_SetPseudonym		(1UL<<5)	/* set player 1's default pseudonym */
#define EF_Open				(1UL<<6)	/* can add more reports */
#define EF_Assassination	(1UL<<7)	/* actual kill/attempt event */
#define EF_Temp				(1UL<<8)
#define EF_Police			(1UL<<9)	/* make/unmake player 1 as police: --> rank field present */
#define EF_MakeWanted		(1UL<<10)	/* make player 1 wanted */
#define EF_MakeUnWanted		(1UL<<11)	/* make player 1 unwanted */
#define EF_Targetting		(1UL<<12)	/* targetting present */
#define EF_SetInactTimer	(1UL<<13)	/* set and start player 1's inactivity timer */
#define EF_Promotion		(1UL<<14)	/* rank field is relative, not absolute */
#define EF_ShowHeadline		(1UL<<15)	/* show event headline */
#define EF_MakeIncompetent	(1UL<<16)	/* make player 1 incompetent */
#define EF_Detarg			(1UL<<17)	/*  */
#define EF_DetargP2			(1UL<<18)	/*  */
#define EF_OnSpecial		(1UL<<19)	/*  */
#define EF_MakeCompetent	(1UL<<20)	/* make player 1 competent */
#define EF_ForceTargets		(1UL<<21)	/* hack --- force alteration of player 1's targets, ignoring their vitality */
#define EF_Reincarnate		(1UL<<22)	/* Bring player back to life */
#define EF_News				(1UL<<23)	/* event has headline/update fields */
#define EF_Wantedness		(1UL<<24)	/* wantedness/police comment field present */
#define EF_Annulled			(1UL<<25)	/* event annulled for scoring purposes (uncredited kill/death) */
#define EF_NoDangerBonus	(1UL<<26)	/* don't award dangerousness bonus or change dangerousness */
#define EF_InvertAutoRes	(1UL<<27)	/* reverse normal resurrection type on player 2 (EF_Assassination and EF_Kill) */
#define EF_DoubleKill		(1UL<<28)	/* --- kill player 1 as well (EF_Assassination and EF_Kill) --- not implemented */
#define EF_SetP2Pseudonym	(1UL<<29)	/* set player 2's default pseudonym (EF_Assassination) */
#define EF_ScoreBonus		(1UL<<30)	/* Scoring bonus present */
#define EF_XXX2				(1UL<<31)

/* Macros and definitions for event flag manipulation */
#define EF_MASK_Pseudonym	(EF_AddPseudonym|EF_SetPseudonym|EF_Assassination)

/* Macros and definitions for automatic player resurrection */
#ifdef DAY_AutoResLimit
 #define OPT_AutoRes (1)
 #define AUTORES(x) (((x) & EF_InvertAutoRes) ^ EF_InvertAutoRes)
#else
 #define OPT_AutoRes (0)
 #define AUTORES(x) (((x) & EF_InvertAutoRes))
 #define DAY_AutoResLimit (0x7FFF)
#endif

/* Report flags */
#define RF_MASK_Auth		((report_flags) (0xFF))	/* XXX */
#define RF_LoadText			(((report_flags)1)<<8)	/* Temporary flag: load report text */
#define RF_Already			(((report_flags)1)<<9)	/* Temporary flag: report already dealt with */
#define RF_Comment			(((report_flags)1)<<10)	/* Report is a comment */
#define RF_Main				(((report_flags)1)<<11)	/* Report is main (toplevel) fragment for event */
#define RF_AutoGen			(((report_flags)1)<<12)	/* Report was automatically generated */
#define RF_Listed			(((report_flags)1)<<13)	/* Show report existence in event overview */
#define RF_Hidden			(((report_flags)1)<<14)	/* Don't show report in news */
#define RF_0_Repro			(((report_flags)1)<<15)	/* Reproduction flag 0 */
#define RF_NoIntro			(((report_flags)1)<<16)	/* Don't present automatic introduction on [Z] inclusion */
#define RF_1_Repro			(((report_flags)1)<<17)	/* Reproduction flag 1 */
#define RF_Priv				(((report_flags)1)<<18)	/* Allow priviledged content in report */

#define RF_Repro			(RF_0_Repro|RF_1_Repro)
#define RF_Normal			(0)							/* Process reports as normal: check HTML and directives */
#define RF_Verbatim			(RF_0_Repro)				/* Print report directly out into news; don't process */
#define RF_Encapsulate		(RF_1_Repro)				/* Encapsulate report in news; no HTML or directives allowed */
#define RF_XXX_Repro		(RF_0_Repro|RF_1_Repro)		/* Unused */

/* XXX XXX --- no longer used? */
#define Change_Pseud		(1)
#define Change_Pass			(2)
#define Change_Attrib		(4)
#define Change_Timer		(8)
#define Change_Events		(16)
#define Change_News			(32)

/* XXX XXX --- old return codes --- convert to ERF_xxx */
#define DOC_WrongUser		(-2)
#define DOC_WrongPass		(-1)
#define DOC_Successful		(0)
#define DOC_WhatToDo 		(6)
#define DOC_ValUnOpen		(7)
#define DOC_SetUnOpen		(9)
#define DOC_BadFPos			(11)
#define DOC_Remaining		(13)
#define DOC_NoLoginForYou	(14)
#define DOC_StatBufEx		(16)
#define DOC_PSMenuWho		(17)
#define DOC_NoPass			(18)
#define DOC_NewsTooLong		(21)
#define DOC_TooManyRep		(23)
#define DOC_Badsscanf		(28)
#define DOC_MailFail		(29)
#define DOC_BadEventParms	(512)
#define DOC_ManyTargets		(519)
#define DOC_TooManyIds		(520)
#define DOC_BadId			(521)
#define DOC_OverMatch		(522)
#define DOC_NoWanted		(523)
#define DOC_HeadlineError	(525)
#define DOC_HistoryError	(526)

/* Error/failure return codes (0-127) */
#define ERF_Void 			(1)		/* --- */
#define ERF_CGIExhaustion 	(2)
#define ERF_CGIManyPairs 	(3)
#define ERF_NoIdGiven 		(4)
#define ERF_fopenEvents		(5)		/* fopen() failed on events file */
#define ERF_fopenUinfo		(8)		/* fopen() failed on uinfo file */
#define ERF_FSwapFailed		(10)
#define ERF_fputsFailed		(12)
#define ERF_BufferExhaust	(15)
#define ERF_MissingPid		(19)
#define ERF_fopenReps		(20)
#define ERF_fopenNews		(22)
#define ERF_BadRepIdx		(24)
#define ERF_BadEvIdx		(25)
#define ERF_NoRefMatch		(26)
#define ERF_BadEvData		(27)
#define ERF_flockFailed		(30)
#define ERF_BadAuth			(31)
#define ERF_BadTread		(32)	/* 32 = unlikely, 33 = early end of line, 34 = no data, 35 = clib error */
#define ERF_BadDateText		(36)	/* failure to convert textual date */
#define ERF_fcntlFailed		(37)
#define ERF_setenvFailed	(38)
#define ERF_execFailed		(39)
#define ERF_ftellFailed		(40)
#define ERF_fseekFailed		(41)
#define ERF_mallocFailed	(42)
#define ERF_freopenErrLog	(43)
#define ERF_tmpfileFailed	(44)
#define ERF_FileError		(45)
#define ERF_fcloseFailed	(46)
#define ERF_BadParms		(47)
#define ERF_InternalError	(48)
#define ERF_Paranoia		(49)
#define ERF_DuplicateIds	(50)
#define ERF_UnknownPlayer	(51)
#define ERF_putcFailed		(52)
#define ERF_fopenFailed		(53)	/* general failures in fopen() */
#define ERF_UnknownAction	(54)
#define ERF_Naughty			(55)	/* --- */
#define ERF_Empty			(56)
#define ERF_printfFailed	(57)
#define ERF_NotOpen			(58)	/* --- */
#define ERF_UnknownRef		(59)
#define ERF_UnknownRepIdx	(60)
#define ERF_freadFailed		(61)
#define ERF_fwriteFailed	(62)
#define ERF_CannotEdit		(63)	/* --- */
#define ERF_TooManyPlayers	(64)
#define ERF_NoPermission	(65)
#define ERF_PassSecurity	(66)
#define ERF_BadUserData		(67)
#define ERF_IncompleteEdit	(68)
#define ERF_BadInput		(69)
#define ERF_EndOfFile		(70)
#define ERF_NoHeader		(71)
#define ERF_BadHeader		(72)
#define ERF_WrongFormat		(73)
#define ERF_fputcFailed		(74)
#define ERF_CantPubAddEv	(75)
#define ERF_CantAddEv		(76)
#define ERF_CantEditEv		(77)
#define ERF_CantPubEditEv	(78)
#define ERF_CantPubDoReps	(79)
#define ERF_NotRepOwner		(80)
#define ERF_BadPubAuth		(81)
#define ERF_Targetting		(82)

/* Marcos for dealing with error/failure return codes (low byte = rc, higher bits = context) */
#define ERF_LOWBITS(x)		((x)&(127))
#define ERF_TYPE(x)			((x)&(255))
#define ERF_CONTEXT(x)		((x>>8))

/* Standard "set rc number and context, then branch to error handler" function-like macro */
#define erf(x,y) do { rc = ((y)<<8)|(x); goto _erf; } while (0)

/* Restrict rank array indices */
#define LIM_RANK(x) (((x) > MAX_rank) ? (MAX_rank + 1) : ((x) < 0) ? 0 : (x))

/***************************************************************************************************************************
** Type, structure and function definitions                                                                               **
***************************************************************************************************************************/

/* Some type pointers to aid in structure definitions */
typedef struct event* event_p;
typedef struct pshort* pshort_p;
typedef struct player* player_p;
typedef struct pstate* pstate_p;
typedef struct pexist* pexist_p;
typedef struct ppseud* ppseud_p;
typedef struct report* report_p;
typedef struct target* target_p;
typedef struct retarg* retarg_p;

/* Short (basic) player information structure... note field naming is compatible with full player struct */
struct pshort
{
	char *name;
	pshort_p link;
	player_p pptr;
	char id[MAX_id];
};

#ifndef PLAYER_EXT_TYPE
 #define PLAYER_EXT_TYPE int
#endif

/* Full, general-purpose player information structure */
struct player
{
	long score, danger;
	char *id;
	char *pass;
	char *name;
	char *coll;
	char *mafia;
	char *addr;
	char *wanted;
	char *pseud;
	char *email;
	target_p asses;
	target_p targs;
	target_p oldts;
	timmy timer;
	unsigned long flags;
	int kills, deaths;
	long rank;
	player_p link;
	PLAYER_EXT_TYPE ext;
};

#ifndef EVENT_EXT_TYPE
 #define EVENT_EXT_TYPE int
#endif

/* Full, general-purpose event information structure */
struct event
{
	char *ref;
	char *id;
	char *pseud;
	retarg_p untarg;
	retarg_p retarg;
	char *vict_pid;
	char *vict_pseud;
	char *news_head;
	char *wantedness;
	long bonus;
	long rank;
	ulong flags;
	timmy when;
	timmy ntimer;
	timmy news_date;
	event_p link;
	EVENT_EXT_TYPE ext;
};

/* Full, general-purpose report information structure */
struct report
{
	char *rid;
	char *pid;
	char *pseud;
	char *text;
	long int fpos, length;
	report_flags flags;
	player_p pptr;
	report_p link_all, link_sub;
	timmy uptim;
};

#ifndef TARGET_EXT_TYPE
 #define TARGET_EXT_TYPE int
#endif

/* Player's target/assassin information structure */
struct target
{
	char *id;
	target_p next;
	TARGET_EXT_TYPE ext;
};

/* Retargetting information structure */
struct retarg
{
	char *assid;
	char *tarid;
	retarg_p next;
};

#endif
/*** The End ***/
