/*****************************************************
**  genmail.c                       >> bin/genmail  **
**  ----------------------------------------------  **
**  TODO: Add player status directives.             **
**        Choose default target/password options.   **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/


#define NO_CGI
#define MAX_fpbuf			65536
#define MAX_statbuf			131072
#define PLAYER_EXT_TYPE 	int

#include "!lib-general.c"
#include "!custom.c"
#include "!lib-uinfo-scan.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/
/* Buffer for input message */
#define	MAX_inbuf	(64*1024)

/* Magic characters */
#define MAGOS "{"
#define MAGOC '{'
#define MAGCC '}'
#define MAGCS "}"

/* Mode flags */
#define MODE_And		(1U<<0)
#define MODE_Debug		(1U<<1)
#define MODE_Verbose	(1U<<2)
#define MODE_NoMail		(1U<<3)

/* Default mode */
static unsigned int mode = MODE_And|MODE_Verbose;

/* Special pointer used to denote default filenames. String irrelevent, but should be non-empty. */
static char* defp = "(default)";

/* Global variables */
static char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos;
static char inbuf[MAX_inbuf+4];

/* Parsed fragment of input text */
struct fragment
{
	struct fragment *next;
	size_t len;
	char *text;
	char type, num, flob;
};

/* Insertation fragment types */
#define FRAG_Text			0
#define FRAG_Pseud			1
#define FRAG_Name			2
#define FRAG_Addr			3
#define FRAG_College		4
#define FRAG_Mafia			5
#define FRAG_Password		6
#define FRAG_Date			7
#define FRAG_From			8
#define FRAG_Email			9
#define FRAG_Numeric		10
#define FRAG_Deadline		11

/* Conditional fragment types (+1 => if, +0 => ifnot/else, +2 => endif, +3 unused) */
#define FRAG_Conditional	32
#define FRAG_IsDead			32
#define FRAG_IsPolice		36
#define FRAG_IsAllocated	40
#define FRAG_HasNewTargs	44
#define FRAG_HasOldTargs	48
#define FRAG_IsWanted		52
#define FRAG_IsIncompetent	56
#define FRAG_IsInactive		60
#define FRAG_IsSpecial		64
#define FRAG_IsToBeMailed	68
#define FRAG_IsListed		72
#define FRAG_IsMailable		76

/* */
#define FLOB_PL		0
#define FLOB_ASS	1
#define FLOB_TAR	2
#define FLOB_OLD	3
#define FLOB_MAX	4

/* Number of possible search terms */
#define MAX_SearchTerms		12

/* Printable names for search terms */
char* sterm_tab[MAX_SearchTerms] =
{
	"Dead", "Police", "XXX", "HasNewTargets", "HasOldTargets",
	"Wanted", "Incompetent", "Inactive", "Special", "Selected", "Listed", "Mailable"
};

/***************************************************************************************************************************
** Evaluation of conditional expressions on players                                                                       **
***************************************************************************************************************************/
unsigned int evalcond(player_p thispptr, player_p *pl[FLOB_MAX], int plim[FLOB_MAX], int type, int n, int f)
{
	char *p, *q;

	/*fprintf(stderr, ">> condition %i on %i (for %s)\n", type, n, (pl[0]) ? pl[0]->id : "!!none!!");*/

	if (n >= plim[f]) return 0;

	switch (type = type & 0xFC)
	{
		/* TRUE if player[f][n] is allocated (ie. as n-th assassin/target/oldtarget for player[0]) */
		case FRAG_IsAllocated: return (pl[f][n]) ? 1 : 0;

		/* TRUE if player[f][n] is unallocated or is dead */
		case FRAG_IsDead: return ((!pl[f][n]) || (pl[f][n]->flags & PF_Dead));

		/* TRUE if player[f][n] is allocated and mailable */
		case FRAG_IsMailable: return ((pl[f][n]) && (pl[f][n]->flags & PF_Mailable));

		/* TRUE if player[f][n] is allocated and has been selected for mailing */
		case FRAG_IsToBeMailed: return ((pl[f][n]) && (pl[f][n]->flags & PF_SendMail));

		/* TRUE if player[f][n] is allocated and is special */
		case FRAG_IsSpecial: return ((pl[f][n]) && (pl[f][n]->flags & PF_Special));

		/* TRUE if player[f][n] is allocated and was in matched in player list given as input */
		case FRAG_IsListed: return ((pl[f][n]) && (pl[f][n]->ext));

		/* TRUE if player[f][n] is allocated and is police */
		case FRAG_IsPolice: return ((pl[f][n]) && (pl[f][n]->flags & PF_Police));

		/* TRUE if player[f][n] is allocated and is wanted */
		case FRAG_IsWanted: return ((pl[f][n]) && (pl[f][n]->flags & PF_Wanted));

		/* TRUE if player[f][n] is allocated and is incompetent */
		case FRAG_IsIncompetent: return ((pl[f][n]) && (pl[f][n]->flags & PF_Incompetent));

		/* TRUE if player[f][n] is allocated and is inactive */
		case FRAG_IsInactive: return ((pl[f][n]) && (pl[f][n]->flags & PF_Inactive));

		/* TRUE if player is allocated and with some different current and old target(s) */
		case FRAG_HasNewTargs:

		/* TRUE if player is allocated and with some identical current and old target(s) */
		/*case FRAG_HasOldTargs:*/
		{
			target_p tar1p, tar2p;
	
			if (f == 0)
			{
				if (thispptr) for (tar1p = thispptr->targs; (tar1p); tar1p = tar1p->next)
				{
					for (tar2p = thispptr->oldts; (tar2p); tar2p = tar2p->next)
					{
						if (streq(tar1p->id, tar2p->id)) break;
					}
					if (!tar2p) return 1;
				}
			}
			else if ((f == FLOB_TAR) || (f == FLOB_OLD))
			{
				if (!((pl[f][n]) && (p = pl[f][n]->id))) return 0;
				for (tar2p = (f == FLOB_TAR) ? thispptr->oldts : thispptr->targs; (tar2p); tar2p = tar2p->next)
				{
					if (streq(p, tar2p->id)) break;
				}
				if (!tar2p) return 1;
			}
			return 0;
		}

		/* Unknown: return TRUE */
		default: return 1;
	}
}

/***************************************************************************************************************************
** Routines to parse input files                                                                                          **
***************************************************************************************************************************/
struct fragment* addfrag(struct fragment** ffr, struct fragment** lfr, char* p, size_t l, char t, char n, char f)
{
	struct fragment* frag;

	if (!(frag = malloc(sizeof(struct fragment)))) return 0;

	if (p) p[l] = '\0';

	frag->text = p;
	frag->len  = l;
	frag->type = t;
	frag->num  = n;
	frag->flob = f;
	frag->next = 0;

	if (!(*ffr)) *ffr = frag;
	if (*lfr) (*lfr)->next = frag;

	/*fprintf(stderr, "addfrag(&%p,&%p,&%p,%u,%i,%i)-->&%p\n", *ffr, *lfr, p, l, t, n, frag);*/
	/*if (p) fprintf(stderr, "       \"%.30s\"", p);*/

	return (*lfr = frag);
}

int parsefrags(FILE *fp, struct fragment** firstfragp)
{
	struct fragment *lastfrag;
	size_t s, x, y, z;
	int rc = 0;


	/* Nothing yet */
	*firstfragp = 0;

	/* Slurp in input file */
	for (s = 0; (!ferror(fp)) && (!feof(fp)); s += x)
	{
		/* Room in buffer? */
		if (s >= MAX_inbuf) erf(ERF_BufferExhaust, 3);

		/* Read as much as possible */
		if ((x = fread(inbuf + s, 1, MAX_inbuf - s, fp)) == 0) continue;

		/* Paranoia */
		else if (/*(x < 0) ||*/ (x > (MAX_inbuf - s))) erf(ERF_freadFailed, 4);
	}

	/* Any residual errors? */
	if (ferror(fp)) erf(ERF_FileError, 5);

	/* Terminate buffer */
	inbuf[s] = '\0';

	/* Scan buffered file for special stuff */
	for (x = y = 0, lastfrag = 0; (x < s); x++)
	{
		char t = 0, n = 0, f = 0;

		/* Important strings begin with magic char */
		if (inbuf[x] != MAGOC) continue;

		/* Use "{{" to get '{' character */
		if (inbuf[x + 1] == MAGOC) x++;

		/* {Date} directive? */
		else if (!strncasecmp(inbuf + x + 1, "date" MAGCS, 5))
		{
			t = FRAG_Date;
			z = x + 5;
		}

		/* {From} directive? */
		else if (!strncasecmp(inbuf + x + 1, "from" MAGCS, 5))
		{
			t = FRAG_From;
			z = x + 5;
		}

		/* {Password} directive? */
		else if (!strncasecmp(inbuf + x + 1, "password" MAGCS, 9))
		{
			t = FRAG_Password;
			z = x + 9;
		}

		/* Normal directives */
		else
		{
			/* Type of directive */
			switch (tolower(inbuf[z = x + 1]))
			{
				/* Insertation directives */
				case 'p': t = FRAG_Pseud; break;
				case 'n': t = FRAG_Name; break;
				case 'a': t = FRAG_Addr; break;
				case 'c': t = FRAG_College; break;
				case 'm': t = FRAG_Mafia; break;
				case 'e': t = FRAG_Email; break;
				case 'b': t = FRAG_Deadline; break;

				/* Ignore unknown directives */
				default : x = z; continue;

	#if 0
				/* "Counting" insertation directives */
				case '#':
				{
					/* What to count? */
					switch (tolower(inbuf[++z]))
					{
						case 'a': n = 1; break;
						case 't': n = 4; break;
						case 'o': n = 7; break;
						case 'n': n = 8; break;

						/* Ignore unknown things */
						default : x = z; continue;
					}
					t = FRAG_Numeric; break;
				}
	#endif

				/* Conditional directives */
				case '/': ++t; /* Fall-through 2 */
				case '?': ++t; /* Fall-through 1 */
				case '!': /* Otherwise, have t=0 */
				{
					/* Which condition? */
					switch (tolower(inbuf[++z]))
					{
						case 'd': t += FRAG_IsDead; break;
						case 'p': t += FRAG_IsPolice; break;
						case 'x': t += FRAG_IsAllocated; break;
						case 'n': t += FRAG_HasNewTargs; break;
						case 'o': t += FRAG_HasOldTargs; break;
						case 's': t += FRAG_IsSpecial; break;
						case 'w': t += FRAG_IsWanted; break;
						case 'i': t += FRAG_IsIncompetent; break;
						case 'c': t += FRAG_IsInactive; break;
						case 'm': t += FRAG_IsToBeMailed; break;
						case 'l': t += FRAG_IsListed; break;
						case 'e': t += FRAG_IsMailable; break;

						/* Ignore unknown conditions */
						default : x = z; continue;
					}
					break;
				}
			}

			/* No numeric suffix allowed? */
			if ((n) && (inbuf[++z] != MAGCC))
			{
				/* Ignore bad directives */
				x = (inbuf[z]) ? (z + 1) : z;
				continue;
			}

			/* No numeric suffix? Assume zero. */
			if ((!n) && (inbuf[++z] != MAGCC))
			{
				int tmp = tolower(inbuf[z]);

				if (tmp == 'a') f = FLOB_ASS;
				else if (tmp == 't') f = FLOB_TAR;
				else if (tmp == 'o') f = FLOB_OLD;
				else f = 0;

				if ((f) && ((inbuf[++z] < '0') || ((n = inbuf[z] - '0') > 9) || (inbuf[++z] != MAGCC))) f = 0;
				
				if (!f)
				{
					/* Ignore bad directives */
					x = (inbuf[z]) ? (z + 1) : z;
					continue;
				}
			}

			/*fprintf(stderr, "foom!\"%s\"\n", inbuf + x);*/
		}

		/* Add previous string fragment */
		if ((x > y) && (!addfrag(firstfragp, &lastfrag, inbuf + y, x - y, FRAG_Text, 0, 0))) erf(ERF_mallocFailed, 20);

		/* Letter following magic char */
		y = x + 1;

		/* Add command fragment */
		if (t)
		{
			if (!addfrag(firstfragp, &lastfrag, inbuf + y, z - y, t, n, f)) erf(ERF_mallocFailed, 21);
			y = 1 + (x = z);
		}
	}

	/* Don't forget the last string fragment */
	if ((x > y) && (!addfrag(firstfragp, &lastfrag, inbuf + y, x - y, FRAG_Text, 0, 0))) erf(ERF_mallocFailed, 22);


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li !! parsefrags(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

int writefrags(FILE *fp, player_p thispptr, player_p *plarr[FLOB_MAX], int plim[FLOB_MAX],
			   player_p plist, struct fragment *fraglist, char* date, char* from)
{
	struct fragment *frag;
	player_p pptr;
	int i, f, ok;
	char* p;


	/* Write out each fragment in turn */
	for (frag = fraglist; (frag); frag = frag->next)
	{
		/*fprintf(stderr, "* frag %p type %i flob %i num %i %s len %lu\n", frag, frag->type,
				frag->flob, frag->num, (frag->type == FRAG_Text) ? "text" : frag->text, (unsigned long) frag->len);*/

		/* No text yet */
		p = 0;

		/* Deal with conditional blocks */
		if (frag->type >= FRAG_Conditional)
		{
			unsigned int a, b;

			/* Ignore out-of-place blocks */
			if ((a = frag->type & 3) > 1) continue;

			/* Evaluate condition to b = 0/1 */
			b = evalcond(thispptr, plarr, plim, frag->type & 0xFC, i = frag->num, f = frag->flob);

			/*fprintf(stderr, "%li:   conditional \"%s\" (%i,%i) a=%i b=%u -> %s\n", uniqid,
					frag->text, frag->type, frag->num, a, b, (a ^ b) ? "unsatisfied" : "satisfied");*/

			/* Match on required condition status */
			if (!(a ^ b)) continue;

			/* Endif and Else conditions */
			b = 2 + (frag->type & 0xFC);
			a = b - 1 - a;

			/* Scan along to end of block */
			for (frag = frag->next; (frag); frag = frag->next)
			{
				/* Corresponding Else or Endif? */
				if ((frag->num == i) && (frag->flob == f) && ((frag->type == a) || (frag->type == b))) break;
			}

			/* Run out of fragments? */
			if (!frag) break;

			/* Otherwise, Else automatically satisfied, or Endif */
			else continue;
		}

		if (frag->flob >= FLOB_MAX) return 999;

		ok = (frag->num < plim[frag->flob]);

		/* What sort of fragment? */
		switch (frag->type)
		{
			case FRAG_Text:
			{
				if (frag->len != fwrite(frag->text, 1, frag->len, fp)) return ERF_fwriteFailed;
				break;
			}

			case FRAG_Password:	p = thispptr->pass; break;
			case FRAG_Date:		p = date; break;
			case FRAG_From:		p = from; break;

			case FRAG_Name:		if ((ok) && (pptr = plarr[frag->flob][frag->num])) p = pptr->name; break;
			case FRAG_Addr:		if ((ok) && (pptr = plarr[frag->flob][frag->num])) p = pptr->addr; break;
			case FRAG_College:	if ((ok) && (pptr = plarr[frag->flob][frag->num])) p = pptr->coll; break;
			case FRAG_Mafia:	if ((ok) && (pptr = plarr[frag->flob][frag->num])) p = pptr->mafia; break;
			case FRAG_Pseud:	if ((ok) && (pptr = plarr[frag->flob][frag->num])) p = pptr->pseud; break;

			case FRAG_Deadline:
			{
				if ((!ok) || !(pptr = plarr[frag->flob][frag->num])) break;
				if (pptr->flags & PF_InactTimer)
				{
					fprintf(fp, "%.2d:%.2d on %s", pptr->timer.hour, pptr->timer.min, fdayname(pptr->timer.day, 0));
				}
				else if (EOF == fputs("None", fp)) return ERF_fputsFailed;
				break;
			}

			case FRAG_Email:
			{
				if ((!ok) || !(pptr = plarr[frag->flob][frag->num]) || (!pptr->email)) break;
				if (EOF == fputs(pptr->email, fp)) return ERF_fputsFailed;
				if (!strchr(pptr->email, '@') && (EOF == fputs(MAILDOMAIN, fp))) return ERF_fputsFailed;
				break;
			}

#if 0
			case FRAG_Numeric:
			{
				unsigned int n = 0, x = (frag->num == 7) ? 1 : 0;
				if (frag->num >= 7) for (i = 0; (i < 3); i++)
				{
					 if ((plarr[4+i]) && (((!plarr[7+i]) || strcmp(plarr[4+i]->id, plarr[7+i]->id)) ^ x)) ++n;
				}
				else for (i = 0; (i < 3); i++) if (plarr[frag->num+i]) ++n;
				if (1 != fprintf(fp, "%u", n)) return ERF_printfFailed;
				break;
			}
#endif
			default: return ERF_InternalError;
		}

		/*if (p) fprintf(stderr, " (\"%s\")\n", p);*/

		/* Write indirected string */
    	if ((p) && (EOF == fputs(p, fp))) return ERF_fputsFailed;
	}

	return 0;
}

/***************************************************************************************************************************
** Main code. Command tail may be "all" or "sel" or anything else for normal action                                       **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	int i, n = 0, nontriv, rc = 0, lockfd = 0, sterm[MAX_SearchTerms], plim[FLOB_MAX];
	unsigned int j;
	time_t thetime;
	player_p plist = 0, pptr, *plarr[FLOB_MAX];
	char *p, thedate[64], *do_input = 0, *do_list = 0, *do_pidlist = 0, *do_tarlist = 0;
	struct fragment *frags = 0;
	FILE* fp = 0, *fp_pidlist = 0, *fp_tarlist = 0;


	/* No search terms, by default */
	for (i = 0; (i < MAX_SearchTerms); i++) sterm[i] = 0;

	/* What time is it? */
	time(&thetime);

	/* Turn it into a valid RFC 822 date */
	strftime(thedate, 64, "%a, %d %b %Y %H:%M:%S GMT", gmtime(&thetime));

	/* File creation umask */
	umask(0117);

	/* Get process id for logging */
	uniqid = (long) getpid();

	/* Logfile header */
	fprintf(stderr, "%li ** Genmail: at RFC822 date \"%s\"\n", uniqid, thedate);

	/* Check command line args */
	for (i = 1; (i < argc) && (p = argv[i]); i++)
	{
		char **ap = 0;
		unsigned int m, mf = 0;

		/* Qualifier? */
		switch (*(p++))
		{
			case '+': m = 1; break;
			case '-': m = 2; break;
			case '^': m = 3; break;
			case '!': m = 4; break;
			case '?': m = 5; break;
			case '/': m = 6; break;
			default : m = 0; --p;
		}

		/* Search switches */
		if (m >= 4)
		{
			/* Which condition? */
			switch (tolower(p[0]))
			{
				case 'd': mf = FRAG_IsDead; break;
				case 'p': mf = FRAG_IsPolice; break;
				case 'n': mf = FRAG_HasNewTargs; break;
				case 'o': mf = FRAG_HasOldTargs; break;
				case 's': mf = FRAG_IsSpecial; break;
				case 'w': mf = FRAG_IsWanted; break;
				case 'i': mf = FRAG_IsIncompetent; break;
				case 'c': mf = FRAG_IsInactive; break;
				case 'm': mf = FRAG_IsToBeMailed; break;
				case 'l': mf = FRAG_IsListed; break;
				case 'e': mf = FRAG_IsMailable; break;

				/* Unknown */
				default: fprintf(stderr, "%li: ! Ignoring unknown parameter: \"%s\"\n", uniqid, argv[i]); continue;
			}

			/* New search term */
			sterm[(mf - FRAG_Conditional) >> 2] = (m >= 6) ? 0 : (mf + m - 4);

			/* Allow "?l=foo"-like constructions to do an automatic "+l=foo" switch */
			if (mf == FRAG_IsListed)
			{
				/* Set filename from (possibly empty) parameter, if present */
				if (p = strchr(p, '=')) do_list = p + 1;

				/* Otherwise, if no previous filename set, and not disabling search term, use default */
				else if ((!do_list) && (m < 6)) do_list = defp;
			}

			/* Done */
			continue;
		}

		/* Normal switches: Only first one or two letters of command switches are matched */
		else switch (tolower(p[0]))
		{
			/* Impose default settings for target mails */
			case '1':
			{
				if (m != 0) goto badqual;
				do_input = FILE_TargMail;
				/* mode and search flags */
				break;
			}

			/* Impose default settings for password mails */
			case '2':
			{
				if (m != 0) goto badqual;
				do_input = FILE_PassMail;
				break;
			}

			/* Special "and"/"or" flag pair */
			case 'a': case 'o':
			{
				if (m <= 1)  { mf = MODE_And; m = (tolower(p[0]) == 'a') ? 1 : 2; break; }
	badqual:	fprintf(stderr, "%li: ! Ignoring badly qualified parameter: \"%s\"\n", uniqid, argv[i]);
				break;
			}

			/* Flags */
			case 'd': mf = MODE_Debug; break;
			case 'v': mf = MODE_Verbose; break;
			case 'n': mf = MODE_NoMail; break;

			/* File parameters */
			case 'm': ap = &do_input; break;
			case 'l': ap = &do_list; break;
			case 'p': ap = &do_pidlist; break;
			case 't': ap = &do_tarlist; break;

			/* Unknown */
			default: fprintf(stderr, "%li: ! Ignoring unknown parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Do flags */
		if (mf) switch (m)
		{
			case 1: mode |=  mf; break;	/* Set flag */
			case 2: mode &= ~mf; break;	/* Clear flag */
			case 3: mode ^=  mf; break;	/* Toggle flag */
			default: fprintf(stderr, "%li: ! Ignoring unqualified parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Handle files parameters */
		if (ap)
		{
			/* If any (possibly empty) filename given, activite mode */
			if (p = strchr(p, '=')) *ap = p + 1;

			/* Unparameterised - activate if appropriate, set to default filename */
			else if ((m <= 1) || ((m == 3) && (*ap == 0))) *ap = defp;

			/* Unparameterised - deactivate if appropriate */
			else *ap = 0;
		}
	}

	/* Defaults: redirect to stdout if no filename given */
	if (do_input == defp) do_input = "";

	/* Default filenames */
	if (do_list == defp) do_list = FILE_DefList;
	if (do_pidlist == defp) do_pidlist = FILE_DefList;
	if (do_tarlist == defp) do_tarlist = FILE_DefTargetList;

	/* Print out options */
	fprintf(stderr, "%li:   %sSearch terms: ", uniqid, (mode & MODE_Debug) ? "+debug " : "");

	/* Logic mode */
	p = (mode & MODE_And) ? " AND " : " OR ";

	/* Print out search terms */
	for (i = nontriv = 0; (i < MAX_SearchTerms); i++) if (sterm[i])
	{
		fprintf(stderr, "%s%s%s", (nontriv++) ? p : "", (sterm[i] & 1) ? "" : "Not", sterm_tab[i]);
	}

	/* Input player list? */
	p = (!do_list) ? "No player list. " : "";

	/* Non mailable-output? */
	if (mode & MODE_NoMail)
	{
		fprintf(stderr, "\n%li:  ! %s** Not generating %s", uniqid,
				p, (do_input) ? "mailable output text **" : "any textual output **\n");
		p = "";
	}

	/* Default file */
	else if (!do_input) do_input = FILE_DefMail;

	/* Process input */
	if (do_input)
	{
		/* Print out input mode */
		fprintf(stderr, "\n%li:   %sTaking input from \"%s\"\n", uniqid, p, (*do_input) ? do_input : "stdin");

		/* Can't read both input files from stdin */
		if ((do_list) && !(*do_list) && !(*do_input)) erf(ERF_BadParms, 20);

		/* Read message text from stdin? */
		if (!(*do_input)) fp = stdin;
	
		/* Open input file */
		else if (!(fp = fopen(do_input, "rb"))) erf(ERF_fopenFailed, 21);
	
		/* Parse input file into fragments */
		if (rc = parsefrags(fp, &frags)) erf(rc, 22);
	
		/* Close input file */
		i = (stdin != fp) && (fclose(fp));
	
		/* Be sure not to try to close it more than once */
		fp = 0; if (i) erf(ERF_fcloseFailed, 23);
	}

	/* Output list of mailed players? */
	if (do_pidlist)
	{
		/* Not allowed to use null filename to indicate stdout */
		if (!do_pidlist[0]) erf(ERF_BadParms, 30);

		/* Open file */
		if (!(fp_pidlist = fopen(do_pidlist, "wb"))) erf(ERF_fopenFailed, 31);

		/* Progress */
		fprintf(stderr, "%li:   Opened output file for list of mailed players: %s\n", uniqid, do_pidlist);
	}

	/* Output list of mailed players' current targets */
	if (do_tarlist)
	{
		/* Not allowed to use null filename to indicate stdout */
		if (!do_tarlist[0]) erf(ERF_BadParms, 32);

		/* Open file */
		if (!(fp_tarlist = fopen(do_tarlist, "wb"))) erf(ERF_fopenFailed, 33);

		/* Progress */
		fprintf(stderr, "%li:   Opened output file for list of mailed targets: %s\n", uniqid, do_tarlist);
	}

	/* List of players given? */
	if (do_list)
	{
		/* List of players on stdin? */
		if (!(*do_list)) fp = stdin;

		/* Open input file */
		else if (!(fp = fopen(do_list, "rb"))) erf(ERF_fopenFailed, 38);
	}

	/* Moof */
	/*fprintf(stderr, "goo\n");*/

	/* Attempt to lock files */
	if (!(lockfd = setflock("Genmail.c "))) erf(ERF_flockFailed, 40);

	/* Read full info for all players */
	if (rc = scanplinfo(SCANPL_AddFull|SCANPL_GetTargs, 0, 0, 0, &plist, statbuf, statbuf_end, &statbuf_pos)) erf(rc, 42);

	/* Release lock */
	if (lockfd) lockfd = setfunlock(lockfd);

	/* Parse list of players */
	if (fp)
	{
		char tmpstr[MAX_id+2];

		/* Message */
		fprintf(stderr, "%li:   Scanning player list %s%s\n", uniqid, (*do_list) ? "from file: " : "on stdin", do_list);

		/* No ids recognised yet */
		n = 0; i = 1;

		/* Read as many whitespace delimited strings as possible */
		while (1 == fscanf(fp, "%"STRING(MAX_id)"s", tmpstr))
		{
			p = (i != 1) ? "\n" : "";

			/* Stop if we encounter an excessively long string */
			if (strlen(tmpstr) >= MAX_id)
			{
				fprintf(stderr, "%s%li: ! Stopped scanning at string \"%s...\" (too long).\n", p, uniqid, tmpstr);
				i = 0; break;
			}

			/* Try to match on player list */
			for (pptr = plist; (pptr); pptr = pptr->link) if (streq(tmpstr, pptr->id)) break;

			/* Unmatched */
			if (!pptr)
			{
				fprintf(stderr, "%s%li:    xx Unknown player \"%s\"\n", p, uniqid, tmpstr);
				i = 1; continue;
			}

			/* Ignore duplicates */
			if (pptr->ext)
			{
				fprintf(stderr, "%s%li:    xx Duplicate player \"%s\"\n", p, uniqid, pptr->id);
				i = 1; continue;
			}

			/* Found one! */
			if ((i != 1) && (i != 11)) { fprintf(stderr, ", %s", pptr->id); ++i; }

			/* Format list neatly (10 per row) */
			else { fprintf(stderr, "%s%li:    -- %s", p, uniqid, pptr->id); i = 2; }

			/* Note that player was in list, and increment count of valid player ids */
			pptr->ext = 1; ++n;
		}

		/* Status report */
		fprintf(stderr, "%s%li:   Found %i distinct valid player ids in list.\n", (i > 1) ? ".\n" : "", uniqid, n);

		/* Close input file */
		i = (stdin != fp) && (fclose(fp));

		/* Be sure not to try to close it more than once */
		fp = 0; if (i) erf(ERF_fcloseFailed, 27);
	}

	/* Logic mode */
	j = (mode & MODE_And) ? 0 : 1;

	/* Run through players */
	for (n = 0, pptr = plist; (pptr); pptr = pptr->link)
	{
		target_p tarp;
		player_p *parrp, *qarrp;

		/* Only mail "Mailable" players */
		if (!(mode & MODE_NoMail) && !(pptr->flags & PF_Mailable)) continue;

		/* Min number of thingies */
		i = 4;

		/* Limits */
		plim[FLOB_PL] = 1;
		for (plim[FLOB_ASS] = 1, tarp = pptr->asses; (tarp); tarp = tarp->next, i++) ++plim[FLOB_ASS];
		for (plim[FLOB_TAR] = 1, tarp = pptr->targs; (tarp); tarp = tarp->next, i++) ++plim[FLOB_TAR];
		for (plim[FLOB_OLD] = 1, tarp = pptr->oldts; (tarp); tarp = tarp->next, i++) ++plim[FLOB_OLD];

/*fprintf(stderr, "%li:   flobs: %i %i %i %i\n", uniqid, plim[FLOB_PL], plim[FLOB_ASS], plim[FLOB_TAR], plim[FLOB_OLD]);*/

		/* Allocate all thingies */
		if (!(parrp = malloc(i * sizeof(player_p)))) erf(ERF_mallocFailed, 70);

		/* Player himself */
		plarr[FLOB_PL]  = parrp;
		plarr[FLOB_PL][0] = pptr;

		/* Individual arrays */
		plarr[FLOB_ASS] = qarrp = 1 + parrp;
		plarr[FLOB_TAR] = plarr[FLOB_ASS] + plim[FLOB_ASS];
		plarr[FLOB_OLD] = plarr[FLOB_TAR] + plim[FLOB_TAR];

		/* Fill in arrays: assassins */
		for (tarp = pptr->asses, *(qarrp++) = 0; (tarp); tarp = tarp->next, ++qarrp)
		{
			/* Null entry? */
			if (!tarp->id) *qarrp = 0;

			/* Search for amongst players */
			else for (*qarrp = plist; (*qarrp); *qarrp = (*qarrp)->link) if (streq(tarp->id, (*qarrp)->id)) break;
		}

		/* Fill in arrays: targets */
		for (tarp = pptr->targs, *(qarrp++) = 0; (tarp); tarp = tarp->next, ++qarrp)
		{
			/* Null entry? */
			if (!tarp->id) *qarrp = 0;

			/* Search for amongst players */
			else for (*qarrp = plist; (*qarrp); *qarrp = (*qarrp)->link) if (streq(tarp->id, (*qarrp)->id)) break;
		}

		/* Fill in arrays: old ones */
		for (tarp = pptr->oldts, *(qarrp++) = 0; (tarp); tarp = tarp->next, ++qarrp)
		{
			/* Null entry? */
			if (!tarp->id) *qarrp = 0;

			/* Search for amongst players */
			else for (*qarrp = plist; (*qarrp); *qarrp = (*qarrp)->link) if (streq(tarp->id, (*qarrp)->id)) break;
		}

		/* Ignore empty search patterns */
		if (nontriv)
		{
			/* Check against search terms */
			for (i = 0; (i < MAX_SearchTerms); i++) if (sterm[i])
			{
				/* Terminate as soon as condition satisfied or falsified */			
				if (j ^ (sterm[i] & 1) ^ evalcond(pptr, plarr, plim, sterm[i] & 0xFC, 0, 0)) break;
			}
	
			/* Is early termination an indicator of success or failure? */
			if (j ^ (i < MAX_SearchTerms)) continue;
		}

		/* Report on progress */
		if (mode & MODE_Verbose)
		{
			if (!(n % 10)) fprintf(stderr, "%s%li:    >> %s", (n) ? "\n" : "", uniqid, pptr->id);
			else fprintf(stderr, ", %s", pptr->id);
		}

		/* Print player id to output list */
		if (fp_pidlist) fprintf(fp_pidlist, "%s\n", pptr->id);

		/* Print player's target information to output list */
		if (fp_tarlist)
		{
			fprintf(fp_tarlist, "%s\t", pptr->id);

			for (tarp = pptr->targs; (tarp); tarp = tarp->next)
			{
				if (tarp->id) fprintf(fp_tarlist, "\t%s", tarp->id);
			}

			fprintf(fp_tarlist, "\n");
		}

		/* Count messages sent */
		++n;

		/* SMTP Mail Commands */
		if (!(mode & MODE_NoMail)) printf("MAIL FROM:" MAILFROM CR LF "RCPT TO:%s%s" CR LF "DATA" CR LF,
										  pptr->email, ((strchr(pptr->email, '@')) ? "" : MAILDOMAIN));

		/* Message headers and body */
		if ((frags) && (rc = writefrags(stdout, pptr, plarr, plim, plist, frags, thedate, MAILFROM_head))) erf(rc, 50);

		/* SMTP Mail Footer */
		if (!(mode & MODE_NoMail)) printf(CR LF "." CR LF);

		/* Free memory */
		free(parrp);
	}

	/* Finish SMTP Session */
	if (!(mode & MODE_NoMail)) fputs("QUIT" CR LF, stdout);

	/* Tidy up after progress text */
	if ((mode & MODE_Verbose) && (n> 0)) fputs(".\n", stderr);


_erf: /***** Error handling *****/

	/* Close temporary open files */
	if ((fp) && (fp != stdout) && (fp != stdin)) fclose(fp);

	/* Close output files */
	if (fp_tarlist) fclose(fp_tarlist);
	if (fp_pidlist) fclose(fp_pidlist);

	/* Debug */
	if (rc) fprintf(stderr, "%li !! main(%i):%i mode=%x\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), mode);

	/* Note success */
	else fprintf(stderr, "%li:   -------> generated %i messages <-------\n", uniqid, n);

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/**** The End! ****/
