/*****************************************************
**  score.c                         >> bin/scoring  **
**  ----------------------------------------------  **
**  TODO: Decide on "reset" options.                **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define NO_CGI
#define MAX_fpbuf			(64*1024)
#define MAX_statbuf			(256*1024)
#define EVENT_EXT_TYPE		int			/* type of  event_p->ext is irrelevent */
#define PLAYER_EXT_TYPE		int			/* type of player_p->ext is irrelevent */
#define PSTATE_EXT_TYPE		int			/* type of pstate_p->ext is irrelevent */
#define PSTATE_PLP_TYPE		player_p	/* type of pstate_p->plp is player_p   */
#define PEXIST_EXT_TYPE		int			/* type of pexist_p->ext is irrelevent */

#include "!lib-general.c"
#include "!custom.c"
#include "!lib-uinfo-edit.c"
#ifdef TEST
 #include "!lib-evolveT.c"
#else
 #include "!lib-evolve.c"
#endif

/***************************************************************************************************************************
** Definitions and declarations                                                                                           **
***************************************************************************************************************************/

/* Mode flags */
#define MODE_UpdateOldTargs	((1UL)<<0)
#define MODE_EvolvePlayers	((1UL)<<1)
#define MODE_All			((1UL)<<2)
#define MODE_InstallCronTab	((1UL)<<3)
#define MODE_StrictTargets	((1UL)<<4)
#define MODE_Reset			((1UL)<<5)

/* Default mode flags */
unsigned long mode = 0;

/* Global buffer for file content */
char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos = statbuf;

/* Special pointer used to denote default filenames. String irrelevent, but should be non-empty. */
static char *defp = "(default)";

/* Event categories */
#define S_Err	(0)
#define S_Att	(1)	/* Attempts */
#define S_Kil	(2)	/* Kills */
#define S_Dth	(3)	/* Deaths */
#define S_Sur	(4)	/* Survivals */

/* Scoring table categories */
int pscdef[SCDEF_MAX_Single][SCDEF_MAX] =
{
	/*                 Against:  ---------Assassins--------- --------- Police ---------- --NonPlayer-- ------ ------ */
	/*                           --Unwanted--- ---Wanted---- --Unwanted--- ---Wanted---- ------------- Umpire - ?? - */
	/* -(Killer)--(NonFatal)--   -Comp- -Inco- -Comp- -Inco- -Comp- -Inco- -Comp- -Inco- -Legl- -Illg- ------ ------ */
	/* Assass Unwanted Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Assass Wanted   Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Assass Wanted   Inco */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Police Unwanted Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Police Wanted   Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Police Wanted   Inco */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/*   Legal Non-Player   */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Illegal Non-Player   */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Umpire               */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Att, S_Att, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* --------------(Fatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Assass Wanted   Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Assass Wanted   Inco */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Police Unwanted Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Police Wanted   Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Police Wanted   Inco */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/*   Legal Non-Player   */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Illegal Non-Player   */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Umpire               */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Kil, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* -(Victim)--(NonFatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Assass Wanted   Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Assass Wanted   Inco */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Unwanted Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Police Wanted   Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Wanted   Inco */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/*   Legal Non-Player   */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Illegal Non-Player   */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Umpire               */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* -(Victim)-----(Fatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Assass Wanted   Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Assass Wanted   Inco */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Unwanted Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Police Wanted   Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Wanted   Inco */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/*   Legal Non-Player   */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Illegal Non-Player   */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Umpire               */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err }
	/* -----------------------  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
};

/***************************************************************************************************************************
** Code to parse input pid and old-target lists                                                                           **
***************************************************************************************************************************/
target_p addnulltarg(target_p *ptarp)
{
	target_p tarp;

	/* Allocate */
	if (!(tarp = malloc(sizeof(struct target)))) return 0;

	/* Empty id to force clearing of fields */
	tarp->id = 0;
	tarp->ext = 0;

	/* Link */
	tarp->next = *ptarp;
	*ptarp = tarp;

	return tarp;
}

editpl_p addtolist(char* p, size_t s, editpl_p *list)
{
	editpl_p edp;
	int i;


	/* Allocate new struct for this player and space for id string */
	if (!(edp = malloc(s + sizeof(struct editpl)))) return 0;

	/* Add to list */
	edp->next = (list) ? (*list) : 0;
	edp->link = 0;
	if (list) *list = edp;

	/* Save player id */
	edp->opid = (s == 0) ? p : strcpy((char *)(edp + 1), p);

	/* Fill in the blanks */
	edp->npass = 0;
	edp->ntimp = 0;
	edp->nname = 0;
	edp->naddr = 0;
	edp->ncoll = 0;
	edp->nrank = 0;
	edp->nmafia = 0;
	edp->nemail = 0;
	edp->npseud = edp->nwant = 0;
	edp->nkills = edp->ndeaths = 0;
	edp->nscore = edp->ndanger = 0;
	edp->ntargs = edp->nasses = edp->noldts = 0;
	edp->nfl_mask = edp->nfl_bits = 0UL;

	/* Return struct pointer */
	return edp;
}

int parsepidlist(char* name, FILE **fpp, editpl_p *list)
{
	char tmpstr[MAX_id+2];
	int i = 1, n = 0;
	editpl_p edp;


	/* Progress */
	fprintf(stderr, "%li:   Reading player id list from %s%s\n", uniqid, (*name) ? "file: " : "stdin", name);

	/* From stdin? */
	if (!(*name)) *fpp = stdin;

	/* From file? */
	else if (!(*fpp = fopen(name, "rb"))) return ERF_fopenFailed;

	/* Read as many whitespace delimited strings as possible */
	while (1 == fscanf(*fpp, "%"STRING(MAX_id)"s", tmpstr))
	{
		char *p = (i != 1) ? "\n" : "";

		/* Stop if we encounter an excessively long string */
		if (strlen(tmpstr) >= MAX_id)
		{
			fprintf(stderr, "%s%li: ! Stopped scanning at string \"%s...\" (too long).\n", p, uniqid, tmpstr);
			i = 0; break;
		}

		/* Paranoia */
		if (!tmpstr[0]) return ERF_Paranoia;

		/* Try to match on list */
		for (edp = *list; (edp); edp = edp->next) if (streq(tmpstr, edp->opid)) break;

		/* Ignore duplicates */
		if (edp)
		{
			fprintf(stderr, "%s%li:    xx Duplicate player \"%s\"\n", p, uniqid, edp->opid);
			i = 1; continue;
		}

		/* Allocate new entry in list */
		if (!(edp = addtolist(tmpstr, 4*(MAX_id+1), list))) return ERF_mallocFailed;

		/* Found one! */
		if ((i != 1) && (i != 11)) { fprintf(stderr, ", %s", edp->opid); ++i; }

		/* Format list neatly (10 per row) */
		else { fprintf(stderr, "%s%li:    -- %s", p, uniqid, edp->opid); i = 2; }

		/* Increment count of valid player ids */
		++n;
	}

	/* Status report */
	fprintf(stderr, "%s%li:   Found %i distinct player ids in list.\n", (i > 1) ? ".\n" : "", uniqid, n);

	/* Close input file */
	i = (stdin != *fpp) && (fclose(*fpp));

	/* Be sure not to try to close it more than once */
	*fpp = 0;

	/* Residual fclose error */
	return (i) ? ERF_fcloseFailed : 0;
}

int parsetarline(FILE *fp, target_p *ptarp, char** id, int* tidy)
{
	char *q, *p, *p_end = fpbuf + MAX_fpbuf - 1, *str;
	int c, i, rc = 0;
	target_p tarp;

	/* Nice formatting */
	q = (*tidy != 1) ? "\n" : ""; *tidy = 0;

	/* Nothing found yet */
	*ptarp = 0;
	*id = 0;

	/* Skip leading whitespace */
	do c = getc(fp); while ((c == '\t') || (c == ' ') || (c == 10) || (c == 13));

	/* Already at EOF */
	if (c == EOF) goto testfp;

	/* Try to read strings */
	for (i = 0, str = fpbuf; ((c != EOF) && (c != 10) && (c != 13)); i++)
	{
		/* Start of string buffer */
		p = str;

		/* Read characters into string */
		while ((c != EOF) && (c != '\t') && (c != ' ') && (c != 13) && (c != 10))
		{
			if (p < p_end) *(p++) = c;
			else erf(0, 2);
			c = getc(fp);
		}

		/* Terminate string */
		*(p++) = '\0';

		/* Allow up to (MAX_id - 1) chars */
		if (p - str > MAX_id) erf(0, 3);
		
		/* Skip trailing whitespace */
		while ((c == '\t') || (c == ' ')) c = getc(fp);

		/* Target ids? */
		if (i > 0)
		{
			/* Allocate struct */
			if (!(tarp = malloc(sizeof(struct target) + p - str))) erf(ERF_mallocFailed, 1);

			/* Copy id */
			tarp->id = strcpy((char*)(tarp) + sizeof(struct target), str);

			/* Clear extension word */
			tarp->ext = 0;

			/* Link */
			tarp->next = *ptarp;
			*ptarp = tarp;
		}

		/* Player id */
		else
		{
			*id = str;
			str = p;
		}
	}

	/* No main player id string found */
	if (i <= 0) erf(0, 4);


	/* No target id strings found */
	else if (i-- < 1)
	{
		/* Warning */
		fprintf(stderr, "%s%li:    tarlist: no target ids for %i \"%s\": clearing list.\n", uniqid, q, i, *id);
#if 0
		/* Allocate empty struct */
		if (!tarp = addnulltarg(ptarp)) erf(ERF_mallocFailed, 2);
#endif
		/* Neatness */
		q = ""; *tidy = 0;
	}

	/* Don't allow trailing content */
	if ((c != EOF) && (c != 10) && (c != 13)) erf(0, 6);

	/* Successful read */
	if (c != EOF) return 0;

testfp: /* Test whether really are at EOF */

	/* Check for errors */
	if (ferror(fp)) erf(ERF_FileError, 10);

	/* Assume EOF */
	return ERF_Void;


_erf: /* Print out useful error messages */

	/* Terminate buffer */
	*((p < p_end) ? p : p_end) = '\0';

	/* Neatness */
	*tidy = 0;

	/* Choose error message */
	if (ERF_TYPE(rc) == 0)
	{
		switch (ERF_CONTEXT(rc))
		{
			case 0: return 0;
			case 1: p = "%s%li:    tarlist error: line too long on id %i: \"%s\": \"%s\"\n"; break;
			case 2: p = "%s%li:    tarlist error: line too long on id %i: \"%s\": \"%s\"\n"; break;
			case 3: p = "%s%li:    tarlist error: id %i too long: \"%s\": \"%s\"\n"; break;
			case 4: p = "%s%li:    tarlist error: missing player id (%i): \"%s\": \"%s\"\n"; break;
			case 5: p = "%s%li:    tarlist error: missing target ids (only %i): \"%s\": \"%s\"\n"; break;
			case 6: p = "%s%li:    tarlist error: trailing text after target id %i: \"%s\": \"%s\"\n"; break;
			default: p = "%s%li:    tarlist error: (%i) \"%s\": \"%s\"\n";
		}
		fprintf(stderr, p, q, uniqid, i, NSTR(*id), NSTR(str));
		return ERF_BadInput;
	}

	/* Report */
	fprintf(stderr, "%s%li !! parsetarline(%i):%i\n", q, uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc)); 
}

int parsetarlist(char* name, FILE **fpp, editpl_p *list)
{
	char* id;
	int j, i = 1, n = 0;
	editpl_p edp, tmplist = 0;
	target_p targs, tarp;


	/* Progress */
	fprintf(stderr, "%li:   Reading target id list from %s%s\n", uniqid, (*name) ? "file: " : "stdin", name);

	/* From stdin? */
	if (!(*name)) *fpp = stdin;

	/* From file? */
	else if (!(*fpp = fopen(name, "rb"))) return ERF_fopenFailed;

	/* Read as many whitespace delimited strings as possible */
	while (!(j = parsetarline(*fpp, &targs, &id, &i)))
	{
		char* p = (i != 1) ? "\n" : "";

		/* Try to match on list */
		for (edp = *list; (edp); edp = edp->next) if (streq(id, edp->opid)) break;

		/* Not on list*/
		if (!edp)
		{
			/* Allocate new entry in list */
			if (!(edp = addtolist(id, 1 + strlen(id), list))) return ERF_mallocFailed;

			/* Found one! */
			if ((i != 1) && (i != 11)) { fprintf(stderr, ", %s", edp->opid); ++i; }

			/* Format list neatly (10 per row) */
			else { fprintf(stderr, "%s%li:    -- %s", p, uniqid, edp->opid); i = 2; }

			/* Increment count of valid player ids */
			++n;
		}

		/* Already on list */
		else
		{
			/* Notification of duplicates */
			i = 1; fprintf(stderr, "%s%li:    xx Duplicate player \"%s\"\n", p, uniqid, edp->opid);
		}

		/* Add target ids */
		if (!edp->noldts) edp->noldts = targs;

		/* Add target ids to end of existing ones */
		else for (tarp = edp->noldts; (tarp); tarp = tarp->next) if (!tarp->next) { tarp->next = targs; break; }

		/* Link into tmp list */
		edp->link = tmplist;
		tmplist = edp;
	}

	/* For each on targetting list */
	for (edp = tmplist; (edp); edp = edp->link)
	{
		/* Empty list --> null targetting */
		if ((!edp->noldts) && (!addnulltarg(&edp->noldts))) return ERF_mallocFailed;
	}

	/* Status report */
	fprintf(stderr, "%s%li:   Found %i new distinct player ids in list.\n", (i > 1) ? ".\n" : "", uniqid, n);

	/* Errors? */
	if ((j) && (j != ERF_Void)) return j;
	
	/* Close input file */
	i = (stdin != *fpp) && (fclose(*fpp));

	/* Be sure not to try to close it more than once */
	*fpp = 0;

	/* Residual fclose error */
	return (i) ? ERF_fcloseFailed : 0;
}

/***************************************************************************************************************************
** Main code. Options taken from command line.                                                                            **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	int i, rc = 0, lockfd = 0;
	time_t thetime;
	timmy thetim, endtim, *etimp = 0;
	struct tm* the_tm;
	char *timestr, *p, *do_tarlist = 0, *q;
	char *do_scores = 0, *do_bodge = 0, *do_output = 0, *do_crontab = 0, *do_pseuds = 0, *do_pidlist = 0;
	FILE *fp_scores = 0, *fp_bodge = 0, *fp_output = 0, *fp_crontab = 0, *fp_pseuds = 0, *fp = 0;
	editpl_p edp, otheredp = 0, edlist = 0;
	pstate_p pslist = 0, pstp;
	unsigned long fl_a = 0, fl_b = 0;


	/* File creation umask */
	umask(0117);

	/* What time is it? */
	time(&thetime);
	the_tm = localtime(&thetime);
	thetim.day  = the_tm->tm_yday - StartingYDay;
	thetim.hour = the_tm->tm_hour;
	thetim.min  = the_tm->tm_min;
	timestr = ctime(&thetime);

	/*fprintf(stderr, "time %i %i %i %i\n", the_tm->tm_yday, the_tm->tm_mday, the_tm->tm_mon, the_tm->tm_wday);*/

	/* Get process id for logging */
	uniqid = (long) getpid();

	/* Check command line args */
	for (i = 1; (i < argc) && (p = argv[i]); i++)
	{
		char **ap = 0;
		unsigned long m, mf = 0;

		/* Qualifier? */
		switch (*(p++))
		{
			case '+': m = 1; break;
			case '-': m = 2; break;
			case '^': m = 3; break;
			default : m = 0; --p;
		}

		/* Command switches: only first letter significant */
		switch (tolower(*(q = p)))
		{
			/* Set time limit for evolution, and enable/disable player update evolution flag if qualified */
			case 'e':
			{
				if (m != 0) mf = MODE_EvolvePlayers;
				if (!(p = strchr(p, '='))) etimp = &thetim;
				else if (!p[1]) etimp = 0;
				else if (str2tim(p+1, &endtim)) etimp = &endtim;
				else fprintf(stderr, "%li: ! Bad time parameter: \"%s\"\n", uniqid, argv[i]);
				break;
			}

			/* Flags */
			case 'u': mf = MODE_UpdateOldTargs; break;
			case 'a': mf = MODE_All; break;
			case 'x': mf = MODE_StrictTargets; break;
			case 'r': mf = MODE_Reset; break;

			/* File parameters */
			case 's': ap = &do_scores; break;
			case 'o': ap = &do_output; break;
			case 'p': ap = &do_pidlist; break;
			case 't': ap = &do_tarlist; break;
			case 'y': ap = &do_pseuds; break;
			case 'b': ap = &do_bodge; break;
			case 'c': ap = &do_crontab; break;

			/* Flag change parameters */
			case 'f':
			{
				unsigned long a, b;
				if (m == 2) fl_a = fl_b = 0;
				else if (m > 1) fprintf(stderr, "%li: ! Ignoring badly qualified switch: \"%s\"\n", uniqid, argv[i]);
				else if ((p = strchr(p, '=')) && (2 == sscanf(p + 1, "&%8lx/&%8lx", &a ,&b)))
				{
					fl_b = (fl_b & (~a)) ^ b;
					fl_a |= a;
				}
				else fprintf(stderr, "%li: ! Ignoring invalid flag bitmasks: \"%s\"\n", uniqid, argv[i]); 
				break;
			}

			/* Unknown */
			default: fprintf(stderr, "%li: ! Ignoring unknown parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Do flags */
		if (mf) switch (m)
		{
			case 1: mode |=  mf; break;	/* Set flag */
			case 2: mode &= ~mf; break;	/* Clear flag */
			case 3: mode ^=  mf; break;	/* Toggle flag */
			default: fprintf(stderr, "%li: ! Ignoring unqualified parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Handle files parameters */
		if (ap)
		{
			/* If any (possibly empty) filename given, activite mode */
			if (p = strchr(p, '=')) *ap = p + 1;

			/* Unparameterised - activate if appropriate, set to default filename */
			else if ((m <= 1) || ((m == 3) && (*ap == 0))) *ap = defp;

			/* Unparameterised - deactivate if appropriate */
			else *ap = 0;
		}

		/* Special crontab file parameter */
		if (ap == &do_crontab)
		{
			if ((*ap == defp) && (tolower(q[1]) == 'i')) mode |= MODE_InstallCronTab;
			else mode &= ~MODE_InstallCronTab;
		}
	}

	/* Defaults: redirect to stdout if no filename given */
	if (do_scores == defp) do_scores = "";
	if (do_output == defp) do_output = "";
	if (do_bodge == defp) do_bodge = "";

	/* Default filenames */
	if (do_pseuds  == defp) do_pseuds  = FILE_NewPseuds;
	if (do_crontab == defp) do_crontab = FILE_CronTab;
	if (do_pidlist == defp) do_pidlist = FILE_DefList;
	if (do_tarlist == defp) do_tarlist = FILE_DefTargetList;

	/* If neither list was passed, assume all players to be fiddled */
	if ((!do_pidlist) && (!do_tarlist)) mode |= MODE_All;

	/* Logfile header */
	fprintf(stderr, "%li ** Scoring: %s%s%s%s%sat %s", uniqid,
			(mode & MODE_Reset) ? "Reset " : "",
			(mode & MODE_EvolvePlayers) ? "EvolvePlayers " : "",
			(mode & MODE_StrictTargets) ? "StrictTargets " : "",
			(mode & MODE_UpdateOldTargs) ? "UpdateOldTargs " : "",
			(mode & MODE_All) ? "(all) " : "", timestr);

	/* Describe evolution information */
	p = (mode & MODE_EvolvePlayers) ? "Evolve game to" : "time limit is";

	/* No time limit */
	if (!etimp) fprintf(stderr, "%li:   Args: %s end of time.\n", uniqid, p);

	/* Time limit set */
	else fprintf(stderr, "%li:   Args: %s %d:%.2d:%.2d%s.\n", uniqid, p,
					 etimp->day, etimp->hour, etimp->min, (etimp == &thetim) ? " (now)" : "");

	/* Some bits cannot be edited */
	if ((fl_a|fl_b) & PF_MASK_NoEdit)
	{
		/* Complain */
		fprintf(stderr, "%li: ! Ignoring protected flag bits &%.8lx/&%.8lx.\n",
				uniqid, fl_a & PF_MASK_NoEdit, fl_b & PF_MASK_NoEdit);

		/* Mask them out now, to save further bother */
		fl_a &= (~PF_MASK_NoEdit);
		fl_b &= (~PF_MASK_NoEdit);
	}

	/* Adjust flags? */
	if (fl_a|fl_b) fprintf(stderr, "%li:   Args: Alter flag bits: AND NOT &%.8lx EOR &%.8lx.\n", uniqid, fl_a, fl_b);

	/* Warn the user if scoring flags have been specified */
	if ((fl_a|fl_b) & PF_MASK_Scoring) fprintf(stderr, "%li: ! Caution: Flag bits &%.8lx/&%.8lx are not static!\n",
											   uniqid, fl_a & PF_MASK_Scoring, fl_b & PF_MASK_Scoring);

	/* Use stdout how many times? */
	i = ((do_scores) && !(*do_scores)) + ((do_bodge) && !(*do_bodge)) + ((do_output) && !(*do_output));

	/* Can't send bodge and breakdown and output to stdout at the same time */
	if (i > 1) erf(ERF_BadParms, 7);

	/* Can't read pidlist and tarlist from stdin at the same time */
	if ((do_pidlist) && (do_tarlist) && !(*do_pidlist) && !(do_tarlist)) erf(ERF_BadParms, 8);

	/* Open crontab file */
	if (do_crontab)
	{
		/* Progress */
		fprintf(stderr, "%li:   Args: Send crontab output to: %s\n", uniqid, do_crontab);

		/* Crontab not allowed on stdout */
		if (!do_crontab[0]) erf(ERF_BadParms, 12);

		/* Open Crontab file */
		if (!(fp_crontab = fopen(do_crontab, "wb"))) erf(ERF_fopenFailed, 13);

		/* Header */
		fprintf(fp_crontab, "###\n### Crontab for "GAME"; Generated at %s###\n", timestr);
	}

	/* Read list of players to which updates should be applied */
	if ((do_pidlist) && (i = parsepidlist(do_pidlist, &fp, &edlist))) erf(i, 17);

	/* Read list of players for which to set specified old targets */
	if ((do_tarlist) && (i = parsetarlist(do_tarlist, &fp, &edlist))) erf(i, 18);

	/* Produce scoring breakdown? */
	if (do_scores)
	{
		/* Progress */
		fprintf(stderr, "%li:   Args: Send scoring breakdown to: %s\n", uniqid, (*do_scores) ? do_scores : "stdout");

		/* To stdout? */
		if (!(*do_scores)) fp_scores = stdout;

		/* To file */
		else if (!(fp_scores = fopen(do_scores, "wb"))) erf(ERF_fopenFailed, 19);

		/* Header */
		fprintf(fp_scores, "Scoring breakdown for "GAME"; Generated at %s\n", timestr);
	}

	/* Produce bodges */
	if (do_bodge)
	{
		/* Progress */
		fprintf(stderr, "%li:   Args: Send bodgedata to: %s\n", uniqid, (*do_bodge) ? do_bodge : "stdout");

		/* To stdout? */
		if (!(*do_bodge)) fp_bodge = stdout;

		/* To file */
		else if (!(fp_bodge = fopen(do_bodge, "wb"))) erf(ERF_fopenFailed, 20);

		/* Header */
		fprintf(fp_bodge, "##\n## Bodgedata for "GAME"; Generated at %s##\n", timestr);
	}

	/* Direct updated uinfo to some specific file? */
	if (do_output)
	{
		/* Progress */
		fprintf(stderr, "%li:   Args: Send output uinfo data to: %s\n", uniqid, (*do_output) ? do_output : "stdout");

		/* To stdout? */
		if (!(*do_output)) fp_output = stdout;

		/* To file */
		else if (!(fp_output = fopen(do_output, "wb"))) erf(ERF_fopenFailed, 21);

		fprintf(fp_output, "###\n### Scoring for \""GAME"\" at %s### Mode: &%lx\n"
				"### Player flags AND NOT &%.8lx EOR &%.8lx\n", timestr, mode, fl_a, fl_b);

		if (do_pidlist) fprintf(fp_output, "### Players: %s\n", (*do_pidlist) ? (do_pidlist) : "< stdin");
		if (do_tarlist) fprintf(fp_output, "### OldTargs: %s\n", (*do_tarlist) ? (do_tarlist) : "< stdin");

		if (!(mode & MODE_EvolvePlayers)) fprintf(fp_output, "###\n");
		else if (!etimp) fprintf(fp_output, "### Evolved to end of time.\n###\n");
		else fprintf(fp_output, "### Evolved to %d:%.2d:%.2d%s.\n###\n",
					 etimp->day, etimp->hour, etimp->min, (etimp == &thetim) ? " (now)" : "");
	}

	/* Attempt to lock files */
	if (!(lockfd = setflock("Score.c "))) erf(ERF_flockFailed, 24);

	/* Evolve game state? */
	if ((fp_scores) || (mode & MODE_EvolvePlayers))
	{
		unsigned long ffl = (mode & MODE_StrictTargets) ? EVOLVE_StrictTargets : 0;

		/* Evolve status to current time */
		if (rc = evolve(EVOLVE_TimersToLimit|EVOLVE_MatchAll|EVLadd_Comments|EVOLVE_InheritPseud|
						EVLadd_ScorePlayer|EVOLVE_Targetting|ffl, 0,0, &pslist, fp_scores, pscdef, 
						0,0,0,0, etimp, 0, statbuf_pos, statbuf_end, &statbuf_pos, 0,0,0,0,0,0)) erf(rc, 25);

		/* Debug */
		/*fprintf(stderr, "%li:   evolved okay.\n", uniqid);*/

		/* Update editing list using evolution results */
		if (mode & MODE_EvolvePlayers)
		{
			/* Run through every evolved player */
			for (pstp = pslist; (pstp); pstp = pstp->next)
			{
				/* Try to match on editing list */
				for (edp = edlist; (edp); edp = edp->next) if (streq(edp->opid, pstp->pid)) break;

				/* Ignore, if not found */
				if ((!edp) && !(mode & MODE_All)) continue;

				/* Or try to add to list */
				else if ((!edp) && !(edp = addtolist(pstp->pid, 0, &edlist))) erf(ERF_mallocFailed, 26);

				/* New scoring info */
				edp->nscore  = &pstp->exist->score;
				edp->nkills  = &pstp->scoretab[S_Kil];
				edp->ndanger = &pstp->danger;
				edp->ndeaths = &pstp->scoretab[S_Dth];
				edp->nrank   = &pstp->exist->rank;

				/* New status info */
				edp->ntimp    = &pstp->exist->timer;
				edp->nfl_mask = PF_MASK_Scoring;
				edp->nfl_bits = pstp->exist->fl & edp->nfl_mask;

				/* New targets and assassins */
				if (!(edp->ntargs = pstp->targs) && (!addnulltarg(&edp->ntargs))) erf(ERF_mallocFailed, 27);
				if (!(edp->nasses = pstp->asses) && (!addnulltarg(&edp->nasses))) erf(ERF_mallocFailed, 28);

				/* New pseudonym/wantedness */
				edp->npseud = (pstp->exist->dps) ? (pstp->exist->dps->text) : "(no pseudonym)";
				edp->nwant = ThisOrThat(pstp->exist->comment, NO_COMMENT);
			}
		}

		/* Footer for scoring breakdown file */
		if (fp_scores)
		{
			if (!etimp) fprintf(fp_scores, "\n*** Entire scoring breakdown (all events) completed ***\n");
			else fprintf(fp_scores, "\n*** Scoring breakdown finished at %d:%.2d:%.2d%s ***\n",
						 etimp->day, etimp->hour, etimp->min, (etimp == &thetim) ? " (now)" : "");
		}

		/* Data to bodgedata file */
		if (fp_bodge)
		{
			pstate_p pstp;
			unsigned long tv1, tv2;
			timmy tmptim;

			/* Run through every evolved player */
			for (i = 0, pstp = pslist; (pstp); pstp = pstp->next)
			{
				if (pstp->exist->fl & (PF_Dead|PF_Police|PF_Special)) continue;

				tv1 = TVAL(pstp->exist->tlact) + (7UL << 16);
				tv2 = TVAL(pstp->exist->timer);

				/*fprintf(stderr, "(%lu,%lu)", tv1, tv2);*/

				if ((pstp->exist->fl & PF_InactTimer) && (tv2 > tv1)) tv1 = tv2;

				TVTM(tmptim, tv1);

				fprintf(fp_bodge, "*h%i\t%.2i:%.2i:%.2i\t%lX\t%s\t%.2i:%.2i:%.2i\n",
						++i, thetim.day, thetim.hour, thetim.min, EF_SetInactTimer, 
						pstp->pid, tmptim.day, tmptim.hour, tmptim.min);
			}
		}
	}

	/* Update the uinfo file */
	if ((mode & (MODE_EvolvePlayers|MODE_UpdateOldTargs)) || (do_tarlist) /*|| (fp_output)*/)
	{
		/* Stuff for resetting scoring */
		long reset_score, reset_danger, reset_rank;
		int reset_kills, reset_deaths;

		/* Default editing options */
		eduf_flags fl = EDUF_FullFix|EDUF_PartialFix;

		/* Ensure that all players are considered for update */
		if ((mode & MODE_All) && !(otheredp = addtolist("(others)", 0, 0))) erf(ERF_mallocFailed, 30);

		/* Reset some fields if not encountered in evolved */
		if ((mode & MODE_Reset) && (otheredp))
		{
			/* ... Reset scores, etc ... */
			*(otheredp->nscore = &reset_score) = INITIAL_SCORE;
			*(otheredp->ndanger = &reset_danger) = INITIAL_DANGER;
			*(otheredp->nrank = &reset_rank) = INITIAL_RANK;
			*(otheredp->nkills = &reset_kills) = 0;
			*(otheredp->ndeaths = &reset_deaths) = 0;
			
			/* Clear target and assassin fields */
			if (!addnulltarg(&otheredp->ntargs)) erf(ERF_mallocFailed, 31);
			if (!addnulltarg(&otheredp->nasses)) erf(ERF_mallocFailed, 32);
		}

		/* Flag editing enabled? */
		if (fl_a|fl_b)
		{
			/* Edit flags (after evolution) */
			for (edp = edlist; (edp); edp = edp->next)
			{
				edp->nfl_mask |= fl_a;
				edp->nfl_bits = (edp->nfl_bits & (~fl_a)) ^ fl_b;
			}

			/* Apply changes for other players */
			if (otheredp)
			{
				otheredp->nfl_mask |= fl_a;
				otheredp->nfl_bits = (otheredp->nfl_bits & (~fl_a)) ^ fl_b;
			}
		}

		/* Arrange for old targets to be updated from current set */
		if (mode & MODE_UpdateOldTargs) fl |= EDUF_UpdateTargets;

       	/* Prepare title comment string */
		if ((!fp_output) && ((statbuf_end - statbuf_pos) > (strlen(timestr) + 64)))
		{
			p = statbuf_pos;
			if (!etimp) sprintf(p, "From scoring past all events at %s", timestr);
			else sprintf(p, "From scoring to %.8d:%.2d:%.2d at %s", etimp->day, etimp->hour, etimp->min, timestr);
			statbuf_pos += 1 + strlen(p);
		}
		else p = 0;

		/* Edit the players */
		if (i = edituinfo(fl, &edlist, otheredp, statbuf_pos, statbuf_end, fp_output, p)) erf(i, 31);

		/* Progress */
		fprintf(stderr, "%li:   Wrote output uinfo data.\n", uniqid);

		/* Footer for scoring breakdown file */
		if (fp_output) fprintf(fp_output, "\n###\n### Updated uinfo data complete.\n###\n");
	}

	/* Not actually writing the changes? */
	else
	{
		/* Progress */
		fprintf(stderr, "%li: ! No output uinfo data writen.\n", uniqid);

		/* Footer for scoring breakdown file */
		if (fp_output) fprintf(fp_output, "\n###\n### No updated uinfo data written.\n###\n");
	}

	/* Finish bodgedata */
	if (fp_bodge) fprintf(fp_bodge, "\n\n###\n### Finished bodgedata.\n###\n");

	/* Write cron tab */
	if (fp_crontab)
	{
		unsigned long tmptv, mintv, reftv;
		struct tm newtm;

		/* In one hour's time at most */
		timmy reftim = addhours(&thetim, 1);

		/* Timeval- */
		reftv = TVAL(reftim);
		mintv = TVAL(thetim);

		/* Run through every evolved player */
		for (pstp = pslist; (pstp); pstp = pstp->next)
		{
			/* Timer active for this player? */
			if (!(pstp->exist->fl & (PF_ResTimer|PF_InactTimer))) continue;

			/* Timer tval */
			tmptv = TVAL(pstp->exist->timer);

			/* Note lowest remaining timer */
			if ((tmptv >= mintv) && (tmptv < reftv))
			{
				reftv = tmptv;
				reftim = pstp->exist->timer;
			}
		}

		/* Message */
		fprintf(stderr, "%li:   Next auto-update time: %.2i:%.2i:%.2i (%s).\n", uniqid,
				reftim.day, reftim.hour, reftim.min, fdayname(reftim.day, 0));
	
		/* New broken-down time */
		newtm = *the_tm;
		newtm.tm_min = reftim.min;
		newtm.tm_hour = reftim.hour;
		newtm.tm_mday += reftim.day - thetim.day;

		/* Correct it */
		(void) mktime(&newtm);
	
		/* ... Write content ... */
		fprintf(fp_crontab, "%d %d %d %i 0 " SYS_Scores " 2>> " FILE_Log " 1>&2 " LF
							"%d %d %d %i 0 " SYS_MkNews " 2>> " FILE_NewsLog " 1>&2 " LF,
				newtm.tm_min, newtm.tm_hour, newtm.tm_mday, the_tm->tm_mon,
				newtm.tm_min, newtm.tm_hour, newtm.tm_mday, the_tm->tm_mon);

		/* Finished writing crontab */
		i = fclose(fp_crontab);

		/* Ensure no further attempts will be made to close file */
		fp_crontab = 0; if (i) erf(ERF_fcloseFailed, 40);

		/* Install the crontab file (do_crontab) */
		if (mode & MODE_InstallCronTab)
		{
			fprintf(stderr, "%li:   Installing crontab file \"%s\".\n", uniqid, FILE_CronTab);
			if (rc = system("crontab -u emn23 " FILE_CronTab)) erf(ERF_InternalError, 41);
		}
	}

	/* Note success */
	fprintf(stderr, "%li:   -------> scoring complete.\n", uniqid);


_erf: /***** Error handling *****/

	/* Close open files */
	if ((fp_bodge)  &&  (fp_bodge != stdout)) fclose(fp_bodge);
	if ((fp_pseuds) && (fp_pseuds != stdout)) fclose(fp_pseuds);
	if ((fp_scores) && (fp_scores != stdout)) fclose(fp_scores);
	if ((fp_output) && (fp_output != stdout)) fclose(fp_output);
	if (fp_crontab) fclose(fp_crontab);

	/* Close temporarily open file */
	if ((fp) && (fp != stdin) && (fp != stdout)) fclose(fp);

	/* Debug */
	if (rc) fprintf(stderr, "%li !! Error %i in scoring main(%i)\n", uniqid, ERF_TYPE(rc), ERF_CONTEXT(rc));

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/**** The end ****/
