/*****************************************************
**  verify.c                         >> bin/verify  **
**  ----------------------------------------------  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define NO_CGI
#define MAX_statbuf			(256*1024)
#define PLAYER_EXT_TYPE		verify_p		/* type of player_p->ext is verify_p */
#define TARGET_EXT_TYPE		unsigned int	/* type of target_p->ext is unsigned int */

typedef struct verify *verify_p;

#include "!lib-general.c"
#include "!custom.c"
#include "!lib-uinfo-read.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Struct for stuff */
struct verify
{
	int cnt_back;
	int cnt_self;
	int cnt_a_inv;
	int cnt_t_inv;
	int cnt_a_val;
	int cnt_t_val;
};

/* List of players */
static player_p plist = 0;
static int playcnt       = 0;
static int playcnt_mail  = 0;
static int playcnt_spec  = 0;
static int playcnt_play  = 0;
static int playcnt_vict  = 0;
static int playcnt_login = 0;
static int playcnt_pol   = 0;
static int playcnt_ass   = 0;
static int playcnt_pol_d = 0;
static int playcnt_ass_d = 0;

/* Targetting flaws */
#define VRF_Followed		((1UL)<<0)
#define VRF_Invalid			((1UL)<<1)
#define VRF_Self			((1UL)<<2)
#define VRF_MissingPlayer	((1UL)<<3)
#define VRF_Valid			((1UL)<<4)

/* Data buffer */
static char statbuf[MAX_statbuf], *statbuf_end = statbuf - 1 + MAX_statbuf, *statbuf_pos = statbuf;

/***************************************************************************************************************************
** Function hooks and helper functions                                                                                    **
***************************************************************************************************************************/
int verify_hook(player_p pp, int state, void* ctx)
{
	/* Allocate space for extra struct */
	if (!(pp->ext = malloc(sizeof(struct verify)))) return -ERF_mallocFailed;

	/* Maintain list of all players */
	pp->link = plist;
	plist = pp;

	/* Number of players */
	++playcnt;

	/* Count things players */
	if (pp->flags & PF_Mailable) ++playcnt_mail;
	if (pp->flags & PF_Playing)  ++playcnt_play;
	if (pp->flags & PF_Victim)   ++playcnt_vict;
	if (pp->flags & PF_Login)    ++playcnt_login;

	/* Specials */
	if (pp->flags & PF_Special)  ++playcnt_spec;

	/* Police */
	else if (pp->flags & PF_Police)
	{
		++playcnt_pol;
		if (pp->flags & PF_Dead) ++playcnt_pol_d;
	}

	/* Assassins */
	else
	{
		++playcnt_ass;
		if (pp->flags & PF_Dead) ++playcnt_ass_d;
	}

	/* Don't call additional function, and don't overwrite data or player struct. */
	return 32+16+8+1;
}

#if 0
int makestatstab__uf()
{
	struct aplayer *pptr;
	int i;

	/* Check for buffer exhaustion */
	if (playcnt >= MAX_players) return -(ERF_BufferExhaust);

	/* Acquire pointer to free player struct */
	pptr = playstab[playcnt++];

	strcpy(pptr->id, p_id);
	for (i = 0; (i < 6); i++) strcpy(pptr->ass[i], p_con[i]);
	pptr->cnt = 0;
	pptr->fl = p_flags;

	return 1;
}
#endif

int verify_aux(player_p p1ptr, target_p tar1p, player_p plist, int m_ass, unsigned int *vrf)
{
	player_p p2ptr;
	target_p tar2p;
	int n, cnt = 0;


	/* Run through list of targets/assassins */
	for (n = 0; (tar1p); tar1p = tar1p->next, n++)
	{
		/* Already followed? */
		if (tar1p->ext & VRF_Followed) continue;

		/* Now followed */
		tar1p->ext |= VRF_Followed; 

		/* Self-allocation */
		if (streq(tar1p->id, p1ptr->id))
		{
			/* Note this self-reference */
			tar1p->ext |= VRF_Self;

			/* Count number of self-refernces */
			++p1ptr->ext->cnt_self;

			/* Flaw detected */
			++cnt; *vrf |= VRF_Self;
		}

		/* Check consistency */
		else
		{
			/* Match player */
			for (p2ptr = plist; (p2ptr); p2ptr = p2ptr->link) if (streq(p2ptr->id, tar1p->id)) break;

			/* Missing player */
			if (!p2ptr)
			{
				/* Note that it does not lead to a player */
				tar1p->ext |= VRF_MissingPlayer;

				/* Count number of invalid assassins */
				if (m_ass) ++p1ptr->ext->cnt_a_inv;

				/* Count number of invalid targets */
				else ++p1ptr->ext->cnt_t_inv;

				/* Flaw detected */
				++cnt; *vrf |= VRF_MissingPlayer;
			}

			/* Match: check backlinkage */
			else
			{
				/* Try to find a valid backlink */
				for (tar2p = (m_ass) ? p2ptr->targs : p2ptr->asses; (tar2p); tar2p = tar2p->next) 
				{
					/* Already followed, or to different player */
					if ((tar2p->ext & VRF_Followed) || (!streq(tar2p->id, p1ptr->id))) continue;

					/* Now followed, and valid both ways */
					tar2p->ext |= VRF_Followed|VRF_Valid;
					tar1p->ext |= VRF_Valid;

					/* Count number of valid assassins/targets */
					if (m_ass)
					{
						++p1ptr->ext->cnt_a_val;
						++p2ptr->ext->cnt_t_val;
					}

					/* Count number of valid targets/assassins */
					else
					{
						++p1ptr->ext->cnt_t_val;
						++p2ptr->ext->cnt_a_val;
					}

					/* Found valid backlink */
					break;
				}

				/* No valid backlink */
				if (!tar2p)
				{
					/* Make note of invalidity */
					tar1p->ext |= VRF_Invalid;

					/* Count number of invalid assassins */
					if (m_ass) ++p1ptr->ext->cnt_a_inv;

					/* Count number of invalid targets */
					else ++p1ptr->ext->cnt_t_inv;

					/* Count number of missing backlinks */
					++p2ptr->ext->cnt_back;

					/* Flaw detected */
					++cnt; *vrf |= VRF_Invalid;
				}
			}
		}
	}

	/* Any flaws? */
	return cnt;
}

/***************************************************************************************************************************
** Main code. Doesn't make use of command line arguments.                                                                 **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	int i, probs, probplyrs, expect, unexpected, rc = 0, lockfd = 0;
	time_t thetime;
	player_p pptr, p2ptr;


	/* File creation umask */
	umask(0117);

	/* What time is it? */
	time(&thetime);

	/* Get process id for logging */
	uniqid = (long) getpid();

	/* Logfile header */
	fprintf(stderr, "%li ** Verify: at %s", uniqid, ctime(&thetime));

	/* Attempt to lock files */
	if (!(lockfd = setflock("Verify.c "))) erf(ERF_flockFailed, 3);
	
	/* Get player info */
	if ((i = uread(0, verify_hook, 0, statbuf, statbuf_end, &statbuf_pos, 0, UREAD_GetTargs))) erf(i, 4);

	/* Report */
	printf("----------------------------------------------------------------------------\n"
		   ">> %3d User(s) in total (%d special, %d normal): %i attackers, %i victims.\n"
		   ">> %3d User(s) allowed to log in to auto umpire. %i Users can be mailed.\n"
		   ">> %3d Assassins: (%d dead, %d alive). %d Police: (%d dead, %d alive)\n"
		   "----------------------------------------------------------------------------\n",
		   playcnt, playcnt_spec, playcnt - playcnt_spec, playcnt_play, playcnt_vict,
		   playcnt_login, playcnt_mail,
		   playcnt_ass, playcnt_ass_d, playcnt_ass - playcnt_ass_d,
		   playcnt_pol, playcnt_pol_d, playcnt_pol - playcnt_pol_d);

	/* Check targetting for each player */
	for (probs = probplyrs = unexpected = 0, pptr = plist; (pptr); pptr = pptr->link)
	{
		unsigned int vrf = 0;
		target_p tarp;

		/* Initialise */
		for (p2ptr = plist; (p2ptr); p2ptr = p2ptr->link)
		{
			p2ptr->ext->cnt_a_val = p2ptr->ext->cnt_a_inv = 0;
			p2ptr->ext->cnt_t_val = p2ptr->ext->cnt_t_inv = 0;
			p2ptr->ext->cnt_self = p2ptr->ext->cnt_back = 0;

			for (tarp = p2ptr->asses; (tarp); tarp = tarp->next) tarp->ext = 0;
			for (tarp = p2ptr->targs; (tarp); tarp = tarp->next) tarp->ext = 0;
		}

		/* Assassins */
		i = verify_aux(pptr, pptr->asses, plist, 1, &vrf);

		/* Remaining targets */
		i += verify_aux(pptr, pptr->targs, plist, 0, &vrf);

		/* Problems for this player? */
		if (i)
		{
			probs += i;
			++probplyrs;
		}

		/* Expected number of targets and assassins */
		expect = ((pptr->flags & PF_Special) || (pptr->flags & PF_Police) || (pptr->flags & PF_Dead)) ? 0 : 3;

		/* Correct number of assassins and targets? */
		if ((pptr->ext->cnt_a_val != expect) || (pptr->ext->cnt_t_val != expect))
		{
			printf(">> %s: (%02X) Expected %i assassins (found %i), and %i targets (found %i).\n",
				   pptr->id, vrf, expect, pptr->ext->cnt_a_val, expect, pptr->ext->cnt_t_val);

			++unexpected;
		}

		/* Any invalid assassins? */
		if (pptr->ext->cnt_a_inv) printf(">> %s: (%02X) Invalid assassins: %i.\n", pptr->id, vrf, pptr->ext->cnt_a_inv);

		/* Any invalid targets? */
		if (pptr->ext->cnt_t_inv) printf(">> %s: (%02X) Invalid targets: %i.\n", pptr->id, vrf, pptr->ext->cnt_t_inv);

		/* Self-allocations? */
		if (pptr->ext->cnt_self) printf(">> %s: (%02X) Self-allocation: %i-fold.\n", pptr->id, vrf, pptr->ext->cnt_self);

		/* Missing backlinks? */
		if (pptr->ext->cnt_back) printf(">> %s: (%02X) Missing backlinks: %i.\n", pptr->id, vrf, pptr->ext->cnt_back);
	}

	printf("----------------------------------------------------------------------------\n"
		   ">> Verification complete: %i targetting problems on %i problematic players.\n"
		   ">> Found %i player(s) with too few or too many targets and/or assassins.\n"
		   "----------------------------------------------------------------------------\n",
		   probs, probplyrs, unexpected);


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li !! Error %i in verify main(%i)\n", uniqid, ERF_TYPE(rc), ERF_CONTEXT(rc));

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}


