/*****************************************************
**  pled.c                             >> pled.cgi  **
**  ----------------------------------------------  **
**  Builds a CGI executable to handle editing of    **
**  user/player information and special events.     **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifndef THIS_CGI
 #define THIS_CGI			"pled.cgi"
#endif

#define MAX_retargs			5

#define MAX_statbuf			(1024*1024)
#define MAX_fpbuf			(64*1024)
#define MAX_cgibuf 			65536
#define MAX_npairs 			600
#define EVENT_EXT_TYPE		int			/* type of  event_p->ext is irrelevent */
#define PEXIST_EXT_TYPE		int			/* type of pexist_p->ext is irrelevent */
#define PSTATE_EXT_TYPE		ppseud_p	/* type of pstate_p->ext is ppseud_p   */
#define PSTATE_PLP_TYPE		player_p	/* type of pstate_p->plp is player_p   */
#define PLAYER_EXT_TYPE		player_p	/* type of player_p->ext is player_p   */

#include <math.h>
#include "!lib-general.c"
#include "!custom.c"
/*#include "!lib-cgi.c"*/
#include "!lib-uinfo-scan.c"
#include "!lib-uinfo-edit.c"
#include "!lib-uinfo-misc.c"
#include "!lib-event-edit.c"
#include "!lib-evolve.c"
#include "!lib-retarg.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Global variables */
static char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos;
static timmy thetim;
static time_t thetime;
static char strb[MAX_id+32], *timestr, *suggestid = 0;

/* Keep track of data to be passed onto chained binaries */
static char* dolaunch = 0;
static FILE* fplaunch = 0;

/* Call any subsidiary programs to rebuild things? */
#define RESC_MkNews	(1)
#define RESC_Scores	(2)

/* Don't, by default */
static int rescore = 0;

/* Turn action types into action bits */
#define ACT_BIT(x)			(1U<<(x))
#define ABIT_Wanted			ACT_BIT(ACT_Wanted)
#define ABIT_Police			ACT_BIT(ACT_Police)
#define ABIT_Pseudonym		ACT_BIT(ACT_Pseudonym)
#define ABIT_Bonus			ACT_BIT(ACT_Bonus)
#define ABIT_Reincarnate	ACT_BIT(ACT_Reincarnate)
#define ABIT_Deadline		ACT_BIT(ACT_Deadline)
#define ABIT_Comment		ACT_BIT(ACT_Comment)
#define ABIT__MASK_Multi	(ABIT_Pseudonym|ABIT_Police|ABIT_Wanted|ABIT_Reincarnate|ABIT_Bonus|ABIT_Deadline|ABIT_Comment)

/* Possible changes to players */
#define CH_Flags		(1<<0)
#define CH_Name			(1<<1)
#define CH_Addr			(1<<2)
#define CH_Coll			(1<<3)
#define CH_Mafia		(1<<4)
#define CH_Email		(1<<5)

/* Warnings */
#define WARN_Good		(0)
#define WARN_NoChanges	(1)
#define WARN__MAX		(2)

/* Allowable characters for passwords (1 good, 0 bad) */
char pwchars[256] =
{
	/****************************************************************************
	** XXX XXX --- Currently, only "safe" HTML chars should be allowed, since  **
	**             password strings are NOT encapsulated when hidden in forms. **
	**             Hence we only allow unaccented letters and digits, and the  **
	**             underscore. Password comparisons are case sensitive.        **
	****************************************************************************/

	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 1,1,1,1,1,1,1,1, 1,1,0,0,0,0,0,0,
	0,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1, 1,1,1,0,0,0,0,1,
	0,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1, 1,1,1,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0
};

/* Possible results of password checks */
#define PW_Good			(0)
#define PW_TooShort		(1)
#define PW_BadChar		(2)
#define PW_TooLong		(3)
#define PW_Missing		(4)
#define PW_Mismatch		(5)
#define PW__MAX			(6)

/* Reasons why password checks fails */
char* badpwstr[PW__MAX] =
{
	"Moof!",
	"Suggested password was too short",
	"Suggested password contains disallowed letters",
	"Suggested password was too long",
	"No suggested password found",
	"Confirmation password did not match suggested password",
};

/* A type to hold a record of problems when adding player status events */
typedef unsigned int bad_actev_type;

/* Things that can go wrong */
#define BAD_actev_PseudMatch	(((bad_actev_type)1)<<0)
#define BAD_actev_BadPseud		(((bad_actev_type)1)<<1)
#define BAD_actev_Boost			(((bad_actev_type)1)<<2)
#define BAD_actev_WantAct		(((bad_actev_type)1)<<3)
#define BAD_actev_MissingMin	(((bad_actev_type)1)<<4)
#define BAD_actev_MissingHour	(((bad_actev_type)1)<<5)
#define BAD_actev_MissingDay	(((bad_actev_type)1)<<6)
#define BAD_actev_PID			(((bad_actev_type)1)<<7)
#define BAD_actev_Missing		(((bad_actev_type)1)<<8)
#define BAD_actev_InvalidMin	(((bad_actev_type)1)<<9)
#define BAD_actev_InvalidHour	(((bad_actev_type)1)<<10)
#define BAD_actev_InvalidDay	(((bad_actev_type)1)<<11)
#define BAD_actev_MissingPseud	(((bad_actev_type)1)<<12)
#define BAD_actev_LongPseud 	(((bad_actev_type)1)<<13)
#define BAD_actev_DefPolice		(((bad_actev_type)1)<<14)
#define BAD_actev_DefReinc		(((bad_actev_type)1)<<15)
#define BAD_actev_BadPolice		(((bad_actev_type)1)<<16)
#define BAD_actev_BadReinc		(((bad_actev_type)1)<<17)
#define BAD_actev_DefWantSt		(((bad_actev_type)1)<<18)
#define BAD_actev_BadWantSt		(((bad_actev_type)1)<<19)
#define BAD_actev__MAX_IDX		(20)
#define BAD_actev__MissingTime	(BAD_actev_MissingMin|BAD_actev_MissingHour|BAD_actev_MissingDay)

/* Descriptions of each problem */
char* doact_badstuff[BAD_actev__MAX_IDX] =
{
	"This pseudonym has already been taken by another player.<br>",
	"You have not supplied a valid pseudonym.<br>",
	"Invalid boost to incompetence deadline.<br>",
	"You have not specified a valid action.<br>",
	"Time (minutes) of event occurance not specified.<br>",
	"Time (hour) of event occurance not specified.<br>",
	"Date (day) of event occurance not specified.<br>",
	"XXX1",
	"XXX2",
	"Invalid time (minutes) of event occurance.<br>",
	"Invalid time (hour) of event occurance.<br>",
	"Invalid date (day) of event occurance.<br>",
	"Missing pseudonym.<br>",
	"Pseudonym too long.<br>",
	"Invalid default police status/rank.<br>",
	"Invalid default reincarnation action.<br>",
	"Invalid police status/rank.<br>",
	"Invalid reincarnation action.<br>",
	"Invalid default criminality status.<br>",
	"Invalid criminality status.<br>"
};

/* Adjustments for events */
typedef unsigned int adj_type;

/* Possible adjustments */
#define ADJ_SetPseud		(((adj_type)1)<<0)
#define ADJ_AddPseud		(((adj_type)1)<<1)
#define ADJ_WantSt			(((adj_type)1)<<2)
#define ADJ_Comment			(((adj_type)1)<<3)
#define ADJ_NoComment		(((adj_type)1)<<4)
#define ADJ_Reincarnate		(((adj_type)1)<<5)
#define ADJ_Police			(((adj_type)1)<<6)
#define ADJ_Deadline		(((adj_type)1)<<7)
#define ADJ_Promotion		(((adj_type)1)<<8)
#define ADJ_Bonus			(((adj_type)1)<<9)
#define ADJ_DefComment		(((adj_type)1)<<10)
#define ADJ_DefBonus		(((adj_type)1)<<11)
#define ADJ_DefBoost		(((adj_type)1)<<12)
#define ADJ_DefPolice		(((adj_type)1)<<13)
#define ADJ_DefReinc		(((adj_type)1)<<14)
#define ADJ_DefWantSt		(((adj_type)1)<<15)

/* Null-terminated list of CGI variable names to exclude from regurgitation if authentication fails */
char* excl_auth[4] =
{
	"pwsecure", "authname", "authpass", 0
};

/* Function declarations */
int send_doc(int w, player_p authp, player_p pptr);

/***************************************************************************************************************************
** Edit user information (admin only) -- does not edit passwords                                                          **
***************************************************************************************************************************/
int doact_edituser(player_p authp, player_p *pl, int n, int state)
{
	int i, j, k, *warn, warns, *ook, ooks, triv, rc = 0;
	unsigned long f;
	char *p;
	editpl_p edp, head = 0, *oedpp = &head;


	/* Array of structures in which to store new data and its validity */
	if (!(edp = malloc(n * sizeof(struct editpl))) || !(ook = malloc(2 * n * sizeof(int)))) erf(ERF_mallocFailed, 2);

	/* Arrap of warnings */
	warn = &ook[n];

	/* Fill it in */
	for (i = ooks = warns = triv = 0; (i < n); i++)
	{
		/* Defaults */
		edp[i].npass = 0;
		edp[i].ntimp = 0;
		edp[i].opid = pl[i]->id;
		edp[i].nname = 0;
		edp[i].naddr = 0;
		edp[i].ncoll = 0;
		edp[i].nmafia = 0;
		edp[i].nemail = 0;
		edp[i].nfl_mask = edp[i].nfl_bits = 0UL;
		edp[i].next = edp[i].link = 0;
		edp[i].npseud = edp[i].nwant = 0;
		edp[i].nkills = edp[i].ndeaths = 0;
		edp[i].nscore = edp[i].ndanger = 0;
		edp[i].nrank = 0;
		ook[i] = warn[i] = 0;

		/* List of all structs */
		if (i > 0) edp[i-1].link = &edp[i];

		/* No changes to targets */
		edp[i].ntargs = edp[i].nasses = 0;

		/* First time... no data entered yet */
		if (!state) continue;

		/* No changes yet */
		k = 0;

		/* Form "pl_7nn" string */
		if (sprintf(strb, "pl_%dn%nn", i+1, &j) < 6) erf(ERF_InternalError, 3);

		/* Get new name "pl_7nn" */
		if ((p = findval(strb, 0)) && strcmp(p, pl[i]->name)) { edp[i].nname = p; ++k; }

		/* Get new addr "pl_7na" */
		strb[j] = 'a'; if ((p = findval(strb, 0)) && strcmp(p, pl[i]->addr)) { edp[i].naddr = p; ++k; }

		/* Get new college "pl_7nc" */
		strb[j] = 'c'; if ((p = findval(strb, 0)) && strcmp(p, pl[i]->coll)) { edp[i].ncoll = p; ++k; }

		/* Get new mafia "pl_7nm" */
		strb[j] = 'm'; if ((p = findval(strb, 0)) && strcmp(p, pl[i]->mafia)) { edp[i].nmafia = p; ++k; }

		/* Get new email "pl_7ne" */
		strb[j] = 'e'; if ((p = findval(strb, 0)) && strcmp(p, pl[i]->email)) { edp[i].nemail = p; ++k; }

		/* Form "pl_7f_xxx" string */
		p = 1 + strcpy(&strb[j-1], "_mask");
		
		/* Get flags bitmask */
		edp[i].nfl_mask = strtoul(findval(strb, 1), 0, 10);

		/* Sanity check */
		edp[i].nfl_mask &= (PF_Admin|PF_Login|PF_Public|PF_Hidden|PF_Retired|PF_Dangerous|
							PF_Playing|PF_Victim|PF_Special|PF_Mailable|PF_SendMail|PF_LongComment);

		/* Get new flags */
		strcpy(p, "pub");      if (findval(strb, 0)) edp[i].nfl_bits |= PF_Public;
		strcpy(p, "play");     if (findval(strb, 0)) edp[i].nfl_bits |= PF_Playing;
		strcpy(p, "vict");     if (findval(strb, 0)) edp[i].nfl_bits |= PF_Victim;
		strcpy(p, "spec");     if (findval(strb, 0)) edp[i].nfl_bits |= PF_Special;
		strcpy(p, "dang");     if (findval(strb, 0)) edp[i].nfl_bits |= PF_Dangerous;
		strcpy(p, "login");    if (findval(strb, 0)) edp[i].nfl_bits |= PF_Login;
		strcpy(p, "admin");    if (findval(strb, 0)) edp[i].nfl_bits |= PF_Admin;
		strcpy(p, "hidden");   if (findval(strb, 0)) edp[i].nfl_bits |= PF_Hidden;
		strcpy(p, "retire");   if (findval(strb, 0)) edp[i].nfl_bits |= PF_Retired;
		strcpy(p, "mailable"); if (findval(strb, 0)) edp[i].nfl_bits |= PF_Mailable;
		strcpy(p, "sendmail"); if (findval(strb, 0)) edp[i].nfl_bits |= PF_SendMail;
		strcpy(p, "longcom");  if (findval(strb, 0)) edp[i].nfl_bits |= PF_LongComment;

		/* Further sanity imposition */
		edp[i].nfl_bits &= edp[i].nfl_mask;

		/* Have flags been changed? */
		if (pl[i]->flags != ((pl[i]->flags & (~edp[i].nfl_mask)) | edp[i].nfl_bits)) ++k;

		/* Validate anything? */
		/* Not at the moment. */

		/* Any fatal errors? */
		if (!warn[i])
		{
			/* Nothing has changed */
			if (!k)
			{
				warn[i] = WARN_NoChanges;
				++triv;
			}

			/* Link valid non-empty change structs into list */
			else
			{
				*oedpp = &edp[i];
				oedpp = &edp[i].next;
			}
		}

		/* Increment count of warnings/failures */
		if (ook[i]) ++ooks;
		if (warn[i]) ++warns;
	}

	/* Check proposed changes */
	if ((state == 2) && (!ooks))
	{
		/* No valid non-empty changes, or some non-trivial warnings */
		if ((!head) || (triv != warns)) state = 1;

		/* Ignoring trivial warnings? */
		else state = ((!triv) || (findval("ignore", 0))) ? 3 : 4;
	}

	/* Carry out the changes */
	if (state == 3)
	{
		/* Title commend for plinfo edit */
		if ((statbuf_end - statbuf_pos) > (strlen(timestr) + 64 + MAX_id))
		{
			p = statbuf_pos;
			sprintf(p, "From player edit by \"%.*s\" at %s", (int)MAX_id, authp->id, timestr);
			statbuf_pos += 1 + strlen(p);
		}
		else p = 0;

		/* Submit them */
		if (i = edituinfo(0U, &head, 0, statbuf_pos, statbuf_end, 0, p)) erf(i, 10);

		/* Success string */
		dolaunch = "<h3><hr><img src=\""TICK_URL"\" align=middle "TICK_SIZ" hspace=8>"
				   "<font color=#FFCCCC>Yay! Dunnit!</font><hr></h3>";

		/* Automatically rebuild news? */
		if (findval("rebuild", 0)) rescore |= RESC_MkNews;

		/* Success! */
		return 0;
	}

	/* Begin HTML */
	printf("Content-type: text/html" LF LF "<HTML><HEAD><TITLE>"GAME" - Edit user(s)</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr>"GAME" - Edit user(s)<hr></h1>"
		   LF "<form method=\"POST\" action=\""CGI_URL"pled.cgi\">"
		   LF "<input type=\"hidden\" name=\"stat_rest\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_vit\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_sort\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_law\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"act\" value=\"%i\">",
		   findval("stat_rest", 1), findval("stat_vit", 1),
		   findval("stat_sort", 1), findval("stat_law", 1), ACT_EditUser);

	/* Print authentication information */
	if (!(authp->flags & PF_Public))
	{
		printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\"><p>Authenticated as <b>%s</b>%s.</p>"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">",
			   authp->id, authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""), authp->pass);
	}

	if (ooks) printf(LF "<h3><hr><img src=\""CROSS_URL"\" align=middle "CROSS_SIZ" hspace=8>"
					 LF "<font color=#FFCCCC>Invalid change%s: %i problem%s found."
					 LF "Please correct.</font><hr></h3>", (ooks > 1) ? "s" : " ", ooks, (ooks > 1) ? "s" : " ");

	if (state == 4) printf(LF "<p><hr><img src=\""SKULL_URL"\" align=left "SKULL_SIZ" hspace=8>"
						   LF "<font color=#FFCCCC>%i of these players would not be affected.</font>"
						   LF "&nbsp;<input type=\"checkbox\" name=\"ignore\" value=\"set\">&nbsp;"
						   LF "Tick box and resubmit to ignore them.<hr clear=left></p>", triv);

	else if (warns) printf(LF "<h3><hr><img src=\""SKULL_URL"\" align=middle "SKULL_SIZ" hspace=8>"
						   LF "<font color=#FFCCCC>For your attention: %i warning%s.</font><hr></h3>",
						   warns, ((warns > 1) ? "s" : " "));

	/* Blurb */
	puts(LF "<p>Should anything go here?</p>"
		 LF "<p><table bgcolor=#AA0000 rows=2 cols=2 border=0 cellspacing=4 cellpadding=4>"
		 LF "<tr><td colspan=2><font color="WHITE">In order for this operation to be completed,"
		 LF "the correct additional security code must be entered.</font></td></tr>"
		 LF "<tr><td align=right><font color="WHITE">Authorization code:&nbsp;</font></td>"
		 LF "<td><input type=\"password\" name=\"pwsecure\" value=\"\"></td></tr></table></p>"
		 LF "<p><table rows=2 cols=1 border=0 cellpadding=4 cellspacing=4 width=100%>");

	/* Deal with each player in turn */
	for (i = 0; (i < n); i++)
	{
		unsigned long fm = 0;

		f = (pl[i]->flags & (~edp[i].nfl_mask)) | edp[i].nfl_bits;

		p = (i & 1) ? RED : BLUE;

		/*fprintf(stderr, "%li:    %i plf=&%lX mask=&%lX bits=&%lX f=&%lX\n", uniqid, i,
				pl[i]->flags, edp[i].nfl_mask, edp[i].nfl_bits, f);*/

		printf(LF "<tr><td bgcolor=%s><input type=\"hidden\" name=\"pl_%did\" value=\"%s\">"
			   LF "<h4><font color="WHITE">&nbsp;<a href=\"plinfo.cgi?id=%s\">%s</a>",
			   p, i+1, pl[i]->id, pl[i]->id, pl[i]->name);

		if (ook[i]) printf(LF "&mdash;<b>Error:%d</b>", ook[i]);
		if (warn[i]) printf(LF "&mdash;<b>Warning:%d</b>", warn[i]);

		printf(LF "</font></h4></td></tr><tr><td bgcolor=%s align=right>"
			   LF "<table border=0 cellspacing=2 cellpadding=0 cols=5 rows=5 width=100%%>"
			   LF "<tr><td align=right><font color="WHITE">%sName%s&nbsp;</font></td>"
			   LF "<td><input type=\"text\" name=\"pl_%dnn\" "OPT_EditUserName" value=\"",
			   p, ((edp[i].nname) ? "<i>" : " "), ((edp[i].nname) ? "~</i>" : ":"), i+1);

		encapsulate(ThisOrThat(edp[i].nname, pl[i]->name), stdout);

		printf("\">&nbsp;</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_admin\"%s value=\"set\">&nbsp;Admin</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_login\"%s value=\"set\">&nbsp;Can login</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_pub\"%s value=\"set\">&nbsp;Public use</td></tr>"
			   LF "<td align=right><font color="WHITE">%sAddress%s&nbsp;</font></td>"
			   LF "<td><input type=\"text\" name=\"pl_%dna\" "OPT_EditUserAddr" value=\"",
			   i+1, ((f & PF_Admin) ? " checked" : " "),
			   i+1, ((f & PF_Login) ? " checked" : " "),
			   i+1, ((f & PF_Public) ? " checked" : "  "),
			   ((edp[i].naddr) ? "<i>" : " "), ((edp[i].naddr) ? "~</i>" : ":"), i+1);

		fm |= (PF_Admin|PF_Login|PF_Public);

		encapsulate(ThisOrThat(edp[i].naddr, pl[i]->addr), stdout);

		printf("\">&nbsp;</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_hidden\"%s value=\"set\">&nbsp;Hidden</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_retire\"%s value=\"set\">&nbsp;Retired</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_dang\"%s value=\"set\">&nbsp;Dangerous (unused)</td></tr>"
			   LF "<tr><td align=right><font color="WHITE">%sCollege%s&nbsp;</font></td>"
			   "<td><input type=\"text\" name=\"pl_%dnc\" "OPT_EditUserColl" value=\"",
   			   i+1, ((f & PF_Hidden) ? " checked" : " "),
   			   i+1, ((f & PF_Retired) ? " checked" : " "),
			   i+1, ((f & PF_Dangerous) ? " checked" : " "),
			   ((edp[i].ncoll) ? "<i>" : " "), ((edp[i].ncoll) ? "~</i>" : ":"), i+1);

		fm |= (PF_Hidden|PF_Retired|PF_Dangerous);

		encapsulate(ThisOrThat(edp[i].ncoll, pl[i]->coll), stdout);

		printf("\">&nbsp;</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_play\"%s value=\"set\">&nbsp;Playing</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_vict\"%s value=\"set\">&nbsp;Valid victim</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_spec\"%s value=\"set\">&nbsp;Special user</td></tr>"
			   LF "<tr><td align=right><font color="WHITE">%sEmail%s&nbsp;</font></td>"
			   LF "<td><input type=\"text\" name=\"pl_%dne\" "OPT_EditUserEmail" value=\"",
   			   i+1, ((f & PF_Playing) ? " checked" : " "),
			   i+1, ((f & PF_Victim) ? " checked" : " "),
			   i+1, ((f & PF_Special) ? " checked" : " "),
			   ((edp[i].nemail) ? "<i>" : " "), ((edp[i].nemail) ? "~</i>" : ":"), i+1);

		fm |= (PF_Playing|PF_Victim|PF_Special);

		encapsulate(ThisOrThat(edp[i].nemail, pl[i]->email), stdout);

		printf("\">&nbsp;</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_longcom\"%s value=\"set\">&nbsp;Long comment</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_mailable\"%s value=\"set\">&nbsp;Mailable</td>"
			   LF "<td><input type=\"checkbox\" name=\"pl_%d_sendmail\"%s value=\"set\">&nbsp;In"
			   LF "next batch</td></tr>"
			   LF "<tr><td align=right><font color="WHITE">%sThingy%s&nbsp;</font></td>"
			   LF "<td colspan=4><input type=\"text\" name=\"pl_%dnm\" "OPT_EditUserMafia" value=\"",
			   i+1, ((f & PF_LongComment) ? " checked" : " "),
			   i+1, ((f & PF_Mailable) ? " checked" : " "),
			   i+1, ((f & PF_SendMail) ? " checked" : "  "),
			   ((edp[i].nmafia) ? "<i>" : " "), ((edp[i].nmafia) ? "~</i>" : ":"), i+1);

		fm |= (PF_LongComment|PF_Mailable|PF_SendMail);

		encapsulate(ThisOrThat(edp[i].nmafia, pl[i]->mafia), stdout);

		printf("\"></td></tr></table><input type=\"hidden\" name=\"pl_%d_mask\" value=\"%lu\"></td></tr>", i+1, fm);
	}

	/* Table footer and submission button */
	printf("</table></p><p><input value=\"Submit changes\" type=\"submit\" name=\"sub_pled\">&nbsp;"
		   "Automatically&nbsp;rebuild&nbsp;news:&nbsp;&nbsp;"
		   "<input type=\"checkbox\" name=\rebuild\"%s value=\"set\"></p>",
		   (findval("rebuild", 0) ? " checked" : " "));

	/* Footer */
	printf(LF "</form><p><hr>"FOOTER" &mdash; Produced at %s</p></BODY></HTML>" LF, timestr);


_erf: /***** Handle errors *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   doact_edituser(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Edit passwords: admin can edit any passwords, other player can edit only their own.                                    **
***************************************************************************************************************************/
int doact_password(player_p authp, player_p *pl, int n, int state, char* sec)
{
	int i, j, *ook, ooks, rc = 0;
	editpl_p edp;
	char *p;
	size_t s;


	/* Array in which to store new passwords and their validity */
	if (!(edp = malloc(n * sizeof(struct editpl))) || !(ook = malloc(n * sizeof(int)))) erf(ERF_mallocFailed, 2);

	/* Fill it in */
	for (i = ooks = 0; (i < n); i++)
	{
		/* Defaults */
		edp[i].npass = 0;
		edp[i].ntimp = 0;
		edp[i].opid = pl[i]->id;
		edp[i].nname = 0;
		edp[i].naddr = 0;
		edp[i].ncoll = 0;
		edp[i].nmafia = 0;
		edp[i].nemail = 0;
		edp[i].nfl_mask = edp[i].nfl_bits = 0UL;
		edp[i].next = edp[i].link = 0;
		edp[i].npseud = edp[i].nwant = 0;
		edp[i].nkills = edp[i].ndeaths = 0;
		edp[i].nscore = edp[i].ndanger = 0;
		edp[i].nrank = 0;
		ook[i] = 0;

		/* List of all structs */
		if (i > 0) edp[i-1].link = edp[i-1].next = &edp[i];

		/* No changes to targets */
		edp[i].ntargs = edp[i].nasses = 0;

		/* First time... no data entered yet */
		if (!state) continue;

		/* Form "pl_7p1" string */
		if (sprintf(strb, "pl_%dp%n1", i+1, &j) < 6) erf(ERF_InternalError, 3);

		/* Get first attempt at password */
		edp[i].npass = findval(strb, 0);

		/* Form "pl_7p2" string for second attempt */
		strb[j] = '2';

		/* Do both attempts match? */
		if ((p = findval(strb, 0)) && (edp[i].npass) && streq(p, edp[i].npass))
		{
			/* Check for invalid characters */
			for (ook[i] = 0, p = edp[i].npass; (*p); p++) if (!pwchars[*p]) { ook[i] = PW_BadChar; break; }

			/* Enforce password length limits */
			if (!ook[i])
			{
				if ((s = p - edp[i].npass) < MIN_PasswordLen) ook[i] = PW_TooShort;
				if (s > MAX_PasswordLen) ook[i] = PW_TooLong;
			}
		}

		/* Neither password given? */
		else if ((!p) && (!edp[i].npass)) ook[i] = PW_Missing;

		/* Password mismatch */
		else ook[i] = PW_Mismatch;

		/* Increment count of failures, and blank invalid password */
		if (ook[i])
		{
			++ooks;
			edp[i].npass = 0;
		}
	}

	/* Carry out the changes */
	if ((state == 2) && (!ooks))
	{
		/* Title commend for plinfo edit */
		if ((statbuf_end - statbuf_pos) > (strlen(timestr) + 64 + MAX_id))
		{
			p = statbuf_pos;
			sprintf(p, "From password edit by \"%.*s\" at %s", (int)MAX_id, authp->id, timestr);
			statbuf_pos += 1 + strlen(p);
		}
		else p = 0;

		/* Submit them */
		if (i = edituinfo(0U, &edp, 0, statbuf_pos, statbuf_end, 0, p)) erf(i, 10);

		/* Success string */
		/*dolaunch = ;*/

		/* Success! Begin HTML */
		printf("Content-type: text/html" LF LF "<HTML><HEAD><TITLE>"GAME" - password(s) changed</TITLE></HEAD>"
			   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
			   LF "<h1><hr>"GAME"<hr></h1>");
	
		/* Print authentication information */
		if (!(authp->flags & PF_Public))
		{
			printf(LF "<p>Was authenticated as <b>%s</b>%s.</p>",
				   authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""));
		}

		printf(LF "<hr><h3><img src=\""TICK_URL"\" align=left "TICK_SIZ" hspace=8>"
			   LF "<font color=#FFCCCC>Password%s successfully changed for%s",
			   ((n > 1) ? "s" : " "), ((n > 1) ? ":</font><br clear=left></h3><ul>"LF"<li>" : " "));

		/* Print player names */
		for (i = j = 0; (i < n); i++)
		{
			printf("%s<a href=\"plinfo.cgi?id=%s\">%s</a>", ((i > 0) ? "</li>"LF"<li>" : ""), pl[i]->id, pl[i]->name);

			/* Has autentication password changed? */
			if (authp == pl[i]) j = 1;
		}

		puts((n > 1) ? "</li></ul><hr>" : "</font><br clear=left><hr></h3>");

		if (findval("passback", 0)) puts("<p><b>Alas! Confirmation of passwords via email is not yet working.</b></p>");

		if (j) printf("<p>Please <a href=\""CGI_URL"front.cgi?id=%s\">login</a> again.</p>", authp->id);

		else
		{
			printf("<p><form method=\"POST\" action=\""CGI_URL"stats.cgi\">"
   				   LF "<input type=\"hidden\" name=\"rest\" value=\"%s\">"
				   LF "<input type=\"hidden\" name=\"vitality\" value=\"%s\">"
				   LF "<input type=\"hidden\" name=\"sort\" value=\"%s\">"
				   LF "<input type=\"hidden\" name=\"law\" value=\"%s\">"
				   LF "<input type=\"hidden\" name=\"act\" value=\"0\">",
				   findval("stat_rest", 1), findval("stat_vit", 1),
				   findval("stat_sort", 1), findval("stat_law", 1));

			for (i = 0; (i < n); i++) printf(LF "<input type=\"hidden\" name=\"id_%s\" value=\"show\">", pl[i]->id);

			if (!(authp->flags & PF_Public))
			{
				printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
					   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">", authp->id, authp->pass);
			}

			puts(LF "<input value=\"Return to full stats page\" type=\"submit\" name=\"resort\">"
				 LF "<input value=\"View these players' stats\" type=\"submit\" name=\"sub_show\"></form></p>");
		}

		printf(LF "<p><hr>"FOOTER" &mdash; Produced at %s</p></BODY></HTML>" LF, timestr);

		/* Success! */
		return 0;
	}

	/* Begin HTML */
	printf("Content-type: text/html" LF LF "<HTML><HEAD><TITLE>"GAME" - Change password(s)</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr>"GAME" - Change Password(s)<hr></h1>"
		   LF "<form method=\"POST\" action=\""CGI_URL"pled.cgi\">"
   		   LF "<input type=\"hidden\" name=\"stat_rest\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_vit\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_sort\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_law\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"act\" value=\"%i\">",
		   findval("stat_rest", 1), findval("stat_vit", 1),
		   findval("stat_sort", 1), findval("stat_law", 1), ACT_Password);

	/* Print authentication information */
	if (!(authp->flags & PF_Public))
	{
		printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\"><p>Authenticated as <b>%s</b>%s.</p>"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">",
			   authp->id, authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""), authp->pass);
	}

	/* Single player title */
	if (n == 1)
	{
		printf(LF "<h3>Change password for <a href=\"plinfo.cgi?id=%s\">%s</a></h3>", pl[0]->id, pl[0]->name);

		if (ooks) printf(LF "<h4><hr><img align=middle src=\""CROSS_URL"\" border=0 width=34 height=34 hspace=8>"
						 LF "<font color=#FFCCCC>Please correct problem: %s.</font><hr></h4>", badpwstr[ook[0]]);
	}

	/* Errors */
	else if (ooks) printf(LF "<h3><hr><img align=middle src=\""CROSS_URL"\" border=0 width=34 height=34 hspace=8>"
						  LF "<font color=#FFCCCC>Bad password%s: %i problem%s found."
						  LF "Please correct.</font><hr></h3>", (ooks > 1) ? "s" : "", ooks, (ooks > 1) ? "s" : "");

	/* Blurb */
	puts(LF "<p>Passwords must contain between "STRING(MIN_PasswordLen)" and "STRING(MAX_PasswordLen)" characters,"
		 LF "and are <b>case sensitive</b>.<br>Only plain letters and numbers, and the underscore <tt>_</tt> are"
		 LF "allowed in them.</p>");

	/* Possible optional security code? */
	if (sec)
	{
		printf(LF "<p><table bgcolor=#AA0000 rows=2 cols=2 border=0 cellspacing=4 cellpadding=4>"
			   LF "<tr><td colspan=2><font color="WHITE">In order for this operation to be completed,"
			   LF "the correct additional security code must be entered.</font></td></tr>"
			   LF "<tr><td align=right><font color="WHITE">Authorization code:&nbsp;</font></td>"
			   LF "<td><input type=\"password\" name=\"pwsecure\" value=\"\"></td></tr></table></p>");
	}

	/* Deal with single player */
	if (n == 1)
	{
		printf(LF "<input type=\"hidden\" name=\"pl_1id\" value=\"%s\">"
			   LF "<p><table rows=4 cols=2 cellpadding=4 cellspacing=4 border=0 bgcolor=#000000>"
			   LF "<tr><td>Enter new password:</td><td><input type=\"password\" name=\"pl_1p1\" value=\"", pl[0]->id);

		if (edp[0].npass) encapsulate(edp[0].npass, stdout);

		printf("\" maxlength="STRING(MAX_PasswordLen)" size="OPT_PasswordSize"></td>"
			   LF "<tr><td>Confirm new password:</td>"
			   "<td><input type=\"password\" name=\"pl_1p2\" maxlength="STRING(MAX_PasswordLen)" value=\"");

		if (edp[0].npass) encapsulate(edp[0].npass, stdout);

		puts("\" size="OPT_PasswordSize"></td></tr>" LF "<tr><td></td><td>");

		if (pl[0]->flags & PF_Mailable)
		{
			printf("<input type=\"checkbox\" name=\"passback\"%svalue=\"set\">&nbsp;"
				   "Send back confirmation of password to player by email" LF,
				   (((state == 0) || (findval("passback", 0))) ? " checked " : " "));
		}
		else puts("Can't send confirmation of password by email.");

		puts("</td></tr><tr><td></td><td>"
			 "<input value=\"Set password\" type=\"submit\" name=\"sub_pled\"></td></tr></table></p>");
	}

	/* Deal with multiple players */
	else
	{
		/* Table header */
		printf(LF "<p><table cols=4 rows=%i border=0 cellpadding=4 cellspacing=4>"
			   LF "<thead bgcolor="GREEN"><td>Player:</td><td>Enter new password:</td>"
			   LF "<td>Confirm new password:</td><td bgcolor=#000000></td></thead>", n);

		/* Deal with each player in turn */
		for (i = 0; (i < n); i++)
		{
			p = (i & 1) ? RED : BLUE;

			printf(LF "<tr><td bgcolor=%s><input type=\"hidden\" name=\"pl_%did\" value=\"%s\">"
				   LF "<font color="WHITE"><b><a href=\"plinfo.cgi?id=%s\">%s</a></b></font></td>"
				   LF "<td bgcolor=%s><input type=\"password\" name=\"pl_%dp1\" value=\"",
				   p, i+1, pl[i]->id, pl[i]->id, pl[i]->name, p, i+1);

			if (edp[i].npass) encapsulate(edp[i].npass, stdout);

			printf("\" maxlength="STRING(MAX_PasswordLen)" size="OPT_PasswordSize"></td>" LF "<td bgcolor=%s>"
				   "<input type=\"password\" name=\"pl_%dp2\" size="OPT_PasswordSize" value=\"", p, i+1);

			if (edp[i].npass) encapsulate(edp[i].npass, stdout);

			if (!ook[i]) puts("\" maxlength="STRING(MAX_PasswordLen)"></td><td bgcolor=#000000></td></tr>");

			else printf("\"></td><td bgcolor=%s><b>%s</b></td></tr>", p, badpwstr[ook[i]]);
		}

		/* Table footer and submission button */
		printf("</table></p><p><input value=\"Set passwords\" type=\"submit\" name=\"sub_pled\">&nbsp;" LF
			   "Automatically&nbsp;email&nbsp;confirmation&nbsp;of&nbsp;passwords&nbsp;to&nbsp;players:&nbsp;"
			   "<input type=\"checkbox\" name=\"passback\"%svalue=\"set\"></p>",
			   ((findval("passback", 0)) ? " checked " : ""));
	}

	/* Footer */
	printf(LF "</form><p><hr>"FOOTER" &mdash; Produced at %s</p></BODY></HTML>" LF, timestr);


_erf: /***** Handle errors *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   doact_password(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Add retargetting events --- Not implemented yet.                                                                       **
***************************************************************************************************************************/
int doact_retarget(player_p authp, player_p *pl, int n, int state)
{
	int i, j, m, rc = 0;
	editev_p nevs, nevp;
	char *p;
	unsigned long fl;
	retarg_p retarps, rtp;


	/* Submit the events? */
	if (state > 1)
	{
		/* Allocate memory for event addition structs */
		if (!(nevs = malloc(n * sizeof(struct editev)))) erf(ERF_mallocFailed, 4);

		/* Event flags */
		fl = EF_Targetting;

		/* Fill it in */
		for (i = m = 0; (i < n); i++)
		{
			/* Empty list */
			retarps = 0;

			/* Get targets */
			if (rc = readretargs__aux(i+1, 't', pl[i], &retarps, strb)) erf(rc, 5);

			/* None found */
			if (!retarps) continue;

			/* Initialise pointer */
			nevp = &nevs[m];

			/* Link into list */
			if (m > 0) nevs[m-1].next = nevp;

			/* End of list */
			nevp->next = 0;
			nevp->link = 0;

			/* Initialise blank struct */
			init_editev_struct(nevp);

			/* Add new event, chronologically */
			nevp->at = thetim;
			nevp->rid = 0;
			nevp->mode = 0;
			nevp->npid = pl[i]->id;
			nevp->nref = 0;
			nevp->oldref = 0;
			nevp->nflsp = &fl;
			nevp->ntimp = &thetim;
			nevp->nretarg = retarps;

			/* Count valid events to add */
			++m;
		}

		/* No valid changes */
		if (!m) state = 1;
	}
	
	/* Add events */
	if (state > 1)
    {
		/* Form comment */
		if ((statbuf_end - (p = statbuf_pos)) > (96 + MAX_id))
		{
			/* Comment string */
			j = sprintf(p, " Added %i retargetting events at realtime %.2i:%.2i:%.2i by auth \"%.*s\":",
						m, thetim.day, thetim.hour, thetim.min, (int)MAX_id, authp->id);
	
			/* Check string length */
			if ((j < 30) || (j > (96+MAX_id))) p = 0;
	
			/* Advance buffer position */
			else statbuf_pos += 1 + j;
		}
	
		/* No comment */
		else p = 0;

		/* Submit them */
		if (i = editevents(0, &nevs[0], &statbuf_pos, statbuf_end, p, 0)) erf(i, 30);

		/* Success string */
		dolaunch = "<h3><hr><img src=\""TICK_URL"\" align=middle "TICK_SIZ" hspace=8>"
				   "<font color=#FFCCCC>Yay! Dunnit!</font><hr></h3>";

		/* Maintain current status */
		rescore |= RESC_Scores;

		/* Success! */
		return 0;
	}

	/* Begin HTML */
	printf("Content-type: text/html" LF LF "<HTML><HEAD><TITLE>"GAME" - Targetting</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr>"GAME" - Targetting<hr></h1>"
		   LF "<form method=\"POST\" action=\""CGI_URL"pled.cgi\">"
   		   LF "<input type=\"hidden\" name=\"stat_rest\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_vit\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_sort\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_law\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"act\" value=\"%i\">",
		   findval("stat_rest", 1), findval("stat_vit", 1),
		   findval("stat_sort", 1), findval("stat_law", 1), ACT_Retarget);

	/* Print authentication information */
	if (!(authp->flags & PF_Public))
	{
		printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\"><p>Authenticated as <b>%s</b>%s.</p>"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">",
			   authp->id, authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""), authp->pass);
	}

	if (state) printf(LF "<h3><hr><img align=middle src=\""CROSS_URL"\" border=0 width=34 height=34 hspace=8>"
						  LF "<font color=#FFCCCC>Something dodgy happened.</font><hr></h3>");

	rc = prtretarg(pl, n, stdout, 0, 0);

	puts("<p><input value=\"Add targetting events\" type=\"submit\" name=\"sub_pled\">&nbsp;</p>");

	printf(LF "</form><p><hr>"FOOTER" &mdash; Produced at %s</p></BODY></HTML>" LF, timestr);

_erf: /***** Handle errors *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   doact_retarget(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Function to "clean" up pseudonyms by stripping leading and trailing whilespace, and collapsing internal whitespace.    **
** Overwrites its input string, and returns null pointer if input or output is empty.                                     **
***************************************************************************************************************************/
char* cleanpseud(char* pin)
{
	char *pout = pin, *p = pin;

	/* Pass null pointers straight through */
	if (!pin) return 0;

	/* Clean up pseudonym */
	while (*pout = *p)
	{
		/* Copy chars */
		while ((*(pout++) = *p) && (*p != ' ') && (*p != '')) ++p;
	
		/* End of string */
		if (!p[0]) { --pout; break; }

		/* Skip leading whitespace entirely */
		if (pout == pin + 1) --pout;

		/* Reduce white space */
		while ((*p == ' ') || (*p == '')) ++p;
	}

	/* Strip trailing white space */
	if ((pout > pin) && ((pout[-1] == ' ') || (pout[-1] == ''))) pout[-1] = '\0';

	/* Return null ptr if empty */
	return (*pin) ? pin : 0;
}

/***************************************************************************************************************************
** Helper functions for adding events                                                                                     **
***************************************************************************************************************************/
void doact_event_aux(unsigned m, char *str, int *colp, char **pin, char* cp[2], unsigned abits, adj_type adj,
					 long rinc, long pol, long boost, long scb, char* pseud, long wantst, char* comment, pstate_p psp)
{
	char *p = "FFFFFF";
	long k;
	unsigned int i;

	/* Initial background colour */
	if (pin) p = *pin;

	/*fprintf(stderr, "%li:   poloaoueete\n", uniqid);*/

	/* Reincarnation */
	if (abits & ABIT_Reincarnate)
	{
		if (colp) printf("<tr><td bgcolor=%s align=right>Reincarnation:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s><select name=\"pl_%srinc\" size=1>", p, str);

		if (m & 1) printf(LF "<option%svalue=\"0\">** As default</option>",
						 ((adj & ADJ_DefReinc) ? " selected " : " "));

		printf(LF "<option%svalue=\"1\">** No change</option>"
			   LF "<option%svalue=\"2\">Reincarnate</option></select>",
			   (!(adj & ADJ_DefReinc) && (!rinc)) ? " selected " : " ",
			   (!(adj & ADJ_DefReinc) && (rinc))  ? " selected " : " ");

		if (!psp) puts("</td></tr>");

		else printf("&nbsp;" LF "Current state: <b>%s</b></td></tr>", (psp->exist->fl & PF_Dead) ? "Dead" : "Alive");
	}

	/* Police and ranks */
	if (abits & ABIT_Police)
	{
		/*fprintf(stderr, "%li:   poloete\n", uniqid);*/

		if (colp) printf("<tr><td bgcolor=%s align=right>Police status/rank:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s><select name=\"pl_%spol\" size=1>", p, str);

		if (adj & ADJ_DefPolice) k = 0;
		else if ((adj & ADJ_Promotion) && (pol >= -1) && (pol <= +1)) k = pol + 4L;
		else if (!(adj & ADJ_Promotion) && (pol <= 0)) k = 1L + (pol == 0);
		else if (!(adj & ADJ_Promotion) && (pol >= 1)) k = 6L + pol;
		else k = 6;

		if (m & 1) printf(LF "<option%svalue=\"\">** As default</option>", ((k == 0) ? " selected " : " "));

		/*fprintf(stderr, "%li:   po1loete\n", uniqid);*/

		printf(LF "<option%svalue=\"x\">** No change</option>"
			   LF "<option%svalue=\"0\">** Not police</option>"
			   LF "<option%svalue=\"+0\">** Police (initial rank)</option>"
			   LF "<option%svalue=\"+1\">** Promote police (+1 rank)</option>"
			   LF "<option%svalue=\"-1\">** Demote police (-1 rank)</option>",
			   (k == 1) ? " selected " : " ", (k == 2) ? " selected " : " ",
			   (k == 4) ? " selected " : " ", (k == 5) ? " selected " : " ", (k == 3) ? " selected " : " ");

		if (k == 6) printf(LF "<option selected value=\"%+li\">** %s police (%+li rank)</option>",
						   pol, (pol >= 0) ? "Promote" : "Demote", pol);

		printf(LF "<option value=\"\">----------------</option>");

		for (i = 1; (i <= MAX_rank); i++) printf(LF "<option%svalue=\"%i\">%s</option>",
												((k == i + 6L) ? " selected " : " "), i, rank_name[i]);
		if (!psp) puts("</select></td></tr>");

		else if (!(psp->exist->fl & PF_Police)) puts("</select>&nbsp;" LF "Currently not police.</td></tr>");

		else printf("</select>&nbsp;" LF "Current rank: <b>%s</b></td></tr>", rank_name[LIM_RANK(psp->exist->rank)]);

		/*fprintf(stderr, "%li:   polo2ete\n", uniqid);*/
	}

	/*fprintf(stderr, "%li:   polo3ete\n", uniqid);*/

	/* Score bonus */
	if (abits & ABIT_Bonus)
	{
		double scbd = scb / (double)(1 << SCORE_SHIFT);

		if (colp) printf("<tr><td bgcolor=%s align=right>Bonus score:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s><input type=\"text\" name=\"pl_%sscb\" value=\"", p, str);

		if (!(adj & ADJ_DefBonus)) printf("%f", scbd);

		if (!psp) puts("\" maxlength=12></td></tr>");

		else printf("\" maxlength=12>&nbsp;" LF "Current score: <b>%f</b></td></tr>",
					psp->score / (double)(1 << SCORE_SHIFT));
	}

	/* Deadlines */
	if (abits & ABIT_Deadline)
	{
		if (colp) printf("<tr><td bgcolor=%s align=right>Inactivity deadline:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s>"
			   "<input type=\"text\" name=\"pl_%sbst_d\" size=4 value=\"", p, str);

		if (!(adj & ADJ_DefBoost)) printf("%li", (boost / 24L));

		printf("\" maxlength=4>&nbsp;&nbsp;days&nbsp; and &nbsp;<input name=\"pl_%sbst_h\" value=\"", str);

		if (!(adj & ADJ_DefBoost)) printf("%li", (boost % 24L));

		puts("\" maxlength=4 size=4 type=\"text\">&nbsp;&nbsp;hours from time of event.");

		if (!psp) puts("</td></tr>");

		else if (!(psp->exist->fl & PF_InactTimer)) puts(LF "No current deadline.</td></tr>");

		else printf(LF "Current deadline: <b>%.2d:%.2d</b> on %s</td></tr>", psp->exist->timer.hour,
					psp->exist->timer.min, fdayname(psp->exist->timer.day, 0));
	}

	/* Pseudonyms */
	if (abits & ABIT_Pseudonym)
	{
		ppseud_p pseudp;

		if (colp) printf("<tr><td bgcolor=%s align=right>Pseudonym:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s><select size=1 name=\"pl_%spso\">"
			   LF "<option%svalue=\"\">** No change</option>"
			   LF "<option%svalue=\"-\">** Set new pseudonym...</option>"
			   LF "<option%svalue=\"+\">** Add new pseudonym...</option>", p, str,
			   (adj & (ADJ_SetPseud|ADJ_AddPseud)) ? " " : " selected ",
			   (adj & ADJ_SetPseud) ? " selected " : " ",
			   (adj & ADJ_AddPseud) ? " selected " : " ");

		if (psp) for (i = 0, k = 1; (i <= 2); i++)
		{
			for (pseudp = psp->allpseuds; (pseudp); pseudp = pseudp->link_pl)
			{
				/*fprintf(stderr, "%li:    eexs! i=%i k=%li fl=&%X \"%s\"\n", uniqid, i, k, pseudp->fl, pseudp->text);*/

				if ((i == 1) && !(pseudp->fl & PSFL_XDefault)) continue;
				else if ((i != 1) && (pseudp->fl & PSFL_XDefault)) continue;
				else if ((i != 1) && (((pseudp->fl & PSFL_XPrior) != 0) ^ ((i & 2) != 0))) continue;
				if (!(k & (2L << i)) && (k & 7))
				{
					fputs(LF "<option value=\"\">--------------------</option>", stdout);
				}
				fputs(LF "<option value=\"", stdout);					
				encapsulate(pseudp->text, stdout);
				fputs("\">", stdout);
				encapslim(pseudp->text, stdout, MAX_PseudDisplay);
				puts((pseudp->fl & PSFL_XDefault) ? " [current]</option>" : "</option>");
				k = (2L << i);
			}
		}

		printf("</select>&nbsp;&nbsp;<input type=\"text\" name=\"pl_%spsn\" value=\"", str);

		if (pseud) encapsulate(pseud, stdout);

		puts("\"></td></tr>");
	}

	/* Wantedness */
	if (abits & ABIT_Wanted)
	{
		k = (adj & ADJ_DefWantSt) ? -1 : wantst;

		if (colp) printf("<tr><td bgcolor=%s align=right>Criminality:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s><select size=1 name=\"pl_%swantst\">", p, str);

		if (m & 1) printf(LF "<option%svalue=\"\">** As default</option>", ((wantst < 0) ? " selected " : " "));

		printf(LF "<option%svalue=\"0\">** No change</option>"
			   LF "<option%svalue=\"1\">Wanted criminal</option>"
			   LF "<option%svalue=\"2\">Incompetent</option>"
			   LF "<option%svalue=\"3\">Unwanted (redeem)</option></select>",
			   ((k == 0) ? " selected " : " "), ((k == 1) ? " selected " : " "),
			   ((k == 2) ? " selected " : " "), ((k == 3) ? " selected " : " "));

		if (!psp) puts("</td></tr>");

		else printf("&nbsp;" LF "Currently: <b>%s</b></td></tr>",
					(k = PF_type(psp->exist->fl)) ? pltypes[k] : "Law-abiding (not wanted)");
	}

	/* Comments */
	if (abits & ABIT_Comment)
	{			   
		if (colp) printf("<tr><td bgcolor=%s align=right>Comments:", p = cp[((*colp)++) & 1]);

		printf("&nbsp;</td>" LF "<td bgcolor=%s><select size=1 name=\"pl_%scomc\">", p, str);

		if (m & 1) printf(LF "<option%svalue=\"\">** As default</option>", ((adj & ADJ_DefComment) ? " selected " : " "));

		printf(LF "<option%svalue=\"1\">** No change</option>"
			   LF "<option%svalue=\"2\">Change comment:</option>",
			   (!(adj & ADJ_DefComment) && !(adj & ADJ_Comment)) ? " selected " : " ",
			   (!(adj & ADJ_DefComment) &&  (adj & ADJ_Comment)) ? " selected " : " ");

		printf("</select>&nbsp;&nbsp;<input type=\"text\" name=\"pl_%scomt\" value=\"", str);

		if (comment) encapsulate(comment, stdout);
		else if ((psp) && (psp->exist->comment)) encapsulate(psp->exist->comment, stdout);

		puts("\"></td></tr>");
	}

	/* Return most recent colour */
	if (pin) *pin = p;
}

int doact_event_hook(event_p ep, int match, void* x5, evolve__ctx_t* ectx)
{
	ppseud_p pseudp;
	pstate_p pstp;

	fprintf(stderr, "%li:   hook\n", uniqid);

	/* Continue to end of all time */
	ectx->endtime = TMAX;

	/* Don't call this hook again */
	ectx->flags = EVOLVE_OnlyPseuds;

	/* Note all pseudonyms so far */
	for (pstp = ectx->plist; (pstp); pstp = pstp->next)
	{
		/* Individual flags */
		pstp->state = EVOLVE_OnlyPseuds;

		/* Default at limit */
		pstp->ext = pstp->exist->dps;

		/* All pseudonyms for players */
		for (pseudp = pstp->allpseuds; (pseudp); pseudp = pseudp->link_pl)
		{
			/* Note prior to limit */
			pseudp->fl |= PSFL_XPrior;

			/* Note defaults at limit */
			if ((pstp->ext) && (!strcasecmp(pstp->ext->text, pseudp->text))) pseudp->fl |= PSFL_XDefault;

			fprintf(stderr, "%li:   hook: \"%s\" pseud=%p \"%.30s\" fl=&%X\n",
					uniqid, pstp->pid, pseudp, pseudp->text, pseudp->fl);
		}
	}

	/* Continue, and recheck term cond */
	return 0;
}

/***************************************************************************************************************************
** Add game events:                                                                                                       **
***************************************************************************************************************************/
int doact_event(player_p authp, char* ipids[], player_p pl[], int n, int state, unsigned int abits)
{
	int i, j, rc = 0, badcnt = 0, badplys, goodplys, admin;
	bad_actev_type *bads, badtype = 0;
	timmy evtim = thetim;
	unsigned long u, *rincs = 0, def_rinc = 0, *comchs = 0;
	long k, def_scb, *scbs = 0, def_boost, *boosts = 0, *police, def_police = -1, *wantsts = 0, def_wantst = 0;
	char *p, *q, **pseuds = 0, *sing = 0, **comments = 0, *def_comment;
	pstate_p *plst, plstlist;
	char *cp[2] = { RED, BLUE };
	adj_type *adjs = 0, def_adj = 0;
	evolve_flags eff;


	/* Handy macros to note when things go wrong */
	#define addbad(x) do { ++badcnt; badtype |= (x); } while (0)
	#define addbads(y,x) do { ++badcnt; bads[(y)] |= (x); } while (0)


	/* Have admin privs? */
	admin = (authp->flags & PF_Admin) != 0;

	/* Allocate an array for player states */
	if (!(plst = malloc(n * sizeof(pstate_p)))) erf(ERF_mallocFailed, 3);

	/* Allocate array for storing validity */
	if (!(bads = malloc(n * sizeof(bad_actev_type)))) erf(ERF_mallocFailed, 4);

	/* Array in which to store adjustments */
	if (!(adjs = malloc(n * sizeof(adj_type)))) erf(ERF_mallocFailed, 5);

	/* Is event time present? (min) */
	if (!(p = findval("actev_m", 0))) addbad(BAD_actev_MissingMin);

	/* If present, require valid event time (min) */
	else if ((u = strtoul(p, &p, 10)) > 59UL)
	{
		evtim.min = 59;
		addbad(BAD_actev_InvalidMin);
	}

	/* Save valid time (min) */
	else evtim.min = (char) u;
	
	/* Require valid event time: hour */
	if (!(p = findval("actev_h", 0))) addbad(BAD_actev_MissingHour);

	/* If present, require valid event time (hour) */
	else if ((u = strtoul(p, &p, 10)) > 23UL)
	{
		evtim.hour = 23;
		addbad(BAD_actev_InvalidHour);
	}
	
	/* Save valid time (hour) */
	else evtim.hour = (char) u;

	/* Require valid event time: day */
	if (!(p = findval("actev_d", 0))) addbad(BAD_actev_MissingDay);

	/* If present, require valid event time (day) */
	else if ((u = strtoul(p, &p, 10)) > ((admin) ? DAY_FullLast : DAY_Last))
	{
		evtim.day = (admin) ? DAY_FullLast : DAY_Last;
		evtim.min = 59;
		evtim.hour = 23;
		addbad(BAD_actev_InvalidDay);
	}

	/* Check both limits (day) */
	else if (u < ((admin) ? DAY_FullFirst : DAY_First))
	{
		evtim.day = (admin) ? DAY_FullFirst : DAY_First;
		evtim.min = 0;
		evtim.hour = 0;
		addbad(BAD_actev_InvalidDay);
	}

	/* Valid time (day) */
	else evtim.day = (char) u;

	/* All time elements missing */
	if (!((~badtype) & BAD_actev__MissingTime))
	{
		badtype |= BAD_actev__MissingTime;
		badcnt -= 3;
	}

	/* All time elements present and correct */
	else if (!badtype) state |= 1;

	/* Score bonuses */
	if (abits & ABIT_Bonus)
	{
		/* Default score bonus */
		def_scb = rint((1 << SCORE_SHIFT) * strtod(findval("pl_def_scb", 1), 0));

		/* Array in which to store bonuses */
		if (!(scbs = malloc(n * sizeof(long)))) erf(ERF_mallocFailed, 8); 
	}

	/* Activity deadlines */
	if (abits & ABIT_Deadline)
	{
		/* Default inactivity boost */
		def_boost = strtol(findval("pl_def_bst_h", 1), 0, 10) + 24 * strtol(findval("pl_def_bst_d", 1), 0, 10);

		/* Array in which to store boosts */
		if (!(boosts = malloc(n * sizeof(long)))) erf(ERF_mallocFailed, 9); 
	}

	/* Reincarnate players */
	if (abits & ABIT_Reincarnate)
	{
		/* Array in which to store new passwords and their validity */
		if (!(rincs = malloc(n * sizeof(unsigned long)))) erf(ERF_mallocFailed, 13);

		/* Default reincarnation action */
		def_rinc = (strtoul(findval("pl_def_rinc", 1), 0, 10) > 1);
	}

	/* Change police and ranks */
	if (abits & ABIT_Police)
	{
		/* Array in which to store new passwords and their validity */
		if (!(police = malloc(n * sizeof(long)))) erf(ERF_mallocFailed, 14);

		/* Acquire default (do nothing) */
		if ((p = findval("pl_def_pol", 0)) && (!streq(p, "x")))
		{
			/* Convert it */
			if ((k = strtol(p, 0, 10)) > MAX_rank) addbad(BAD_actev_DefPolice);

			/* Promotions */
			else if ((p[0] == '+') || (p[0] == '-'))
			{
				if (k < MAX_rank) addbad(BAD_actev_DefPolice);
				else { def_police = k; def_adj |= ADJ_Promotion; }
			}

			/* Absolute ranks */
			else
			{
				if (k < -1) addbad(BAD_actev_DefPolice);
				else { def_police = k; def_adj &= (~ADJ_Promotion); }
			}
		}
	}

	/* Defaults for changes in wantedness */
	if (abits & ABIT_Wanted)
	{
		/* Array in which to store criminality statuses */
		if (!(wantsts = malloc(n * sizeof(long)))) erf(ERF_mallocFailed, 11);

		/* Default state change */
		if (((k = strtol(findval("pl_def_wantch", 1), 0, 10)) >= 0) && (k <= 3)) def_wantst = k;

		/* Invalid - use 0 (do nothing) */
		else addbad(BAD_actev_DefWantSt);
	}

	/* Comments (case history) */
	if (abits & ABIT_Comment)
	{
		/* Array in which to store comments */
		if (!(comments = malloc(n * sizeof(char*)))) erf(ERF_mallocFailed, 12);

		/* Array in which to store comment changes */
		if (!(comchs = malloc(n * sizeof(unsigned long)))) erf(ERF_mallocFailed, 13);

		/* Default comment? */
		if (strtoul(findval("pl_def_comc", 1), 0, 10) > 1) def_adj |= ADJ_Comment;

		/* Comment string */
		def_comment = findval("pl_def_comt", 1);
	}

	/* No defaults for pseudonym changes */
	if (abits & ABIT_Pseudonym)
	{
		/* Array in which to store new passwords and their validity */
		if (!(pseuds = malloc(n * sizeof(char*)))) erf(ERF_mallocFailed, 14);
	}

	/* Initialise arrays */
	for (i = 0, badplys = goodplys = 0; (i < n); i++)
	{
	 	adjs[i] = 0;
	 	bads[i] = 0;
	 	plst[i] = 0;

		/* Score bonuses */
		if (abits & ABIT_Bonus)
		{
			/* Form "pl_7scb" string */
			if (sprintf(strb, "pl_%dscb", i+1) < 7) erf(ERF_InternalError, 20);

			/* Use default if none given */
			if (p = findval(strb, 0)) scbs[i] = rint((1 << SCORE_SHIFT) * strtod(p, 0));

			/* Use specific bonus */
			else { scbs[i] = def_scb; adjs[i] |= ADJ_DefBonus; }

			/* Ignore events with zero bonus */
			if (scbs[i]) adjs[i] |= ADJ_Bonus;
		}
	
		/* No defaults for pseudonym changes */
		if (abits & ABIT_Pseudonym)
		{
			/* Form "pl_7psn" string */
			if (sprintf(strb, "pl_%dpsn", i+1) < 7) erf(ERF_InternalError, 22);

			/* "New" pseudonym string */
			pseuds[i] = cleanpseud(findval(strb, 0));

			/* Form "pl_7pso" string */
			if (sprintf(strb, "pl_%dpso", i+1) < 7) erf(ERF_InternalError, 23);
	
			/* Do nothing? (k = 3) */
			if (!(p = cleanpseud(findval(strb, 0)))) k = 3;

			/* Add/set "new" pseud (k = 1 or 2) */
			else if ((k = streq(p, "-")) || (streq(p, "+"))) k += 1;

			/* Use selected pseud as new string (k = 0) */
			else pseuds[i] = p;

			/* Check validity of pseudonyms */
			if (k < 3)
			{
				/* Warn against missing pseud text */
				if (!pseuds[i]) addbads(i, BAD_actev_MissingPseud);

				/* Be careful... */
				else if (streq(pseuds[i], "-") || streq(pseuds[i], "+")) addbads(i, BAD_actev_BadPseud);

				/* Warn against overly long pseudonyms */
				else if (strlen(pseuds[i]) > MAX_PseudLength) addbads(i, BAD_actev_LongPseud);

				/* Valid */
				adjs[i] |= (k == 1) ? ADJ_AddPseud : ADJ_SetPseud;
			}
		}

		/* Reincarnation */
		if (abits & ABIT_Reincarnate)
		{
			/* Form "pl_7rinc" string */
			if (sprintf(strb, "pl_%drinc", i+1) < 8) erf(ERF_InternalError, 26);

			/* Specific setting */
			if ((p = findval(strb, 0)) && ((u = strtoul(p, 0, 10)) > 0)) rincs[i] = (u > 1);

			/* Default setting */
			else { rincs[i] = def_rinc; adjs[i] |= ADJ_DefReinc; }

			/* Reincarnate? */
			if (rincs[i]) adjs[i] |= ADJ_Reincarnate;

			fprintf(stderr, "%li:   reinc \"%s\" %lX rinc[%i]=%u", uniqid,
					pl[i]->id, adjs[i] & ADJ_Reincarnate, i, rincs[i]);
		}

		/* Police ranks */
		if (abits & ABIT_Police)
		{
			/* Form "pl_7rank" string */
			if (sprintf(strb, "pl_%dpol", i+1) < 7) erf(ERF_InternalError, 26);

			/* Default action */
			adjs[i] = (adjs[i] & (~ADJ_Promotion)) | ADJ_DefPolice | (def_adj & ADJ_Promotion);
			police[i] = def_police;

			/* Acquire specific setting */
			if (p = findval(strb, 0))
			{
				/* Do nothing */
				if (streq(p, "x")) { police[i] = -1; adjs[i] &= (~ADJ_Promotion|ADJ_DefPolice); }

				/* Convert it */
				else if ((k = strtol(p, 0, 10)) > MAX_rank) addbads(i, BAD_actev_BadPolice);
	
				/* Promotions */
				else if ((p[0] == '+') || (p[0] == '-'))
				{
					if (k < MAX_rank) addbads(i, BAD_actev_BadPolice);
					else { police[i] = k; adjs[i] = (adjs[i] & (~ADJ_DefPolice)) | ADJ_Promotion; }
				}
	
				/* Absolute ranks */
				else
				{
					if (k < 0) addbads(i, BAD_actev_BadPolice);
					else { police[i] = k; adjs[i] &= (~ADJ_Promotion|ADJ_DefPolice); }
				}
			}

			/* Changes? */
			if ((adjs[i] & ADJ_Promotion) || (police[i] >= 0)) adjs[i] |= ADJ_Police;
		}

		/* Activity deadlines */
		if (abits & ABIT_Deadline)
		{
			/* Form "pl_7bst_h" string */
			if (sprintf(strb, "pl_%ibst_%nh", i+1, &k) < 9) erf(ERF_InternalError, 27);

			/* Read strings (also "pl_7bst_d" string) */
			p = findval(strb, 0); strb[k] = 'd'; q = findval(strb, 0);

			/* Specific boost */
			if ((p) || (q)) 
			{
				boosts[i] = ((p) ? strtol(p, 0, 10) : 0L) + 24 * ((q) ? strtol(q, 0, 10) : 0L);
				adjs[i] &= (~ADJ_DefBoost);
			}

			/* Default boost */
			else { boosts[i] = def_boost; adjs[i] |= ADJ_DefBoost; }

			/* Set a deadline? */
			if (boosts[i]) adjs[i] |= ADJ_Deadline;
		}

		/* Changes in wantedness */
		if (abits & ABIT_Wanted)
		{
			/* Form "pl_7wantst" string */
			if (sprintf(strb, "pl_%dwantst", i+1) < 10) erf(ERF_InternalError, 29);

			/* Read string, convert */
			if (p = findval(strb, 0)) k = strtol(p, 0, 10);

			/* Use default */
			if ((!p) || (k < 0)) { wantsts[i] = def_wantst; adjs[i] |= ADJ_DefWantSt; }

			/* Valid status */
			else if (k < 4) wantsts[i] = k;

			/* Invalid status */
			else { wantsts[i] = 0; addbads(i, BAD_actev_BadWantSt); }

			/* Change wantedness? */
			if (wantsts[i] > 0) adjs[i] |= ADJ_WantSt;
		}

		/* Comments */
		if (abits & ABIT_Comment)
		{
			/* Form "pl_7comt" string */
			if (sprintf(strb, "pl_%dcom%nt", i+1, &k) < 8) erf(ERF_InternalError, 31);

			/* Save comment */
			comments[i] = findnameval(strb, 1+32, 0);

			/* Form "pl_7comc" string */
			strb[k] = 'c'; 

			/* Specified value */
			if ((u = strtoul(findval(strb, 1), 0, 10)) > 1)
			{
				/* Use it */
				adjs[i] |= ADJ_Comment;

				/* Don't give current comment */
				if (!comments[i]) comments[i] = "";
			}

			/* Default value */
			else if (u == 0)
			{
				/* Any valid such? */
				adjs[i] |= ADJ_DefComment | ((def_adj & ADJ_Comment) ? ADJ_Comment : 0);
			}
		}

		/* Count number of players with problems */
		if (bads[i]) ++badplys;

		/* Count number of players with valid changes */
		else if (adjs[i]) ++goodplys;
	}

	/* Debug message */
	fprintf(stderr, "%li:   doact_event one auth=\"%s\" abits=&%X n=%i of %i badcnt=%i badplys=%i badtype=&%X\n",
			uniqid, authp->id, abits, goodplys, n, badcnt, badplys, badtype);

	/* Enable hook for pseudonyms, and don't record unless required */
	eff = (abits & ABIT_Pseudonym) ? (EVOLVE_FuncEndCond|EVOLVE_FuncExhaust|EVOLVE_AllPseuds) : 0;

	/* Don't record comments unless required */
	if (abits & ABIT_Wanted) eff |= EVOLVE_Comments;

	/* Debug message */
	fprintf(stderr, "%li:   doact_event evolve eff=&%lX to %.2i:%.2i:%.2i.\n",
			uniqid, eff, evtim.day, evtim.hour, evtim.min);

	/* Evolve to get state at time of events */
	if (rc = evolve(eff|EVOLVE_InheritPseud|EVOLVE_TimersToLimit,
					ipids, &plst, &plstlist, 0,0,0,0,0,0, &evtim, 0,
					statbuf_pos, statbuf_end, &statbuf_pos, 0,0,0, doact_event_hook, pseuds, 0)) erf(rc, 45);	

	/* Valid date? */
	if (state & 1)
	{
		/* ... ... */
	}

	/* ... More checks ... */
	
	/* Debug message */
	fprintf(stderr, "%li:   doact_event two auth=\"%s\" abits=&%X n=%i of %i badcnt=%i badplys=%i badtype=&%X\n",
			uniqid, authp->id, abits, goodplys, n, badcnt, badplys, badtype);

	/* Carry out the changes, if valid */
	if ((state == 3) && (!badcnt) && (goodplys > 0))
	{
		editev_p nevs, nevp;
		unsigned long *fls;
		timmy *tims;

		/* Allocate memory for timers */
		if (!(tims = malloc(goodplys * sizeof(timmy)))) erf(ERF_mallocFailed, 49);

		/* Allocate memory for event addition structs */
		if (!(nevs = malloc(goodplys * sizeof(struct editev)))) erf(ERF_mallocFailed, 50);

		/* Allocate memory for flags */
		if (!(fls = malloc(goodplys * sizeof(unsigned long)))) erf(ERF_mallocFailed, 51);

		/* Fill them in */
		for (i = j = 0; (i < n); i++)
		{
			/* Event flags */
			fls[j] = 0;

			/* No changes */
			if (!adjs[i]) continue;

			/* Paranoia: don't exceed buffer */
			if (j >= goodplys) erf(ERF_InternalError, 52);

			/* Initialise pointer and default timer */
			nevp = &nevs[j];
			tims[j] = thetim;

			/* Link into list */
			if (j > 0) nevs[j-1].next = nevp;

			/* End of list */
			nevp->next = 0;
			nevp->link = 0;

			/* Initialise blank struct */
			init_editev_struct(nevp);

			/* Add new event, chronologically */
			nevp->at = evtim;
			nevp->rid = 0;
			nevp->mode = 0;
			nevp->npid = pl[i]->id;
			nevp->nref = 0;
			nevp->oldref = 0;
			nevp->nflsp = &fls[j];
			nevp->ntimp = &tims[j];

			/* Score bonuses */
			if (adjs[i] & ADJ_Bonus)
			{
				fls[j] |= EF_ScoreBonus;
				nevp->nbonusp = &scbs[i];
			}

			/* Pseudonyms */
			if (adjs[i] & (ADJ_AddPseud|ADJ_SetPseud))
			{
				nevp->npseud = pseuds[i];
				fls[j] |= (adjs[i] & ADJ_SetPseud) ? EF_SetPseudonym : EF_AddPseudonym;
			}

			/* Reincarnation */
			if (adjs[i] & ADJ_Reincarnate) fls[j] |= EF_Reincarnate;

			/* Police rank */
			if (adjs[i] & ADJ_Police)
			{
				if (adjs[i] & ADJ_Promotion) fls[j] |= EF_Promotion;
				fls[j] |= EF_Police;
				nevp->nrankp = &police[i];
			}

			/* Deadlines */
			if (adjs[i] & ADJ_Deadline)
			{
				*(nevp->ntimp) = addhours(&evtim, (int)boosts[i]);
				fls[j] |= EF_SetInactTimer;
			}

			/* Deadlines */
			if (adjs[i] & ADJ_Comment)
			{
				nevp->nwanted = (adjs[i] & ADJ_DefComment) ? def_comment : comments[i];
				fls[j] |= EF_Wantedness;
			}

			/* Wanted state */
			if (adjs[i] & ADJ_WantSt)
			{
				switch (wantsts[i])
				{
					case 1: fls[j] |= EF_MakeWanted|EF_MakeCompetent; break;
					case 2: fls[j] |= EF_MakeWanted|EF_MakeIncompetent; break;
					case 3: fls[j] |= EF_MakeUnWanted|EF_MakeCompetent; break;
				}
			}

			/* Structure built */
			j++;
		}

		/* Paranoia: check number correspond */
		if (j != goodplys) erf(ERF_InternalError, 54);

		/* Form comment */
		if ((statbuf_end - (p = statbuf_pos)) > (96 + MAX_id))
		{
			/* Comment string */
			j = sprintf(p, " Added %i action events at realtime %.2i:%.2i:%.2i by auth \"%.*s\":",
						goodplys, thetim.day, thetim.hour, thetim.min, (int)MAX_id, authp->id);
	
			/* Check string length */
			if ((j < 20) || (j > (96+MAX_id))) p = 0;
	
			/* Advance buffer position */
			else statbuf_pos += 1 + j;
		}
	
		/* No comment */
		else p = 0;

		/* Submit them */
		if (i = editevents(0, &nevs[0], &statbuf_pos, statbuf_end, p, 0)) erf(i, 55);

		/* Success string */
		dolaunch = "<h3><hr><img src=\""TICK_URL"\" align=middle "TICK_SIZ" hspace=8>"
				   "<font color=#FFCCCC>Yay! Dunnit!</font><hr></h3>";

		/* Automatically rebuild news? */
		if ((!admin) || (findval("rebuild", 0))) rescore |= RESC_MkNews;

		/* Maintain current status */
		rescore |= RESC_Scores;

		/* Success! */
		return 0;
	}

	/* Title */
	p = (abits == ABIT_Pseudonym) ? "Add/change pseudonym" : "Add player events";

	/* Only one field per player, or more than one? */
	if (abits == ABIT_Bonus) sing = "Bonus score";
	else if (abits == ABIT_Deadline) sing = "New inactivity deadline";
	else if (abits == ABIT_Reincarnate) sing = "Reincarnate";
	else if (abits == ABIT_Police) sing = "Status and rank";
	else if (abits == ABIT_Pseudonym) sing = "New (or old) pseudonym";
	else if (abits == ABIT_Wanted) sing = "New criminality";
	else if (abits == ABIT_Comment) sing = "New comments";

	/* Otherwise, output HTML form */
	printf("Content-type: text/html" LF LF "<HTML><HEAD><TITLE>"GAME" - %s</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr>"GAME" - %s<hr></h1>"
		   LF "<form method=\"POST\" action=\""CGI_URL"pled.cgi\">"
		   LF "<input type=\"hidden\" name=\"stat_rest\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_vit\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_sort\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"stat_law\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"act\" value=\"%i\">"
		   LF "<input type=\"hidden\" name=\"abits\" value=\"%X\">",
		   p, p ,findval("stat_rest", 1), findval("stat_vit", 1),
		   findval("stat_sort", 1), findval("stat_law", 1), ACT_Multi, abits);

	/* Print authentication information */
	printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\"><p>Authenticated as <b>%s</b>%s.</p>"
		   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">",
		   authp->id, authp->name, ((authp->flags & PF_Admin) ? " (admin)" : ""), authp->pass);

	if (badcnt) printf(LF "<h3><hr><img src=\""CROSS_URL"\" align=middle "CROSS_SIZ" hspace=8>"
					   LF "<font color=#FFCCCC>Invalid events: %i problem%s found: &%lX"
					   LF "Please correct.</font><hr></h3>", badcnt, (badcnt > 1) ? "s" : " ", badtype);

	else if (state == 3) printf(LF "<p><hr><img src=\""SKULL_URL"\" align=left "SKULL_SIZ" hspace=8>"
								LF "<font color=#FFCCCC>No changes have been specified.</font><hr clear=left></p>");

	puts("<p>Time at which to apply changes:&nbsp;");

	/* Selector for time and date of event */
	printdatesel("actev_d", "actev_h", "actev_m", &evtim, admin);

	puts("&nbsp;<input value=\"Show player info for that time\" type=\"submit\" name=\"sub_etim\"></p>");

	/* Initial colour index */		
	j = 0;

#if 0
	/* Special display for single player pseud changes */
	if (0 /*(n == 1) && (abits == ABIT_Pseudonym)*/)
	{
	}

	/* Normal admin/multiplayer display */
	else
#endif
	{
		/* Defaults, if more than one player */
		if ((n > 1) && (abits != ABIT_Pseudonym))
		{
			/* Title for table of defaults */
			puts(LF "<h3>&quot;Default&quot; action:</h3>"
				 LF "<p><table cols=2 cellpadding=4 cellspacing=4 border=0 "/*width=100%%*/">");
	
			/* Print default form selectors */
			doact_event_aux(0, "def_", &j, 0, cp, abits & (~ABIT_Pseudonym), def_adj,
							def_rinc, def_police, def_boost, def_scb, 0, def_wantst, def_comment, 0);
	
			/* Footer for table of defaults */
			puts("</table></p>" LF"<h3>Player-specific actions:</h3>"
				 LF "<p>Use these to override the default settings.</p>");
		}

		/* Table header and titles - single field */
		if (sing) printf(LF "<p><table cols=2 border=0 cellpadding=4 cellspacing=4 "/*width=100%%*/">"
						 LF "<thead><td bgcolor="GREEN">Player:</td><td bgcolor="GREEN">%s:</td></thead>", sing);
	
		/* Table header: multi-field layout */
		else printf(LF "<p><table cols=1 border=0 cellpadding=4 cellspacing=4 width=100%%>");
	
		/* Deal with each player in turn */
		for (i = 0; (i < n); i++)
		{
			/* Form player label */
			sprintf(strb, "%i", i+1);
	
			/* Begin row for each player */
			if (sing) printf(LF "<tr bgcolor=%s><td><b>", p = cp[(j++) & 1]);
	
			/* Block title for each player */
			else printf(LF "<tr><td bgcolor="GREEN"><h4>&nbsp;");
	
			/* Player id and name */
			printf(LF "<input type=\"hidden\" name=\"pl_%did\" value=\"%s\">"
				   LF "<font color="WHITE"><a href=\"plinfo.cgi?id=%s\">%s</a></font>",
				   i+1, pl[i]->id, pl[i]->id, pl[i]->name);
	
			/* Block for each player */
			if (!sing) puts("</h4><br><table bgcolor=#000000 border=0 cellspacing=2 cellpadding=4 cols=2 width=100%>");
				   
			/*fprintf(stderr, "%li:   eek %i %i %s", uniqid, i, n, pl[i]->id);*/

			/* Print form content */
			doact_event_aux(1, strb, ((sing) ? 0 : &j), &p, cp, abits, adjs[i],
							(rincs) ? rincs[i] : 0, (police) ? police[i] : 0,
							(boosts) ? boosts[i] : 0, (scbs) ? scbs[i] : 0,
							(pseuds) ? pseuds[i] : 0, (wantsts) ? wantsts[i] : 0,
							(comments) ? comments[i] : 0, plst[i]);
				   
			/* End block */
			if (!sing) puts("</table></td></tr>");
		}
	
		/* Table footer and submission button */
		printf(LF "</table></p><p><input value=\"Add events\" type=\"submit\" name=\"sub_pled\">"
			   "<input value=\"Check, but don't add yet\" type=\"submit\" name=\"sub_chk\"></p>");
	}

	/* Offer admin the chance to not automatically rebuild news */
	if (admin) printf(LF "<p>Automatically rebuild news:&nbsp;&nbsp;"
					  "<input type=\"checkbox\" name=\"rebuild\"%s value=\"set\"></p>",
					  findval("rebuild", 0) ? " checked" : " ");

	/* Footer */
	printf(LF "</form><p><hr>"FOOTER" &mdash; Produced at %s</p></BODY></HTML>" LF, timestr);


_erf: /***** Handle errors *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   doact_event(%i):%i abits=&%X state=%i\n", uniqid,
					ERF_CONTEXT(rc), ERF_TYPE(rc), abits, state);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Main code. Doesn't make use of command line arguments.                                                                 **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	char *authname = "", *authpass = "", *ipids[MAX_CheckAct+2], *sec = 0;
	int i, n = 0, rc = 0, lockfd = 0, action = ACT_None, state = 0;
	player_p pptr, authp = 0, *plp = 0, whichpptr = 0;
	pshort_p pshlist = 0;
	struct tm* the_tm;
	unsigned int abits = 0;
	scanplinfo_flags scanflags;


	/* What time is it? */
	time(&thetime);
	the_tm = localtime(&thetime);
	thetim.day  = the_tm->tm_yday - StartingYDay;
	thetim.hour = the_tm->tm_hour;
	thetim.min  = the_tm->tm_min;
	timestr = ctime(&thetime);

	/* Unbuffered */
	setbuf(stdout, 0);

	/* Output HTTP headers */
	/* printf("Content-type: text/html" LF LF); */

	/* File creation umask */
	umask(0117);

	/*printf("Content-type: text/html" LF LF);
	while ((i = getchar()) != EOF) putc(i, stdout);
	printf("\n\nfoo\n\n");*/

	/* Make note in log and redirect stderr there */
	if (rc = seterrlog("Pled.c", timestr)) exit(send_doc(rc, 0, 0));

	/* Parse CGI input */
	if ((i = parsecgi()) > 0) erf(i, 1);

	/* No input given */
	else if (i == 0) erf(ERF_Void, 2);

	/* If no authentication name given, assume public user */
	if (!(authname = findval("authname", 0))) authname = PUB_ID;

	/* Authentication password */
	else authpass = findval("authpass", 1);

	/* Force pseudonym action */
	if (findval("sub_pseud", 0)) action = ACT_Pseudonym;

	/* Force password action */
	else if (findval("sub_password", 0)) action = ACT_Password;

	/* Was an action specified? */
	else action = (int)strtol(findval("act", 1), 0, 10);

	/* Invalid action? */
	if ((action <= ACT_None) || (action >= ACT__LIM)) erf(ERF_UnknownAction, 3);

	/* Use action bitfield? */
	if (action == ACT_Multi)
	{
		/* Were action bits specified? */
		abits = (unsigned int)strtoul(findval("abits", 1), 0, 16);

		/* Filter bits */
		abits &= ABIT__MASK_Multi;

		/* If only single bit set, covert back to action number */
		if (abits == ABIT_Pseudonym) action = ACT_Pseudonym;

		/* If empty, assume all bits set */
		else if (!abits) abits = ABIT__MASK_Multi;
	}

	/* Set bitfield appropriately */
	else abits = ACT_BIT(action) & ABIT__MASK_Multi;

	/* No player ids other than authentication ones */
	ipids[n = 0] = authname;

	/* Form array of player ids */
	while (1)
	{
		/* Form "pl_1id" string */
		if (sprintf(strb, "pl_%did", ++n) < 6) erf(ERF_InternalError, 4);

		/* End of search */
		if (!(ipids[n] = findval(strb, 0))) break;

		/* Check array bounds, and sanity of query */
		if (n >= (MAX_CheckAct + 1)) erf(ERF_TooManyPlayers, 5);

		/* Check for duplicates */
		for (i = 1; (i < n); i++) if (streq(ipids[n], ipids[i])) erf(ERF_DuplicateIds, 6);
	}

	/* Terminate array */
	ipids[n] = 0;

	/* No players */
	if (n <= 1) erf(ERF_NoIdGiven, 10);

	/* First presentation of form? */
	state = findval("sub_pled", 0) ? 2 : 0;

	/* Claim lockfile */
	if (!(lockfd = setflock("Pled.c "))) erf(ERF_flockFailed, 11);

	/* Read targets, sometimes */
	scanflags = SCANPL_AddShort | (action == ACT_Retarget) ? SCANPL_GetTargs : 0;

	/* Read player info */
	if (rc = scanplinfo(scanflags, ipids, &plp, &pshlist, 0, statbuf, statbuf_end, &statbuf_pos)) erf(rc, 12);

	/* Check allocation and get authentication player info */
	if ((!plp) || !(authp = plp[0])) erf(ERF_InternalError, 13);

	/* Invalid authentication? */
	if ((!authp->name) || !(authp->flags & PF_Login) || (!(authp->flags & PF_Public) && !streq(authp->pass, authpass)))
	{
		authp = 0;
		erf(ERF_BadAuth, 14);
	}

	/* Public authentication? */
	else if (authp->flags & PF_Public)
	{
 		if (n == 2) suggestid = ipids[1];
		erf(ERF_BadPubAuth, 15);
	}

	/* Players can edit their own passwords or pseudonyms */
	i = (action == ACT_Password) || ((OPT_ChangePseud) && (action == ACT_Pseudonym));

	/* Check permissions */
	if ((!i) || (n != 2) || (!streq(authp->id, ipids[1])))
	{
		/* Non-admin types can only act on theirselves */
		if (!(authp->flags & PF_Admin))
		{
			if (n == 2) suggestid = ipids[1];
			erf(ERF_NoPermission, 16);
		}

		/* Additional security when editing other players' passwords or user info */
		if ((action == ACT_Password) || (action == ACT_EditUser))
		{
			/* Point at a (possibly null) string whenever password might be needed */
			sec = findval("pwsecure", 1);
	
			/* Enforce only when submitting changes */
			if ((state) && (strcmp(sec, PWSECURE))) erf(ERF_PassSecurity, 17);
		}
	}

	/* Check for unknown players */
	for (i = 1; (i < n); i++) if (!(pptr = plp[i]) || (!pptr->name)) erf(ERF_UnknownPlayer, 18);

	/* Acting on single player? */
	whichpptr = (n == 2) ? plp[1] : 0;

	/* Debug */
	fprintf(stderr, "%li:   main act=%d on n=%i under auth \"%s\".\n", uniqid, action, n, authp->id);

	/* Set passwords */
	if (action == ACT_Password) rc = doact_password(authp, &plp[1], n - 1, state, sec);

	/* Edit users */
	else if (action == ACT_EditUser) rc = doact_edituser(authp, &plp[1], n - 1, state);

	/* Retargetting */
	else if (action == ACT_Retarget) rc = doact_retarget(authp, &plp[1], n - 1, state);

	/* Add events */
	else rc = doact_event(authp, &ipids[1], &plp[1], n - 1, state, abits);

	/* Sucessful, finished, launch stats page */
	if ((!rc) && (dolaunch))
	{
		/* Prepare temporary file for chained stdin */
		if (!(fplaunch = prepforstdin())) erf(ERF_tmpfileFailed, 20);

		/* Encode special display type */
		urlencode(fplaunch, "sub_show", "AutoGen", 1);

		/* Include players in special stats display afterwards */
		for (i = 1; (i < n); i++)
		{
			fputs("&id_", fplaunch);
			urlencode(fplaunch, plp[i]->id, "show", 1);
		}

		/* Encode old presentation method */
		urlencode(fplaunch, "sort",     findval("stat_sort", 1), 0);
		urlencode(fplaunch, "rest",     findval("stat_rest", 1), 0);
		urlencode(fplaunch, "law",      findval("stat_law",  1), 0);
		urlencode(fplaunch, "vitality", findval("stat_vit",  1), 0);

		/* Encode action type */
		fprintf(fplaunch, "&action=%i&abits=%X", action - 1, abits);

		/* Encode authname */
		urlencode(fplaunch, "authname", authp->id, 0);

		/* Encode password */
		urlencode(fplaunch, "authpass", authp->pass, 0);

		/* Encode "successful result" string */
		urlencode(fplaunch, "result", dolaunch, 0);
	}

	/* Lose lock and run "scoring" or "mknews" */
	if ((!rc) && ((rescore) || (fplaunch)))
	{
		/* Release the lock */
		setfunlock(lockfd); lockfd = 0;

		/* Rescore if necessary */
		if (rescore & RESC_Scores) system(SYS_Scores" 2> "FILE_ScoreLog" 1>&2 &");

		/* Rebuild news if necessary */
		if (rescore & RESC_MkNews) system(SYS_MkNews" 2> "FILE_NewsLog" 1>&2 &");
	}

	/* Take off! */
	if ((!rc) && (fplaunch)) erf(launchwithstdin(CGI_ROOT "stats.cgi", fplaunch), 30);


_erf: /***** Error handling *****/

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Debug */
	if (rc) fprintf(stderr, "%li:   main(%i):%i authname=\"%s\" authpass=\"%s\" %s act=%i abits=&%X\n",
					uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), authname, authpass, ((authp) ? "good " : ""), action, abits);

	/* Only retain low bits of normal rc's */
	if (rc) send_doc(ERF_LOWBITS(rc), authp, whichpptr);

	return 0;
}

/***************************************************************************************************************************
** Generate various special documents.                                                                                    **
***************************************************************************************************************************/
int send_doc(int w, player_p authp, player_p pptr)
{
	int i, rc = 0;
	char *p, *q;


	/* Output HTTP headers */
    printf("Content-type: text/html" LF LF);

	/* Which error? */
	switch (w)
	{
		/* Pass on */
		default: error_doc(w); return 0;

		case ERF_BadAuth:
		p = "Bad authentication";
		q = "<P>Please enter your password correctly.</p>";
		break;

		case ERF_BadPubAuth:
		p = "Not authenticated";
		q = "<P>Please select yourself from the list and enter your password correctly.</p>";
		break;

		case ERF_NoPermission:
		p = "Insufficiently priviledged authentication";
		q = "<P>Please select a more appropriate user from the list and enter the correct password.</p>";
		break;

		case ERF_PassSecurity:
		p = "Wrong security code";
		q = "<P>Please type in the correct security code, and reenter your password.</p>";
		break;
	}

	printf(LF "<HTML><HEAD><TITLE>" GAME " Assassins' Automatic Umpire: Error</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr>Error: Unable to complete request<hr></h1>"
		   LF "<h3><img src=\"" CROSS_URL "\" align=left " CROSS_SIZ " hspace=8>"
		   LF "<font color=#FFCCCC>%s</font><br clear=left></h3>"
		   LF "<form method=\"POST\" action=\"" CGI_URL THIS_CGI "\">%s"
		   LF "<p><table cols=2 border=0 cellpadding=0 cellspacing=4>", p, q);

	i = ((authp) && (authp->flags & PF_Observer)) ? 0 : OPT_IDMenuMode;

	if (pptr) printf(LF "<tr><td align=right>Acting on:&nbsp;</td><td><b>%s</b></td></tr>",
					 (!i) ? pptr->name : (pptr->pseud) ? pptr->pseud : pptr->id);

	puts(LF "<tr><td align=right>Authenticate as:&nbsp;</td><td><select size=1 name=\"authname\">");

	p = (suggestid) ? suggestid : (authp) ? (authp->id) : PUB_ID;

	if (i = makeidmenu(p, PF_Public|PF_Login, PF_Login, i))
	{
		fprintf(stderr, "%li:   rc %d from makeidmenu(3) in send_doc(%i)\n", uniqid, i, w);
	}

	puts("</select></td></tr>" LF "<tr><td align=right>Password:&nbsp;</td>"
		 LF "<td><input name=\"authpass\" type=\"password\" size=24 maxlength=64></td></tr>");

	if (!(i = (w != ERF_PassSecurity)))
	{
		printf(LF "<tr><td align=right>Security code:&nbsp;</td>"
			   LF "<td><input type=\"password\" name=\"pwsecure\" value=\"");
		encapsulate(findval("pwsecure", 1), stdout);
		printf("\"></td></tr>");
	}

	regurgitate(excl_auth + i);

	puts(LF "<tr><td></td><td><input value=\"(Re)authenticate and resubmit\" type=\"submit\" name=\"reauth\">"
		 LF "</td></tr></table></p></form><hr><p>" FOOTER "<hr></p></body></html>");


_erf: /***** Error handling *****/

	if (rc) fprintf(stderr, "%li:   send_doc(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));
	return 0;
}

/*** The End ***/
