/*****************************************************
**  lib-general.c                        (library)  **
**  ----------------------------------------------  **
**  Updated to use new time system.                 **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifdef included__lib_general
 #error "** lib-general multiply #included **"
#else
 #define included__lib_general fromfile

#define MALLOC_CHECK_ 1

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>

#include "system.h"
#include "typedefs.h"

/***************************************************************************************************************************
** Some function definitions                                                                               **
***************************************************************************************************************************/

int gextract(char** fromptr, char** toptr, char* toend, char** name, char** val);
int pextract(int* length, char** toptr, char* toend, char** name, char** val);
int tread(FILE* fp, char* to, char* end, char** tonext);
int thopp(FILE* fp, int* howmuch);
void encapsulate(char* p, FILE* fp);
void htmlify(char* p, FILE* fp);
void toeol(FILE* fp);
void error_doc(int which);

/***************************************************************************************************************************
** Global variables, tables, definitions and declarations                                                                 **
***************************************************************************************************************************/

/* Invoked actions */
#define ACT_None			0
#define ACT_Password		1
#define ACT_EditUser		2
#define ACT_Wanted			3
#define ACT_Police			4
#define ACT_Reincarnate		5
#define ACT_Retarget		6
#define ACT_Pseudonym		7
#define ACT_Bonus			8
#define ACT_Deadline		9
#define ACT_Comment			10
#define ACT_Multi			11
#define ACT__LIM			12

/* Size of action table */
#define MAX_act				11

/* Invoked action descriptions */
char* acttab[MAX_act] =
{
	"Change password",
	"Edit player information",
	"Alter wantedness",
	"Change police rank",
	"Reincarnate as player",
	"Maximise targetting",
	"Add/Change pseudonym",
	"Award bonus points",
	"Set inactivity deadline",
	"Change player comments",
	"Edit player status"
};

extern char **environ;

const char* pltypes[16] =
{
	"Player", "***(Error)***", "Wanted Criminal", "Wanted Incompetent",
	"Police", "***(Error)***", "Corrupt Police", "Incompetent Police",
	"Special", "?", "?", "?", "?", "?", "?", "?"
};
const char* daynames[7] =
{
		"Friday", "Saturday", "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday"
};
const char* shrtdaynames[9] =
{
		"XX", "Thurs", "Fri", "Sat", "Sun", "Mon", "Tues", "Wed", "ZZ"
};
const char* entity_table[256] =
{
		"","","","","","","","","","",0,"","",0,"","","","","","","","","","","","","","","","","","",
		0,0,"&quot;",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,"&lt;",0,"&gt;",0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "",
		0,0,0,0,0,0,0,0,0,0,0,0,0,"&trade;",0,0,0,0,0,0,"&lquot;","&rquot;",0,0,"&mdash;","&ndash;",0,0,0,0,0,0,
		"&nbsp;",  "&iexcl;", "&cent;",  "&pound;", "&curren", "&yen;",   "&brvbar;","&sect;",
		"&uml;",   "&copy;",  "&ordf;",  "&laquo;", "&not;",   "&shy;",   "&reg;",   "&macr;",
		"&deg;",   "&plusmn;","&sup2;",  "&sup3;",  "&acute;", "&micro;", "&para;",  "&middot;",
		"&cedil;", "&sup1;",  "&ordm;",  "&raquo;", "&frac14;","&frac12;","&frac34;","&iquest;",
		"&Agrave;","&Aacute;","&Acirc;", "&Atilde;","&Auml;",  "&Aring;", "&AElig;", "&Ccedil;",
		"&Egrave;","&Eacute;","&Ecirc;", "&Euml;",  "&Igrave;","&Iacute;","&Icirc;", "&Iuml;",
		"&ETH;",   "&Ntilde;","&Ograve;","&Oacute;","&Ocirc;", "&Otilde;","&Ouml;",  "&times;",
		"&Oslash;","&Ugrave;","&Uacute;","&Ucirc;", "&Uuml;",  "&Yacute;","&para;",  "&szlig;",
		"&agrave;","&aacute;","&acirc;", "&atilde;","&auml;",  "&aring;", "&aelig;", "&ccedil;",
		"&egrave;","&eacute;","&ecirc;", "&euml;",  "&igrave;","&iacute;","&icirc;", "&iuml;",
		"&eth;",   "&ntilde;","&ograve;","&oacute;","&ocirc;", "&otilde;","&ouml;",  "&divide",
		"&oslash;","&ugrave;","&uacute;","&ucirc;", "&uuml;",  "&yacute;","&THORN;", "&yuml;"
};

char* nullstr = "";
long uniqid = 0;

time_t epochtime;
timeval thetime = 0;

#ifdef MAX_fpbuf
 char fpbuf[MAX_fpbuf], *fpend = fpbuf + MAX_fpbuf;
#endif

/* Include CGI library, if needed */
#ifndef NO_CGI
 #ifndef included__lib_cgi
  #include "!lib-cgi.c"
 #endif
#endif

/***************************************************************************************************************************
** File format version checking code.                                                                                     **
***************************************************************************************************************************/
int chkheader(FILE *fp, int chkt, char *bp, int* verp, long *pfpos, unsigned long *parmp, long *comend)
{
	size_t i, s;
	char *p;
	int c;


	/* Initialise version number */
	if (verp) *verp = CHK_NoHeader;

	/* Ignore comments prior to header */
	while (((c = fgetc(fp)) == '#')) toeol(fp);

	/* Check for file errors */
	if (ferror(fp)) return ERF_FileError;
	if (feof(fp)) return ERF_EndOfFile;

	/* Position of beginning of header */
	if ((comend) && ((*comend = ftell(fp) - 1) == -2L)) return ERF_ftellFailed;

	/* Header must start with '!' character */
	if ((bp[0] = c) != '!') return (verp) ? 0 : ERF_NoHeader;

	/* Read input file header */
	for (p = bp + 1, i = chk[chkt].len - 1; (i > 0); i -= s, p += s)
	{
		if ((s = fread(p, 1, i, fp)) > i) return ERF_Paranoia;
		if (ferror(fp)) return ERF_FileError;
		if (feof(fp)) return ERF_EndOfFile;
	}

	/* Check first few characters against string */
	if (strncasecmp(bp, chk[chkt].str1, s = strlen(chk[chkt].str1))) return ERF_BadHeader;

	/* Read exactly (CHK_DIG_ver) decimal digits of format version */
	for (p = bp + s, s = 0, i = 0; (i < CHK_DIG_ver); i++)
	{
		if ((*p >= '0') && (*p <= '9')) s = (s * 10) + (*(p++) - '0');
		else return ERF_BadHeader;
	}

	/* Store the version number, if requested */
	if (verp) *verp = (int) s;

	/* Otherwise, complain if doesn't match the current version */
	else if (s != CHK_VER) return ERF_WrongFormat;

	/* Check next few characters against second string */
	if (strncasecmp(p, chk[chkt].str2, i = strlen(chk[chkt].str2))) return ERF_BadHeader;

	/* Extra parameter? */
	if ((chkt == CHK_events) || (chkt == CHK_reps))
	{
		unsigned long parm = 0;

		/* Read exactly (CHK_DIG_parm) decimal digits of parameter */
		for (p += i, i = 0; (i < CHK_DIG_parm); i++)
		{
			if ((*p >= '0') && (*p <= '9')) parm = (parm * 10) - '0' + *(p++);
			else return ERF_BadHeader;
		}

		/* Save value */
		if (parmp) *parmp = parm;

		/* Check next few characters against third string */
		if (strncasecmp(p, chk[chkt].str3, i = strlen(chk[chkt].str3))) return ERF_BadHeader;
	}

	/* Sanity check */
	if ((p - bp) + i != chk[chkt].len) return ERF_InternalError;

	/* Position of data immediately after header */
	if ((pfpos) && ((*pfpos = ftell(fp)) == -1L)) return ERF_ftellFailed;

	/* Header is valid */
	return 0;
}

int prtheader(FILE *fp, int chkt, unsigned long parm)
{
	int i;

	switch (chkt)
	{
		case CHK_uinfo:
		{
			i = fprintf(fp, "%s%.2i%s", chk[chkt].str1, CHK_VER, chk[chkt].str2);
			break;
		}
		case CHK_events:
		case CHK_reps:
		{
			i = fprintf(fp, "%s%.2i%s%."STRING(CHK_DIG_parm)"lu%s",
						chk[chkt].str1, CHK_VER, chk[chkt].str2, parm, chk[chkt].str3);
			break;
		}
		default: return ERF_InternalError;
	}
	return (chk[chkt].len == i) ? 0 : ERF_printfFailed;
}

/***************************************************************************************************************************
** Low-level file analysis/input routines. The toeol() function skips all characters until EOL or EOF is reported.        **
** The tread() function skips initial tabs, and reads subsequent data into the buffer delimited by (to),(end) until a     **
** tab, EOL or EOF is encountered. (*tonext) is pointed to the character following the terminator of the buffered data.   **
** It returns (0) if data read up to tab, (1) if data was read up to EOL/EOF, (2) if no data was read, (3) on error.      **
** thopp() is as tread, but does not store the data; if non-zero, (*howmuch) is set to required buffer size for data.     **
** The toeolpos() function is as toeol(), but stores the file position of the EOL marker at its second arg (if non-zero). **
***************************************************************************************************************************/
int tskip(FILE* fp)
{
	int c;

	/* Skip initial tabs */
	while ((c = fgetc(fp)) == '\t');

	/* Error, or EOF? */
	if (c == EOF) return (ferror(fp)) ? 3 : 2;

	/* Put back the excess char */
	return (EOF == ungetc(c, fp)) ? 3 : (c == '\n') ? 1 : 0;
}
int tread(FILE* fp, char* to, char* end, char** tonext)
{
	int c;

	/* Skip initial tabs */
	while ((c = fgetc(fp)) == '\t');

	/* No data to read */
	switch (c)
	{
		case '\n' : ungetc(c, fp);
		case EOF  : return 2;
	}

	/* Read data into buffer */
	while ((c != EOF) && (c != '\n') && (c != '\t') && (to < end))
	{
		*(to++) = c;
		c = fgetc(fp);
	}

	/* Buffer exhaustion */
	if (to >= end) return 3;

	/* Terminate */
	*to = '\0';

	/* Remaining free space in buffer */
	if (tonext) *tonext = to + 1;

	/* Tidy up */
	switch (c)
	{
		case '\t' : return 0;
		case '\n' : ungetc(c, fp);
		case EOF  : return 1;
	}

	/* This shouldn't ever happen */
	return 3;
}
int thopp(FILE* fp, int* howmuch)
{
	int c, n = 0;

	/* Skip initial tabs */
	while ((c = fgetc(fp)) == '\t');

	/* No data to read */
	switch (c)
	{
		case '\n' : ungetc(c, fp);
		case EOF  : return 2;
	}

	/* Read data into buffer */
	while ((c != EOF) && (c != '\n') && (c != '\t')) { n++; c = fgetc(fp); }

	/* How large a buffer would be needed? */
	if (howmuch) *howmuch = n + 1;

	/* Tidy up */
	switch (c)
	{
		case '\t' : return 0;
		case '\n' : ungetc(c, fp);
		case EOF  : return 1;
	}

	/* This shouldn't ever happen */
	return 3;
}
void toeol(FILE* fp)
{
	int c;
	while (('\n' != (c = fgetc(fp))) && (c != EOF));
}
int toeolpos(FILE* fp, long* posp)
{
	int c;
	while (('\n' != (c = fgetc(fp))) && (c != EOF));
	if (posp)
	{
		if (-1L == (*posp = ftell(fp))) return ERF_ftellFailed;
		if (c == '\n') --(*posp);
	}
	if ((c == EOF) && (ferror(fp))) return ERF_FileError;
	return 0;
}

/***************************************************************************************************************************
** Optimised file-to-file byte copier: does (len) bytes from (fin) to (fout) using a buffer (bp) of size (bsiz),          **
** optionally seeking first to (inseek) in (fin) if (inseek) != -1. There is a check facility for those of a paranoid     **
** disposition: if (incheck) != -1, the file position of (fin) after use must match the value of (incheck).               **
** If (len) is -1, then bytes are copied until EOF is reached.                                                            **
***************************************************************************************************************************/
int ftofcpy(FILE* fin, FILE* fout, long len, char* bp, long bsiz, long inseek, long incheck)
{
	long left;
	size_t dec, ltmp;


	/* Debug: Print parameters */
	/*fprintf(stderr, "%li:   ftofcpy(len=%li,bsiz=%li,inseek=%li,incheck=%li) ferror_in=%i feof_in=%i.\n",
			uniqid, len, bsiz, inseek, incheck, ferror(fin), feof(fin));*/

	/* "To EOF" mode? */
	if ((left = len) == -1L) left = bsiz;

	/* Sanity check */
	if (left < 0) return ERF_BadParms;

	/* Optional: seek in input file */
	if ((inseek != -1L) && (fseek(fin, inseek, SEEK_SET))) return ERF_fseekFailed;

	/* Debug: Initialisation */
	/*fprintf(stderr, "%li:   ftofcpy init %li of %li: inseek=%li ferror_in=%i feof_in=%i.\n",
			uniqid, left, len, inseek, ferror(fin), feof(fin));*/

	/* Copy (fin) -> (fout) piecewise */
	while ((left > 0) && (!ferror(fin)))
	{
		/* Premature EOF is an error, unless in "To EOF" mode */
		if (feof(fin))
		{
			if (len == -1L) goto loaf;
			else break;
		}

		/* Read up to bsiz bytes */
		dec = fread(bp, 1, (size_t)((left < bsiz) ? left : bsiz), fin);

		/* Write them out */
		if (/*(dec < 0) ||*/ ((dec > 0) && ((ltmp = fwrite(bp, 1, dec, fout)) != dec)))
		{
			fprintf(stderr, "%li:   Internal failure: fwrite at %li of %li: dec=%lu ferror_out=%i ltmp=%lu.\n",
					uniqid, left, len, (unsigned long) dec, ferror(fout), (unsigned long) ltmp);

			/* Output failure */
			return ERF_fwriteFailed;
		}

		/* Reduce count of bytes left */
		if (len != -1L) left -= dec;
	}

	/* Trap read errors */
	if (left != 0)
	{
		fprintf(stderr, "%li:   Internal failure: fread at %li of %li: dec=%lu ferror_in=%i feof_in=%i.\n",
				uniqid, left, len, (unsigned long) dec, ferror(fin), feof(fin));

		/* Input failure */
		return ERF_freadFailed;
	}

loaf: /* Skipped read error reporting if hit EOF in "To EOF" mode */

	/* Paranoid mode: check input file position is where it should be */
	if (incheck != -1L)
	{
		/* Get current pos */
		if ((left = ftell(fin)) == -1L) return ERF_ftellFailed;

		/* Mismatch? */
		if (left != incheck) return ERF_InternalError;
	}

	/* Success */
	return 0;
}

/***************************************************************************************************************************
** Spit out the string (p) in a form suitable for inclusion in HTML documents, allowing "&;" entities through unscathed.  **
** The htmlify() function allows "<b>" and "<i>" tags through, strips out superfluous tags, and ensures correct nesting,  **
** whilst encapsulate() escapes even "<b>" and "<i>" tags.                                                                **
***************************************************************************************************************************/
void encapsulate(char* p, FILE *fp)
{
	int ch;
	while (ch = *(p++)) if (!entity_table[ch]) putc(ch, fp); else fputs(entity_table[ch], fp);
}
void encapslim(char* p, FILE *fp, int lim)
{
	int ch, cnt, len;

	if ((len = strlen(p)) > lim)
	{
		cnt = 2 * lim / 3;
		while ((cnt-- > 0) && (ch = *(p++))) if (!entity_table[ch]) putc(ch, fp); else fputs(entity_table[ch], fp);
		fputs("...", fp);
		p += len - lim;
	}
	while (ch = *(p++)) if (!entity_table[ch]) putc(ch, fp); else fputs(entity_table[ch], fp);
}
void htmlify(char* p, FILE *fp)
{
	int c;
	int bold = 0, italic = 0;
	enum { NoTag, BoldTag, ItalicTag } order = NoTag;

	/* Encode newlines and linefeeds */
	while (c = *(p++)) if ((c < 32) && (p[0] < 32) && (p[0] > 0))
	{
		fputs("</p>" LF "<p>", fp); p++;
	}

	/* turn characters other than "<" into html entities if necessary */
	else if (c != '<')
	{
		if (!entity_table[c]) putc(c, fp); else fputs(entity_table[c], fp);
	}

	/* line break tag */
	else if (((p[0]=='b')||(p[0]=='B')) && ((p[1]=='r')||(p[1]=='R')) && (p[2]=='>'))
	{
		fputs("<br>", fp); p += 3;
	}

	/* opening paragraph tag */
	else if (((p[0]=='p')||(p[0]=='P')) && (p[1]=='>'))
	{
		fputs("<p>", fp); p += 2;
	}

	/* closing paragraph tag */
	else if ((p[0]=='/') && ((p[1]=='p')||(p[1]=='P')) && (p[2]=='>'))
	{
		fputs("</p>", fp); p += 3;
	}

	/* Start of "bold" tag */
	else if (((p[0]=='b')||(p[0]=='B')) && (p[1]=='>'))
	{
		if (!bold++)
		{
			if (order == NoTag) order = BoldTag;
			fputs("<B>", fp);
		}
		p += 2;
	}

	/* End of "bold" tag */
	else if ((p[0]=='/') && ((p[1]=='b')||(p[1]=='B')) && (p[2]=='>'))
	{
		if ((bold) && !(--bold))
		{
			if ((order == bold) && (italic)) { fputs("</I></B><I>", fp); order = ItalicTag; }
			else { fputs("</B>", fp); if (order == BoldTag) order = NoTag; }
		}
		p += 3;
	}

	/* Start of "italic" tag */
	else if (((p[0]=='i')||(p[0]=='I')) && (p[1]=='>'))
	{
		if (!italic++)
		{
			if (order == NoTag) order = ItalicTag;
			fputs("<I>", fp);
		}
		p += 2;
	}

	/* End of "italic" tag */
	else if ((p[0]=='/') && ((p[1]=='i')||(p[1]=='I')) && (p[2]=='>'))
	{
		if ((italic) && !(--italic))
		{
			if ((order == italic) && (bold)) { fputs("</B></I><B>", fp); order = BoldTag; }
			else { fputs("</I>", fp); if (order == ItalicTag) order = NoTag; }
		}
		p += 3;
	}

	/* Otherwise encode the "<" character as an html entity */
	else fputs(entity_table[c], fp);

	/* Close any "bold" and/or "italic" tags left open */
	if (order == ItalicTag) fputs(((bold) ? "</B></I>" : "</I>"), fp);
	else if (order == BoldTag) fputs(((italic) ? "</I></B>" : "</B>"), fp);
}

/***************************************************************************************************************************
** timeval(): Converts string (p) from "[d]dddhhmm" into the returned timeval; returns zero if given bad input.           **
** addhours(): Adds an integer number of hours onto existing timeval; returns zero if new timeval invalid.                **
***************************************************************************************************************************/
int str2u32b(u32b* v, char* p, indx min, indx max)
{
	indx n;

	/* Loop through digits */
	for (n = 0, *v = 0; ((p[n] >= '0') && (p[n] <= '9')); ++n)
	{
		/* Too many digits */
		if (n > max) return 1;

		/* Accumulate value */
		*v = (p[n] - '0') + (10 * (*v));
	}

	/* Badly terminated */
	if (p[n] != '\0') return 2;

	/* Lower digit limits */
	if (n < min) return 3;

	/* Okay */
	return 0;
}
timeval str2tim(char* p)
{
	u16b d;
	byte h, m;
	indx n = 0;

	/* Check digits and calculate length */
	while ((p[n] >= '0') && (p[n] <= '9')) ++n;

	/* Check length and validity of last digit */
	if ((p[n]) || (n < TDIG_Min) || (n > TDIG_Max)) return 0;

	/* Minute value (last two digits) */
	m = (p[n-1] - '0') + 10 * (p[n-2] - '0');

	/* Hour value (middle two digits) */
	h = (p[n-3] - '0') + 10 * (p[n-4] - '0');

	/* Day value (initial digits) */
	for (n -= 4, d = 0; (n > 0); --n) d = *(p++) - '0' + 10 * d;

	/* Combined timeval (zero if bad, but that's not likely, is it?)*/
	return TVAL(d, h, m);
}
timeval addhours(timeval tbase, u32b hours)
{
	if (hours > (T_d_MAX * 24)) return 0;

	hours += T_h(tbase);

	return TVAL((T_d(tbase) + (hours / (u32b)24)), (hours % 24), T_m(tbase));
}
const char* fdayname(u16b d, char* form)
{
	static char timestrfbuf[MAX_timestrfbuf];
	struct tm t = { 0, 0, 0, StartingMDay, StartingMonth, StartingYear-1900, StartingWDay, StartingYDay, 0 };

	t.tm_mday += d;
	t.tm_wday += d;
	t.tm_yday += d;

	(void) mktime(&t);
	if (!form) form = "%A, %d %B";
	return (strftime(timestrfbuf, MAX_timestrfbuf, form, &t)) ? timestrfbuf : "[Unconverted]";
}
const char* wdayname(u16b d)
{
	return daynames[(StartingWDay + d) % 7];
}
const char* perid(timeval t)
{
	struct period* p = &periods[0];

	while (t > p->tmax) p++;

	return (p->index);
}
struct period* per(timeval t)
{
	struct period* p = &periods[0];

	while (t > p->tmax) p++;

	return (p);
}
timeval strings2tim(char* dp, char* hp, char* mp)
{
	u32b d, h, m;

	/* Any null pointers? */
	if ((!dp) || (!hp) || (!mp)) return 0;

	/* Convert day: too many digits? */
	if (str2u32b(&d, dp, TDIG_Min - 4, TDIG_Max - 4)) return 0;

	/* Convert hour: too many digits? */
	if (str2u32b(&h, hp, 1, 2)) return 0;

	/* Convert min: too many digits? */
	if (str2u32b(&m, mp, 1, 2)) return 0;

	/* Convert into timeval - return 0 if invalid at this point */
	return TVAL(d, h, m);
}

/***************************************************************************************************************************
** Produce and output HTML document data for the various situations indicated by (which)... or not, at the moment.        **
***************************************************************************************************************************/
void error_doc(int which)
{
	printf(LF "<hr></html><HTML><HEAD><TITLE>" GAME " Assassins' Game: GCI Query Error</TITLE></HEAD>"
		   LF "<BODY bgcolor=#DDDDFF text=#000000 link=#0000FF vlink=#FF0000 alink=#FF00FF><h1>Bad luck, old chap!</h1>"
		   LF "<p>An error occured whilst processing your query!"
		   LF "Please email <a href=\"mailto:emn23@hermes.cam.ac.uk\">Ed Nokes (emn23)</a>"
		   LF "&mdash; take care to inform him of the magic number, which is <code>%d</code>.</p>"
		   LF "<hr><p>If I happen to be looking, here's some useful information for me:</p>"
		   LF "<p><font color=#880000><code>Return code: <b>%d</b>."
		   LF "<br>Process id: <b>%li</b>."
		   LF "<br>Clib errno: <b>%i</b>, strerror: &quot;<b>%s</b>&quot;.</code>",
		   which, which, uniqid, errno, strerror(errno));

#ifdef ACORN
	{	_kernel_oserror *ke = _kernel_last_oserror();

		if (ke) printf(LF "<br>_kernel_last_oserror: %p = <b>%i</b>: &quot;<b>%s</b>&quot;", ke, ke->errnum, ke->errmess);
	}
#endif

	puts(LF "</code></b></font></p>" LF "<P>");

#ifndef NO_CGI
	{	int i;

		printf("</P>" LF "<p>Initial query data extraction yielded:</p><p><UL>" LF);
		for (i = 0; (i < npairs); i++)
		{
			printf(LF "<li>(%d) <code><b>%s =</b> ", i, name[i]);
			encapsulate(val[i], stdout);
			puts("</code></li>");
		}
		puts("</UL>");
	}
#endif /* NO_CGI */

	printf(LF "</p><hr><p>There's almost certainly no point in trying again if the fault is at this end."
		   LF "If the fault is with your browser, hit it and try again.</p></BODY></HTML>" LF);
}

/***************************************************************************************************************************
** File locking, to prevent multiple simultaneous access to datafiles.                                                    **
***************************************************************************************************************************/
int setflock(char* p)
{
#ifdef ACORN
	return 1;
#else
	int fd = open(FILE_Lock, O_RDWR|O_CREAT, S_IRWXU|S_IRWXG);
	if ((fd != -1) && (!flock(fd, LOCK_EX)))
	{
		(void)write(fd, p, strlen(p));
		fcntl(fd, F_SETFD, FD_CLOEXEC);
		return fd+1;
	}
	if  (fd != -1) close(fd);
	return 0;
#endif
}
int setfunlock(int fd)
{
#ifndef ACORN
	flock(fd-1, LOCK_UN);
	(void)write(fd-1, "...unlocked.\n", 13);
	close(fd-1);
#endif
	return 0;
}

/***************************************************************************************************************************
** Print player status                                                                                                    **
***************************************************************************************************************************/
void printpstatus(unsigned long flags, timeval timer, char* con[9])
{
	if (flags & PF_Special) puts("Special");
	else if (flags & PF_Retired) puts("Retired");
	else if (!(flags & PF_Playing)) puts("Non-player");
	else if (flags & PF_Dead)
	{
		if ((!timer) || !(flags & PF_ResTimer)) printf("%s - Dead", pltypes[PF_type(flags)]);
		else printf("%s - Dead until "TDISP"", pltypes[PF_type(flags)], Thm(timer));
	}
	else
	{
		printf("%s - Alive", pltypes[PF_type(flags)]);
		if (flags & PF_Inactive) puts(" - Inactive");
	    else if (flags & PF_InactWarning) puts(" - Almost inactive");
		if ((timer) && (flags & PF_InactTimer))
		{
			printf(" - Deadline: "TDISP" %s)", Thm(timer), fdayname(T_d(timer), "%d/%b"));
		}
	}

#if 0
	if ((p) && (flags & (PF_ResTimer|PF_InactTimer|PF_Inactive)))
	{
		printf("%s %s/"TDISP")", p, fdayname(timer->day, "%d/%b"), Thm(timer));
	}
	if ((con) && (strcmp(con[3],con[6]) || strcmp(con[4],con[7]) || strcmp(con[5],con[8]))) puts("<br>Targets outdated");
#endif
}

/***************************************************************************************************************************
** Bits                                                                                                                   **
***************************************************************************************************************************/
#if 0
char* str(const char *fmt, ...)
{
	int n, size = STR_INITIAL;
	char *p;
	va_list ap;

	if (!(p = malloc(size))) return 0;

	while (1)
	{
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf(p, size, fmt, ap);
		va_end(ap);

		/* If that worked, return the string. */
		if ((n > -1) && (n < size)) return p;

		/* Else try again with more space. */
		if (n > -1) size = n+1;   /* glibc 2.1, precisely what is needed */
		else size *= 2;           /* glibc 2.0, twice the old size */

		if (!(p = realloc(p, size))) return 0;
	}
}
#endif
void urlencode(FILE* fp, char* name, char* value, int first)
{
	char* p;
	int i, c;

	/* Separate strings */
	if (!first) fputc('&', fp);

	/* First encode name, then value */
	for (p = name, i = 2; (p) && (i--); p = value)
	{
		/* Place the '=' sign between */
		if (!i) fputc('=', fp);

		/* Process each character */
		while (c = *(p++))
		{
			/* Quote characters */
			if ((c < 32) || (c == '+') || (c == '=') || (c == '%') || (c == '&') || (c >= 127))
			{
				fprintf(fp, "%%%2.2X", (unsigned int) c);
			}

			/* Encode space as '+', remaining characters unquoted */
			else fputc((c == 32) ? '+' : c, fp);
		}
	}
}
FILE* prepforstdin(void)
{
	FILE *fp;

#ifdef ACORN
	return (fp = fopen("RAM:$/tmp.txt", "wb+"));
#else
	/* Close stdin (fd 0) */
	if (fclose(stdin)) return 0;

    /* Open a temporary file (as fd 0) */
	else return (fp = tmpfile());
#endif
}
int launchwithstdin(char* p, FILE* fp)
{
	long int fpos;
	char clen[32];

	/* Get the file length */
	if ((fpos = ftell(fp)) == -1L) return (ERF_ftellFailed);

	/* Back to start of file */
	if (fseek(fp, 0, SEEK_SET)) return (ERF_fseekFailed);

	/* Create the content-length string */
	sprintf(clen, "%li", fpos);

	/* Kill the query string variable */
	setenv("QUERY_STRING", "", 1);

	/* Set the content-length environmental variable */
	if (setenv("CONTENT_LENGTH", clen, 1)) return (ERF_setenvFailed);

	/* Set the content-type environmental variable */
	if (setenv("CONTENT_TYPE", "application/x-www-form-urlencoded", 1)) return (ERF_setenvFailed);

	/* Set the request-methed environmental variable */
	if (setenv("REQUEST_METHOD", "POST", 1)) return (ERF_setenvFailed);

	/* Clear the close-on-exec flag */
	if (fcntl(0, F_SETFD, 0)) return (ERF_fcntlFailed);

	/* Execute another program */
#ifndef ACORN
	if (execl(p, p, (char *)0)) return (ERF_execFailed);
#else
	if (my_execl(p, fp, p, 0)) return (ERF_execFailed);
#endif

	/* Superfluous */
	return 0;
}

unsigned int myrand(unsigned int lim)
{
	return (lim * rand() / (RAND_MAX+1.0));
}

/***************************************************************************************************************************
** Build a <form><select>-style date and time selector, named with (dayt),(hourt),(mint). Set flag for "full" day range.  **
***************************************************************************************************************************/
void printdatesel(char *dayt, char* hourt, char *mint, timeval dt, int full)
{
	u16b i;

	/* Time limits */
    timeval tmin = (full) ? TIME_DateSelFullMin : TIME_DateSelMin;
    timeval tmax = (full) ? TIME_DateSelFullMax : TIME_DateSelMax;

	/* Default above range? */
	if (dt > tmax) dt = tmax;

	/* Default below range? */
	else if ((dt) && (dt < tmin)) dt = tmin;

	printf("<select name=\"%s\" size=1>", hourt);

	for (i = 0; i < 24; i++) printf(LF"<option value=\"%d\"%s>%02d</option>", i, ((dt)&&(T_h(dt))) ? " selected" : "", i);

	printf("</select> : <select name=\"%s\" size=1>", mint);

	for (i = 0; i < 60; i++) printf(LF"<option value=\"%d\"%s>%02d</option>", i, ((dt)&&(T_m(dt))) ? " selected" : "", i);

	printf("</select> &nbsp; <select name=\"%s\" size=1>", dayt);

	for (i = T_d(tmin); (i < T_d(tmax)); i++)
	{
		printf(LF"<option value=\"%d\"%s>%s</option>", i, ((i == T_d(dt)) ? " selected" : ""), fdayname(i, 0));
	}

	puts("</select>");
}

/***************************************************************************************************************************
** Initialise stuff.                                                                                                      **
***************************************************************************************************************************/
int geninit(char** tstrp, mode_t umaskval, uint logging, char *title)
{
	struct tm *thetm;
	char *p;


	/* Process id (or other identifier) */
	uniqid = (long) getpid();

	/* Get epoch time */
	time(&epochtime);

	/* Time string for current time */
	p = ctime(&epochtime);

	/* Redirect stderr to logfile? */
	if (logging)
	{
		/* Temporary file creation umask for logfile */
		umask(0117);

		/* Reopen file descriptor */
		if (!freopen(FILE_Log, "ab", stderr)) return (-ERF_freopenErrLog);

		/* No buffering */
		setbuf(stderr, 0);
	}

	/* Set file creation umask */
	umask(umaskval);

	/* Title */
	if (title) fprintf(stderr, "%li ** %s at %s", uniqid, title, (p) ? p : "[internal error - ctime failed]\n");

	/* Invalid ctime? */
	if (!p) return ERF_InternalError;

	/* Return ctime string for current time? */
	if (tstrp) *tstrp = p;

	/* Convert to broken down time */
	thetm = localtime(&epochtime);

	/* Don't yet allow games to span multiple years */
	if (thetm->tm_year != (StartingYear - 1900)) return ERF_TimeKeeping;

	/* Invalid timeval? */
	if (!(thetime = TVAL((thetm->tm_yday - StartingYDay), thetm->tm_hour, thetm->tm_min))) return ERF_BadTimeVal;

	/* Seed random number generator */
	srand(uniqid * (unsigned int) epochtime);

	/* Success */
	return 0;
}


#endif
/*** End of file ***/
