/*****************************************************
**  flock.c                           >> bin/flock  **
**  ----------------------------------------------  **
**  Build an executable which can be run to grab a  **
**  lock on the datafiles, for safe manual updates  **
**  to the system. Lock is held until a keypress    **
**  is received; often handy to send to background  **
**  with ctrl-Z whilst working on datafiles.        **
**  ----------------------------------------------  **
**  converted to new time system                    **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define NO_CGI

#include "!lib-general.c"

int main(int argc, char* argv[])
{
	int rc = 0, lockfd = 0;
	char* thetimestr;


	/* Initialse stuff: time, file creation umask, pid, no title, don't redirect stderr to log */
	if (rc = geninit(&thetimestr, 0117, 0, 0)) erf(rc, 1);

	/* Attempt to lock files */
	if (!(lockfd = setflock("Lock.c "))) erf(ERF_flockFailed, 2);

	/* Logfile header */
	fprintf(stderr, "%li ** Locked: at %s", uniqid, thetimestr);

	puts("Terminate stdin to release lock.");

	/* Maintain lock until key pressed */
	while (EOF != getchar());
	
	puts("Unlocked.");

_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li !! Error %s (%i) in main(%i)\n", uniqid, ERF_STR(rc), ERF_TYPE(rc), ERF_CONTEXT(rc));

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/*** The end ***/
