/*****************************************************
**  lib-event-read.c                     (library)  **
**  ----------------------------------------------  **
**  Converted to new time system?                   **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifdef included__lib_event_read
 #error "** lib-event-read multiply #included **"
#else
 #define included__lib_event_read fromfile

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Get targetting information */
#define EVREAD_Targetting	(1<<0)

/* Macro: evread using the temporary buffer */
#ifdef MAX_fpbuf
 #define evlist(x,y,z) evread((x),(y),fpbuf,fpend,0,(z),0)
#endif

/* Function-pointer prototypes for hooks in evread() */
typedef int (*evread_efn) (event_p data, int state, FILE* fp, void* context);
typedef int (*evread_rfn) (event_p data, char* repid, int state, void* context);

/* Context structure */
struct getevinfo__ctx
{
	char* ref;
	event_p* epp;
	report_p lastrep, *rpp;
};

/***************************************************************************************************************************
** Process the list of events. Returns nonzero result on failure. Calls (entry_f)() for each event in file order with     **
** state set to (0). (ev_f)() may return with bit 1 set to instruct this function not to process anymore events.          **
** Calls (rep_f)() for each report id after (entry_f)(), provided that returned with bit 0 clear; after processing        **
** all report ids, calls (entry_f)() again with state set to (1) provided first (entry_f) returned bit 2 set.             **
** (rep_f)() has bit 0 set on first call, bit 1 set on last call, bit 2 set for state rather than game events.            **
***************************************************************************************************************************/
int evread(evread_efn event_f, evread_rfn rep_f, char* buf, char* bufend, char** bufused, void* ctx, int mode)
{
	FILE* fp = 0;
	char *p, *bp, *er = 0;
	int i, j, k, l, c, rc = 0;
	event_p ep = 0;


	/* Ensure at least bytes free in buffer */
	if ((bufend - buf) < chk[CHK_events].len) erf(ERF_BufferExhaust, 60);

	/* Open player info file */
	if (!(fp = fopen(FILE_Events, "rb"))) erf(ERF_fopenEvents, 0);

	/* Check file header (ignoring initial comments) */
	if (i = chkheader(fp, CHK_events, buf, 0, 0, 0, 0)) erf(i, 61);

	/* Scan through for valid lines */
	while ((c = fgetc(fp)) != EOF)
	{
		/* Ignore lines not beginning with "*" character */
		switch (c)
		{
			case '*'  : break;
			default   : toeol(fp);
			case '\n' : continue;
		}
		
		/* Current buffer use */
		bp = buf;

		/* Claim memory for an event struct (if none already available) */
		if ((!ep) && !(ep = malloc(sizeof(struct event)))) erf(ERF_mallocFailed, 2);

		/* Clear opaque extension word */
		ep->ext = 0;

		/* No retargetting yet */
		ep->retarg = ep->untarg = 0;

		/* Skip initial tabs; read ref field */
		if (i = tread(fp, er = ep->ref = bp, bufend, &bp)) erf(ERF_BadTread+i, 80);

		/* Read textual event time field */
		if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 81);

		/* Convert to date format */
		if (!(ep->when = str2tim(bp))) erf(ERF_BadDateText, 82);

		/* Read and convert flags field */
		if (i = tread(fp, bp, bufend, 0)) erf(ERF_BadTread+i, 83);
		ep->flags = strtoul(bp, 0, 16);

		/*fprintf(stderr, "%li:   event: %s flags &%lX\n", uniqid, ep->ref, ep->flags);*/

		/* Read id field */
		if (i = tread(fp, ep->id = bp, bufend, &bp)) erf(ERF_BadTread+i, 84);

		/* Read new timer field */
		if ((i = tread(fp, bp, bufend, 0)) > 1)	erf(ERF_BadTread+i, 85);

		/* Convert to date format */
		if (!(ep->ntimer = str2tim(bp))) erf(ERF_BadDateText, 86);

		/* Read scoring bonus (optional) */
		if (ep->flags & EF_ScoreBonus)
		{
			/* Read field, report errors */
			if ((i) || ((i = tread(fp, bp, bufend, 0)) > 1)) erf(ERF_BadTread+i, 87);
			ep->bonus = atol(bp);
		}
		else ep->bonus = 0;

		/* Read police rank (optional) */
		if (ep->flags & EF_Police)
		{
			/* Read field, report errors */
			if ((i) || ((i = tread(fp, bp, bufend, 0)) > 1)) erf(ERF_BadTread+i, 87);
			ep->rank = atol(bp);

			/*fprintf(stderr, "%li:   police event: %s rank %li\n", uniqid, ep->ref, ep->rank);*/
		}
		else ep->rank = 0;

		/* Read pseudonym (optional) */
		if (ep->pseud = (ep->flags & EF_MASK_Pseudonym) ? (bp) : 0)
		{
			/* Read field, report errors */
			if ((i) || ((i = tread(fp, bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 88);
		}

		/* Handle retargetting */
		if (ep->flags & EF_Targetting) while (1)
		{
			retarg_p retarp, *tarlist;

			/* Read */
			if ((i) || ((i = tread(fp, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 70);

			/* End of list */
			if (p[0] == '-') break;

			/* Don't process? */
			if (!(mode & EVREAD_Targetting)) continue;

			/* Allocate storage */
			if (!(retarp = malloc(sizeof(struct retarg)))) erf(ERF_mallocFailed, 71);

			/* Assassin? */
			if (p[0] == 'U') tarlist = &ep->untarg;

			/* Target? */
			else if (p[0] == 'T') tarlist = &ep->retarg;

			/* Invalid data */
			else erf(ERF_InternalError, 72);

			/* Assassins' id */
			retarp->assid = p + 1;
			
			/* Separate string */
			if (!(p = strchr(p + 1, ':'))) erf(ERF_InternalError, 73);

			/* Terminate */
			p[0] = '\0';

			/* Targets' id */
			retarp->tarid = p + 1;

			/* Valid ids */
			if (*(retarp->assid) <= 32) erf(ERF_InternalError, 74);
			if (*(retarp->tarid) <= 32) erf(ERF_InternalError, 75);

			/* Link into list */
			retarp->next = *tarlist;
			*tarlist = retarp;
		}

		/* Read 2 assassination fields (optional) */
		if (ep->flags & EF_Assassination)
		{
			if ((i) || ((i = tread(fp, ep->vict_pid = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 89);
			if ((i) || ((i = tread(fp, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 90);
			ep->vict_pseud = SESC(p);
		}
		else ep->vict_pid = ep->vict_pseud = 0;

		/* Read and parse 2 news fields (optional) */
		if (ep->flags & EF_News)
		{
			if ((i) || ((i = tread(fp, bp, bufend, 0)) > 1)) erf(ERF_BadTread+i, 91);
			if (!(ep->news_date = str2tim(bp))) erf(ERF_BadDateText, 92);
			if ((i) || ((i = tread(fp, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 93);
			ep->news_head = SESC(p);
		}
		else { ep->news_head = 0; ep->news_date = 0; }

		/* Read wantedness field (optional) */
		if (ep->flags & EF_Wantedness)
		{
			if ((i) || ((i = tread(fp, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 94);
			ep->wantedness = SESC(p);
		}

		/* Mix flags: bit 0 => repids follow, bit 1 => no repids, bit 2 => second call */
		l = (i == 0) ? 1 : 2;
	
		/* Call supplied function for events */
		k = (!event_f) ? 0 : event_f(ep, l, fp, ctx);

		/* Handle thrown errors - pass on modulus of return code */
		if (k < 0) erf(-k, 95);

		/* Loop for report ids */
		if ((!(k & 1)) && (rep_f) && (l & 1))
		{
			char* np;

			/* Read current repid field */
			for (c = l; ((j = tread(fp, bp, bufend, &np)) < 2); c &= ~1)
			{
				/* Call supplied function, return error if thrown */
				if ((i = rep_f(ep, bp, ((j) ? (c + 2) : (c)), ctx)) < 0) erf(-i, 96);

				/* Don't overwrite buffer */
				if (i & 8) bp = np;

				/* Allow early (or forced) termination */
				if ((i & 1) || (j != 0)) break;
			}

			/* Handle any file access errors */
			if (j > 1) erf(ERF_BadTread+j, 97);
		}

		/* Add terminating null */
		if (++bp < bufend) bp[-1] = '\0'; else erf(ERF_BufferExhaust, 22);

		/* Call the supplied function again? If buffer and struct are not being overwritten,
		** their contents are guaranteed safe once this call returns a non-negative value. */
		if ((k & 4) && ((j = event_f(ep, (l | 4), fp, ctx)) < 0)) erf(-j, 23);

		/* Retain buffer if required */
		if (k & 8) buf = bp;

		/* Free traget allocations? */
		if ((ep) && !(k & 32))
		{
			retarg_p retarp, nretarp;

			for (retarp = ep->retarg; (retarp); retarp = nretarp) { nretarp = retarp->next; free(retarp); }
			for (retarp = ep->untarg; (retarp); retarp = nretarp) { nretarp = retarp->next; free(retarp); }

			ep->retarg = ep->untarg = 0;
		}

		/* Preserve this event struct? */
		if (k & 16) ep = 0;

		/* Early return if required */
		if (k & 2) break;

		/* Next please */
		toeol(fp);

		/* Paranoia */
		er = 0;
	}

	/* Successful termination */
	if (bufused) *bufused = buf;

_erf: /***** Error handling *****/

	if (fp) fclose(fp);
	if (rc <= 0) return (rc);

	/* Debug */
	fprintf(stderr, "%li !! evread(%i):%i on \"%s\"\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), STR(er));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Acquire event information by reference                                                                                 **
***************************************************************************************************************************/
int getevinfo__ef(event_p ep, int state, FILE* fp, void* context)
{
	struct getevinfo__ctx* ctx = context;

	/* No match - try the next one */
	if (strcmp(ctx->ref, ep->ref)) return (1);

	/* Could allocate new event struct and use it */
	/*if (!(*ctx->epp = malloc(sizeof(struct event)))) return -ERF_mallocFailed;*/
	/***(ctx->epp) = *ep;*/

	/* Use internal event structs */
	*(ctx->epp) = ep;

	/*fprintf(stderr, "%li ... %s", uniqid, ep->ref);*/

	/* No need to search any more events */
	return (0+2+8+16+32);
}

int getevinfo__rf(event_p ep, char* repid, int state, void* context)
{
	struct getevinfo__ctx* ctx = context;
	report_p rp;

	/* Allocate report structure */
	if (!(rp = malloc(sizeof(struct report)))) return -(ERF_mallocFailed);

	/* Link into existing list */
	if (ctx->lastrep) ctx->lastrep->link_all = rp;

	/* First in chain */
	else *(ctx->rpp) = rp;

	/* Place at end of list */
	(ctx->lastrep = rp)->link_all = 0;

	/* Store report id, etc. */
	rp->rid = repid;
	rp->pid = 0;
	rp->text = 0;
	rp->fpos = 0L;
	rp->pptr = 0;
	rp->pseud = 0;
	rp->flags = 0UL;
	rp->length = 0L;
	rp->link_sub = 0;
	rp->uptim = 0;

	/* All repids */
	return (8);
}

int getevinfo(char* ref, event_p* epp, report_p* rpp, char** buf_pos, char* buf_end, int mode)
{
	struct getevinfo__ctx ctx;

	/* Must provide some way of returning the information */
	if (!epp) return ERF_BadParms;

	/* No match yet */
	*(ctx.epp = epp) = 0;

	/* Interested in related reports? */
	if (ctx.rpp = rpp) *(rpp) = ctx.lastrep = 0;

	/* Desired event ref */
	ctx.ref = ref;

	/* Scan through event list to find correct position for new event */
	return (evread(getevinfo__ef, (rpp) ? (getevinfo__rf) : (0), *buf_pos, buf_end, buf_pos, &ctx, mode));
}

#endif
/*** The End ***/
