/*****************************************************
**  lib-event-edit.c                     (library)  **
**  ----------------------------------------------  **
**  converted to new time system?                   **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifdef included__lib_event_edit
 #error "** lib-event-edit multiply #included **"
#else
 #define included__lib_event_edit fromfile
 #undef IFDEB
 #undef IFWARN
 #define IFDEB(x) do { /*(x);*/ } while (0)
 #define IFWARN(x) do { /*(x);*/ } while (0)

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Type to hold event editing control flags */
typedef unsigned int edev__flags;

/* Event editing control glags */
#define EDEV_Edit		(((edev__flags)1)<<0)	/* Edit existing event, otherwise add around existing event */
#define EDEV_Before		(((edev__flags)1)<<1)	/* Add before existing event, otherwise add after it */
#define EDEV_RemoveRid	(((edev__flags)1)<<2)	/* Dis-associate report id from event, otherwise add it --- Not used */
#define EDEV_GetRids	(((edev__flags)1)<<3)	/* Return list of report ids when editing events */
#define EDEV_OnlyRids	(((edev__flags)1)<<4)	/* Edit list of report ids for event, but not event itself */

/* Handy typedef */
typedef struct editev* editev_p;

/* A structure to hold event content changes */
struct editev
{
	edev__flags mode;
	char *oldref;				/* source event ref, or null to add by time */
	timeval at;					/* source time (if oldref null) */
	char *nref;					/* null => use oldref, or automatically allocate */
	char *npid;
	char *npseud;
	retarg_p nretarg;
	retarg_p nuntarg;
	char *nvict_pid;
	char *nvict_pseud;
	char *nnews_head;
	timeval nnews_date;
	char *nwanted;
	unsigned long* nflsp;
	timeval ntimp;
	long* nbonusp;
	long* nrankp;
	char* rid;					/* if non-null, report id to add or delete */
	editev_p next, link;		/* for initial linked list and internal use */
	event_p ep;
	report_p reps;
};

/***************************************************************************************************************************
** Helper function: Initialise a mostly empty event edit structure.                                                       **
***************************************************************************************************************************/
void init_editev_struct_part(editev_p edp)
{
	int i;

	/* Initialise in rest of struct */
	edp->nretarg = 0;
	edp->nuntarg = 0;
	edp->nref = 0;
	edp->npid = 0;
	edp->npseud = 0;
	edp->nvict_pid = 0;
	edp->nvict_pseud = 0;
	edp->nnews_head = 0;
	edp->nnews_date = 0;
	edp->nwanted = 0;
	edp->nflsp = 0;
	edp->ntimp = 0;
	edp->nbonusp = 0;
	edp->nrankp = 0;
}

void init_editev_struct(editev_p edp, char *oref, edev__flags mode, editev_p next, editev_p link, timeval at, char* rid)
{
	init_editev_struct_part(edp);

    edp->mode = mode;
	edp->next = next;
	edp->link = link;
	edp->rid = rid;
	edp->oldref = oref;
	edp->at = at;
}

/***************************************************************************************************************************
** Helper function: Keeps track of report id structs                                                                      **
** Maintains a double chain: start at (edp->reps) and follow (link_all) fields for reports in correct order.              **
** (edp->reps) -> (first), (first->link_all) -> (second), ... , (nth->link_all) -> 0.                                     **
** Also: for cyclic list of reports in backward order (latest first) start with (first->link_sub) -> (nth) and follow     **
** the (link_sub) fields: (nth->link_sub) -> (n-1th), (n-1th->link_sub) -> (n-2nd), ... , (second->link_sub) -> (first)   **
** but be careful with the boundary conditions: don't follow (first->link_sub) back to the "last" entry in the list.      **
** eg: if (last = edp->reps) for (rp = last = last->link_sub, edp->reps->link_sub = 0; (rp); rp = rp->link_sub) ...       **
** or: if (rp = edp->reps) do { rp = rp->link_sub; ... } while (rp != edp->reps);                                         **
***************************************************************************************************************************/
report_p editevents_rep(editev_p edp, char* p)
{
	report_p rptr;

	/* Allocate space for struct */
	if (!(rptr = malloc(sizeof(struct report)))) return 0;

	/* Add as start of chain */
	if (!edp->reps)
	{
		rptr->link_all = 0;
		rptr->link_sub = rptr;
		edp->reps = rptr;
	}

	/* Link onto end of chain: */
	else
	{
		rptr->link_all = 0;
		rptr->link_sub = edp->reps->link_sub;
		edp->reps->link_sub->link_all = rptr;
		edp->reps->link_sub = rptr;
	}

	/* Only partially initialise struct */
	rptr->rid = p;
	rptr->flags = 0;

	/* Success */
	return rptr;
}

/***************************************************************************************************************************
** Helper functions                                                                                                       **
***************************************************************************************************************************/
int editevents_init(event_p* epp, char* p, timeval t)
{
	event_p ep;
	int i;

	/* Paranoia */
	if ((!p) || (!t)) return ERF_BadParms;

	/* Allocate storage space */
	if (!(ep = malloc(sizeof(struct event)))) return ERF_mallocFailed;

	/* Given event ref */
	ep->ref = p;

	/* Given event time */
	ep->when = t;

	/* Clear it (no previous data) */
	ep->ext = 0;
	ep->link = 0;
	ep->rank = 0;
	ep->bonus = 0;
	ep->flags = 0;

	/* Set textual fields as null */
	ep->id = 0;
	ep->pseud = 0;
	ep->vict_pid = 0;
	ep->news_head = 0;
	ep->vict_pseud = 0;
	ep->wantedness = 0;

	/* Clear timers */
	ep->news_date = 0;
	ep->ntimer = 0;

	/* Clear target allocation fields */
	ep->retarg = 0;
	ep->untarg = 0;

	/* Save it */
	*epp = ep;

	/* Success */
	return 0;
}

int editevents_aux(FILE* fp, editev_p edp, int m, struct event **evpp)
{
	event_p ep = 0;
	int i, rc = 0;
	unsigned long oef = 0;
	char *p;


	/* Editing an existing event? Ensure (opf) is non-zero. */
	if ((evpp) && (ep = *evpp)) oef = ep->flags | EF_Temp;

	/* Need to allocate an event structure? */
	else if (i = editevents_init(&ep, edp->nref, edp->at)) erf(i, 2);

	/* Update ref, but don't change existing event times */
	if (edp->nref) ep->ref = edp->nref;

	/* Update textual fields from the editing structure */
	if (edp->npseud) ep->pseud = edp->npseud;
	if (edp->nwanted) ep->wantedness = edp->nwanted;
	if (edp->nvict_pid) ep->vict_pid = edp->nvict_pid;
	if (edp->nnews_head) ep->news_head = edp->nnews_head;
	if (edp->nvict_pseud) ep->vict_pseud = edp->nvict_pseud;

	/* Update player id from the editing structure if specified */
	if (edp->npid) ep->id = edp->npid;
	else if (!ep->id) erf(ERF_BadParms, 4);

	/* Update flags from the editing structure if specified */
	if (edp->nflsp) ep->flags = *edp->nflsp;
	else if (!oef) erf(ERF_BadParms, 5);

	/* Update timer from the editing structure if specified */
	if (edp->ntimp) ep->ntimer = edp->ntimp;
	else if (!ep->ntimer) erf(ERF_BadParms, 6);

	/* Optional bonus score field */
	if (ep->flags & EF_ScoreBonus)
	{
		if (edp->nbonusp) ep->bonus = *edp->nbonusp;
		else if (!oef) erf(ERF_BadParms, 7);
		else if (!(oef & EF_ScoreBonus)) erf(ERF_BadParms, 8);
	}
	else ep->bonus = 0;

	/* Optional police rank field */
	if (ep->flags & EF_Police)
	{
		if (edp->nrankp) ep->rank = *edp->nrankp;
		else if (!oef) erf(ERF_BadParms, 16);
		else if (!(oef & EF_Police)) erf(ERF_BadParms, 17);
	}
	else ep->rank = 0;

	/* Check pseudonym text is present, if required */
	if (!(ep->flags & EF_MASK_Pseudonym)) ep->pseud = 0;
	else if (!ep->pseud) erf(ERF_BadParms, 9);

	/* Check comment text is present, if required */
	if (!(ep->flags & EF_Wantedness)) ep->wantedness = 0;
	else if (!ep->wantedness) erf(ERF_BadParms, 10);

	/* Optional assassination fields */
	if (ep->flags & EF_Assassination)
	{
		if (!ep->vict_pid) erf(ERF_BadParms, 11);
		if (!ep->vict_pseud) erf(ERF_BadParms, 12);
	}
	else ep->vict_pid = ep->vict_pseud = 0;

	/* Optional News fields */
	if (ep->flags & EF_News)
	{
		if (!ep->news_head) erf(ERF_BadParms, 13);
		if (edp->nnews_date) ep->news_date = edp->nnews_date;
		else if (!ep->news_date) erf(ERF_BadParms, 14);
	}
	else { ep->news_head = 0; ep->news_date = 0; }

	/* Update target deallocations */
	if (!(ep->flags & EF_Targetting)) ep->untarg = 0;
	else if (edp->nuntarg) ep->untarg = edp->nuntarg;

	/* Update target reallocations */
	if (!(ep->flags & EF_Targetting)) ep->retarg = 0;
	else if (edp->nretarg) ep->retarg = edp->nretarg;

	/* Leading newline */
	if ((m == 3) && (EOF == fputc('\n', fp))) erf(ERF_fputcFailed, 18);

	/* Write compulsary fields */
	i = fprintf(fp, "*%s\t"TFMT"\t%lX\t%s\t"TFMT"", ep->ref, Tdhm(ep->when), ep->flags, ep->id, Tdhm(ep->ntimer));

	/* Check for errors */
	if (i < 24) erf(ERF_printfFailed, 19);

	/* Write score bonus field */
	if (ep->flags & EF_ScoreBonus)
	{
		if (2 > fprintf(fp, "\t%li", ep->bonus)) erf(ERF_printfFailed, 20);
	}

	/* Write police rank field */
	if (ep->flags & EF_Police)
	{
		if (2 > fprintf(fp, "\t%li", ep->rank)) erf(ERF_printfFailed, 21);
	}

	/* Write pseudonym field */
	if (ep->flags & EF_MASK_Pseudonym)
	{
		if (2 > fprintf(fp, "\t%s", ep->pseud)) erf(ERF_printfFailed, 22);
	}

	/* Write retargetting fields */
	if (ep->flags & EF_Targetting)
	{
		retarg_p retarp;

		for (retarp = ep->untarg; (retarp); retarp = retarp->next) if (retarp->assid)
		{
			if (5 > fprintf(fp, "\tU%s:%s", retarp->assid, retarp->tarid)) erf(ERF_printfFailed, 23);
		}

		for (retarp = ep->retarg; (retarp); retarp = retarp->next) if (retarp->assid)
		{
			if (5 > fprintf(fp, "\tT%s:%s", retarp->assid, retarp->tarid)) erf(ERF_printfFailed, 23);
		}

		if (2 != fprintf(fp, "\t-")) erf(ERF_printfFailed, 23);
	}

	/* Write assassination fields */
	if (ep->flags & EF_Assassination)
	{
		if (4 > fprintf(fp, "\t%s\t%s%s", ep->vict_pid, ESC(ep->vict_pseud))) erf(ERF_printfFailed, 24);
	}
		
	/* Write news fields */
	if (ep->flags & EF_News)
	{
		i = fprintf(fp, "\t"TFMT"\t%s%s", Tdhm(ep->news_date), ESC(ep->news_head));

		if (i < 11) erf(ERF_printfFailed, 25);
	}

	/* Write wantedness fields */
	if (ep->flags & EF_Wantedness)
	{
		if (2 > fprintf(fp, "\t%s%s", ESC(ep->wantedness))) erf(ERF_printfFailed, 26);
	}

	/* Special mode */
	if (m > 0)
	{
		/* Add initial report id immediately? */
		if ((edp->rid) && !(edp->mode & EDEV_RemoveRid))
		{
			/* Save report ids? */
			if ((edp->mode & EDEV_GetRids) && (!editevents_rep(edp, edp->rid))) erf(ERF_mallocFailed, 27);

			/* Write it out */
			if (2 > fprintf(fp, "\t%s", edp->rid)) erf(ERF_printfFailed, 28);
		}		

		/* Terminating newline? */
		if ((m == 1) && (EOF == fputc('\n', fp))) erf(ERF_fputcFailed, 29);
	}

	/* Debug */
	IFWARN(fprintf(stderr, "%li:   --[%ce]--> eref=\"%s\" efl=&%lx p1id=\"%s\" at "TFMT" vict=\"%s\"\n",
				   uniqid, (oef) ? '~' : '+', ep->ref, ep->flags, ep->id, Tdhm(ep->when), STR(ep->vict_pid)));

	/* Return event struct, if requested */
	if (evpp) *evpp = ep;

	/* Otherwise, free it */
	else free(ep);


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li !! editevents_aux(%i):%i edp=%p evpp=%p->%p m=%i ep=%p oef=&%lx\n",
					uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), edp, evpp, (evpp) ? (*evpp) : 0, m, ep, oef);

	/* Finished */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** New event adder/editor function                                                                                        **
***************************************************************************************************************************/
int editevents(edev__flags mode, editev_p list, char** bufp, char* bufend, char* comnt, FILE* fred)
{
	int rc = 0, n = -1, i, j, c;
	FILE *fin = 0, *fout = 0;
	editev_p rlist, clist, tlist, blist, edp, *pedp, edp_a, edp_e;
	unsigned long idx;
	long lmin, lmax, ltmp, lcom;
	char *p, *bp, *buf, *eref = 0;
	timeval ewhen, tv;


	/* Ensure input is non-empty */
	if (!list) erf(ERF_BadParms, 3);

	/* Ensure minimum number of bytes free in buffer */
	if ((bufend - (buf = *bufp)) < chk[CHK_events].len) erf(ERF_BufferExhaust, 4);

	/* Ensure there are no ctrl chars in comment string */
	if (p = comnt) 
	{
		while (*p >= 32) p++;
		if (*p != '\0') *p = '\0';
	}
	else comnt = "";

	/* Open old events files */
	if (!(fin = fopen(FILE_Events, "rb"))) erf(ERF_fopenEvents, 6);

	/* Check file header (ignoring initial comments) and get free index */
	if (i = chkheader(fin, CHK_events, buf, 0, &lmin, &idx, &lcom)) erf(i, 7);

	/* Sort out input list */
	for (clist = rlist = 0, n = 0, edp = list; (edp); edp = edp->next, ++n)
	{
		/* Not yet found or allocated */
		edp->ep = 0;
		edp->reps = 0;

		/* Edit event report ids but not event data */
		if (edp->mode & EDEV_OnlyRids)
		{
			/* Require an event ref */
			if (!edp->oldref) erf(ERF_BadParms, 8);

			/* For convenience, enforce certain mode flags */
			edp->mode = (edp->mode & (~EDEV_Before)) | EDEV_Edit;

			/* Initialise rest of struct */
			init_editev_struct_part(edp);
		}

		/* Add to list of events to add/edit by reference */
		if (edp->oldref)
		{
			/* NB: List ordering is reversed, but this is corrected again later */
			edp->link = rlist;
			rlist = edp;

			/* Reuse old ref if editing event */
			if ((!edp->nref) && (edp->mode & EDEV_Edit)) edp->nref = edp->oldref;
			
			/* Sanity check: can't remove rids from new events */
			if (!(edp->mode & EDEV_Edit) && (edp->mode & EDEV_RemoveRid)) erf(ERF_BadParms, 9);
		}

		/* Otherwise, add to cronological list of dated event additions */
		else
		{
			/* Time of this event */
			tv = edp->at;

			/* Search for correct chronological position in list */
			for (edp_a = clist, edp_e = 0; (edp_a); edp_a = (edp_e = edp_a)->link) if (edp_a->at > tv) break;

			/* Add in at start of clist */
			if (!edp_e) clist = edp;

			/* Add in after edp_e in clist */
			else edp_e->link = edp;

			/* But before edp_a */
			edp->link = edp_a;

			/* Sanity check: can't remove rids from new events */
			if (edp->mode & EDEV_RemoveRid) erf(ERF_BadParms, 10);
		}

		/* Allocate new refs */
		if (!edp->nref)
		{
			if ((bufend - buf) < MAX_eref) erf(ERF_BufferExhaust, 12);
			if (2 > sprintf(buf, "e%lu", idx)) erf(ERF_printfFailed, 13);
			buf = (edp->nref = buf) + 1 + strlen(buf);
			++idx;
		}
	}

	/* Redirect output to specified file */
	if (fred) fout = fred;

	/* Open new events files */
	else if (!(fout = fopen(FILE_NewEvents, "wb"))) erf(ERF_fopenEvents, 15);

	/* Copy initial comments */
	if (i = ftofcpy(fin, fout, lcom, fpbuf, MAX_fpbuf, 0, lcom)) erf(i, 16);

	/* Back to position just after header */
	if (fseek(fin, lmin, SEEK_SET)) erf(ERF_fseekFailed, 17);

	/* Write current file header */
	if (i = prtheader(fout, CHK_events, idx)) erf(i, 18);

	/* Debug: no buffering */
	/* setbuf(fout, 0); */

	/* Track comments and timeval */
	lcom = -1L; tv = TMIN;

	/* Scan through for valid lines */
	while ((c = fgetc(fin)) != EOF)
	{
		/* Ignore lines not beginning with "*" character */
		switch (c)
		{
			case '*'  : break;
			case '#'  : if ((lcom == -1L) && (-1L == (lcom = ftell(fin)))) erf(ERF_ftellFailed, 20);
			default   : toeol(fin);
			case '\n' : if (c != '#') lcom = -1L;
						continue;
		}

		/* Current buffer use */
		bp = buf;

		/* File position of char after '*' */
		if (-1L == (lmax = ftell(fin))) erf(ERF_ftellFailed, 29);

		/* Skip initial tabs; read ref field */
		if (i = tread(fin, eref = bp, bufend, &bp)) erf(ERF_BadTread+i, 30);

		/* Read textual event time field */
		if (i = tread(fin, bp, bufend, 0)) erf(ERF_BadTread+i, 31);

		/* Convert to date format */
		if (!(ewhen = str2tim(bp))) erf(ERF_BadDateText, 32);

		/* Position just after date */
		if (-1L == (ltmp = ftell(fin))) erf(ERF_ftellFailed, 33);

		/* Debug - an event record was found on input */
		IFDEB(fprintf(stderr, "%li:   editev found \"%s\":"TFMT" at %li (now %li) with lcom=%li lmin=%li.\n",
					  uniqid, eref, Tdhm(ewhen), lmax, ltmp, lcom, lmin));

		/* Initialisation for scan: no event-edits or pre/post-event-additions found yet */
		edp_e = tlist = blist = 0;
		pedp = &rlist;

		/* Scan through list for events referencing this one */
		while (edp_a = *pedp)
		{
			/* Referenced */
			if (streq(edp_a->oldref, eref))
			{
				/* Remove from list */
				*pedp = edp_a->link;

				/* Add to list of events to add immediately after this event */
				if (!(edp_a->mode & EDEV_Edit))
				{
					/* Select pre/post list */
					editev_p *tmpedpp = (edp_a->mode & EDEV_Before) ? &blist : &tlist;

					/* NB: Reverse order of list again, so we end up with the original ordering */
					edp_a->link = *tmpedpp;
					*tmpedpp = edp_a;
				}

				/* Collect first edit for this event */
				else if (!edp_e) edp_e = edp_a;

				/* Complain about multiple edits on this event */
				else erf(ERF_BadParms, 36);
			}

			/* Not referenced -- keep on list */
			else pedp = &edp_a->link;
		}

		/* Debug - anything to do? */
		IFDEB(fprintf(stderr, "%li:   editev \"%s\": edits %p, pre %p, post %p.\n", uniqid, eref, edp_e, blist, tlist));

		/* Anything to add before this event record? */
		if ((j = ((clist) && ((tv = ewhen) > clist->at))) || (blist))
		{
			/* No proceeding comments? */
			if (lcom == -1L) lcom = lmax;

			/* Paranoia */
			if (--lcom < lmin) erf(ERF_InternalError, 40);

			/* Anything to copy? */
			else if (lcom > lmin)
			{
				/* Copy up to start of preceeding comments (if any) or this event */
				if (i = ftofcpy(fin, fout, lcom - lmin, fpbuf, MAX_fpbuf, lmin, lcom)) erf(i, 41);

				/* Have now transferred up to beginning of preceeding comments */
				lmin = lcom;
			}

			/* Debug - some event(s) to insert */
			IFDEB(fprintf(stderr, "%li:   editev \"%s\" prior chrono %p direct %p.\n", uniqid, eref, clist, blist));

			/* Insert events chronologically before this one */
			if (j)
			{
				/* Comment */
				fprintf(fout, "#-autoinserted-cron:%s\n", comnt);
	
				/* Add the new events */
				for (edp = clist; (edp) && (edp->at < tv); edp = edp->link)
				{
					/* Output new event (edp->ep = 0 on entry) and trailing newline */
					if (i = editevents_aux(fout, edp, 1, &edp->ep)) erf(i, 42);
				}
	
				/* Update list */
				clist = edp;

				/* Additional newline, to separate blocks */
				if (EOF == fputc('\n', fout)) erf(ERF_fputcFailed, 43);
			}

			/* Insert events to be added directly prior to this one */
			if (blist)
			{
				/* Comment */
				fprintf(fout, "#-autoinserted-before-\"%s\":%s\n", eref, comnt);
	
				/* Add the new events */
				for (edp = blist; (edp); edp = edp->link)
				{
					/* Force event time to same value as referenced event */
					edp->at = ewhen;

					/* Output new event (edp->ep = 0 on entry) and trailing newline */
					if (i = editevents_aux(fout, edp, 1, &edp->ep)) erf(i, 44);
				}

				/* Additional newline, to separate blocks */
				if (EOF == fputc('\n', fout)) erf(ERF_fputcFailed, 45);
			}
		}

		/* Skip over event content */
		if ((!edp_e) && (tlist))
		{
			/* Back to position just after date */
			if (fseek(fin, ltmp, SEEK_SET)) erf(ERF_fseekFailed, 50);

			/* Set lmax --> newline char at end of line, go onto next line */
			if (i = toeolpos(fin, &lmax)) erf(i, 51);

			/* Position at end of line */
			/*if (-1L == (lmax = ftell(fin))) erf(ERF_ftellFailed, 51);*/
		}

		/* Don't copy the '*' at beginning of event record */
		else if (edp_e) --lmax;

		/* Copy comments prior to event, and possibly event itself */
		if ((edp_e) || (tlist))
		{
			/* Debug - an event record was found on input */
			IFDEB(fprintf(stderr, "%li:   editev COM \"%s\":"TFMT" at lmax=%li ltmp=%li lcom=%li lmin=%li.\n",
						  uniqid, eref, Tdhm(ewhen), lmax, ltmp, lcom, lmin));

			/* Paranoia */
			if (lmax < lmin) erf(ERF_InternalError, 52);

			/* Copy previous content */
			else if (lmax > lmin)
			{
				/* Copy up to start or end of this event */
				if (i = ftofcpy(fin, fout, lmax - lmin, fpbuf, MAX_fpbuf, lmin, lmax)) erf(i, 53);

				/* Have now transferred up to beginning, or possibly end, of event */
				lmin = lmax;
			}
		}

		/* Nothing more to do for this one: onto next line */
		else toeol(fin);

		/* Edit the event */
		if (edp = edp_e)
		{
			/* Event struct pointer, to be filled in. */
			event_p ep = 0;

			/* Debug - Editing an existing event */
			IFWARN(fprintf(stderr, "%li:   editev \"%s\" collating event edit data; %p.\n", uniqid, eref, edp));

			/* Allocate structure into which to read event data */
			if (i = editevents_init(&ep, eref, ewhen)) erf(i, 60);

			/* Back to position just after date */
			if (fseek(fin, ltmp, SEEK_SET)) erf(ERF_fseekFailed, 61);

			/* Read and convert flags field */
			if (i = tread(fin, bp, bufend, 0)) erf(ERF_BadTread+i, 62);
			ep->flags = strtoul(bp, 0, 16);
	
			/* Read player id field */
			if (i = tread(fin, ep->id = bp, bufend, &bp)) erf(ERF_BadTread+i, 63);
	
			/* Read new timer field */
			if ((i = tread(fin, bp, bufend, 0)) > 1) erf(ERF_BadTread+i, 64);
	
			/* Convert to date format */
			if (!(ep->ntimer = str2tim(bp))) erf(ERF_BadDateText, 65);
	
			/* Read scoring bonus (optional) */
			if (ep->flags & EF_ScoreBonus)
			{
				/* Read field, report errors */
				if ((i) || ((i = tread(fin, bp, bufend, 0)) > 1)) erf(ERF_BadTread+i, 66);
				ep->bonus = atol(bp);
			}
	
			/* Read police rank (optional) */
			if (ep->flags & EF_Police)
			{
				/* Read field, report errors */
				if ((i) || ((i = tread(fin, bp, bufend, 0)) > 1)) erf(ERF_BadTread+i, 67);
				ep->rank = atol(bp);
			}

			/* Read pseudonym (optional) */
			if (ep->flags & EF_MASK_Pseudonym)
			{
				/* Read field, report errors */
				if ((i) || ((i = tread(fin, ep->pseud = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 68);
			}
	
			/* Handle retargetting (optional) */
			while (ep->flags & EF_Targetting)
			{
				retarg_p retarp, *tarlist;
	
				/* Read */
				if ((i) || ((i = tread(fin, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 68);
	
				/* End of list */
				if (p[0] == '-') break;
	
				/* Allocate storage */
				if (!(retarp = malloc(sizeof(struct retarg)))) erf(ERF_mallocFailed, 121);
	
				/* Assassin? */
				if (p[0] == 'U') tarlist = &ep->untarg;
	
				/* Target? */
				else if (p[0] == 'T') tarlist = &ep->retarg;
	
				/* Invalid data */
				else erf(ERF_InternalError, 122);
	
				/* Assassins' id */
				retarp->assid = p + 1;
				
				/* Separate string */
				if (!(p = strchr(p + 1, ':'))) erf(ERF_InternalError, 123);
	
				/* Terminate */
				p[0] = '\0';
	
				/* Targets' id */
				retarp->tarid = p + 1;
	
				/* Valid ids */
				if (*(retarp->assid) <= 32) erf(ERF_InternalError, 124);
				if (*(retarp->tarid) <= 32) erf(ERF_InternalError, 125);
	
				/* Link into list */
				retarp->next = *tarlist;
				*tarlist = retarp;
			}
	
			/* Read 2 assassination fields (optional) */
			if (ep->flags & EF_Assassination)
			{
				if ((i) || ((i = tread(fin, ep->vict_pid = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 69);
				if ((i) || ((i = tread(fin, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 70);
				ep->vict_pseud = SESC(p);
			}
	
			/* Read and parse 2 news fields (optional) */
			if (ep->flags & EF_News)
			{
				if ((i) || ((i = tread(fin, bp, bufend, 0)) > 1)) erf(ERF_BadTread+i, 71);
				if (!(ep->news_date = str2tim(bp))) erf(ERF_BadDateText, 72);
				if ((i) || ((i = tread(fin, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 73);
				ep->news_head = SESC(p);
			}
	
			/* Read wantedness field (optional) */
			if (ep->flags & EF_Wantedness)
			{
				/* Read field, report errors */
				if ((i) || ((i = tread(fin, p = bp, bufend, &bp)) > 1)) erf(ERF_BadTread+i, 74);
				ep->wantedness = SESC(p);
			}

			/* Print comment to output */
			fprintf(fout, (edp->mode & EDEV_OnlyRids) ? "#-edited-repid-list:%s\n" : "#-edited:%s\n", comnt);

			/* Output event with no trailing newline */
			if (j = editevents_aux(fout, edp, 0, &ep)) erf(j, 78);

			/* Copy existing report ids */
			for (j = c = 0, p = bp; (i == 0); j++)
			{
				/* Read a report id */
				if ((i = tread(fin, bp, bufend, &p)) > 1) erf(ERF_BadTread+i, 79);

				/* Copy this id over to output? */
				if (!(edp->mode & EDEV_RemoveRid) || !(edp->rid) || (!streq(bp, edp->rid)))
				{
					/* Write out */
					if (EOF == fputc('\t', fout)) erf(ERF_fputcFailed, 81);
					if (EOF == fputs(bp, fout)) erf(ERF_fputsFailed, 82);

					/* Preserve report ids? */
					if (edp->mode & EDEV_GetRids)
					{
						/* Allocate, link, fill in */
						if (!editevents_rep(edp, bp)) erf(ERF_mallocFailed, 83);

						/* Don't overwrite rid */
						bp = p;
					}
				}

				/* Remove this id? */
				else
				{	++c;
					IFWARN(fprintf(stderr, "%li:   --- remove \"%s\" rid[%i|%i] \"%s\".\n", uniqid, eref, j, c, bp));
				}
			}

			/* Add new report id? */
			if ((edp->rid) && !(edp->mode & EDEV_RemoveRid))
			{
				/* Write out the new id after existing ones */
				if (2 > fprintf(fout, "\t%s", edp->rid)) erf(ERF_printfFailed, 85);

				/* Debug - added a report id? */
				IFWARN(fprintf(stderr, "%li:   editev \"%s\" adding rid \"%s\".\n", uniqid, eref, edp->rid));

				/* Preserve report ids? Allocate struct, link, fill in, if so. */
				if ((edp->mode & EDEV_GetRids) && (!editevents_rep(edp, edp->rid))) erf(ERF_mallocFailed, 86);
    		}

			/* Tried to remove non-existent or duplicated rid */
			else if ((edp->rid) && (c != 1))
			{
				/* Print message */
				fprintf(stderr, "%li !! editev \"%s\" remove rid \"%s\": found %i times.\n", uniqid, eref, edp->rid, c);
			}

			/* Save event pointer */
			edp->ep = ep;

			/* Position at end of line */
			/*if (-1L == (lmin = ftell(fin))) erf(ERF_ftellFailed, 84);*/

			/* Get position at end of line, move onto next line */		
			if (i = toeolpos(fin, &lmin)) erf(i, 87);

			/* Preserve buffered text */
			buf = bp;
		}

		/* Debuf - any prior events to add? */
		IFDEB(fprintf(stderr, "%li:   editev \"%s\" adding %p prior events...\n", uniqid, eref, tlist));

		/* Comment */
		if (tlist) fprintf(fout, "\n#-autoinserted-after-\"%s\":%s", eref, comnt);

		/* Add events immediately afterwards */
		for (edp = tlist; (edp); edp = edp->link)
		{
			/* Force event time to same value as preceeding event */
			edp->at = ewhen;

			/* Output new event (edp->ep = 0 on entry) with leading but no trailing newline */
			if (i = editevents_aux(fout, edp, 3, &edp->ep)) erf(i, 88);
		}

		/* Debug - Finished with processing associated with this event record */
		IFDEB(fprintf(stderr, "%li:   editev finished with \"%s\".\n", uniqid, eref));

		/* Forget event ref */
        eref = 0;

		/* No preceeding comments now */
		lcom = -1L;

		/* Test for errors on the input stream */
		if (i = ferror(fin)) erf(ERF_FileError, 86);

		/* Optimisation: stop searching if no more changes to make */
		if ((!rlist) && (!clist)) break;
	}

	/* Debug - About to copy last chunk */
	IFWARN(fprintf(stderr, "%li:   editev copying last chunk from %li.\n", uniqid, lmin));

	/* Copy remaining chunk of fin to fout, in a paranoid manner */
	if (i = ftofcpy(fin, fout, -1L, fpbuf, MAX_fpbuf, lmin, -1L)) erf(i, 90);

	/* Debug - Adding events to end of file? */
	IFDEB(fprintf(stderr, "%li:   editev writing remaining cron-events %p.\n", uniqid, clist));

	/* Comment */
	if (clist) fprintf(fout, "\n#-autoinserting-atend:%s\n", comnt);

	/* Write any remaining chronological events */
	for (edp = clist; (edp); edp = edp->link)
	{
		/* Check that these really should be after the last event */
		if (edp->at < tv) erf(ERF_Paranoia, 91);

		/* Output new event (edp->ep = 0 on entry) and trailing newline */
		if (i = editevents_aux(fout, edp, 1, &edp->ep)) erf(i, 92);
	}

	/* Update list */
	clist = edp;

	/* Debug - Any events missed? */
	IFWARN(fprintf(stderr, "%li:   editev writing complete with clist=%p rlist=%p.\n", uniqid, clist, rlist));

	/* Note buffer space used */
	if (bufp) *bufp = buf;

	/* Close input files */
	fclose(fin);
	fin = 0;

	/* Some changes were not committed? */
	if ((clist) || (rlist)) erf(ERF_IncompleteEdit, 93);

	/* Overwrite existing file? */
	if (!fred)
	{
		/* Close file */
		fclose(fout);
		fout = 0;

		/* Remove very old uinfo file */
		remove(FILE_OldEvents);
	
		/* Use output file as new uinfo file, rename old file away */
		if (rename(FILE_Events, FILE_OldEvents) || rename(FILE_NewEvents, FILE_Events)) erf(ERF_FSwapFailed, 96);
	}

	/* ... */


_erf: /***** Error handling *****/

	/* Debug */
	if (rc)
	{
		fprintf(stderr, "%li !! editevents(%i):%i", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));
		if (fin) fprintf(stderr, " fin_err=%i fin_eof=%i fin_pos=%li", ferror(fin), feof(fin), ftell(fin));
		if (fout) fprintf(stderr, " fout_err=%i fout_eof=%i fout_pos=%li", ferror(fout), feof(fout), ftell(fout));
		fputc('\n', stderr);
	}

	/* Tidy up */
	if (fin) fclose(fin);
	if (fout) fclose(fout);

	/* Finished */
	return (ERF_LOWBITS(rc));
}

#undef IFDEB
#endif
/*** The End ***/
