/*****************************************************
**  stats.c                   >> cgi-bin/stats.cgi  **
**  ----------------------------------------------  **
**  ----------------------------------------------  **
**  Updated to use new time system.                 **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define MAX_statbuf			(1024*1024)
#define MAX_fpbuf			65536
#define MAX_cgibuf 			65536
#define MAX_npairs 			600
#define EVENT_EXT_TYPE		int			/* type of  event_p->ext is irrelevent */
#define PLAYER_EXT_TYPE		player_p	/* type of player_p->ext is player_p   */

#include "!lib-general.c"
#include "!custom.c"
/*#include "!lib-cgi.c"*/
#include "!lib-uinfo-read.c"
#include "!lib-uinfo-misc.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Things to do to players */
#define SEL_None		0
#define SEL_SelAll		1
#define SEL_SelPrev		2
#define SEL_ClearSel	3

/* Type definitions for sorting/restriction functions */
typedef player_p (*sortfp) (player_p pptr);
typedef int (*restfp) (player_p pptr);

/* Global variables used in sorting or restriction functions */
static int sort, rest, vitality, law;
static sortfp sortf;
static restfp restf;
static unsigned long vitmask, vittest;

/* Complete and restricted lists of players */
static player_p playlist = 0, initpptr = 0;

/* A type to hold information on things that went wrong */
typedef unsigned int bad_type;

#define BAD_Action		((bad_type)(1<<0))
#define BAD__MAX_IDX	1

/* Descriptions of things that went wrong */
char* bad_desc[BAD__MAX_IDX] =
{
	"Invalid action specified."
};

int send_doc(int w, player_p authp);

/***************************************************************************************************************************
** Sorting functions and table                                                                                            **
***************************************************************************************************************************/
#define MAX_sorttab 9

int strlastcmp(char* p, char* q)
{
	char* t;
	while (t = strchr(p, ' ')) p = t + 1;
	while (t = strchr(q, ' ')) q = t + 1;
	return strcmp(p, q);
}
int strlastcasecmp(char* p, char* q)
{
	char* t;
	while (t = strchr(p, ' ')) p = t + 1;
	while (t = strchr(q, ' ')) q = t + 1;
	return strcasecmp(p, q);
}
player_p sort_by_score(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (qp->score >= pptr->score)); qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_danger(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (qp->danger >= pptr->danger)); qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_status(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr;
		 ((qp) && ((qp->flags & PF_MASK_Sorting) >= (pptr->flags & PF_MASK_Sorting)));
		 qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_name(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (strcasecmp(qp->name, pptr->name) <= 0)); qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_lastname(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (strlastcasecmp(qp->name, pptr->name) <= 0)); qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_id(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (strcmp(qp->id, pptr->id) <= 0)); qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_college(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (strcasecmp(qp->coll, pptr->coll) <= 0)); qp = qp->ext) pp = qp;
	return pp;
}
player_p sort_by_pseudo(player_p pptr)
{
	player_p qp, pp;
	for (pp = 0, qp = initpptr; ((qp) && (strcasecmp(qp->pseud, pptr->pseud) <= 0)); qp = qp->ext) pp = qp;
	return pp;
}
struct { char* desc; sortfp func; } sorttab[MAX_sorttab] =
{
	{ "Sort by status",         sort_by_status   },
	{ "Sort by score",          sort_by_score    },
	{ "Sort by dangerousness",  sort_by_danger   },
	{ "Sort by name",           sort_by_name     },
	{ "Sort by last name",      sort_by_lastname },
	{ "Sort by id",             sort_by_id       },
	{ "Sort by college",        sort_by_college  },
	{ "Sort by pseudonym",      sort_by_pseudo   },
	{ "Unsorted",               0                }
};

/***************************************************************************************************************************
** Restriction functions and table                                                                                        **
***************************************************************************************************************************/
#define MAX_resttab 6
#define MAX_lawtab 5

struct { char* desc; unsigned long mask, test; } lawtab[MAX_lawtab] =
{
	{ "Types of",        0,                        0                        },
	{ "Unwanted",        PF_Wanted,                0                        },
	{ "Wanted",          PF_Wanted,                PF_Wanted                },
	{ "Criminal",        PF_Wanted|PF_Incompetent, PF_Wanted                },
	{ "Incompetent",     PF_Wanted|PF_Incompetent, PF_Wanted|PF_Incompetent }
};
int rest_to_assassins(player_p pptr)
{
	return ((pptr->flags & (PF_Playing|PF_Police|PF_Special)) == (PF_Playing));
}
int rest_to_police(player_p pptr)
{
	return ((pptr->flags & (PF_Playing|PF_Police|PF_Special)) == (PF_Playing|PF_Police));
}
int rest_to_inact(player_p pptr)
{
	return ((pptr->flags & (PF_Playing|PF_Inactive|PF_Special)) == (PF_Playing|PF_Inactive));
}
int rest_to_inact_ass(player_p pptr)
{
	return ((pptr->flags & (PF_Playing|PF_Inactive|PF_Police|PF_Special)) == (PF_Playing|PF_Inactive));
}
int rest_to_inact_pol(player_p pptr)
{
	return ((pptr->flags & (PF_Playing|PF_Inactive|PF_Police|PF_Special)) == (PF_Playing|PF_Inactive|PF_Police));
}
struct { char* desc; restfp func; } resttab[MAX_resttab] =
{
	{ "Players",                 0					  },
	{ "Assassins",               rest_to_assassins	  },
	{ "Police",                  rest_to_police	      },
	{ "Inactive players",        rest_to_inact		  },
	{ "Inactive assassins",      rest_to_inact_ass	  },
	{ "Inactive police",         rest_to_inact_pol	  }
};
int rest_show(player_p pptr)
{
	char sid[MAX_id + 4];

	sprintf(sid, "id_%s", pptr->id);
	return (findval(sid, 0)) ? 1 : 0;
}
char* vittab[3] = { "All", "Living", "Dead" };

/***************************************************************************************************************************
** Transfer control to other CGI apps to do things. (act) has been validated.                                             **
***************************************************************************************************************************/
struct psort { char* pid; unsigned long x; struct psort *next, *prev; };

int invokeaction(int act)
{
	struct psort *head = 0, *tail = 0, *this, *that;
	int n, i, rc = 0;
	FILE *fp;
	char *p;


	/* Search for selected players of form "id_xxx" */
	for (n = i = 0; (i < npairs); i++)
	{
		/* Pattern match? */
		if (strncmp("id_", name[i], 3)) continue;

		/* Check id suffix */
		if (!name[i][3]) erf(ERF_MissingPid, 1);

		/* Paranoia: too many players found? */
		if (n++ >= MAX_CheckAct) erf(ERF_TooManyPlayers, 2);

		/* New struct */
		if (!(this = malloc(sizeof(struct psort)))) erf(ERF_mallocFailed, 3);

		/* Save player id */
		this->pid = name[i] + 3;

		/* Sorting value */
		this->x = strtoul(val[i], 0, 10);

		/* Ensure correct ordering of data: optimised for "already correct" case by starting from tail-end */
		for (that = tail; (that); that = that->prev) if (that->x <= this->x) break;

		/* Add at head of doubly-linked list */
		if (!that)
		{
			if (this->next = head) head->prev = this;
			this->prev = 0;
			head = this;
			if (!tail) tail = this;
		}

		/* Add after 'that' entry in list */
		else
		{
			if (this->next = that->next) that->next->prev = this;
			that->next = this;
			this->prev = that;
			if (that == tail) tail = this;
		}
	}

	/* No ids found? */
	if (n <= 0) erf(ERF_NoIdGiven, 4);

	/* Prepare temporary file for chained stdin */
	if (!(fp = prepforstdin())) erf(ERF_tmpfileFailed, 5);

	/* Encode authentication */
	urlencode(fp, "authname",  findval("authname", 1), 1);
	urlencode(fp, "authpass",  findval("authpass", 1), 0);
	urlencode(fp, "stat_vit",  findval("vitality", 1), 0);
	urlencode(fp, "stat_rest", findval("rest",     1), 0);
	urlencode(fp, "stat_sort", findval("sort",     1), 0);
	urlencode(fp, "stat_law",  findval("law",      1), 0);

	/* Action type */
	fprintf(fp, "&act=%i", /*(act < ACTIONDIV) ?*/ (act + 1) /*: (act + 1 - ACTIONDIV)*/);

	/* Pass on sorted player ids */
	for (i = 1, this = head; (this); this = this->next, i++)
	{
		fprintf(fp, "&pl_%i", i);
		urlencode(fp, "id", this->pid, 1);
	}

	/* Debug */
	fprintf(stderr, "%li:   invoking act=%d with n=%i\n", uniqid, act, n);

	/* Pass control onto which other executable? */
	p = /*(act < ACTIONDIV) ?*/ CGI_ROOT "pled.cgi" /*: CGI_ROOT "event.cgi"*/;

	/* Output HTTP headers */
    /*printf("Content-type: text/plain" LF LF);*/

	/* Transfer control */
	if (!(rc = launchwithstdin(p, fp))) erf(ERF_InternalError, 6);


_erf: /***** Handle errors *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   invokeaction(%i):%i type %d\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), act);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Function hooks to build statistics table. We use the pp->link fields to maintain the list of all players, and pp->ext  **
** to maintain the sorted and restricted list.                                                                            **
***************************************************************************************************************************/
int makestatstab__uf(player_p pp, int state, void* ctx)
{
	player_p qptr;


	/* Maintain separate list of all players */
	pp->link = playlist;
	playlist = pp;

	/* Ignore hidden players and check restrictions */
	if ((pp->flags & PF_Hidden) || ((pp->flags & vitmask) != vittest) || ((restf) && (!restf(pp)))) pp->ext = 0;
	
	/* Sort and link into restricted list behind 'qptr' */
	else if ((sortf) && (qptr = sortf(pp))) 
	{
		pp->ext   = qptr->ext;
		qptr->ext = pp;
	}

	/* Or at head of restricted list */
	else
	{
		pp->ext  = initpptr;
		initpptr = pp;
	}

	/* Don't call additional function, and don't overwrite data or player struct. */
	return 32+16+8+1;
}

/* Print out the bits of HTML that allow selection of the sorting method */
void printsortform(int dvit, int dlaw, int drest, int dsort)
{
	int i;
	char *p;

	if (OPT_EasySort) p = "<p>Show ";

	else p = LF "<table rows=5 cols=2 border=0 cellspacing=2 cellpadding=2>"
			 LF "<tr><td bgcolor="GREEN">Vitality:&nbsp;</td>" LF "<td bgcolor="RED">";

	printf("%s<select name=\"vitality\" size=1>", p);

	for (i = 0; (i < 3); i++) printf("<option value=\"%d\"%s>%s</option>\n", i,
									  (i == dvit) ? " selected" : "", vittab[i]);

	if (!OPT_EasySort)
	{
		puts("</select></td></tr>"
			 LF "<tr><td bgcolor="GREEN">Lawfulness:&nbsp;</td>"
			 LF "<td bgcolor="BLUE"><select name=\"law\" size=1>");

		for (i = 0; (i < MAX_lawtab); i++) printf("<option value=\"%d\"%s>%s</option>\n", i,
											  (i == dlaw) ? " selected" : "", lawtab[i].desc);

		puts("</select></td></tr>"
			 LF "<tr><td bgcolor="GREEN">Player&nbsp;type:&nbsp;</td>"
			 LF "<td bgcolor="RED"><select name=\"rest\" size=1>");

		for (i = 0; (i < MAX_resttab); i++) printf("<option value=\"%d\"%s>%s</option>\n", i,
												   (i == drest) ? " selected" : "", resttab[i].desc);

		p = "</td></tr>" LF "<tr><td bgcolor="GREEN">Sorting:&nbsp;</td>" LF "<td bgcolor="RED">";
	}

	else p = " &nbsp; ";

	printf("</select>%s<select name=\"sort\" size=1>", p);

	for (i = 0; (i < MAX_sorttab); i++) printf("<option value=\"%d\"%s>%s</option>\n", i,
											   (i == dsort) ? " selected" : "", sorttab[i].desc);

	puts((OPT_EasySort) ? "</select> &nbsp;" : "</select></td></tr></table>");
}

/***************************************************************************************************************************
** Main code. Doesn't make use of command line arguments.                                                                 **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos;
	char *thetimestr, *authname = "", *authpass = "", *p;
	int colour, i, j, admin, rc = 0, lockfd = 0, action = 0, sel = SEL_None;
	player_p pptr, authp = 0;
	bad_type bad = 0;


	/* Output HTTP headers */
	/* printf("Content-type: text/html" LF LF); */

	/* Initialse stuff: time, file creation umask, pid, make note in log and redirect stderr there */
	if ((rc = geninit(&thetimestr, 0117, 1, "Stats")) < 0) exit(send_doc(-rc, 0));

	/* Logged error handler now established */
	else if (rc) erf(rc, 1);

	/* Parse CGI input */
	if ((i = parsecgi()) > 0) erf(i, 2);

	/* No input given */
	/*else if (i == 0) erf(ERF_Void, 2);*/

	/* Check security */
	/*if (strcmp(findval("security", 1), SECURITY)) exit(insecure(2+32, "stats"));*/

	/* If no authentication name given, assume public user */
	if (!(authname = findval("authname", 0))) authname = PUB_ID;

	/* Authentication password */
	else authpass = findval("authpass", 1);

	/* Note whether action was valid */
	i = 1;

	/* Selected action */
	if (!(p = findval("action", 0)) || (!isdigit(*p)) || ((action = atoi(p)) < 0) || (action >= MAX_act)) i = action = 0;

	/* Invoke action */
	if (findval("sub_act", 0))
	{
		/* Action type was valid */
		if (i) erf(invokeaction(action), 0);

		/* Otherwise, invalid, so ignore */
		sel = SEL_None;
		bad |= BAD_Action;
	}

	/* Resort, do nothing else */
	else if (findval("resort", 0) || findval("sub_show", 0)) sel = SEL_None;

	/* Select all on new list? */
	else if (findval("sub_selall", 0)) sel = SEL_SelAll;

	/* Select all on prev *and* new list? */
	else if (findval("sub_selprev", 0)) sel = SEL_SelPrev;

	/* Clear selection? */
	else if (findval("sub_clearsel", 0)) sel = SEL_ClearSel;

	/* Otherwise, don't fiddle with selection */
	else sel = SEL_None;

	/* Claim lockfile */
	if (!(lockfd = setflock("Stats.c "))) erf(ERF_flockFailed, 3);

	/* Sorting method specified? */
	j = (p = findval("sort", 0)) ? (atoi(p)) : (0);

	/* Choose sorting function */
	for (sort = (MAX_sorttab - 1); (sort > 0) && (j != sort); sort--); sortf = sorttab[sort].func;

	/* Sorting restriction specified? */
	rest = (p = findval("rest", 0)) ? (atoi(p)) : (0);

	/* Lawfulness restriction specified? */
	law = (p = findval("law", 0)) ? (atoi(p)) : (0);
	
	/* Vitality restriction specified? */
	if (((vitality = (p = findval("vitality", 0)) ? atoi(p) : (0)) < 0) || (vitality > 2)) vitality = 0;

	/* Special view: only given players */
	if (findval("sub_show", 0))
	{
		restf = rest_show;
		vitmask = vittest = 0;
	}

	/* Normal restriction/lawfulness functions */
	else
	{
		/* Choose restriction function */
		for (j = rest, rest = (MAX_resttab - 1); (rest > 0) && (j != rest); rest--);
		restf = resttab[rest].func;

		/* Choose lawfulness modifier */
		for (j = law, law = (MAX_lawtab - 1); (law > 0) && (j != law); law--);

		/* Set restriction bitmasks */
		vitmask = lawtab[law].mask | ((vitality > 0) ? (PF_Dead) : (0));
		vittest = lawtab[law].test | ((vitality > 1) ? (PF_Dead) : (0));
	}

	/* Read statistics data */
	if (rc = uread(0, makestatstab__uf, 0, statbuf, statbuf_end, &statbuf_pos, 0, UREAD_GetTargs)) erf(rc, 4);

	/* Check authentication */
	for (authp = playlist; (authp); authp = authp->link) if (!strcmp(authp->id, authname)) break;

	/* Invalid authentication? */
	if ((!authp) || (!authp->name) || !(authp->flags & PF_Login)
								   || (!(authp->flags & PF_Public) && strcmp(authp->pass, authpass)))
	{
		authp = 0;
		erf(ERF_BadAuth, 5);
	}

	/*fprintf(stderr, "%li:   main i=%i act=%d under auth \"%s\".\n", uniqid, i, sel, authp->id);*/

	/* Could do evolution to arbitrary times here */

	/* Admin privs? */
	admin = ((authp->flags & PF_Admin) ? 2 : 0) | ((authp->flags & PF_Observer) ? 1 : 0);

	/* If not priv'd, ignore some earlier warnings */
	if (admin < 2) bad &= 0;

	puts("Content-type: text/html" LF
		 LF "<HTML><HEAD><TITLE>"GAME" Statistics Page</TITLE></HEAD>"
		 LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		 LF "<h1>"GAME" Player Statistics</h1><p>Click on players' names for more detailed "
		 LF "information and statistics...</p>"
		 LF "<form method=\"POST\" action=\""CGI_URL"stats.cgi\">");

	if (bad)
	{
		puts(LF "<h3><img src=\""CROSS_URL"\" align=left "CROSS_SIZ" hspace=8><font color=#FFCCCC>");

		for (i = 0; (i < BAD__MAX_IDX); i++) if (bad & (1 << i)) puts(bad_desc[i]);

		puts("</font><br clear=left></h3>");
	}

	if (!(authp->flags & PF_Public))
	{
		printf(LF "<input type=\"hidden\" name=\"authname\" value=\"%s\"><p>Authenticated as <b>%s</b>%s.</p>"
			   LF "<input type=\"hidden\" name=\"authpass\" value=\"%s\">",
			   authp->id, authp->name, ((admin & 2) ? " (admin)" : ""), authp->pass);
	}

	if (!(admin & 1)) puts(LF "<p><b>Only those playing are allowed to view the names and addresses of players.</b></p>");

	if (admin > 0)
	{
		puts("<p>");
		printsortform(vitality, law, rest, sort);
		puts("<input value=\"Resort\" type=\"submit\" name=\"resort\"></p>");
	}

	/* Force default sorting */
	else
	{
		player_p pp;

		/* Defaults */
		law = 0;
		vitality = 0;
		sort = MAX_sorttab - 1;

		/* Restriction function */
		restf = resttab[rest = 0].func;

		/* Set restriction bitmasks */
		vitmask = lawtab[law].mask | ((vitality > 0) ? (PF_Dead) : (0));
		vittest = lawtab[law].test | ((vitality > 1) ? (PF_Dead) : (0));

		/* Resort players */
		for (initpptr = 0, pp = playlist; (pp); pp = pp->link)
		{
			/* Ignore hidden players and check restrictions */
			if ((pp->flags & PF_Hidden) || ((pp->flags & vitmask) != vittest) || ((restf) && (!restf(pp)))) continue;

			/* Place in list */	
			pp->ext = initpptr;
			initpptr = pp;
		}
	}

 #if defined(TMPSTR1) || defined(TMPSTR2) || defined(TMPSTR3)
  #error "*** TMPSTR and/or TMPINT already #defined ***"
 #endif
 #if OPT_ShowScore
  #if OPT_ShowShortComments
   #define TMPSTR1 "<td>Water:</td><td>Score:</td>"
   #define TMPSTR2 "11"
   #define TMPSTR3 "6"
  #else
   #define TMPSTR1 "<td>Score:</td>"
   #define TMPSTR2 "10"
   #define TMPSTR3 "5"
  #endif
 #else
  #if OPT_ShowShortComments
   #define TMPSTR1 "<td>Water:</td>"
   #define TMPSTR2 "10"
   #define TMPSTR3 "5"
  #else
   #define TMPSTR1 ""
   #define TMPSTR2 "9"
   #define TMPSTR3 "4"
  #endif
 #endif

	if (admin > 1)
	{
		printf("<p><table cols="TMPSTR2" cellspacing=4 cellpadding=4 border=0><thead>"
			   LF "<td></td><td>Name</td><td>College</td><td>Address:</td>"TMPSTR1"<td>Kills:</td>"
			   LF "<td>Assassins:</td><td>Targets:</td><td>Status:</td><td>Pseudonym:</td></thead>");
	}
	else if (admin == 1)
	{
		printf("<p><table cols="TMPSTR2" cellspacing=4 cellpadding=4 border=0><thead>" LF
			   "<td>Name</td><td>College</td><td>Address:</td>"TMPSTR1"<td>Kills:</td><td>Deaths:</td>"
			   "<td>Status:</td><td>Pseudonym:</td></thead>");
	}
	else
	{
		printf("<p><table cols="TMPSTR3" cellspacing=4 cellpadding=4 border=0><thead>" LF
			   "<td>Pseudonym:</td>"TMPSTR1"<td>Kills:</td><td>Deaths:</td><td>Status:</td></thead>");
	}

 #undef TMPSTR1
 #undef TMPSTR2
 #undef TMPSTR3

	/* Players who passed the restriction */
	for (pptr = initpptr, i = 0, colour = 0; (pptr); pptr = pptr->ext, i++)
	{
		/* Ignore hidden players */
		if (pptr->flags & PF_Hidden) continue;
	
		/* Alternate background colour */
		p = (colour = 1 - colour) ? BLUE : RED;
	
		/* Checkbox to allow selection of player */
		if (admin > 1)
		{
			char sid[MAX_id + 4];

			sprintf(sid, "%cd_%s", (sel == SEL_SelPrev) ? 'o' : 'i', pptr->id);
	
			if (!((j = (sel == SEL_SelAll)) || (sel == SEL_ClearSel))) j = (findval(sid, 0) != 0);

			printf(LF "<tr><td valign=middle><input type=\"checkbox\" name=\"id_%s\" value=\"%d\"%s></td>"
				   LF "<input type=\"hidden\" name=\"od_%s\" value=\"%d\">",
				   pptr->id, i, (j) ? " checked" : "", pptr->id, i);
		}

		/* Ignore issues of selection if un-priv'd */
		else fputs(LF "<tr>", stdout);

		/* Link to plinfo */
		printf(LF "<td bgcolor=%s><font color="WHITE"><a href=\"plinfo.cgi?id=%s\">", p, pptr->id);

		/* Print pseud */
		if (admin == 0)
		{
			encapsulate(pptr->pseud, stdout);
			puts("</a></font></td>");
		}

		/* Print name, college, address */
		else printf("%s</a></font></td>" LF "<td bgcolor=%s><font color="WHITE">%s</font></td>"
					LF "<td bgcolor=%s><font color="WHITE">%s</font></td>",
					pptr->name, p, pptr->coll, p, pptr->addr);

	#if OPT_ShowShortComments
	{
		char *tmpptr, *wibble, *bibble, *tmpstring;

		bibble = "???";
		
		if (tmpstring = malloc(strlen(pptr->mafia) + 1))
		{
			for (tmpptr = tmpstring, wibble = pptr->mafia; (*wibble); wibble++) *(tmpptr++) = tolower(*wibble);
	
			*tmpptr = 0;
	
			if (strstr(tmpstring, "no water")) bibble = "NWW";
	
			else if (strstr(tmpstring, "full water")) bibble = "Full";
	
			else if (strstr(tmpstring, "with care")) bibble = "WWC";
			
            free(tmpstring);
		}

		printf(LF "<td bgcolor=%s><font color="WHITE">%s</font></td>", p, bibble);
	}
	#endif

	#if OPT_ShowScore

		/* Print score, dangerousness */
		printf(LF "<td bgcolor=%s><font color="WHITE"><b>%li</b>"
	#if OPT_ShowDanger
				"(%0.2lf)"
	#endif
				"</font></td>", p, ((pptr->score + SCORE_ROUND) >> SCORE_SHIFT)
	#if OPT_ShowDanger
			   , (pptr->danger / ((double)(1<<DANGER_SHIFT)))
	#endif
			   );
	#endif

		/* Lookup assassin/target names from id's */
		if (admin > 1)
		{
			target_p tarp;
			player_p qptr;

			/* Print kills, deaths */
			printf(LF "<td bgcolor=%s><font color="WHITE"><b>%i&nbsp;/&nbsp;%i</b></font></td>"
				   LF "<td bgcolor=%s><font color="WHITE">", p, pptr->kills, pptr->deaths, p);

			for (tarp = pptr->asses, j = 0; (tarp); tarp = tarp->next, j++)
			{
				if ((tarp->id) && (strcmp(tarp->id, NON_ID)))
				{
					for (qptr = playlist; (qptr); qptr = qptr->link) if (!strcmp(qptr->id, tarp->id)) break;
					if (!qptr) printf("%s<b>[Unknown:%s]</b>", (j) ? "<br>" : "", tarp->id);
					else printf("%s<a href=\"plinfo.cgi?id=%s\">%s</a>", (j) ? "<br>" : "", qptr->id, qptr->name);
				}
				else puts("<b>Unallocated</b>");
			}
			if (!j) puts("<b>No assassins!</b>");
	
			printf("</font></td><td bgcolor=%s><font color="WHITE">", p);

			for (tarp = pptr->targs, j = 0; (tarp); tarp = tarp->next, j++)
			{
				if ((tarp->id) && (strcmp(tarp->id, NON_ID)))
				{
					for (qptr = playlist; (qptr); qptr = qptr->link) if (!strcmp(qptr->id, tarp->id)) break;
					if (!qptr) printf("%s<b>[Unknown:%s]</b>", (j) ? "<br>" : "", tarp->id);
					else printf("%s<a href=\"plinfo.cgi?id=%s\">%s</a>", (j) ? "<br>" : "", qptr->id, qptr->name);
				}
				else puts("<b>Unallocated</b>");
			}
			if (!j) puts("<b>No targets!</b>");

			printf("</font></td><td bgcolor=%s><font color="WHITE">", p);
		}

		/* Print kills, deaths */
		else printf(LF "<td bgcolor=%s><font color="WHITE"><b>%i</b></font></td>"
					LF "<td bgcolor=%s><font color="WHITE"><b>%i</b></font></td>"
					LF "<td bgcolor=%s><font color="WHITE">", p, pptr->kills, p, pptr->deaths, p);

		printpstatus(pptr->flags, pptr->timer, 0);
	
		if (admin > 0)
		{
			printf("</font></td><td bgcolor=%s><font color="WHITE">", p);
			encapsulate(pptr->pseud, stdout);
		}

		puts("</font></td>");
	}

	puts("</table></p>");

	/* Admin options: control selection, add events, edit user info, etc */
	if (admin > 1)
	{
		puts("<p><input value=\"Clear selection\" type=\"submit\" name=\"sub_clearsel\">"
			 "<input value=\"Select listed\" type=\"submit\" name=\"sub_selprev\">"
			 "<input value=\"Select all\" type=\"submit\" name=\"sub_selall\">&nbsp; or:"
			 "&nbsp;&nbsp;<input value=\"Act on selected players\" type=\"submit\" name=\"sub_act\">"
			 "&nbsp;to&nbsp;<select name=\"action\" size=1>");

		for (i = 0; (i < MAX_act); i++)
		{
			printf("<option value=\"%d\"%s>%s</option>\n", i, (i == action) ? " selected" : "", acttab[i]);
		}

		puts("</select></p>");
	}

	printf(LF "</form><p align=right><hr>Produced at %s</p></BODY></HTML>" LF, thetimestr);


_erf: /***** Error handling *****/

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Debug */
	if (rc) fprintf(stderr, "%li:   main(%i):%i:%s authname=\"%s\" authpass=\"%s\" %s\n", uniqid,
					ERF_CONTEXT(rc), ERF_TYPE(rc), ERF_STR(rc), authname, authpass, ((authp) ? "good " : ""));

	/* Only retain low bits of normal rc's */
	if (rc) send_doc(ERF_LOWBITS(rc), authp);

	return 0;
}

/***************************************************************************************************************************
** Generate various special documents.                                                                                    **
***************************************************************************************************************************/
char* regurg_excl[7] = { "authname", "authpass", "sort", "law", "rest", "vitality", 0 };

int send_doc(int w, player_p authp)
{
	int rc = 0;
	char* p;

	/* Output HTTP headers */
    printf("Content-type: text/html" LF LF);

	/* Pass on most errors */
	if (w != ERF_BadAuth)
	{
		error_doc(w);
		return 0;
	}

	puts(LF "<HTML><HEAD><TITLE>" GAME ": Player listing</TITLE></HEAD>"
		 LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink="WHITE">"
		 LF "<h1><hr>Player Listing<hr></h1><form action=\""CGI_URL"stats.cgi\" method=\"POST\">"
		 LF "<h3><img align=left src=\""CROSS_URL"\" border=0 width=34 height=34 hspace=8>"
		 LF "<font color=#FFCCCC>Authentication failed!</font></h3><p>"
		 LF "<p>Authenticate as: &nbsp; <select size=1 name=\"authname\">");

	if (rc = makeidmenu(((p = findval("authname", 0)) ? p : PUB_ID), PF_Login, PF_Login, OPT_IDMenuMode)) erf(rc, 1);

	puts("</select> &nbsp; password: <input name=\"authpass\" type=\"password\" size=24 maxlength=64></p><p>");	 

	printsortform(atoi(findval("vitality", 1)), atoi(findval("law", 1)),
				  atoi(findval("rest", 1)),     atoi(findval("sort", 1)));

	puts("</p><p><input value=\"Submit\" type=\"submit\"></p>");

	regurgitate(regurg_excl);

	puts("</form></BODY></HTML>");

	return 0;


_erf: /***** Error handling *****/

	fprintf(stderr, "%li:   send_doc(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));
	return 0;
}

