/*****************************************************
**  lib-evolve.c                         (library)  **
**  ----------------------------------------------  **
**  Provides a routine to process events from the   **
**  events.txt file until some limiting condition   **
**  is met, maintaining and updating various bits   **
**  of state per player, and globally, throughout   **
**  this "time-evolution" of the game. This forms   **
**  the core of all programs that need more state   **
**  info than the snapshot in the uinfo.txt file.   **
**  The routine is highly configurable and of the   **
**  greatest importance to the system: take care!   **
**  ----------------------------------------------  **
**  converted to new time system?                   **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#ifdef included__lib_evolve
 #error "** lib-evolve multiply #included **"
#else
 #define included__lib_evolve fromfile

#ifndef included__lib_event_read
 #include "!lib-event-read.c"
#endif

/* Avoid messyness with maths library on non-Acorn platforms which may define a clashing erf() function */
#ifndef ACORN
 double exp(double x) { return 1 + x + 0.5*x*x + x*x*x/6.0; }
#endif

/***************************************************************************************************************************
** Definitions                                                                                                            **
***************************************************************************************************************************/

#define EVOLVE_Init				(((evolve_flags)1)<<0)	/* Special */
#define EVOLVE_TimersToLimit	(((evolve_flags)1)<<1)	/* Update timers to limit after all relevent events processed */
#define EVOLVE_MustMatch		(((evolve_flags)1)<<2)	/* Cause an error if any pids not on initial list appear */
#define EVOLVE_MatchAll			(((evolve_flags)1)<<3)	/* Add all players not on initial list anyway */
#define EVOLVE_FuncAssass		(((evolve_flags)1)<<4)	/* Call hook for assassination events */
#define EVOLVE_FuncEndCond		(((evolve_flags)1)<<5)	/* Call hook when events pass end condition */
#define EVOLVE_FuncExhaust		(((evolve_flags)1)<<6)	/* Call hook upon running out of events before end condition */
#define EVOLVE_FuncNews			(((evolve_flags)1)<<7)	/* Call hook for newsworthy events */
#define EVOLVE_FuncOnTick		(((evolve_flags)1)<<24)	/* Call hook when time changes */
#define EVOLVE_InheritPseud		(((evolve_flags)1)<<25)	/* Default pseudonym is inherited over reincarnation */
#define EVOLVE_Prior			(((evolve_flags)1)<<26)	/* Ignore events matching final ref */
#define EVOLVE_Exclusive		(((evolve_flags)1)<<27)	/* Ignore events happening at end time */
#define EVOLVE_BacklinkPlayers	(((evolve_flags)1)<<28)	/* */
#define EVOLVE_OnlyPseuds		(((evolve_flags)1)<<29)	/* */
#define EVOLVE_Targetting		(((evolve_flags)1)<<30)	/* Do retargettings */
#define EVOLVE_StrictTargets	(((evolve_flags)1)<<31)	/* Be intollerent */

#define EVOLVE_Comments			(((evolve_flags)1)<<8)	/* Init: Record changes to comment (wantedness) field */
#define EVOLVE_AllExists		(((evolve_flags)1)<<9)	/* Init: Don't overwrite previous existences */
#define EVOLVE_AllPseuds		(((evolve_flags)1)<<11)	/* Init: Collect all pseudonyms */
#define EVOLVE_NoPseuds			(((evolve_flags)1)<<12)	/* Init: Don't bother with pseudonyms at all */
#define EVOLVE_NoTargetting		(((evolve_flags)1)<<13)	/* Init: Don't do retargettings */
#define EVOLVE_ScoreExists		(((evolve_flags)1)<<14)	/* Init: Maintain scoring table for each existence */
#define EVOLVE_ScorePlayer		(((evolve_flags)1)<<15)	/* Init: Maintain overall scoring table for player */

#define EVLadd_Comments			(((evolve_flags)1)<<16)	/* Add: Record changes to comment (wantedness) field */
#define EVLadd_AllExists		(((evolve_flags)1)<<17)	/* Add: Don't overwrite previous existences */
#define EVLadd_AllPseuds		(((evolve_flags)1)<<19)	/* Add: Collect all pseudonyms */
#define EVLadd_NoPseuds			(((evolve_flags)1)<<20)	/* Add: Don't bother with pseudonyms at all */
#define EVLadd_NoTargetting		(((evolve_flags)1)<<21)	/* Add: Don't do retargettings */
#define EVLadd_ScoreExists		(((evolve_flags)1)<<22)	/* Add: Maintain scoring table for each player existence */
#define EVLadd_ScorePlayer		(((evolve_flags)1)<<23)	/* Add: Maintain overall scoring table for player */

#define SCDEF_Comp_Norm_Ass		(0)
#define SCDEF_Inco_Norm_Ass		(1)
#define SCDEF_Comp_Want_Ass		(2)
#define SCDEF_Inco_Want_Ass		(3)
#define SCDEF_Comp_Norm_Pol		(4)
#define SCDEF_Inco_Norm_Pol		(5)
#define SCDEF_Comp_Want_Pol		(6)
#define SCDEF_Inco_Want_Pol		(7)
#define SCDEF_LegalNonPlayer	(8)
#define SCDEF_IllegalNonPlayer	(9)
#define SCDEF_Special			(10)
#define SCDEF_Unknown			(11)
#define SCDEF_None				(12)
#define SCDEF_MAX				(13)			  /* 13 ... 26 are fatal versions of the above */
#define SCDEF_MAX_Cols			(13)
#define SCDEF_Fatal			SCDEF_MAX
#define SCDEF_MAX_Global	(2*SCDEF_MAX)  		  /* Global table has fatal/non-fatal duplicates (only first index) */
#define SCDEF_Victim		SCDEF_MAX_Global
#define SCDEF_MAX_Single	(2*SCDEF_MAX_Global)  /* Individual tables have Assassin/Victim duplicates ( "" "" "" ) */

#ifndef PEXIST_EXT_TYPE
 #error "<<< PEXIST_EXT_TYPE undefined >>>"
#endif
#ifndef PSTATE_EXT_TYPE
 #error "<<< PSTATE_EXT_TYPE undefined >>>"
#endif
#ifndef PSTATE_PLP_TYPE
 #error "<<< PSTATE_PLP_TYPE undefined >>>"
#endif

typedef unsigned long evolve_flags;

typedef unsigned int pseud_flags;

#define PSFL_Mine			(((pseud_flags)1)<<0)
#define PSFL_Used			(((pseud_flags)1)<<1)
#define PSFL_Temp			(((pseud_flags)1)<<2)
#define PSFL_NotYet			(((pseud_flags)1)<<3)
#define PSFL_Later			(((pseud_flags)1)<<4)
#define PSFL_XDefault		(((pseud_flags)1)<<5)
#define PSFL_XPrior			(((pseud_flags)1)<<6)

typedef unsigned int evolve_notes;

#define NB_ScoreBonus		(((evolve_notes)1)<<0)
#define NB_ScoreKill		(((evolve_notes)1)<<1)
#define NB_ScoreIllegal		(((evolve_notes)1)<<2)
#define NB_ScTabFatal		(((evolve_notes)1)<<3)
#define NB_ScTabAttempt		(((evolve_notes)1)<<4)

/* Structure to hold a pseudonym */
struct ppseud
{
	char* text;
	ppseud_p link_ex, link_pl;
	pseud_flags fl;
};

/* Structure to hold state for each existence ("life") of the player */
struct pexist
{
	pstate_p psp;
	pexist_p link_all, link_pl;
	unsigned long fl;
	long score;
	long rank;
	char *comment;
	char *killedby;
	ppseud_p pseuds, dps;
	timeval timer, twant, born, died, tpol, tlact;
	PEXIST_EXT_TYPE ext;
	int *scoretab;
};

/* Structure to hold complete, historical player state during evolution */
struct pstate
{
	PSTATE_EXT_TYPE ext;
	PSTATE_PLP_TYPE plp;
	long score;
	long danger;
	char *cps;
	char pid[MAX_id];
	target_p targs, asses;
	unsigned long pflags;
	evolve_flags state;
	pstate_p next;
	ppseud_p allpseuds;
	struct pexist firstexist;
	pexist_p exist;
	int *scoretab;
	ppseud_p chkpseud;
};

typedef struct evolve__ctx evolve__ctx_t;

typedef int (*evolve_fn) (event_p x1, int i, void* x6, evolve__ctx_t* x7);

typedef int (*evolve_evfn) (event_p x1, pstate_p x2, pstate_p x3, int x4, int x5, void* x6, evolve__ctx_t* x7);

typedef int (*evolve_repfn) (event_p ep, char* rid, int state, void* ctx, evolve__ctx_t* evolvectx);

/* Structure to hold evolve__xxx helper function context */
struct evolve__ctx
{
	timeval endtime, lasttime;
	u32b elapsed;
	evolve_flags flags;
	pstate_p plist;
	pexist_p elist;
	int scmax_s;
	int *scoretab;
	int (*scdef_s)[SCDEF_MAX];
	int (*scdef_t)[SCDEF_MAX];
	int existcnt;
	evolve_fn func;
	evolve_evfn evfunc, secfunc;
	evolve_repfn repfunc;
	void* fctx;
	PSTATE_PLP_TYPE pl_list;
	event_p eptr;
	char* endref;
	int matched;
	evolve_notes notes;
	FILE* scbrkfp;
	ppseud_p chkpseudlist;
};

struct evolve__results
{
	pexist_p elist;
	int existcnt;
};

/***************************************************************************************************************************
** Declarations                                                                                                           **
***************************************************************************************************************************/
pstate_p emptypidx[1] = { 0 };

/***************************************************************************************************************************
** Helper functions                                                                                                       **
***************************************************************************************************************************/
#ifdef TEST
double dangerbonus(long danger)
{
	return 1.0; /*exp(0.15 * danger / ((double)(1<<DANGER_SHIFT)));*/
}
#else
double dangerbonus(long danger)
{
	/*return exp(0.15 * danger / ((double)(1<<DANGER_SHIFT)));*/
	return 1 + (danger / ((double) (7 << DANGER_SHIFT)));
}
#endif

/* Add existences upon resurrection/reincarnation */
int resurrect(pstate_p pls, struct evolve__ctx* ctx, pexist_p* ppexpp, timeval when, unsigned long fff)
{
	pexist_p exp = pls->exist, pexp = (ppexpp) ? (*ppexpp) : pls->exist;
	int i, rc = 0;


	/* Paranoia */
	if (!when) erf(ERF_InternalError, 1);

	/* Optional: Allocate new existence structure, retain old */
	if (pls->state & EVOLVE_AllExists)
	{
		/* Size of scoring table, or zero if none required */
		i = (!(pls->state & EVOLVE_ScoreExists)) ? 0 : (ctx->scmax_s + 1) * sizeof(int);

		/* Allocate new structure */
		if (!(exp = malloc(i + sizeof(struct pexist)))) erf(ERF_mallocFailed, 5);

		/* Point to any scoring table */
		exp->scoretab = (!i) ? 0 : (int*)(((char*) exp) + sizeof(struct pexist));

		/* Increment count of existence structures */
		ctx->existcnt++;

		/* Initialise meta-data */
		exp->psp = pls;

		/* No extension word */
		exp->ext = 0;

		/* Copy previous flags */
		exp->fl = (pexp->fl &= ~PF_OffTheHook);

		/* Set "off the hook" flag */
		if (fff) pexp->fl |= PF_OffTheHook;

		/* No previous comment */
		if (!pexp->comment) exp->comment = 0;

		/* Or copy old comment (can only happen if ctx->flags EVOLVE_Comments set) */
		else if (exp->comment = malloc(strlen(pexp->comment)+1)) strcpy(exp->comment, pexp->comment);

		/* Failure */
		else erf(ERF_mallocFailed, 5);

		/* Inherit ranks */
		exp->rank = pexp->rank;

		/* Option: inherit most recent default pseudonym */
		if ((ctx->flags & EVOLVE_InheritPseud) && !(pls->state & EVOLVE_NoPseuds) && ((pexp->dps)))
		{
			/* Allocate new structure, but no new text storage */
			if (!(exp->dps = malloc(sizeof(struct ppseud)))) erf(ERF_mallocFailed, 6);

			/* Link directly to previous storage */
			(exp->pseuds = exp->dps)->text = pexp->dps->text;

			/* Pseud flags */
			exp->dps->fl = 0;

			/* Add as only member of existence chain, don't link into player chain (already there) */
			exp->dps->link_ex = exp->dps->link_pl = 0;
		}

		/* Otherwise, start pseudonymless (don't inherit previous) */
		else exp->pseuds = exp->dps = 0;

		/* Link into player state chain */
		exp->link_pl = pls->exist;
		pls->exist = exp;

		/* Link into entire existence chain */
		exp->link_all = ctx->elist;
		ctx->elist = exp;

		/* Retain timer value */
		exp->timer = pexp->timer;
	}

	/* Otherwise, overwrite old structure */
	else
	{
		/* Free old "killed by" text */
		if (exp->killedby) free(exp->killedby);

		/* Collecting all pseudonyms... */
		if (!(pls->state & EVOLVE_NoPseuds) && (pls->state & EVOLVE_AllPseuds))
		{
			/*fprintf(stderr, "\nzzz-A %s %p", pls->pid, exp->dps);*/
	
			/* Don't inherit, so no pseudonyms for this existence */
			if (!(ctx->flags & EVOLVE_InheritPseud)) exp->pseuds = exp->dps = 0;

			/* Otherwise, isolate inherited default pseudonym (if any) on existence list */
			else if (exp->pseuds = exp->dps) exp->dps->link_ex = 0;
		}

		/* Not collecting all pseudonyms, and not inheriting previous default pseudonym */
		else if (!(pls->state & EVOLVE_NoPseuds) && !(ctx->flags & EVOLVE_InheritPseud))
		{
			/*fprintf(stderr, "\nzzz-B %s %p", pls->pid, exp->dps);*/

			/* Paranoia - default pseud should match first in existence and player pseud lists */
			if ((exp->pseuds != exp->dps) || (pls->allpseuds != exp->pseuds)) erf(ERF_Paranoia, 11);

			/* Paranoia - should only ever be a single entry in existence and player pseud lists */
			if ((exp->dps) && ((exp->dps->link_ex) || (exp->dps->link_pl))) erf(ERF_Paranoia, 12);

			/* Delink and free from both lists */
			if (exp->dps) free(exp->dps);
			pls->allpseuds = exp->dps = exp->pseuds = 0;
		}

		/*fprintf(stderr, "\nzzz-C %s %p", pls->pid, exp->dps);*/

		/* Previous existence has been overwritten */
		pexp = 0; if (ppexpp) *ppexpp = 0;
	}

	/* Not "killed by" anyone, yet (but retain old wantedness) */
	exp->killedby = 0;

	/* Time of inception of this existence */
	exp->born = exp->tlact = when;

	/* Time of becoming police */
	if (exp->fl & PF_Police) exp->tpol = when;

	/* Set "redeemed" flag */
	exp->fl = (exp->fl & (~PF_Redeemed)) | ((fff) ? PF_Redeemed : 0);

	/* Zero existence scores, etc. */
	exp->score = 0;

	/* Initialise scoring table */
	if (exp->scoretab) for (i = 0; (i <= ctx->scmax_s); i++) exp->scoretab[i] = 0;


_erf: /***** Error handling *****/

	/* Debug */
	if (rc)fprintf(stderr, "%li:   resurrect(%i):%i %s "TDEBfmt"\n", uniqid,
				   ERF_CONTEXT(rc), ERF_TYPE(rc), pls->pid, TDEBparm(when));

	/* Only retain low bits of normal rc's */
	return (-ERF_LOWBITS(rc));
}

/* Classify events by outcome and participant type */
int pgenus(unsigned long evf, player_p pp, unsigned long expf, char* pid)
{
	/*fprintf(stderr, "[%s]", pid);*/

    /* Player not known */
	if (!pid) return (SCDEF_Unknown);

	/* Full player flags available */
    /*else if (pp)
    {
		if    (pp->flags & PF_IllegalNP) return (SCDEF_IllegalNonPlayer);
		else if (pp->flags & PF_LegalNP) return (SCDEF_LegalNonPlayer);
		else if (pp->flags & PF_Special) return (SCDEF_Special);
		else return (PF_genus(expf));
    }*/

    /* Full player flags not available - assume not special if id not recognised */
    else
    {
		if      (!strcmp(pid, ILL_ID)) return (SCDEF_IllegalNonPlayer);
		else if (!strcmp(pid, LEG_ID)) return (SCDEF_LegalNonPlayer);
    	else if (!strcmp(pid, UMP_ID)) return (SCDEF_Special);
    	else if (!strcmp(pid, PUB_ID)) return (SCDEF_Special);
    	else return ((int) PF_genus(expf));
    }
}

/* Deal with pseudonym changes: (setit) bit 0 => set default */
int evolve__pseud(pexist_p exp, char* ps, int setit)
{
	ppseud_p pseudp;

	/* Collect pseudonyms only */
	if (exp->psp->state & EVOLVE_OnlyPseuds)
	{
		/* Already used by this player? */
		for (pseudp = exp->psp->allpseuds; (pseudp); pseudp = pseudp->link_pl) if (!strcasecmp(pseudp->text, ps)) break;

		/* Not already used */
		if (!pseudp)
		{
			/* Allocate structure for pseudonym */
			if (!(pseudp = malloc(sizeof(struct ppseud) + 1 + strlen(ps)))) return (ERF_mallocFailed);
	
			/* Store text */
			strcpy((pseudp->text = ((char*) pseudp) + sizeof(struct ppseud)), ps);

			/* Link in at head of chain for this player */
			pseudp->link_pl = exp->psp->allpseuds;
			exp->psp->allpseuds = pseudp;

			/* Not yet seen */
			pseudp->fl = PSFL_NotYet;
			pseudp->link_ex = 0;

			fprintf(stderr, "%li:   onlypseud added \"%s\" pseud=%p \"%.30s\" fl=&%u\n",
					uniqid, exp->psp->pid, pseudp, ps, pseudp->fl);
		}

		else fprintf(stderr, "%li:   onlypseud match \"%s\" pseud=%p \"%.30s\" fl=&%u\n",
					 uniqid, exp->psp->pid, pseudp, ps, pseudp->fl);

		/* Pseudonym to be used */
		pseudp->fl |= PSFL_Later;

		/* Success */
		return 0;
	}

	/* Not interested... */
	if ((exp->psp->state & EVOLVE_NoPseuds) || !((exp->psp->state & EVOLVE_AllPseuds) || (setit & 1))) return 0;

	/* Already used in this existence? */
	for (pseudp = exp->pseuds; (pseudp); pseudp = pseudp->link_ex) if (!strcasecmp(pseudp->text, ps)) break;

	/*if (pseudp) fprintf(stderr, "(match!!1)");*/

	/* No, a new one for this existence? */
	if (!pseudp)
	{
		/*fprintf(stderr, "(Not found in exist)");*/

		/* Lose previous pseudonym for this existence if not collecting them all */
		if (!(exp->psp->state & EVOLVE_AllPseuds))
		{
			/* Paranoia - existence pseud list should only hold the default pseudonym (or be empty) */
			if (exp->pseuds != exp->dps) return (ERF_Paranoia);

			/* Was there it empty? */
			else if (exp->pseuds)
			{
				/* Paranoia - should only ever be a single entry in existence pseud list in this case */
				if (exp->pseuds->link_ex) return (ERF_Paranoia);

				/* Delink from player pseud list - should be head of list or duplicate of earlier entry */
				if (exp->psp->allpseuds == exp->pseuds) exp->psp->allpseuds = exp->pseuds->link_pl;

				/* Paranoia - if a duplicate of another in player list, should not appear in list itself */
				else if (exp->pseuds->link_pl) return (ERF_Paranoia);

				/* Delink and free from list for this existence */
				free(exp->pseuds);
				exp->dps = exp->pseuds = 0;
			}
		}

		/*fprintf(stderr, "(Lala)");*/

		/* Already used by this player? (in an earlier existence) */
		for (pseudp = exp->psp->allpseuds; (pseudp); pseudp = pseudp->link_pl) if (!strcasecmp(pseudp->text, ps)) break;

		/* Not already used */
		if (!pseudp)
		{
			/*fprintf(stderr, "(Not already used)");*/

			/* Allocate structure for pseudonym */
			if (!(pseudp = malloc(sizeof(struct ppseud) + 1 + strlen(ps)))) return (ERF_mallocFailed);
	
			/* Store text */
			strcpy((pseudp->text = ((char*) pseudp) + sizeof(struct ppseud)), ps);

			/* Link in at head of chain for this player */
			pseudp->link_pl = exp->psp->allpseuds;
			exp->psp->allpseuds = pseudp;
		}

		/* Used in earlier existence */
		else
		{
			ppseud_p prevpp = pseudp;

			/*fprintf(stderr, "(already used before)");*/

			/* Allocate structure for pseudonym */
			if (!(pseudp = malloc(sizeof(struct ppseud)))) return (ERF_mallocFailed);

			/* Don't link in to chain for this player (will already be in there) */
			pseudp->link_pl = 0;
			pseudp->text = prevpp->text;
		}
	
		/* Link into chain for this existence */
		pseudp->fl = 0;
		pseudp->link_ex = exp->pseuds;
		exp->pseuds = pseudp;
	}

	/*fprintf(stderr, "(finit)");*/

	/*fprintf(stderr, "go(%i,%s)", setit, ps);*/

	/* Make this the default for this existence */
	if (setit & 1) exp->dps = pseudp;

	/* Success */
	return (0);
}

/* Check for and handle timer expiration: timers go off when their value is attained or exceeded */
int evolve__timer(pstate_p plst, struct evolve__ctx* ctx, timeval et, u32b elapsed)
{
	unsigned long pt;
	int i;

	/* Dangerousness decays with time */
	/*if ((pt = elapsed - (et & 0xFF) - 60 * ((et >> 8) & 0xFF) - 1440 * (et >> 16)) > 0)
	{
		plst->danger = (long) (plst->danger / (pow(2.0, pt / ((double) HALFLIFE))));
	}*/

	/* Timer timeval */
	pt = plst->exist->timer;

	/* Due to resurrect? Has timer over-ticked? */
	if ((plst->exist->fl & PF_ResTimer) && (pt <= et))
	{
		/* Resurrect the player */
		if (i = resurrect(plst, ctx, 0, pt, 0)) return (i);

		/* Clear the timer */
		plst->exist->timer = 0;

		/* No longer dead, no longer due for resurrection */
		plst->exist->fl &= ~(PF_Dead|PF_ResTimer);
	}

	/* Deal with player's inactivity timer */
	if (plst->exist->fl & PF_InactTimer)
	{
		timeval timee;

		/* Mark player an incompetently inactive */
		if (pt <= et) plst->exist->fl = (plst->exist->fl /*& ~(PF_InactTimer)*/) | (PF_Inactive|PF_InactWarning);
		
		/* Work out player's inactivity warning time */
		else if (!(timee = addhours(et, HOURS_InactWarning))) return ERF_BadTimeVal;

		/* Or warn about player's inactivity */
		else if (pt <= timee) plst->exist->fl |= (PF_InactWarning);
	}

	return 0;
}

/* Add a new player to list with state flags (fl) */
pstate_p evolve__addpstate(char* id, evolve_flags fl, struct evolve__ctx* ctx, char* chkpseud)
{
	pstate_p psp;
	ppseud_p pseudp = 0;
	int i;

	/* Allocate storage space */
	if (!(psp = malloc(sizeof(struct pstate)))) return 0;

	/* Allocate storage space */
	if ((chkpseud) && !(pseudp = malloc(sizeof(struct ppseud)))) return 0;

	/* No per-player scoring table? */
	if (!(fl & EVOLVE_ScorePlayer)) psp->scoretab = 0;

	/* Allocate per-player scoring table? */
	else if (!(psp->scoretab = malloc((ctx->scmax_s + 1) * sizeof(int)))) return 0;

	/* Initialise it */
	else for (i = 0; (i <= ctx->scmax_s); i++) psp->scoretab[i] = 0;

	/* No per-existence scoring table? */
	if (!(fl & EVOLVE_ScoreExists)) psp->firstexist.scoretab = 0;

	/* Allocate per-existence scoring table? */
	else if (!(psp->firstexist.scoretab = malloc((ctx->scmax_s + 1) * sizeof(int)))) return 0;

	/* Initialise it */
	else for (i = 0; (i <= ctx->scmax_s); i++) psp->firstexist.scoretab[i] = 0;

	/* Initialise flags/state */
	psp->state = fl;

	/* Copy player id */
	strncpy(psp->pid, id, MAX_id-1);
	psp->pid[MAX_id-1] = '\0';

	/* No targets yet */
	psp->targs = 0;
	psp->asses = 0;

	/* Link in at head of list */
	psp->next = ctx->plist;
	ctx->plist = psp;

	/* Link existence into player state structure */
	psp->exist = &psp->firstexist;
	psp->exist->psp  = psp;
	psp->exist->link_pl = 0;

	/* Link into overall existence list */
	psp->exist->link_all = ctx->elist;
	ctx->elist = psp->exist;
	ctx->existcnt++;

	/* Associate with external player list entry */
	for (psp->plp = ctx->pl_list; (psp->plp); psp->plp = psp->plp->link) if (streq(psp->plp->id, id)) break;

	/* Could complain if a flag set and match was found on external list... but we have no way of returning an rc */
	/* if ((fl & EVOLVE_Moo) && (ctx->pl_list) && (!psp->plp)) return 0; */

	/* No extension word */
	psp->ext = 0;
	psp->exist->ext = 0;

#ifdef EVOLVE_ALLOW_EXT_BACKLINKS
	/* Backlinks from external player structures to state structures */
	if ((fl & EVOLVE_BacklinkPlayers) && (psp->plp)) psp->plp->ext = psp;
#endif

	/* Initial player flags */
	psp->pflags = INITIAL_FLAGS;

	if (!strcmp(id, UMP_ID) || !strcmp(id, PUB_ID)) psp->pflags |= PF_Special;
	else if (!strcmp(id, LEG_ID)) psp->pflags |= PF_Special|PF_LegalNP;
	else if (!strcmp(id, ILL_ID)) psp->pflags |= PF_Special|PF_IllegalNP;
	else psp->pflags |= PF_Playing;

	/* Initial player flags */
	psp->exist->fl = psp->pflags;

	/* Initial score */
	psp->exist->score = psp->score = INITIAL_SCORE;
	psp->danger = INITIAL_DANGER;
	psp->exist->rank = INITIAL_RANK;

	/* No wantedness comment or killer identity */
	psp->exist->comment  = 0 /* "No comment" */;
	psp->exist->killedby = 0 /* "*** No-one ***" */;

	/* No pseudonyms yet */
	psp->allpseuds = psp->exist->pseuds = psp->exist->dps = 0;
	psp->cps = 0;

	/* Pseudonym to check? */
	if (psp->chkpseud = pseudp)
	{
		pseudp->fl = 0;
		pseudp->text = chkpseud;
		pseudp->link_pl = 0;
		pseudp->link_ex = ctx->chkpseudlist;
		ctx->chkpseudlist = pseudp;
	}

	/* Reset timers */
	psp->exist->timer = psp->exist->twant = psp->exist->tpol = 0;
	psp->exist->born = psp->exist->died = psp->exist->tlact = 0;

	/* Success */
	return (psp);
}

/* Wrapper for report id jobby */
int evolve__wrapper(event_p ep, char* p, int state, void* context)
{
	struct evolve__ctx* ctx = context;

	/* Early termination if no hook present */
	return (ctx->repfunc) ? (*ctx->repfunc)(ep, p, state, ctx->fctx, ctx) : 1;
}

/* Check validity of pseuds */
void evolve__chkpseuds(char* text, pstate_p psp, struct evolve__ctx* ctx)
{
	ppseud_p pseudp, *opseudpp;

	/* Do test for each checkable pseudonym */
	for (pseudp = *(opseudpp = &ctx->chkpseudlist); (pseudp); pseudp = pseudp->link_ex)
	{
		/* Match: already used */
		if (!strcasecmp(pseudp->text, text))
		{
			/* Already used by this player, or by another */
			pseudp->fl |= ((psp) && (pseudp == psp->chkpseud)) ? PSFL_Mine : PSFL_Used;

			/* Optimisation: remove from list when no more information to gain */
			if (!((~pseudp->fl) & (PSFL_Mine|PSFL_Used))) *opseudpp = pseudp->link_ex;
		}

		/* No match: retain */
		else opseudpp = &pseudp->link_ex;
	}
}

/* Match players on list */
int evolve__find(char* pid, pstate_p *pspp, struct evolve__ctx* ctx, int m)
{
	/* Try to match first player id in list */
	for (*pspp = ctx->plist; (*pspp); *pspp = (*pspp)->next) if (!strcmp(pid, (*pspp)->pid)) break;

	/* Missing... */
	if (!(*pspp))
	{
		/* Option: complain if player missing from initial list */
		if (ctx->flags & EVOLVE_MustMatch) return ERF_MissingPid;

		/* Option: maintain state for players not on initial list */
		else if ((ctx->flags & EVOLVE_MatchAll) && (m))
		{
			/* Try to allocate and link new 'pstate' struct */
			if (!(*pspp = evolve__addpstate(pid, ctx->flags, ctx, 0))) return ERF_mallocFailed;
		}
	}

	/* Success */
	return 0;
}

/* Handle targetting allocations/deallocations */
int evolve__dotargs(retarg_p retarp, int m, struct evolve__ctx* ctx, char* tit)
{
	target_p tarp, *ptarp;
	pstate_p apsp, tpsp;
	int rc = 0;


	/* Deal with each allocation/deallocation */
	while (retarp)
	{
		/* Paranoia */
		if ((!retarp->assid) || (!retarp->tarid)) erf(ERF_InternalError, 1);

		/*fprintf(stderr, "%li:    %s\"%s\"->\"%s\"\n", uniqid, (m) ? "add":"rem", retarp->assid, retarp->tarid);*/

		/* Try to match assassin's id in list */
		if (rc = evolve__find(retarp->assid, &apsp, ctx, (ctx->flags & EVOLVE_NoTargetting) ? 0 : 1)) erf(rc, 2);

		/* Try to match target's id in list */
		if (rc = evolve__find(retarp->tarid, &tpsp, ctx, (ctx->flags & EVOLVE_NoTargetting) ? 0 : 1)) erf(rc, 3);

		/* Allocate on assassins' target list */
		if ((m) && (apsp))
		{
			/* Allocate space for struct */
			if (!(tarp = malloc(sizeof(struct target) + 1 + strlen(retarp->tarid)))) erf(ERF_mallocFailed, 4);

			/* Add to list*/
			tarp->ext = 0;
			tarp->id = strcpy(((char*)tarp) + sizeof(struct target), retarp->tarid);
			tarp->next = apsp->targs;
			apsp->targs = tarp;
		}

		/* Deallocate from assassins' target list */
		if ((!m) && (apsp))
		{
			/* Find existing allocation of assassin -> target */
			for (ptarp = &apsp->targs; (tarp = *ptarp); ptarp = &(*ptarp)->next)
			{
				/* Match? */
				if (streq(retarp->tarid, tarp->id)) break;
			}

			/* Found: deallocate */
			if (tarp)
			{
				*ptarp = tarp->next;
				free(tarp);
			}

			/* Complain: unable to deallocate */
			else if (ctx->flags & EVOLVE_StrictTargets) erf(ERF_InternalError, 5);

			/* Warning: unable to deallocate */
			else fprintf(stderr, "%li: ! Removal of non-existant allocation: ass:\"%s\": \"%s\"->\"%s\" (%p,%p)\n",
						 uniqid, tit, retarp->assid, retarp->tarid, apsp, tpsp);
		}

		/* Allocate on target's assassin list */
		if ((m) && (tpsp))
		{
			/* Allocate space for struct */
			if (!(tarp = malloc(sizeof(struct target) + 1 + strlen(retarp->assid)))) erf(ERF_mallocFailed, 6);

			/* Add to list*/
			tarp->ext = 0;
			tarp->id = strcpy(((char*)tarp) + sizeof(struct target), retarp->assid);
			tarp->next = tpsp->asses;
			tpsp->asses = tarp;
		}

		/* Update targets' assassin allocations */
		if ((!m) && (tpsp))
		{
			/* Find existing allocation of assassin <- target */
			for (ptarp = &tpsp->asses; (tarp = *ptarp); ptarp = &(*ptarp)->next)
			{
				/* Match? */
				if (streq(retarp->assid, tarp->id)) break;
			}

			/* Found: deallocate */
			if (tarp)
			{
				*ptarp = tarp->next;
				free(tarp);
			}

			/* Complain: unable to deallocate */
			else if (ctx->flags & EVOLVE_StrictTargets) erf(ERF_InternalError, 7);

			/* Warning: unable to deallocate */
			else fprintf(stderr, "%li: ! Removal of non-existant allocation: tar:\"%s\": \"%s\"->\"%s\" (%p,%p)\n",
						 uniqid, tit, retarp->assid, retarp->tarid, apsp, tpsp);
		}

		/* Next in list */
		retarp = retarp->next;
	}


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   evolve__dotargs(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/* Handle targetting block deallocations for player */
int evolve__detarg(pstate_p pstp, struct evolve__ctx* ctx, char* tit)
{
	target_p tar1p, tar2p, *ptarp;
	pstate_p ps2p;


	/* Delinkage of targets */
	for (tar1p = pstp->targs; (tar1p); tar1p = tar2p)
	{
		/* Find relevent target player struct */
		for (ps2p = ctx->plist; (ps2p); ps2p = ps2p->next) if (!strcmp(tar1p->id, ps2p->pid)) break;

		/* Found? */
		if (ps2p)
		{
			/* Find existing allocation of assassin -> target */
			for (ptarp = &ps2p->asses; (tar2p = *ptarp); ptarp = &tar2p->next)
			{
				/* Match? */
				if (streq(pstp->pid, tar2p->id)) break;
			}

			/* Found: deallocate */
			if (tar2p)
			{
				*ptarp = tar2p->next;
				free(tar2p);
			}

			/* Complain: unable to deallocate */
			else if (ctx->flags & EVOLVE_StrictTargets) return ERF_InternalError;

			/* Warning: unable to deallocate */
			else fprintf(stderr, "%li: ! Dealloc of non-existant allocation: ass:\"%s\": \"%s\"->\"%s\" (%p,%p)\n",
						 uniqid, tit, pstp->pid, ps2p->pid, pstp, ps2p);
		}

		/* Next */
		tar2p = tar1p->next;

		/* Free this entry */
		free(tar1p);
	}

	/* Delinkage of assassins */
	for (tar1p = pstp->asses; (tar1p); tar1p = tar2p)
	{
		/* Find relevent assassin player struct */
		for (ps2p = ctx->plist; (ps2p); ps2p = ps2p->next) if (!strcmp(tar1p->id, ps2p->pid)) break;

		/* Found? */
		if (ps2p)
		{
			/* Find existing allocation of target -> assassin */
			for (ptarp = &ps2p->targs; (tar2p = *ptarp); ptarp = &tar2p->next)
			{
				/* Match? */
				if (streq(pstp->pid, tar2p->id)) break;
			}

			/* Found: deallocate */
			if (tar2p)
			{
				*ptarp = tar2p->next;
				free(tar2p);
			}

			/* Complain: unable to deallocate */
			else if (ctx->flags & EVOLVE_StrictTargets) return ERF_InternalError;

			/* Warning: unable to deallocate */
			else fprintf(stderr, "%li: ! Dealloc of non-existant allocation: tar:\"%s\": \"%s\"<-\"%s\" (%p,%p)\n",
						 uniqid, tit, pstp->pid, ps2p->pid, pstp, ps2p);
		}

		/* Next */
		tar2p = tar1p->next;

		/* Free this entry */
		free(tar1p);
	}

	/* All targets and assassins removed */
	pstp->targs = pstp->asses = 0;

	/* Finished */
	return 0;
}

/* Deal with raw events */
int evolve__ef(event_p ep, int state, FILE* fp, void* context)
{
	static int pg1, pg2;
	static pstate_p p1s, p2s;
	pstate_p pstp;
	pexist_p pexp, exp;
	struct evolve__ctx* ctx = context;
	unsigned long fff;
	int i, keep = 0, rc = 0;


	/*fprintf(stderr, " ev[%s]", ep->ref);*/
	/*fprintf(stderr, "**[%s|%s|%02i:%02i:%02i|%s]\n",
			ep->ref, ep->id, ep->when.day, ep->when.hour, ep->when.min,
			(ep->pseud)?ep->pseud:"nopseud");*/

	/* Normal (first) call for this event */
	if (!(state & 4))
	{
		/* Time has changed */
		if ((ctx->flags & EVOLVE_FuncOnTick) && (ep->when != ctx->lasttime))
		{
			/* Call hook? */
			if ((i = (*ctx->func)(ep, -1, ctx->fctx, ctx)) < 0) erf(-i, 30);
		}

		/* Callbacks may introduce new limits: repeat checks until limits exhausted */
		while (1)
		{
			/* This is event *after* inclusive event ref match */
			if (ctx->matched == 1) ctx->matched = 2;
	
			/* Time limit exceeded */
			else if (ep->when > ctx->endtime) ctx->matched = 4;
	
			/* Event ref match */
			else if ((ctx->endref) && streq(ctx->endref, ep->ref))
			{
				keep = 8+16+32;
				ctx->eptr = ep;
				ctx->matched = (ctx->flags & EVOLVE_Prior) ? 3 : 1;
			}
	
			/* Time limit reached */
			else if ((ep->when == ctx->endtime) && (ctx->flags & EVOLVE_Exclusive)) ctx->matched = 3;
	
			/* Not past end condition? Process event */
			if (ctx->matched <= 1) break;

			/* Terminate if no callback required */
			if (!(ctx->flags & EVOLVE_FuncEndCond)) return 1+2+keep;
	
			/* Perform optional callback, may alter some context */
			if ((i = (*ctx->func)((ctx->matched == 2) ? ctx->eptr : ep, ctx->matched, ctx->fctx, ctx)) < 0) erf(-i, 30);

			/* Reset matching context for check under (new) conditions, stop if required */
			if ((ctx->matched = i) >= 1) return 1+2+keep;
		}

		/* Set lasttime to time of most recently processed event */
		ctx->lasttime = ep->when;

		/* Update timers, if not only collecting pseuds */
		if (!(ctx->flags & EVOLVE_OnlyPseuds))
		{
			/* Evolve outstanding timers up to time of this event */
			for (pstp = ctx->plist; (pstp); pstp = pstp->next)
			{
				/* Deal with timer tickover */
				if (rc = evolve__timer(pstp, ctx, ep->when, ctx->elapsed)) erf(rc, 53);
			}
	
			/* Update elapsed time value */
			ctx->elapsed = ELAPTIME(ep->when);
		}

		/* Ignore event? Ask for next, if so */
		if (ep->flags & EF_Ignore) return 1+keep;

		/* Try to match first player id in list */
		if (i = evolve__find(ep->id, &p1s, ctx, 1)) erf(i, 2);

		/* Handle retargeting events */
		if (ctx->flags & EVOLVE_Targetting)
		{
			/* Deallocate targets */
			if ((ep->flags & EF_Targetting) && (i = evolve__dotargs(ep->untarg, 0, ctx, ep->ref))) erf(i, 51);

			/* Deallocate all of player's links */
			if ((ep->flags & EF_Detarg) && (p1s) && (i = evolve__detarg(p1s, ctx, ep->ref))) erf(i, 52);
		}

		/* Check pseudonyms against p1 pseudonym */
		if ((ep->flags & EF_MASK_Pseudonym) && (ctx->chkpseudlist)) evolve__chkpseuds(ep->pseud, p1s, ctx);

		/* Update p1 state */
		if (p1s)
		{
			/*fprintf(stderr, " ev[%s] p1s=%s", ep->ref, p1s->pid);*/

			/* Only collect pseudonyms? */
			if (ctx->flags & EVOLVE_OnlyPseuds)
			{
				/* Player 1's pseudonym */
				if ((ep->flags & EF_MASK_Pseudonym) && (rc = evolve__pseud(p1s->exist, ep->pseud, 0))) erf(rc, 99);

				/* No more processing */
				goto collected1;
			}

			/* No current pseudonym yet */
			p1s->cps = 0;

			/* Acquire current existence data (that prior to event) */
			p1s->pflags = (pexp = exp = p1s->exist)->fl;

			/* Update and activate player's timing device */
			if (ep->flags & (EF_SetInactTimer))
			{
				exp->fl = (exp->fl & ~(PF_Inactive|PF_InactWarning)) | (PF_InactTimer);
				exp->timer = ep->ntimer;
			}

			/* Optional: Record changes in comment field */
			if ((p1s->state & EVOLVE_Comments) && (ep->flags & EF_Wantedness))
			{
				/* Lose old wantedness (comment) */
				if (exp->comment) free(exp->comment);

				/* No new wantedness (comment) */
				if (!ep->wantedness[0]) exp->comment = 0;

				/* Copy new wantedness (comment) */
				else if (exp->comment = malloc(strlen(ep->wantedness)+1)) strcpy(exp->comment, ep->wantedness);

				/* Track errors */
				else erf(ERF_mallocFailed, 14);
			}

			/* XXX --- do we need to tidy up normal resurrection? */

			/* Deal with new existences (reincarnation or making wanted->unwanted) */
			if ((ep->flags & EF_Reincarnate) || (fff = ((ep->flags & EF_MakeUnWanted) && (exp->fl & PF_Wanted))))
			{
				/*fprintf(stderr, "\n>>Reincarnate [%s|%s|%02i:%02i:%02i|%s] %lX:",
						ep->ref, ep->id, ep->when.day, ep->when.hour, ep->when.min,
						(ep->pseud)?ep->pseud:"nopseud", p1s->state);*/

				/* Not annulled? */
				if (!(ep->flags & EF_Annulled))
				{
					if (rc = resurrect(p1s, ctx, &pexp, ep->when, fff)) erf(rc, 15);
					exp = p1s->exist;
				}

				/* No longer dead? */ 
				if (ep->flags & EF_Reincarnate) exp->fl &= (~PF_Dead);
			}

			/* Add given pseudonym to list of all pseudonyms if required, and set as default if required */
			if (ep->flags & EF_MASK_Pseudonym)
			{
				/*fprintf(stderr, "(fl:%lx)", ep->flags & EF_MASK_Pseudonym);*/

				if (rc = evolve__pseud(exp, ep->pseud, (ep->flags & EF_SetPseudonym) ? 1 : 0)) erf(rc, 16);
				p1s->cps = ep->pseud;
			}

			/* Use this or default pseudonym (if any) as current */
			else p1s->cps = (exp->dps) ? (exp->dps->text) : 0;

			/* No longer wanted */
			if (ep->flags & EF_MakeUnWanted) exp->fl &= ~PF_Wanted;

			/* Change in police status */
			if (ep->flags & EF_Police)
			{
				/* Make player, not police? */
				if ((ep->rank <= 0) && !(ep->flags & EF_Promotion))
				{
					exp->fl &= (~PF_Police);

					/*fprintf(stderr, "%li:   [%s] notpol \"%s\"\n", uniqid, ep->ref, p1s->pid);*/
				}

				/* Make police */
				else
				{
					/*fprintf(stderr, "%li:   [%s] mkpol \"%s\" rank=%li &%lX\n",
							uniqid, ep->ref, p1s->pid, ep->rank, ep->flags & EF_Promotion);*/

					/* New to the force, or null promotion -> set default rank */
					if (!(exp->fl & PF_Police) || ((ep->flags & EF_Promotion) && (ep->rank == 0))) exp->rank = DEF_rank;

					/* New rank */
					exp->rank = ep->rank + ((ep->flags & EF_Promotion) ? exp->rank : 0);

					/* Check for excessive demotion */
					if (exp->rank < 1) exp->rank = 1;
					
					/* Time at which made police */
					if (!(exp->fl & PF_Police)) exp->tpol = ep->when;

					/* Set player flag */
					exp->fl |= PF_Police;
				}
			}

			/* Make player competent */
			if (ep->flags & EF_MakeCompetent) exp->fl &= ~PF_Incompetent;

			/* Make player wanted */
			if (ep->flags & EF_MakeWanted)
			{
				/* Set flags */
				exp->fl |= PF_Wanted;

				/* Record time of becoming wanted */
				/* if ((!p1s->pflags & PF_Wanted)) */ exp->twant = ep->when;
			}

			/* Make player incompetent */
			if (ep->flags & EF_MakeIncompetent) exp->fl |= PF_Incompetent;

			/* Award scoring bonuses */
			if (ep->flags & EF_ScoreBonus)
			{
				if (ctx->scbrkfp) fprintf(ctx->scbrkfp, "%s: bonus for %s:(%li,%li) ", ep->ref, ep->id,
										  p1s->score >> SCORE_SHIFT, p1s->danger >> DANGER_SHIFT);

				ctx->notes |= NB_ScoreBonus;
				p1s->score += ep->bonus;

				if (ctx->scbrkfp) fprintf(ctx->scbrkfp, "(%+li) => (%li,%li)\n", ep->bonus >> SCORE_SHIFT,
								  p1s->score >> SCORE_SHIFT, p1s->danger >> DANGER_SHIFT);
			}
		}

		/* Only collect pseudonyms */
		else if (ep->flags & EF_MASK_Pseudonym) goto collected1;

		/* Genus of assassin */
		if (!p1s) pg1 = pgenus(ep->flags, 0, 0, 0);

		/* Attribute kills which cause players to go wanted to criminal activity */
		else pg1 = pgenus(ep->flags, 0, (p1s->pflags | (p1s->exist->fl & PF_Wanted)), p1s->pid);

		/* Dealt with p1. Now p2, if any */
		collected1:

		/* Deal with assassination events */
		if (ep->flags & EF_Assassination)
		{
			/* Time of last activity */
			if (p1s) p1s->exist->tlact = ep->when;

			/* Try to match second player id in list */
			for (p2s = ctx->plist; (p2s); p2s = p2s->next) if (!strcmp(ep->vict_pid, p2s->pid)) break;

			/* Option: complain if player missing from initial list */
			if ((!p2s) && (ctx->flags & EVOLVE_MustMatch)) erf(ERF_MissingPid, 3);

			/* Option: maintain state for players not on initial list */
			else if ((!p2s) && (ctx->flags & EVOLVE_MatchAll))
			{
				/* Try to allocate and link new 'pstate' struct */
				if (!(p2s = evolve__addpstate(ep->vict_pid, ctx->flags, ctx, 0))) erf(ERF_mallocFailed, 4);
			}

			/* Acquire current existence data (that prior to event) */
			if ((p2s) && (p2s != p1s)) p2s->pflags = p2s->exist->fl;

			/* Check pseudonyms against p2 pseudonym */
			if (ctx->chkpseudlist) evolve__chkpseuds(ep->vict_pseud, p2s, ctx);

			/* Update p2 state */
			if (p2s)
			{
				/* Handle retargeting events */
				if ((ctx->flags & EVOLVE_Targetting) && (ep->flags & EF_DetargP2))
				{
					/* Deallocate all of player's links */
					if (i = evolve__detarg(p2s, ctx, ep->ref)) erf(i, 53);
				}

				/* Make default? */
				i = ((ep->flags & EF_SetP2Pseudonym) && !(ctx->flags & EVOLVE_OnlyPseuds)) ? 1 : 0;

				/* Add given pseudonym to list of all pseudonyms if required, and set as default if required */
				if (rc = evolve__pseud(p2s->exist, ep->vict_pseud, i)) erf(rc, 17);
				
				/* Only collect pseudonyms? */
				if (ctx->flags & EVOLVE_OnlyPseuds) goto flobble;

				/* Use this or default pseudonym (if any) as current */
				p2s->cps = ep->vict_pseud;

				/*fprintf(stderr, " prekill[%s]", ep->ref);*/

				/* Kill p2 (last comparison should deal with true flags from uinfo file) */
				if ((ep->flags & EF_Kill) && !(ep->flags & EF_OnSpecial) && !(p2s->exist->fl & PF_Special))
				{
					timeval tmptim;
					
					/* Work out res time */
					if (!(tmptim = addhours(ep->when, reshours(ep->when)))) erf(ERF_BadTimeVal, 5);
	
					/* Note fact and time of death */
					p2s->exist->fl |= PF_Dead;
					p2s->exist->died = ep->when;
	
					/* Resurrect victim? */
					if ((AUTORES(ep->flags)) && (tmptim < TIME_AutoResLimit))
					{
						p2s->exist->timer = tmptim;
						p2s->exist->fl |= PF_ResTimer;
						/*fprintf(stderr, " setres[%s]", ep->ref);*/
					}
				}
			}

			/* Only collect pseudonyms? */
			else if (ctx->flags & EVOLVE_OnlyPseuds) return (keep+1);

			/* Genus of victim */
			if (!p2s) pg2 = pgenus(ep->flags, 0, 0, 0);

			/* Known victim */
			else pg2 = pgenus(ep->flags, 0, p2s->pflags, p2s->pid);

			/*fprintf(stderr, " hoo[%s]", ep->ref);*/

			/* Handle player scores if event not annulled */
			if ((p1s) && (p2s) && (ep->flags & EF_Kill) && !(ep->flags & EF_Annulled) && !(p1s->exist->fl & PF_Special))
			{
				/* Player on player kills */
				if (!(p2s->exist->fl & PF_Special))
				{
					/* Fundamental points loss */
					long x = p2s->score / 6;

					/* Points gain, may differ from loss if dangerousness bonus not disabled */
					long y = (ep->flags & EF_NoDangerBonus) ? x : (long)(x * dangerbonus(p2s->danger));

					/* Scoring breakdown: initial */
					if (ctx->scbrkfp)
					{
						fprintf(ctx->scbrkfp, "%s: p1=%s:(%li,%li) on p2=%s:(%li,%li) (+) %li (-) %li ",
								ep->ref, ep->id, p1s->score >> SCORE_SHIFT, p1s->danger >> DANGER_SHIFT,
								ep->vict_pid, p2s->score >> SCORE_SHIFT, p2s->danger >> DANGER_SHIFT,
								y >> SCORE_SHIFT, x >> SCORE_SHIFT);
					}

					/* Scoring gains/losses */
					p2s->score -= x;
					p1s->score += y;

					/* Changes to dangerousness, may be disabled for event */
					if (!(ep->flags & EF_NoDangerBonus))
					{
						p1s->danger += (1 << DANGER_SHIFT);
						p2s->danger /= 2;
					}

					/* Scoring breakdown: final */
					if (ctx->scbrkfp)
					{
						fprintf(ctx->scbrkfp, "=> p1=(%li,%li) p2=(%li,%li)\n",
								p1s->score >> SCORE_SHIFT, p1s->danger >> DANGER_SHIFT,
								p2s->score >> SCORE_SHIFT, p2s->danger >> DANGER_SHIFT);
					}

					/* Note that scoring has happened */
					ctx->notes |= NB_ScoreKill;
				}

				/* Illegal NP kill */
				else if (p2s->exist->fl & PF_IllegalNP)
				{
					p1s->score -= (10 << SCORE_SHIFT);
					ctx->notes |= NB_ScoreIllegal;
				}
			}

			/*fprintf(stderr, " wibble[%s]", ep->ref);*/

			/* Update scoring tables if event not annulled */
			if (((ctx->scdef_s) || (ctx->scdef_t)) && !(ep->flags & EF_Annulled))
			{
				/* Note fatality */
				int ff1 = pg1 | ((ep->flags & EF_Kill) ? (SCDEF_Fatal) : 0);
				int ff2 = pg2 | ((ep->flags & EF_Kill) ? (SCDEF_Fatal) : 0);

				/*fprintf(stderr, " scoretab[%s] for %s pg1=%i pg2=%i\n", ep->ref, p1s->pid, pg1, pg2);*/
	
				/* Update global scoring table */
				if (ctx->scoretab) ++ctx->scoretab[ctx->scdef_t[ff1][pg2]];

				/* Note that scoring tables may have changed */
				ctx->notes |= ((ep->flags & EF_Kill) ? (NB_ScTabFatal) : (NB_ScTabAttempt));

				/* Update assassin's scoring table */
				if ((p1s) && (p1s->scoretab)) ++p1s->scoretab[i = ctx->scdef_s[ff1][pg2]];

				/*if (!i) fprintf(stderr, "%li:   scnull1(%s) %lx %lx\n", uniqid, p1s->pid, ff1, pg2);*/

				/* Update victim's scoring table */
				if ((p2s) && (p2s->scoretab)) ++p2s->scoretab[i = ctx->scdef_s[ff2+SCDEF_Victim][pg1]];

				/*if (!i) fprintf(stderr, "%li:   scnull2(%s) %lx %lx\n", uniqid, p2s->pid, pg2+SCDEF_Victim, ff1);*/

				/* Update assassin's scoring table for this existence */
				if ((p1s) && (pexp) && (pexp->scoretab)) ++pexp->scoretab[ctx->scdef_s[ff1][pg2]];

				/* Update victim's scoring table for this existence */
				if ((p2s) && (p2s->exist->scoretab)) ++p2s->scoretab[ctx->scdef_s[ff2+SCDEF_Victim][pg1]];

				/*fprintf(stderr, " sctab[%s]", ep->ref);*/
			}

			/* Call back user function */
			if (ctx->flags & EVOLVE_FuncAssass)
			{
				/*fprintf(stderr, " callback[%s]", ep->ref);*/

				/* Give up if something bad happens */
				if ((i = (*ctx->evfunc)(ep, p1s, p2s, pg1, pg2, ctx->fctx, ctx)) < 0) erf(-i, 18);
			}
		}

		/* Not as assassination */
		else pg2 = SCDEF_None;

		/* Skipped some P2 stuff? */
		flobble:

		/* Handle retargeting events */
		if ((ctx->flags & EVOLVE_Targetting) && (ep->flags & EF_Targetting))
		{
			/* Reallocate targets */
			if (rc = evolve__dotargs(ep->retarg, 1, ctx, ep->ref)) erf(rc, 53);
		}

		/* Only collecting pseuds? */
		if (ctx->flags & EVOLVE_OnlyPseuds) return (keep+1);

		/* Call back user function if event newsworthy */
		if ((ep->flags & EF_News) && (ctx->flags & EVOLVE_FuncNews))
		{
			/* Give up if something bad happens */
			if ((i = (*ctx->evfunc)(ep, p1s, p2s, pg1, pg2, ctx->fctx, ctx)) < 0) erf(-i, 19);
		}

		/* Next, please */
		i = ((ctx->repfunc) ? 0 : 1) | ((ctx->secfunc) ? 4 : 0);

		/*fprintf(stderr, " fin[%s]=%i\n", ep->ref, i);*/

		/* evread() status */
		return i+keep;
	}

	/* Second call */
    else
    {
		/* Call hook, and give up if something bad happens */
		if ((ctx->secfunc) && ((i = (*ctx->secfunc)(ep, p1s, p2s, pg1, pg2, ctx->fctx, ctx)) < 0)) erf(-i, 21);

		/* Reset hooks for convenience */
		ctx->secfunc = 0; ctx->repfunc = 0;

		/* Next please */
    	return 0+keep;
    }


_erf: /***** Error handling *****/

	/* No error, or reserved error */
	if (rc <= 0) rc = ERF_InternalError;

	/* Debug */
	if (rc) fprintf(stderr, "%li:   evolve__ef(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (-ERF_LOWBITS(rc));
}


/***************************************************************************************************************************
** Step through events from start to (endtime), whilst updating and maintaining player and game state. If (endref) is     **
** nonnull, will also stop if an event with that reference is reached. Set EVOLVE_Exclusive to exclude events happening   **
** on the time limit, and/or EVOLVE_Prior to exclude the event specified by reference. If (meep) is non-null, and the     **
** referenced (endref) event is found, the relevent 'event' struct is preserved and a pointer returned as (*meep). If     **
** (rettime) is non-null, the effective end time of the evolution is returned as (*rettime).                              **
** if (ipids) is non-zero, it is taken as an array of "interesting player ids", and a 'pstate' structure preallocated for **
** each on list. If (pidx) is additionally non-zero, a pointer to an array of pointers to the 'interesting players' is    **
** stored in (*pidx) -- duplicates and empty lists are dealt with as expected. The head of the entire linked list of all  **
** processed players is placed in (*pout) -- all of the 'interesting players' will be there, and possibly others.         **
***************************************************************************************************************************/
int evolve(evolve_flags flags, char** ipids, pstate_p** pidx, pstate_p* pout, FILE* scbrkfp,
		   int (*scdef_s)[SCDEF_MAX], int (*scdef_t)[SCDEF_MAX], int* sctab_tint, char* endref,
		   event_p* meep, timeval endtime, timeval *rettime, char* buf, char* bufend, char** bufused,
		   evolve_evfn evfn, void* fctx, PSTATE_PLP_TYPE pl_list, evolve_fn func, char** chkpseuds,
		   struct evolve__results *erout)
{
	int rc = 0, nin = 0, i, j;
	struct evolve__ctx ctx;
	pstate_p psp;
	timeval tv;


	/* Referenced event not yet found */
	if (meep) *meep = 0;

	/* Ignore empty endrefs */
	if ((ctx.endref = endref) && !(*endref)) ctx.endref = 0;

	/* Go right to end of file if no time limit given, else have  */
	/* time limit on evolution (ie, determine state at this time) */
	ctx.endtime = (!endtime) ? TMAX : endtime;

	/*if (endtime) fprintf(stderr, "moo: "TFMT"\n", Tdhm(endtime));*/

	/* Single player scoring table definition */
	if (ctx.scdef_s = scdef_s)
	{
		/* Count size of scoring table proper */
		for (i = 0, ctx.scmax_s = 0; (i < SCDEF_MAX_Single); i++) for (j = 0; (j < SCDEF_MAX); j++) 
		{
			/* Store maximal index of array */
			if (ctx.scmax_s < scdef_s[i][j]) ctx.scmax_s = scdef_s[i][j];
		}
	}

	/* No global scoring table definition? */
	if (!(ctx.scdef_t = scdef_t)) ctx.scoretab = 0;

	/* No table proper supplied? */
	else if (!(ctx.scoretab = sctab_tint)) ctx.scdef_t = 0;

	/* Clear the global scoring table proper */
	else for (i = 0; (i < SCDEF_MAX_Global); i++) for (j = 0; (j < SCDEF_MAX); j++) sctab_tint[scdef_t[i][j]] = 0;

	/* External player list to link against */
	ctx.pl_list = pl_list;

	/* Scoring breakdown output stream */
	ctx.scbrkfp = scbrkfp;

	/* No players in results list initially (could pass in a list as parameter to go here) */
	ctx.plist = 0;

	/* No existences, yet */
	ctx.elist = 0;
	ctx.existcnt = 0;
	ctx.chkpseudlist = 0;

	/* Nothing earlier has happened */
	ctx.lasttime = 0;
	ctx.elapsed = 0;
	ctx.matched = 0;
	ctx.notes = 0;
	ctx.eptr = 0;

#ifndef EVOLVE_ALLOW_EXT_BACKLINKS
	/* Bad interaction between compile-time and run-time options */
	if (flags & EVOLVE_BacklinkPlayers) erf(ERF_BadParms, 7);
#endif

	/* Complain about parameter/flag mismatches: callbacks desired, but no function passed */
	if ((flags & (EVOLVE_FuncAssass|EVOLVE_FuncNews)) && (!evfn)) erf(ERF_BadParms, 8);
	if ((flags & (EVOLVE_FuncEndCond|EVOLVE_FuncExhaust|EVOLVE_FuncOnTick)) && (!func)) erf(ERF_BadParms, 9);

	/* Individual scoring desired, but no scoring table definition passed */
	if ((flags & (EVOLVE_ScorePlayer|EVOLVE_ScoreExists)) && (!scdef_s)) erf(ERF_BadParms, 10);
	if ((flags & (EVLadd_ScorePlayer|EVLadd_ScoreExists)) && (!scdef_s)) erf(ERF_BadParms, 11);

	/* Sort out contradictory options */
	if ((flags & EVOLVE_MustMatch) && (flags & EVOLVE_MatchAll)) erf(ERF_BadParms, 14);
	if ((flags & EVOLVE_AllPseuds) && (flags & EVOLVE_NoPseuds)) erf(ERF_BadParms, 15);
	if ((flags & EVLadd_AllPseuds) && (flags & EVLadd_NoPseuds)) erf(ERF_BadParms, 16);

	/* Sort out flag for extra additions */
	ctx.flags = (flags &~ 0x0000FF01) | ((flags & 0x00FF0000) >> 8);

	/* Flags for initial entries */
	flags = (flags &~ 0x00FF0000) | EVOLVE_Init;

	/* Interested in returning linked list? */
	if (pout) *pout = ctx.plist;

 	/* Count initial players */
	if (ipids) while (ipids[nin]) nin++;

	/* Were there any in list? */
	if (nin)
	{
		/* Allocate index array for them, if required */
		if ((pidx) && (!(*pidx = malloc(nin * sizeof(pstate_p))))) erf(ERF_mallocFailed, 2);

		/* Deal with each initial player id */
		for (nin = 0; (ipids[nin]); nin++)
		{
			/* Non-empty name? */
			if (*ipids[nin])
			{
				/* Search list of already initialised players for matching id */
				for (psp = ctx.plist; (psp); psp = psp->next) if (!strcmp(psp->pid, ipids[nin])) break;
	
				/* Not already allocated; proceed to initialise a new structure */
				if (!psp)
				{
					/* Check a pseudonym? */
					char* chkpseud = (chkpseuds) ? chkpseuds[nin] : 0;

					/* Find space for struct, initialise to blank state */
					if (!(psp = evolve__addpstate(ipids[nin], flags, &ctx, chkpseud))) erf(ERF_mallocFailed, 3);
				}
			}

			/* Ignore empty names */
			else psp = 0;

			/* Place pointer in index, if required */
			if (pidx) (*pidx)[nin] = psp;
		}
	}

	/* Return pointer to blank index if initial list empty */
	else if (pidx) *pidx = emptypidx;

	/* Callback functions and context */
	ctx.fctx = fctx;
	ctx.func = func;
	ctx.evfunc = evfn;
	ctx.repfunc = 0;
	ctx.secfunc = 0;

	/* Bother with targetting? */
	i = (ctx.flags & EVOLVE_Targetting) ? EVREAD_Targetting : 0;

	/* Step through events, helper function collects data */
	if (rc = evread(evolve__ef, evolve__wrapper, buf, bufend, bufused, &ctx, i)) erf(rc, 4);

	/* Time limit will be taken as time of referenced event (if found) */
	tv = (ctx.eptr) ? ctx.eptr->when : ctx.endtime;

	/* Sanity check: can't evolve back before last processed event */
	if (tv < ctx.lasttime) tv = ctx.lasttime;

	fprintf(stderr, "%li:   moof &%lX\n", uniqid, ctx.flags);

	/* Option: Evolve timers right up to time limit */
	if (ctx.flags & EVOLVE_TimersToLimit)
	{
		/* Deal with timer tickovers */
		for (psp = ctx.plist; (psp); psp = psp->next)
		{
			/* Evolve each timer up to limit */
			if (rc = evolve__timer(psp, &ctx, tv, ctx.elapsed)) erf(rc, 5);
		}

		/* Update elapsed time value */
		ctx.elapsed = ELAPTIME(tv);
	}

	/* Return effective end time of evolution */
	if (rettime) *rettime = tv;

	/* Ran out of events prior to passing end condition */
	if ((ctx.matched <= 1) && (ctx.flags & EVOLVE_FuncExhaust))
	{
		/* Optional callback, may alter some context */
		if ((i = (*ctx.func)(0, 0, ctx.fctx, &ctx) < 0)) erf(-i, 6);
	}

	/* Returning linked list? */
	if (pout) *pout = ctx.plist;

	/* Returning referenced event struct? */
	if (meep) *meep = ctx.eptr;

	/* Referenced event struct no longer needed? */
	else if (ctx.eptr) free(ctx.eptr);
 
	/* Pass back some extra information */
	if (erout)
	{
		erout->elist = ctx.elist;
		erout->existcnt = ctx.existcnt;
	}


_erf: /***** Error handling *****/

	/* No error, or reserved error */
	if (rc <= 0) return (rc);

	/* Debug */
	fprintf(stderr, "%li:   evolve(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

#endif
/*** The end ***/
