/*****************************************************
**  genmail.c                       >> bin/genmail  **
**  ----------------------------------------------  **
**  Generates textual output suitable for, eg. the  **
**  targetting or password emails, from current     **
**  game status. Can also generate lists of player  **
**  information according to selection criteria.    **
**  ----------------------------------------------  **
**  TODO: Add player status directives.             **
**        Choose default target/password options.   **
**  ----------------------------------------------  **
**  Converted to new time system                    **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2003 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define NO_CGI
#define MAX_fpbuf			65536
#define MAX_statbuf			131072
#define PLAYER_EXT_TYPE 	int

#include "!lib-general.c"
#include "!custom.c"
#include "!lib-uinfo-scan.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Buffer for input message */
#define	MAX_inbuf (64*1024)

/* Magic characters */
#define MAGOS "{"
#define MAGOC '{'
#define MAGCC '}'
#define MAGCS "}"

/* Mode flags */
#define MODE_And		(1U<<0)
#define MODE_Debug		(1U<<1)
#define MODE_Verbose	(1U<<2)
#define MODE_NoMail		(1U<<3)

/* Default mode */
static unsigned int mode = MODE_And|MODE_Verbose;

/* Special pointer used to denote default filenames. String irrelevent, but should be non-empty. */
static char* defp = "(default)";

/* Global variables */
static char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos;
static char inbuf[MAX_inbuf+4];

/* Parsed fragment of input text */
struct fragment
{
	struct fragment *next;
	size_t len;
	char *text;
	char type, num, flob;
};

/* Insertation fragment types */
#define FRAG_Text			0
#define FRAG_Pseud			1
#define FRAG_Name			2
#define FRAG_Addr			3
#define FRAG_College		4
#define FRAG_Mafia			5
#define FRAG_Password		6
#define FRAG_Date			7
#define FRAG_From			8
#define FRAG_Email			9
#define FRAG_Numeric		10
#define FRAG_Deadline		11
#define FRAG_ID				12

/* Looping fragment types */
#define FRAG_InitLoop		30
#define FRAG_IterLoop		31

/* Conditional fragment types (+1 => if, +0 => ifnot/else, +2 => endif, +3 unused) */
#define FRAG_Conditional	32
#define FRAG_IsDead			32
#define FRAG_IsPolice		36
#define FRAG_IsAllocated	40
#define FRAG_SameOldTarget	44
#define FRAG_IsMailable		48
#define FRAG_IsWanted		52
#define FRAG_IsIncompetent	56
#define FRAG_IsInactive		60
#define FRAG_IsAlmostInactive   64
#define FRAG_IsSpecial		68
#define FRAG_IsToBeMailed	72
#define FRAG_IsListed		76
#define FRAG_HasSome		80	/* Special... should be last in list: integers < FLOB_MAX may be added sometimes */
#define FRAG_CondLim		84

/* Mask to remove lowest bits of conditanal fragment types */
#define SMASK	0xFC

/* Set of flobs */
struct flobset
{
	int lim;
	player_p* pl;
	char* stuff;
	int cur;
	int cnt;
};

/* Types of flob */
#define FLOB_PL		0
#define FLOB_ASS	1		/* Current assassins */
#define FLOB_TAR	2		/* Current targets */
#define FLOB_OLD	3		/* Last known targets */
#define FLOB_NEW	4		/* New targets since last known */
#define FLOB_SAME	5		/* Same targets as last known */
#define FLOB_EX		6		/* Ex-targets from last known */
#define FLOB_MAX	7

/* Number of possible search terms */
#define MAX_SearchTerms		((FRAG_CondLim - FRAG_Conditional) / 4)
#define MAX_ExSrchTerms		(MAX_SearchTerms + FLOB_MAX - 1)

/* Printable names for search terms */
char* sterm_tab[MAX_ExSrchTerms] =
{
	"Dead", "Police", "[Allocated]", "SameOldTarget", "Mailable",
	"Wanted", "Incompetent", "Inactive", "AlmostInact", "Special",
        "Selected", "Listed", "HasSome", "HasAss", "HasTar", "HasOldTar",
        "HasNewTar", "HasSameTar", "HasExTar"
};

/***************************************************************************************************************************
** Preparation of player information/linkage arrays                                                                       **
***************************************************************************************************************************/
int chkarr(struct flobset arr[FLOB_MAX], int max, int c)
{
	if (( arr[FLOB_PL].stuff >= arr[FLOB_ASS].stuff) ||
		(arr[FLOB_ASS].stuff >= arr[FLOB_TAR].stuff) ||
		(arr[FLOB_TAR].stuff >= arr[FLOB_OLD].stuff) ||
		(arr[FLOB_OLD].stuff >=  arr[FLOB_PL].stuff + max)) return 1;

	if (( arr[FLOB_PL].pl >= arr[FLOB_ASS].pl) ||
		(arr[FLOB_ASS].pl >= arr[FLOB_TAR].pl) ||
		(arr[FLOB_TAR].pl >= arr[FLOB_OLD].pl) ||
		(arr[FLOB_OLD].pl >=  arr[FLOB_PL].pl + max)) return 1;

	if (c > 1)
	{
		if (( arr[FLOB_OLD].pl >= arr[FLOB_SAME].pl) ||
			(arr[FLOB_SAME].pl >=   arr[FLOB_PL].pl + max)) return 1;

		if (( arr[FLOB_OLD].stuff >= arr[FLOB_SAME].stuff) ||
			(arr[FLOB_SAME].stuff >=   arr[FLOB_PL].stuff + max)) return 1;
	}

	if (c > 2)
	{
		if ((arr[FLOB_SAME].pl >= arr[FLOB_NEW].pl) ||
			( arr[FLOB_NEW].pl >=  arr[FLOB_PL].pl + max)) return 1;

		if ((arr[FLOB_SAME].stuff >= arr[FLOB_NEW].stuff) ||
			( arr[FLOB_NEW].stuff >=  arr[FLOB_PL].stuff + max)) return 1;

		if ((arr[FLOB_SAME].lim >= arr[FLOB_PL].pl + max -  arr[FLOB_OLD].pl) ||
			( arr[FLOB_NEW].lim >= arr[FLOB_PL].pl + max - arr[FLOB_SAME].pl)) return 1;

		if ((arr[FLOB_SAME].lim >= arr[FLOB_PL].stuff + max -  arr[FLOB_OLD].stuff) ||
			( arr[FLOB_NEW].lim >= arr[FLOB_PL].stuff + max - arr[FLOB_SAME].stuff)) return 1;

		if ((arr[FLOB_SAME].lim >= max) || (arr[FLOB_NEW].lim >= max)) return 1;
	}

	if (c > 3)
	{
		if ((arr[FLOB_NEW].pl >= arr[FLOB_EX].pl) ||
			( arr[FLOB_EX].pl >= arr[FLOB_PL].pl + max)) return 21;

		if (arr[FLOB_EX].lim >= arr[FLOB_PL].pl + max -  arr[FLOB_NEW].pl) return 22;

		if ((arr[FLOB_NEW].stuff >= arr[FLOB_EX].stuff) ||
			( arr[FLOB_EX].stuff >= arr[FLOB_PL].stuff + max)) return 23;

		if (arr[FLOB_EX].lim >= arr[FLOB_PL].stuff + max -  arr[FLOB_NEW].stuff) return 24;

		if (arr[FLOB_EX].lim >= max) return 25;
	}

	return 0;
}

void freearr(struct flobset arr[FLOB_MAX])
{
	/*fprintf(stderr, "freearr( %p %p )\n", arr[FLOB_PL].pl, arr[FLOB_PL].stuff);*/

	if (arr[FLOB_PL].pl) free(arr[FLOB_PL].pl);

	/*fprintf(stderr, "moo ");*/

	if (arr[FLOB_PL].stuff) free(arr[FLOB_PL].stuff);

	arr[FLOB_PL].pl = 0;
	arr[FLOB_PL].stuff = 0;	

	/*fprintf(stderr, "freearr() successful\n");*/
}

int initarr(player_p pptr, struct flobset arr[FLOB_MAX], player_p plist)
{
	target_p tarp;
	player_p *parrp, qptr;
	char *pcharr;
	int i, j, max;

	/*fprintf(stderr, "%li:   initarr(\"%s\") ", uniqid, pptr->id);*/

	/* Limits */
	for (arr[FLOB_ASS].lim = 1, tarp = pptr->asses; (tarp); tarp = tarp->next) arr[FLOB_ASS].lim++;
	for (arr[FLOB_TAR].lim = 1, tarp = pptr->targs; (tarp); tarp = tarp->next) arr[FLOB_TAR].lim++;
	for (arr[FLOB_OLD].lim = 1, tarp = pptr->oldts; (tarp); tarp = tarp->next) arr[FLOB_OLD].lim++;

	/* Minimum size of an array */
	arr[FLOB_SAME].lim = arr[FLOB_NEW].lim = arr[FLOB_EX].lim = arr[FLOB_PL].lim = 1;

	/* Maximum number of thingies */
	max = i = 2 + arr[FLOB_ASS].lim + (arr[FLOB_TAR].lim + arr[FLOB_OLD].lim) * 2;

	/* Allocate some thingies (array of player pointers) */
	if (!(parrp = malloc(i * sizeof(player_p)))) return ERF_mallocFailed;

	/* Allocate more thingies (array of chars) */
	if (!(pcharr = malloc(i * sizeof(char)))) return ERF_mallocFailed;

	/*fprintf(stderr, " with max=%i->%p ( %p %p )", max, maxp, parrp, pcharr);*/

	/* Player himself */
	(arr[FLOB_PL].pl  = parrp)[0] = pptr;
	(arr[FLOB_PL].stuff = pcharr)[0] = 0;

	/* Individual arrays of player pointers */
	arr[FLOB_ASS].pl = 1 + parrp;
	arr[FLOB_TAR].pl = arr[FLOB_ASS].pl + arr[FLOB_ASS].lim;
	arr[FLOB_OLD].pl = arr[FLOB_TAR].pl + arr[FLOB_TAR].lim;
	arr[FLOB_SAME].pl = arr[FLOB_OLD].pl + arr[FLOB_OLD].lim;

	/* Individual arrays of status chars */
	arr[FLOB_ASS].stuff = 1 + pcharr;
	arr[FLOB_TAR].stuff = arr[FLOB_ASS].stuff + arr[FLOB_ASS].lim;
	arr[FLOB_OLD].stuff = arr[FLOB_TAR].stuff + arr[FLOB_TAR].lim;
	arr[FLOB_SAME].stuff = arr[FLOB_OLD].stuff + arr[FLOB_OLD].lim;

	/* Paranoia */
	if (chkarr(arr, max, 1)) fprintf(stderr, "[***** chkarr *****]");

	/* Initial entries */
	arr[FLOB_ASS].pl[0] = 0;
	arr[FLOB_TAR].pl[0] = 0;
	arr[FLOB_OLD].pl[0] = 0;
	arr[FLOB_SAME].pl[0] = 0;

	/* Clear the array of chars */
	memset(pcharr, 0, max);

	/* Fill in arrays: assassins */
	for (tarp = pptr->asses, i = 1; (tarp); tarp = tarp->next, ++i)
	{
		/* Paranoia */
		if ((i >= arr[FLOB_ASS].lim) || (i >= max)) fprintf(stderr, "[***** bounds 1 *****]");

		/* Null entry? */
		if (!tarp->id) qptr = 0;

		/* Search for amongst players */
		else for (qptr = plist; (qptr); qptr = qptr->link) if (streq(tarp->id, qptr->id)) break;

		/* Store result */
		arr[FLOB_ASS].pl[i] = qptr;
	}

	/* Fill in arrays: current targets */
	for (tarp = pptr->targs, i = 1; (tarp); tarp = tarp->next, ++i)
	{
		/* Paranoia */
		if ((i >= arr[FLOB_TAR].lim) || (i >= max)) fprintf(stderr, "[***** bounds 2 *****]");

		/* Null entry? */
		if (!tarp->id) qptr = 0;

		/* Search for amongst players */
		else for (qptr = plist; (qptr); qptr = qptr->link) if (streq(tarp->id, qptr->id)) break;

		/* Store result */
		arr[FLOB_TAR].pl[i] = qptr;
	}

	/* Fill in arrays: last set of targets */
	for (tarp = pptr->oldts, i = 1; (tarp); tarp = tarp->next, ++i)
	{
		/* Paranoia */
		if ((i >= arr[FLOB_OLD].lim) || (i >= max)) fprintf(stderr, "[***** bounds 3 *****]");

		/* Null entry? */
		if (!tarp->id) qptr = 0;

		/* Search for amongst players */
		else for (qptr = plist; (qptr); qptr = qptr->link) if (streq(tarp->id, qptr->id)) break;

		/* Store result */
		if (arr[FLOB_OLD].pl[i] = qptr)
		{
			/* Check against current targets */
			for (j = 1; (j < arr[FLOB_TAR].lim); j++) if (arr[FLOB_TAR].pl[j] == arr[FLOB_OLD].pl[i])
			{
				/* Enforce one-to-one-ness */
				if (!(arr[FLOB_TAR].stuff[j])) break;
			}

			/* Valid match */
			if (j < arr[FLOB_TAR].lim)
			{
				arr[FLOB_TAR].stuff[j] = 1;
				arr[FLOB_OLD].stuff[i] = 1;

				/* Count targets that are both current and previous */
				++arr[FLOB_SAME].lim;
				/*fprintf(stderr, "[++]");*/
			}
		}
	}

	/* Paranoia */
	if (chkarr(arr, max, 2)) fprintf(stderr, "[***** chkarr 2 *****]");

	/* Initialise sub arrays of player pointers */
	arr[FLOB_NEW].pl = arr[FLOB_SAME].pl + arr[FLOB_SAME].lim;
	arr[FLOB_NEW].stuff = arr[FLOB_SAME].stuff + arr[FLOB_SAME].lim;

	/* Sort current targets into sub arrays of "same" and "new" ones */
	for (i = 1, j = 0; (i < arr[FLOB_TAR].lim); i++)
	{
		/* Paranoia */
		if ((i >= arr[FLOB_TAR].lim) || (i >= max)) fprintf(stderr, "[* bounds 3 %i %i %i *]", i, arr[FLOB_TAR].lim, max);
		if ((j >= arr[FLOB_SAME].lim) || (j >= max)) fprintf(stderr, "[* bounds 4 %i %i %i *]", j,arr[FLOB_SAME].lim,max);

		/* Same as previously known target */
		if (arr[FLOB_TAR].stuff[i])
		{
			arr[FLOB_SAME].pl[++j] = arr[FLOB_TAR].pl[i];
			arr[FLOB_SAME].stuff[j] = arr[FLOB_TAR].stuff[i];
		}

		/* New target */
		else arr[FLOB_NEW].pl[arr[FLOB_NEW].lim++] = arr[FLOB_TAR].pl[i];

		/* Paranoia */
		if (chkarr(arr, max, 3)) fprintf(stderr, "[***** chkarr 3 *****]");
	}

	/* Initialise sub arrays for ex-targets */
	arr[FLOB_EX].pl = arr[FLOB_NEW].pl + arr[FLOB_NEW].lim;
	arr[FLOB_EX].stuff = arr[FLOB_NEW].stuff + arr[FLOB_NEW].lim;

	/* Paranoia */
	if (i = chkarr(arr, max, 4)) fprintf(stderr, "[***** chkarr 4 i=%i *****]", i);


	/* Extract ex-targets from previous set */
	for (i = 1, j = 0; (i < arr[FLOB_OLD].lim); i++)
	{
		/* Paranoia */
		if (j >= max) fprintf(stderr, "[* bounds 6 %i %i *]", j, max);

		/* Not same as current target */
		if (!arr[FLOB_OLD].stuff[i]) arr[FLOB_EX].pl[++j] = arr[FLOB_OLD].pl[i];
	}

	/* Limits for last subarray */
	arr[FLOB_EX].lim = j + 1;
	arr[FLOB_EX].pl[0] = 0;
	arr[FLOB_NEW].pl[0] = 0;
	

	/* Paranoia: Sanity check */
	if ((arr[FLOB_EX].lim - 2 + arr[FLOB_NEW].lim + 2 * arr[FLOB_SAME].lim) != (arr[FLOB_TAR].lim + arr[FLOB_OLD].lim))
	{
		/* Bounds don't match up */
		fprintf(stderr, "%li !! Bounds mismatch in initarr() on \"%s\": %i+%i+%i %i+%i", uniqid, pptr->id,
				arr[FLOB_EX].lim, arr[FLOB_NEW].lim, arr[FLOB_SAME].lim, arr[FLOB_TAR].lim, arr[FLOB_OLD].lim);

		/* Error! */
		return ERF_InternalError;
	}

	/* Initialisation */
	for (i = 0; (i < FLOB_MAX); i++)
	{
		/* Initialise "current" indexes */
		arr[i].cur = arr[i].cnt = 0;

		/* Count number of non-null entries */
		if (i > FLOB_PL) for (j = 1; (j < arr[i].lim); j++) if (arr[i].pl[j]) ++arr[i].cnt;
	}

	/*fprintf(stderr, " ininarr:internals:(%p,%p)\n", arr[FLOB_PL].pl, arr[FLOB_PL].stuff);*/

	/* Done */
	return 0;
}

/***************************************************************************************************************************
** Evaluation of conditional expressions on players                                                                       **
***************************************************************************************************************************/
unsigned int evalcond(player_p thispptr, struct flobset arr[FLOB_MAX], int type, uint n, uint f)
{
	/* Debug */
	/*fprintf(stderr, ">> condition %i on %i (for %s)\n", type, n, (pl[0]) ? pl[0]->id : "!!none!!");*/

	/* Invalid choice of linkset? */
	if (f >= FLOB_MAX)
	{
		/* Debug */
		fprintf(stderr, "%li: ! Invalid evalcond: %i [%u,%u] \"%s\" %p\n", uniqid,
		 				type, f, n, (arr[0].pl[0]) ? arr[0].pl[0]->id : "!!none!!", thispptr);
		return 0;
	}

	/* Current number within set? */
	if ((n == 0) && (f != FLOB_PL)) n = arr[f].cur;

	/* Valid number within set? */
	if (n >= arr[f].lim) return 0;

	/* Evaluate */	
	switch (type = type & SMASK)
	{
		/* TRUE if player[f][n] is allocated (ie. as n-th assassin/target/oldtarget for player[0]) */
		case FRAG_IsAllocated: return (arr[f].pl[n]) ? 1 : 0;

		/* TRUE if player[f][n] is unallocated or is dead */
		case FRAG_IsDead: return ((!arr[f].pl[n]) || (arr[f].pl[n]->flags & PF_Dead));

		/* TRUE if player[f][n] is allocated and mailable */
		case FRAG_IsMailable: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_Mailable));

		/* TRUE if player[f][n] is allocated and has been selected for mailing */
		case FRAG_IsToBeMailed: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_SendMail));

		/* TRUE if player[f][n] is allocated and is special */
		case FRAG_IsSpecial: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_Special));

		/* TRUE if player[f][n] is allocated and was in matched in player list given as input */
		case FRAG_IsListed: return ((arr[f].pl[n]) && (arr[f].pl[n]->ext));

		/* TRUE if player[f][n] is allocated and is police */
		case FRAG_IsPolice: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_Police));

		/* TRUE if player[f][n] is allocated and is wanted */
		case FRAG_IsWanted: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_Wanted));

		/* TRUE if player[f][n] is allocated and is incompetent */
		case FRAG_IsIncompetent: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_Incompetent));

		/* TRUE if player[f][n] is allocated and is inactive */
		case FRAG_IsInactive: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_Inactive));

                /* TRUE if player[f][n] is allocated and nearly inactive */
                case FRAG_IsAlmostInactive: return ((arr[f].pl[n]) && (arr[f].pl[n]->flags & PF_InactWarning));
                                      
		/* TRUE if player[f][n] is allocated with and is current and old target */
		case FRAG_SameOldTarget: return ((arr[f].pl[n]) && (arr[f].stuff[n]));

		/* TRUE if player[f] has non-null entries */
		case FRAG_HasSome: return (arr[f].cnt > 0);

		/* Unknown: return TRUE */
		default: return 1;
	}
}

/***************************************************************************************************************************
** Routines to parse input files                                                                                          **
***************************************************************************************************************************/
struct fragment* addfrag(struct fragment** ffr, struct fragment** lfr, char* p, size_t l, char t, char n, char f)
{
	struct fragment* frag;

	if (!(frag = malloc(sizeof(struct fragment)))) return 0;

	if (p) p[l] = '\0';

	frag->text = p;
	frag->len  = l;
	frag->type = t;
	frag->num  = n;
	frag->flob = f;
	frag->next = 0;

	if (!(*ffr)) *ffr = frag;
	if (*lfr) (*lfr)->next = frag;

	/*fprintf(stderr, "addfrag(&%p,&%p,&%p,%u,%i,%i)-->&%p\n", *ffr, *lfr, p, l, t, n, frag);*/
	/*if (p) fprintf(stderr, "       \"%.30s\"", p);*/

	return (*lfr = frag);
}

int parsefrags(FILE *fp, struct fragment** firstfragp)
{
	struct fragment *lastfrag;
	size_t s, x, y, z;
	int rc = 0;


	/* Nothing yet */
	*firstfragp = 0;

	/* Slurp in input file */
	for (s = 0; (!ferror(fp)) && (!feof(fp)); s += x)
	{
		/* Room in buffer? */
		if (s >= MAX_inbuf) erf(ERF_BufferExhaust, 3);

		/* Read as much as possible */
		if ((x = fread(inbuf + s, 1, MAX_inbuf - s, fp)) == 0) continue;

		/* Paranoia */
		else if (/*(x < 0) ||*/ (x > (MAX_inbuf - s))) erf(ERF_freadFailed, 4);
	}

	/* Any residual errors? */
	if (ferror(fp)) erf(ERF_FileError, 5);

	/* Terminate buffer */
	inbuf[s] = '\0';

	/* Scan buffered file for special stuff */
	for (x = y = 0, lastfrag = 0; (x < s); x++)
	{
		/* Assume neither playerset nor playernumber is needed */
		char t = 0, n = 0, f = 0, req = 0;

		/* Important strings begin with magic char */
		if (inbuf[x] != MAGOC) continue;

		/* Use "{{" to get '{' character */
		if (inbuf[x + 1] == MAGOC) x++;

		/* {Date} directive? */
		else if (!strncasecmp(inbuf + x + 1, "date" MAGCS, 5))
		{
			t = FRAG_Date;
			z = x + 5;
		}

		/* {From} directive? */
		else if (!strncasecmp(inbuf + x + 1, "from" MAGCS, 5))
		{
			t = FRAG_From;
			z = x + 5;
		}

		/* {Password} directive? */
		else if (!strncasecmp(inbuf + x + 1, "password" MAGCS, 9))
		{
			t = FRAG_Password;
			z = x + 9;
		}

		/* Normal directives */
		else
		{
			/* Assume both playerset and playernumber needed */
			req = 1+4+8;			

			/* Type of directive */
			switch (tolower(inbuf[z = x + 1]))
			{
				/* Insertation directives */
				case 'p': t = FRAG_Pseud; break;
				case 'n': t = FRAG_Name; break;
				case 'a': t = FRAG_Addr; break;
				case 'c': t = FRAG_College; break;
				case 'm': t = FRAG_Mafia; break;
				case 'e': t = FRAG_Email; break;
				case 'b': t = FRAG_Deadline; break;
				case 'i': t = FRAG_ID; break;

				/* "Counting" insertation directives - non-null playerset needed */
				case '#': t = FRAG_Numeric; req = 1+2; break;

				/* Ignore unknown directives */
				default : x = z; continue;

				/* Looping directives */
				case '<': t = FRAG_InitLoop; req = 1+2; break;
				case '>': t = FRAG_IterLoop; req = 1+2; break;

				/* Conditional directives */
				case '/': ++t; /* Fall-through 2 */
				case '?': ++t; /* Fall-through 1 */
				case '!': /* Otherwise, have t=0 */
				{
					/* Which condition? */
					switch (tolower(inbuf[++z]))
					{
						case 'd': t += FRAG_IsDead; break;
						case 'p': t += FRAG_IsPolice; break;
						case 'x': t += FRAG_IsAllocated; break;
						case 's': t += FRAG_IsSpecial; break;
						case 'w': t += FRAG_IsWanted; break;
						case 'i': t += FRAG_IsIncompetent; break;
						case 'c': t += FRAG_IsInactive; break;
                                                case 'v': t += FRAG_IsAlmostInactive; break;
						case 'm': t += FRAG_IsToBeMailed; break;
						case 'l': t += FRAG_IsListed; break;
						case 'e': t += FRAG_IsMailable; break;

						/* Has some entries in list? */
						case 't': t += FRAG_HasSome; req = 1+2; break;

						/* On current and old target list? */
						case 'b': t += FRAG_SameOldTarget; req = 1+2+4; break;

						/* Ignore unknown conditions */
						default : x = z; continue;
					}
					break;
				}
			}

			/* Choice of playerset required? */
			if (req & 1)
			{
				/* If none given, default is (FLOB_PL = 0) from initial declarations */
				if (inbuf[++z] != MAGCC) switch (tolower(inbuf[z]))
				{
					case 'a': f = FLOB_ASS;  break;
					case 't': f = FLOB_TAR;  break;
					case 'o': f = FLOB_OLD;  break;
					case 'n': f = FLOB_NEW;  break;
					case 's': f = FLOB_SAME; break;
					case 'x': f = FLOB_EX;   break;

					/* Correct for early end-of-string */
					case '\0': --z; break;
				}

				/* Is FLOB_PL an invalid choice? */
				if ((req & 2) && (f == FLOB_PL)) goto ig_bad;
			}
			
			/* Numeric suffix desired? Default is 0 from initial declaration */
			if ((req & 4) && !((req & 8) && (f == FLOB_PL)))
            {
				/* No suffix present? */
				if (inbuf[++z] == MAGCC)
				{
					/* Demand that a suffix be present? */
					if (req & 16) goto ig_bad;
				}

				/* If one is present, is it valid? */
				else if ((inbuf[z] < '0') || ((n = inbuf[z] - '0') > 9)) goto ig_bad;

				/* Do we allow 0 as a valid choice? */
				if ((req & 32) && (n == 0)) goto ig_bad;
			}

			/* Rubbish on the end of the directive? */
			if ((inbuf[z] != MAGCC) && ((inbuf[z] == '\0') || (inbuf[++z] != MAGCC)))
			{
				/* Ignore bad directives */
	ig_bad:		x = (inbuf[z]) ? (z + 1) : z;
				continue;
			}

			/* Debug */
			/*fprintf(stderr, "foom!\"%s\"\n", inbuf + x);*/
		}

		/* Add previous string fragment */
		if ((x > y) && (!addfrag(firstfragp, &lastfrag, inbuf + y, x - y, FRAG_Text, 0, 0))) erf(ERF_mallocFailed, 20);

		/* Letter following magic char */
		y = x + 1;

		/* Add command fragment */
		if (t)
		{
			if (!addfrag(firstfragp, &lastfrag, inbuf + y, z - y, t, n, f)) erf(ERF_mallocFailed, 21);
			y = 1 + (x = z);
		}
	}

	/* Don't forget the last string fragment */
	if ((x > y) && (!addfrag(firstfragp, &lastfrag, inbuf + y, x - y, FRAG_Text, 0, 0))) erf(ERF_mallocFailed, 22);


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li !! parsefrags(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Write fragmented input back out with substitution                                                                      **
***************************************************************************************************************************/
int writefrags(FILE *fp, player_p thispptr, struct flobset arr[FLOB_MAX],
			   player_p plist, struct fragment *fraglist, char* date, char* from)
{
	struct fragment *frag, *loops[FLOB_MAX], *loopends[FLOB_MAX];
	player_p pptr;
	int i, f, ok, n, rc = 0;
	char* p;


	/* No loops yet */
	for (i = 0; (i < FLOB_MAX); i++) loops[i] = loopends[i] = 0;

	/* Write out each fragment in turn */
	for (frag = fraglist; (frag); frag = frag->next)
	{
		/*fprintf(stderr, "* frag %p type %i flob %i num %i %s len %lu\n", frag, frag->type,
				frag->flob, frag->num, (frag->type == FRAG_Text) ? "text" : frag->text, (unsigned long) frag->len);*/

		/* No text yet */
		n = frag->num;
		p = 0;

		/* Sanity check */
		if ((f = frag->flob) >= FLOB_MAX) erf(ERF_InternalError, 1);

		/* Deal with loops: opening */
		if (frag->type == FRAG_InitLoop)
		{
			/* Already in loop? Ignore if so. */
			if (loops[f]) continue;

			/* Note start of loop */
			loops[f] = frag;

			/* End not yet known */
			loopends[f] = 0;

			/* Process loop condition */
			goto doloop;
		}

		/* Loop iteration */
		if (frag->type == FRAG_IterLoop)
		{
			/* In a loop? Ignore if not. */
			if (loops[f])
			{
				/* Note end of loop */
				loopends[f] = frag;

	doloop:		/* Find next valid player */
				while (++arr[f].cur < arr[f].lim) if (arr[f].pl[arr[f].cur]) break;
	
				/* End of loop... no more players in list */
				if (arr[f].cur >= arr[f].lim)
				{
					arr[f].cur = 0;
					loops[f] = 0;

					/* Use cached loop endpoint? */
					if (loopends[f]) frag = loopends[f];

					/* Search forward for first relevent loopend */
					else for (frag = frag->next; (frag); frag = frag->next)
					{
						/* Matching loop iterator? */
						if ((frag->type == FRAG_IterLoop) && (frag->flob == f)) break;
					}
				}

				/* Back to start of loop */
				else frag = loops[f];
			}

			/* Run out of fragments? */
			if (!frag) break;

			/* Process "next" fragment (ie. after immediately after start or end of loop) */
			continue;
		}

		/* Deal with conditional blocks */
		if (frag->type >= FRAG_Conditional)
		{
			unsigned int a, b;

			/* Ignore out-of-place blocks */
			if ((a = frag->type & 3) > 1) continue;

			/* Evaluate condition to b = 0/1 */
			b = evalcond(thispptr, arr, frag->type & SMASK, n, f);

			/*fprintf(stderr, "%li:   conditional \"%s\" (%i,%i) a=%i b=%u -> %s\n", uniqid,
					frag->text, frag->type, frag->num, a, b, (a ^ b) ? "unsatisfied" : "satisfied");*/

			/* Match on required condition status */
			if (!(a ^ b)) continue;

			/* Endif and Else conditions */
			b = 2 + (frag->type & SMASK);
			a = b - 1 - a;

			/* Scan along to end of block */
			for (frag = frag->next; (frag); frag = frag->next)
			{
				/* Corresponding Else or Endif? */
				if ((frag->num == n) && (frag->flob == f) && ((frag->type == a) || (frag->type == b))) break;
			}

			/* Run out of fragments? */
			if (!frag) break;

			/* Otherwise, Else automatically satisfied, or Endif */
			else continue;
		}

		/* Use current number within set? */
		if ((n == 0) && (f != FLOB_PL)) n = arr[f].cur;

		/* Number is valid within set? */
		ok = (n < arr[f].lim);

		/* What sort of fragment? */
		switch (frag->type)
		{
			/* Plain text */
			case FRAG_Text:
			{
				if (frag->len != fwrite(frag->text, 1, frag->len, fp)) erf(ERF_fwriteFailed, 6);
				break;
			}

			/* Strings for mail recipient */
			case FRAG_Password:	p = thispptr->pass; break;
			case FRAG_Date:		p = date; break;
			case FRAG_From:		p = from; break;

			/* Strings for a given player */
			case FRAG_ID:		if ((ok) && (pptr = arr[f].pl[n])) p = pptr->id; break;
			case FRAG_Name:		if ((ok) && (pptr = arr[f].pl[n])) p = pptr->name; break;
			case FRAG_Addr:		if ((ok) && (pptr = arr[f].pl[n])) p = pptr->addr; break;
			case FRAG_College:	if ((ok) && (pptr = arr[f].pl[n])) p = pptr->coll; break;
			case FRAG_Mafia:	if ((ok) && (pptr = arr[f].pl[n])) p = pptr->mafia; break;
			case FRAG_Pseud:	if ((ok) && (pptr = arr[f].pl[n])) p = pptr->pseud; break;

			/* Current intactivity deadline */
			case FRAG_Deadline:
			{
				if ((!ok) || !(pptr = arr[f].pl[n])) break;
				if (pptr->flags & PF_InactTimer)
				{
					fprintf(fp, ""TDISP" on %s", Thm(pptr->timer), fdayname(T_d(pptr->timer), 0));
				}
				else if (EOF == fputs("None", fp)) erf(ERF_fputsFailed, 7);
				break;
			}

			/* Email address */
			case FRAG_Email:
			{
				if ((!ok) || !(pptr = arr[f].pl[n]) || (!pptr->email)) break;
				if (EOF == fputs(pptr->email, fp)) return ERF_fputsFailed;
				if (!strchr(pptr->email, '@') && (EOF == fputs(MAILDOMAIN, fp))) erf(ERF_fputsFailed, 8);
				break;
			}

			/* Count number of valid entries in a certain set */
			case FRAG_Numeric:
			{
				fprintf(fp, "%d", arr[f].cnt);
				break;
			}

			/* Paranoia: catch invalid types */
			default: erf(ERF_InternalError, 9);
		}

		/*if (p) fprintf(stderr, " (\"%s\")\n", p);*/

		/* Write indirected string */
    	if ((p) && (EOF == fputs(p, fp))) erf(ERF_fputsFailed, 10);
	}


_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li !! writefrags(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/***************************************************************************************************************************
** Main code. Command tail contains options.                                                                              **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	int i, n = 0, nontriv, rc = 0, lockfd = 0, sterm[MAX_ExSrchTerms];
	unsigned int j;
	player_p plist = 0, pptr;
	char *p, thedate[64], *do_input = 0, *do_list = 0, *do_pidlist = 0, *do_tarlist = 0;
	struct flobset arr[FLOB_MAX];
	struct fragment *frags = 0;
	FILE* fp = 0, *fp_pidlist = 0, *fp_tarlist = 0;


	/* No search terms, by default */
	for (i = 0; (i < MAX_ExSrchTerms); i++) sterm[i] = 0;

	/* Debug: no buffering on stdout */
	setbuf(stdout, 0);
	
	/* Initialse stuff: time, file creation umask, pid, no title, don't redirect stderr to log */
	if (i = geninit(0, 0117, 0, 0)) erf(i, 1);

	/* Turn it into a valid RFC 822 date */
	strftime(thedate, 64, "%a, %d %b %Y %H:%M:%S GMT", gmtime(&epochtime));

	/* Logfile header */
	fprintf(stderr, "%li ** Genmail: at RFC822 date \"%s\"\n", uniqid, thedate);

	/* Check command line args */
	for (i = 1; (i < argc) && (p = argv[i]); i++)
	{
		char **ap = 0;
		unsigned int m, mf = 0;

		/* Qualifier? */
		switch (*(p++))
		{
			case '+': m = 1; break;
			case '-': m = 2; break;
			case '^': m = 3; break;
			case '!': m = 4; break;
			case '?': m = 5; break;
			case '/': m = 6; break;
			default : m = 0; --p;
		}

		/* Search switches */
		if (m >= 4)
		{
			/* Special stuff */
			j = 0;

			/* Which condition? */
			switch (tolower(p[0]))
			{
				case 'd': mf = FRAG_IsDead; break;
				case 'p': mf = FRAG_IsPolice; break;
				case 's': mf = FRAG_IsSpecial; break;
				case 'w': mf = FRAG_IsWanted; break;
				case 'i': mf = FRAG_IsIncompetent; break;
				case 'c': mf = FRAG_IsInactive; break;
                                case 'v': mf = FRAG_IsAlmostInactive; break;
				case 'm': mf = FRAG_IsToBeMailed; break;
				case 'l': mf = FRAG_IsListed; break;
				case 'e': mf = FRAG_IsMailable; break;

				/* Unknown */
	iguknparm:	default: fprintf(stderr, "%li: ! Ignoring unknown parameter: \"%s\"\n", uniqid, argv[i]); continue;

				/* Special case: specify type of targets */
				case 't':
				{
					/* Which type? */
					switch (tolower(p[1]))
					{
						case 'a': j = MAX_SearchTerms -1 + FLOB_ASS;  break;
						case 't': j = MAX_SearchTerms -1 + FLOB_TAR;  break;
						case 'o': j = MAX_SearchTerms -1 + FLOB_OLD;  break;
						case 'n': j = MAX_SearchTerms -1 + FLOB_NEW;  break;
						case 's': j = MAX_SearchTerms -1 + FLOB_SAME; break;
						case 'x': j = MAX_SearchTerms -1 + FLOB_EX;   break;
						default: goto iguknparm;
					}
					mf = FRAG_HasSome;
				}
			}

			/* New search term */
			sterm[(j) ? (j) : ((mf - FRAG_Conditional) >> 2)] = (m >= 6) ? 0 : (mf + m - 4);

			/* Allow "?l=foo"-like constructions to do an automatic "+l=foo" switch */
			if (mf == FRAG_IsListed)
			{
				/* Set filename from (possibly empty) parameter, if present */
				if (p = strchr(p, '=')) do_list = p + 1;

				/* Otherwise, if no previous filename set, and not disabling search term, use default */
				else if ((!do_list) && (m < 6)) do_list = defp;
			}

			/* Done */
			continue;
		}

		/* Normal switches: Only first one or two letters of command switches are matched */
		else switch (tolower(p[0]))
		{
			/* Impose default settings for target mails */
			case '1':
			{
				if (m != 0) goto badqual;
				do_input = FILE_TargMail;
				/* mode and search flags */
				break;
			}

			/* Impose default settings for password mails */
			case '2':
			{
				if (m != 0) goto badqual;
				do_input = FILE_PassMail;
				break;
			}

			/* Special "and"/"or" flag pair */
			case 'a': case 'o':
			{
				if (m <= 1)  { mf = MODE_And; m = (tolower(p[0]) == 'a') ? 1 : 2; break; }
	badqual:	fprintf(stderr, "%li: ! Ignoring badly qualified parameter: \"%s\"\n", uniqid, argv[i]);
				break;
			}

			/* Flags */
			case 'd': mf = MODE_Debug; break;
			case 'v': mf = MODE_Verbose; break;
			case 'n': mf = MODE_NoMail; break;

			/* File parameters */
			case 'm': ap = &do_input; break;
			case 'l': ap = &do_list; break;
			case 'p': ap = &do_pidlist; break;
			case 't': ap = &do_tarlist; break;

			/* Unknown */
			default: fprintf(stderr, "%li: ! Ignoring unknown parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Do flags */
		if (mf) switch (m)
		{
			case 1: mode |=  mf; break;	/* Set flag */
			case 2: mode &= ~mf; break;	/* Clear flag */
			case 3: mode ^=  mf; break;	/* Toggle flag */
			default: fprintf(stderr, "%li: ! Ignoring unqualified parameter: \"%s\"\n", uniqid, argv[i]);
		}

		/* Handle files parameters */
		if (ap)
		{
			/* If any (possibly empty) filename given, activite mode */
			if (p = strchr(p, '=')) *ap = p + 1;

			/* Unparameterised - activate if appropriate, set to default filename */
			else if ((m <= 1) || ((m == 3) && (*ap == 0))) *ap = defp;

			/* Unparameterised - deactivate if appropriate */
			else *ap = 0;
		}
	}

	/* Defaults: redirect to stdout if no filename given */
	if (do_input == defp) do_input = "";

	/* Default filenames */
	if (do_list == defp) do_list = FILE_DefList;
	if (do_pidlist == defp) do_pidlist = FILE_DefList;
	if (do_tarlist == defp) do_tarlist = FILE_DefTargetList;

	/* Print out options */
	fprintf(stderr, "%li:   %s%sSearch terms: ", uniqid,
			(mode & MODE_Debug) ? "Debug; " : "", (mode & MODE_Verbose) ? "Verbose; " : "");

	/* Logic mode */
	p = (mode & MODE_And) ? " AND " : " OR ";

	/* Print out search terms */
	for (i = nontriv = 0; (i < MAX_ExSrchTerms); i++) if (sterm[i])
	{
		fprintf(stderr, "%s%s%s", (nontriv++) ? p : "", (sterm[i] & 1) ? "" : "Not", sterm_tab[i]);
	}

	/* Input player list? */
	p = (!do_list) ? "No player list. " : "";

	/* Non mailable-output? */
	if (mode & MODE_NoMail)
	{
		fprintf(stderr, "\n%li:  ! %s** Not generating %s", uniqid,
				p, (do_input) ? "mailable output text **" : "any textual output **\n");
		p = "";
	}

	/* Default file */
	else if (!do_input) do_input = FILE_DefMail;

	/* Process input */
	if (do_input)
	{
		/* Print out input mode */
		fprintf(stderr, "\n%li:   %sTaking input from \"%s\"\n", uniqid, p, (*do_input) ? do_input : "stdin");

		/* Can't read both input files from stdin */
		if ((do_list) && !(*do_list) && !(*do_input)) erf(ERF_BadParms, 20);

		/* Read message text from stdin? */
		if (!(*do_input)) fp = stdin;
	
		/* Open input file */
		else if (!(fp = fopen(do_input, "rb"))) erf(ERF_fopenFailed, 21);
	
		/* Parse input file into fragments */
		if (rc = parsefrags(fp, &frags)) erf(rc, 22);
	
		/* Close input file */
		i = (stdin != fp) && (fclose(fp));
	
		/* Be sure not to try to close it more than once */
		fp = 0; if (i) erf(ERF_fcloseFailed, 23);
	}

	/* Output list of mailed players? */
	if (do_pidlist)
	{
		/* Not allowed to use null filename to indicate stdout */
		if (!do_pidlist[0]) erf(ERF_BadParms, 30);

		/* Open file */
		if (!(fp_pidlist = fopen(do_pidlist, "wb"))) erf(ERF_fopenFailed, 31);

		/* Progress */
		fprintf(stderr, "%li:   Opened output file for list of mailed players: %s\n", uniqid, do_pidlist);
	}

	/* Output list of mailed players' current targets */
	if (do_tarlist)
	{
		/* Not allowed to use null filename to indicate stdout */
		if (!do_tarlist[0]) erf(ERF_BadParms, 32);

		/* Open file */
		if (!(fp_tarlist = fopen(do_tarlist, "wb"))) erf(ERF_fopenFailed, 33);

		/* Progress */
		fprintf(stderr, "%li:   Opened output file for list of mailed targets: %s\n", uniqid, do_tarlist);
	}

	/* List of players given? */
	if (do_list)
	{
		/* List of players on stdin? */
		if (!(*do_list)) fp = stdin;

		/* Open input file */
		else if (!(fp = fopen(do_list, "rb"))) erf(ERF_fopenFailed, 38);
	}

	/* Moof */
	/*fprintf(stderr, "goo\n");*/

	/* Attempt to lock files */
	if (!(lockfd = setflock("Genmail.c "))) erf(ERF_flockFailed, 40);

	/* Read full info for all players */
	if (rc = scanplinfo(SCANPL_AddFull|SCANPL_GetTargs, 0, 0, 0, &plist, statbuf, statbuf_end, &statbuf_pos)) erf(rc, 42);

	/* Release lock */
	if (lockfd) lockfd = setfunlock(lockfd);

	/* Parse list of players */
	if (fp)
	{
		char tmpstr[MAX_id+2];

		/* Message */
		fprintf(stderr, "%li:   Scanning player list %s%s\n", uniqid, (*do_list) ? "from file: " : "on stdin", do_list);

		/* No ids recognised yet */
		n = 0; i = 1;

		/* Read as many whitespace delimited strings as possible */
		while (1 == fscanf(fp, "%"STRING(MAX_id)"s", tmpstr))
		{
			p = (i != 1) ? "\n" : "";

			/* Stop if we encounter an excessively long string */
			if (strlen(tmpstr) >= MAX_id)
			{
				fprintf(stderr, "%s%li: ! Stopped scanning at string \"%s...\" (too long).\n", p, uniqid, tmpstr);
				i = 0; break;
			}

			/* Try to match on player list */
			for (pptr = plist; (pptr); pptr = pptr->link) if (streq(tmpstr, pptr->id)) break;

			/* Unmatched */
			if (!pptr)
			{
				fprintf(stderr, "%s%li:    xx Unknown player \"%s\"\n", p, uniqid, tmpstr);
				i = 1; continue;
			}

			/* Ignore duplicates */
			if (pptr->ext)
			{
				fprintf(stderr, "%s%li:    xx Duplicate player \"%s\"\n", p, uniqid, pptr->id);
				i = 1; continue;
			}

			/* Found one! */
			if ((i != 1) && (i != 11)) { fprintf(stderr, ", %s", pptr->id); ++i; }

			/* Format list neatly (10 per row) */
			else { fprintf(stderr, "%s%li:    -- %s", p, uniqid, pptr->id); i = 2; }

			/* Note that player was in list, and increment count of valid player ids */
			pptr->ext = 1; ++n;
		}

		/* Status report */
		fprintf(stderr, "%s%li:   Found %i distinct valid player ids in list.\n", (i > 1) ? ".\n" : "", uniqid, n);

		/* Close input file */
		i = (stdin != fp) && (fclose(fp));

		/* Be sure not to try to close it more than once */
		fp = 0; if (i) erf(ERF_fcloseFailed, 27);
	}

	/* Logic mode */
	j = (mode & MODE_And) ? 0 : 1;

	/* Run through players */
	for (n = 0, pptr = plist; (pptr); pptr = pptr->link)
	{
		target_p tarp;

		/* Debug */
		if (mode & MODE_Debug) fprintf(stderr, " (id:\"%s\",&%lx) ", STR(pptr->id), pptr->flags);

		/* Only mail "Mailable" players */
		if (!(mode & MODE_NoMail) && !(pptr->flags & PF_Mailable)) continue;

		/* Initialise arrays */
		if (i = initarr(pptr, arr, plist)) erf(i, 70);

		/* Ignore empty search patterns */
		if (nontriv)
		{
			/* Check against search terms */
			for (i = 0; (i < MAX_ExSrchTerms); i++) if (sterm[i])
			{
				/* Choice of set, if relevent, else player to be emailed */
				uint flob = (i < MAX_SearchTerms) ? 0 : (i + 1 - MAX_SearchTerms);

				/* Terminate as soon as condition satisfied or falsified */			
				if (j ^ (sterm[i] & 1) ^ evalcond(pptr, arr, sterm[i] & SMASK, 0, flob)) break;
			}
	
			/* Is early termination an indicator of success or failure? */
			if (j ^ (i < MAX_ExSrchTerms)) goto donext;
		}

		/* Report on progress */
		if (mode & MODE_Verbose)
		{
			if (!(n % 10)) fprintf(stderr, "%s%li:    >> %s", (n) ? "\n" : "", uniqid, pptr->id);
			else fprintf(stderr, ", %s", pptr->id);
		}

		/* Print player id to output list */
		if (fp_pidlist) fprintf(fp_pidlist, "%s\n", pptr->id);

		/* Print player's target information to output list */
		if (fp_tarlist)
		{
			fprintf(fp_tarlist, "%s\t", pptr->id);

			for (tarp = pptr->targs; (tarp); tarp = tarp->next)
			{
				if (tarp->id) fprintf(fp_tarlist, "\t%s", tarp->id);
			}

			fprintf(fp_tarlist, "\n");
		}

		/* Count messages sent */
		++n;

		/* SMTP Mail Commands */
		if (!(mode & MODE_NoMail)) printf("MAIL FROM:" MAILFROM CR LF "RCPT TO:%s%s" CR LF "DATA" CR LF,
										  pptr->email, ((strchr(pptr->email, '@')) ? "" : MAILDOMAIN));

		/* Message headers and body */
		if ((frags) && (rc = writefrags(stdout, pptr, arr, plist, frags, thedate, MAILFROM_head))) erf(rc, 50);

		/* SMTP Mail Footer */
		if (!(mode & MODE_NoMail)) printf(CR LF "." CR LF);

donext:	/* Debug */
		if (mode & MODE_Debug) fprintf(stderr, " (done:\"%s\") ", STR(pptr->id));

		/* Free memory */
		freearr(arr);
	}

	/* Finish SMTP Session */
	if (!(mode & MODE_NoMail)) fputs("QUIT" CR LF, stdout);

	/* Tidy up after progress text */
	if ((mode & MODE_Verbose) && (n> 0)) fputs(".\n", stderr);


_erf: /***** Error handling *****/

	/* Close temporary open files */
	if ((fp) && (fp != stdout) && (fp != stdin)) fclose(fp);

	/* Close output files */
	if (fp_tarlist) fclose(fp_tarlist);
	if (fp_pidlist) fclose(fp_pidlist);

	/* Debug */
	if (rc) fprintf(stderr, "%li !! Error %s (%i) in main(%i) mode=%x\n", uniqid,
					ERF_STR(rc), ERF_TYPE(rc), ERF_CONTEXT(rc), mode);

	/* Note success */
	else fprintf(stderr, "%li:   -------> generated %i messages <-------\n", uniqid, n);

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	return (ERF_LOWBITS(rc));
}

/**** The End! ****/

#if 0
oldcode()
{
	/* TRUE if player is allocated and with some identical current and old target(s) */
	/*case FRAG_HasOldTargs:*/
	{
		target_p tar1p, tar2p;
	
		if (f == 0)
		{
			if (thispptr) for (tar1p = thispptr->targs; (tar1p); tar1p = tar1p->next)
			{
				for (tar2p = thispptr->oldts; (tar2p); tar2p = tar2p->next)
				{
					if (streq(tar1p->id, tar2p->id)) break;
				}
				if (!tar2p) return 1;
			}
		}
		else if ((f == FLOB_TAR) || (f == FLOB_OLD))
		{
			if (!((pl[f][n]) && (p = pl[f][n]->id))) return 0;
			for (tar2p = (f == FLOB_TAR) ? thispptr->oldts : thispptr->targs; (tar2p); tar2p = tar2p->next)
			{
				if (streq(p, tar2p->id)) break;
			}
			if (!tar2p) return 1;
		}
		return 0;
	}
/* blah */
	{
		unsigned int n = 0, x = (frag->num == 7) ? 1 : 0;
		if (frag->num >= 7) for (i = 0; (i < 3); i++)
		{
			 if ((plarr[4+i]) && (((!plarr[7+i]) || strcmp(plarr[4+i]->id, plarr[7+i]->id)) ^ x)) ++n;
		}
		else for (i = 0; (i < 3); i++) if (plarr[frag->num+i]) ++n;
		if (1 != fprintf(fp, "%u", n)) return ERF_printfFailed;
		break;
	}
}
#endif
