/*****************************************************
**  plinfo.c                 >> cgi-bin/plinfo.cgi  **
**  ----------------------------------------------  **
**  ----------------------------------------------  **
**  Updated to use new time system.                 **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  he author accepts no responsibility or          **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**  ----------------------------------------------  **
*****************************************************/

#define MAX_fpbuf			65536
#define MAX_cgibuf 			4096
#define MAX_npairs 			32
#define MAX_statbuf			(1024*1024)
#define EVENT_EXT_TYPE		int			/* type of  event_p->ext is irrelevent */
#define PLAYER_EXT_TYPE		int			/* type of player_p->ext is irrelevent */
#define PSTATE_EXT_TYPE		summary_p	/* type of pstate_p->ext is summary_p */
#define PSTATE_PLP_TYPE		player_p	/* type of pstate_p->plp is player_p */
#define PEXIST_EXT_TYPE		eshort_p	/* type of pexist_p->ext is eshort_p */

typedef struct eshort* eshort_p;
typedef struct summary* summary_p;

#include "!lib-general.c"
#include "!custom.c"
/*#include "!lib-cgi.c"*/
#include "!lib-uinfo-scan.c"
#include "!lib-uinfo-misc.c"
#include "!lib-evolve.c"

/***************************************************************************************************************************
** Declarations and definitions                                                                                           **
***************************************************************************************************************************/

/* Non-fatal wrong-goings */
typedef u32b bad_things;

/* Flags for non-fatal things that can go wrong */
#define BAD_InvalidDate		(((bad_things)1)<<0)
#define BAD_DateRange		(((bad_things)1)<<1)

/* Event summary flags */
typedef u32b eshort_flags;

/* Lowest four flags must not be altered */
#define ESHORT_Fatal			(((eshort_flags)1)<<0)
#define ESHORT_VictDead			(((eshort_flags)1)<<1)
#define ESHORT_AssDead			(((eshort_flags)1)<<2)
#define ESHORT_Victim			(((eshort_flags)1)<<3)
#define ESHORT_Redeemed			(((eshort_flags)1)<<4)
#define ESHORT_GoComp			(((eshort_flags)1)<<5)
#define ESHORT_GoInco			(((eshort_flags)1)<<6)
#define ESHORT_GoWanted			(((eshort_flags)1)<<7)
#define ESHORT_NoDangerBonus	(((eshort_flags)1)<<8)
#define ESHORT_XXX				(((eshort_flags)1)<<9)

/* Structure to hold per-event summaries */
struct eshort
{
	eshort_p next;
	pstate_p other;
	timeval ewhen;
	char eref[MAX_eref];
	int thispg, otherpg;
	eshort_flags eshf;
};

/* Structure to hold context for evolution callback */
struct recev__ctx
{
	pstate_p psp;
	char* pid;
	summary_p sump;
};

/* Global scoretable event categories */
#define S_Err	(0)
#define S_Att	(1)	/* Attempts */
#define S_Kil	(2)	/* Kills */
#define S_Dth	(3)	/* Deaths */
#define S_Sur	(4)	/* Survivals */
#define S_Leg	(5)	/* Legal NP kills */
#define S_Ill	(6)	/* Illegal NP kills */
#define S_Non	(7)	/* NP attempts */
#define S_MAX	(8)

/* Scoring table category matrix */
int player_scdef[SCDEF_MAX_Single][SCDEF_MAX] =
{
	/*                 Against:  ---------Assassins--------- --------- Police ---------- --NonPlayer-- ------ ------ */
	/*                           --Unwanted--- ---Wanted---- --Unwanted--- ---Wanted---- ------------- Umpire - ?? - */
	/* -(Killer)--(NonFatal)--   -Comp- -Inco- -Comp- -Inco- -Comp- -Inco- -Comp- -Inco- -Legl- -Illg- ------ ------ */
	/* Assass Unwanted Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Assass Wanted   Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Assass Wanted   Inco */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Police Unwanted Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Police Wanted   Comp */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Police Wanted   Inco */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/*   Legal Non-Player   */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Illegal Non-Player   */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Umpire               */ { S_Att, S_Err, S_Att, S_Att, S_Att, S_Err, S_Att, S_Att, S_Non, S_Non, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* --------------(Fatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Assass Wanted   Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Assass Wanted   Inco */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Police Unwanted Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Police Wanted   Comp */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Police Wanted   Inco */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/*   Legal Non-Player   */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Illegal Non-Player   */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Umpire               */ { S_Kil, S_Err, S_Kil, S_Kil, S_Kil, S_Err, S_Kil, S_Kil, S_Leg, S_Ill, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* -(Victim)--(NonFatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Assass Wanted   Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Assass Wanted   Inco */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Unwanted Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Police Wanted   Comp */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Police Wanted   Inco */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/*   Legal Non-Player   */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Illegal Non-Player   */ { S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Err, S_Sur, S_Sur, S_Sur, S_Sur, S_Sur, S_Err },
	/* Umpire               */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* -(Victim)-----(Fatal)--  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
	/* Assass Unwanted Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Assass Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Assass Wanted   Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Assass Wanted   Inco */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Unwanted Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Unwanted Inco */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Sur, S_Err },
	/* Police Wanted   Comp */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Police Wanted   Inco */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/*   Legal Non-Player   */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Illegal Non-Player   */ { S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Err, S_Dth, S_Dth, S_Dth, S_Dth, S_Sur, S_Err },
	/* Umpire               */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* Unknown              */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err },
	/* None                 */ { S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err, S_Err }
	/* -----------------------  ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ ------ */
};

/* Per-player summary event categories */
#define P_Err		(0)
#define P_Annulled	(1)
#define P_PUB		(2)	/* Need to be priv'd to see summary columns before this */
#define P_Att		(2)	/* Attempts on other players */
#define P_Kil		(3)	/* Kills of other players */
#define P_Dth		(4)	/* Deaths at hands of other players */
#define P_Sur		(5)	/* Survived attempts by other players */
#define P_MAX		(6)

/* Table of column title for per-player summary */
char* summarytit[P_MAX] =
{
	"<b>Erroneous events</b>",
	"<b>Annulled events</b>",
	"Attempts on",
	"Kills of",
	"Killed by",
	"Survived attempts by"
};

/* Table to translate from scoretable categories into per-player summary categories */
int summarytab[S_MAX] = 
{
	P_Err, P_Sur, P_Dth, P_Kil, P_Att, P_Dth, P_Dth, P_Sur
};

/* Structure to hold per-player summaries */
struct summary
{
	pstate_p pstp;
	summary_p next;
	int tab[P_MAX];
};

static char statbuf[MAX_statbuf], *statbuf_end = statbuf + MAX_statbuf, *statbuf_pos;
static player_p plist = 0;

char* excl_list[4] =
{
	"id", "authname", "authpass", 0
};

int send_doc(int w, player_p authp);

/***************************************************************************************************************************
** Event and termination hooks for evolution function                                                                     **
***************************************************************************************************************************/
int recev(event_p ep, pstate_p ps1p, pstate_p ps2p, int pg1, int pg2, void* context, evolve__ctx_t* ectx)
{
	struct recev__ctx* ctx = context;
	pstate_p otherpsp;
	eshort_p eshp;


	/* No player state struct noted yet */
	if (!ctx->psp)
	{
		/* Match on first player id? */
		if ((ps1p) && (streq(ctx->pid, ps1p->pid))) ctx->psp = ps1p;

		/* Second player id? */
		else if ((ps2p) && (streq(ctx->pid, ps2p->pid))) ctx->psp = ps2p;
	}

	/* Does event involve selected player? */
	if ((!ctx->psp) || ((otherpsp = ps2p, (ps1p != ctx->psp)) && (otherpsp = ps1p, ((ps2p != ctx->psp))))) return 0;

	/* Allocate event summary struct (don't need to test (ctx->psp->exist) here, I think)*/
	if (!(eshp = malloc(sizeof(struct eshort)))) return -(ERF_mallocFailed);

	/*fprintf(stderr, "*(%s)*", ep->ref);*/

	/* Link into the list of events for this existence */
	eshp->next = ctx->psp->exist->ext;
	ctx->psp->exist->ext = eshp;

	/* Store link to identity of other player */
	eshp->other = otherpsp;

	/* Event was fatal? */
	eshp->eshf = (ep->flags & EF_Kill) ? ESHORT_Fatal : 0;

	/* Nature of participants */
	eshp->thispg  = (ps1p == ctx->psp) ? pg1 : pg2;
	eshp->otherpg = (ps1p == ctx->psp) ? pg2 : pg1;

	/* Store time of event */
	eshp->ewhen = ep->when;

	/* No dangerousness bonus to event */
	if (ep->flags & EF_NoDangerBonus) eshp->eshf |= ESHORT_NoDangerBonus;

	/* More event flags if selected player was assassin */
	if (ps1p == ctx->psp)
	{
		/* Went wanted? */
		if ((ps1p->exist->fl & PF_Wanted) && !(ps1p->pflags & PF_Wanted)) eshp->eshf |= ESHORT_GoWanted;

		/* Went unwanted? */
		if (!(ps1p->exist->fl & PF_Wanted) && (ps1p->pflags & PF_Wanted)) eshp->eshf |= ESHORT_Redeemed;

		/* Went incompetent? */
		if ((ps1p->exist->fl & PF_Incompetent) && !(ps1p->pflags & PF_Incompetent)) eshp->eshf |= ESHORT_GoInco;

		/* Went competent? */
		if (!(ps1p->exist->fl & PF_Incompetent) && (ps1p->pflags & PF_Incompetent)) eshp->eshf |= ESHORT_GoComp;
	}

	/* Selected player was victim */
	else eshp->eshf |= ESHORT_Victim;

	/* Were participants already dead? */
	if (ps1p->pflags & PF_Dead) eshp->eshf |= ESHORT_AssDead;
	if (ps2p->pflags & PF_Dead) eshp->eshf |= ESHORT_VictDead;

	/*fprintf(stderr, "%li   vpid=%s pvfl=%lx vfl=%lx\n", uniqid, p2sp->pid, p2sp->pflags, p2sp->exist->fl);*/

	/* Copy event ref */
	strncpy(eshp->eref, ep->ref, MAX_eref-1);
	eshp->eref[MAX_eref-1] = '\0';

	/* Other player is known... add event info to per-player summary */
	if (otherpsp)
	{
		summary_p sump;
		int i;

		/* No existing summary struct? */
		if (!(sump = otherpsp->ext))
		{
			/* Allocate new summary struct */
			if (!(sump = malloc(sizeof(struct summary)))) return -(ERF_mallocFailed);

			/* Link into chain */
			sump->next = ctx->sump;
			ctx->sump = otherpsp->ext = sump;

			/* Note which player summary applies to */
			sump->pstp = otherpsp;

			/* Initialise table */
			for (i = 0; (i < P_MAX); i++) sump->tab[i] = 0;
		}

		/* Take account of fatality and victimhood */
		i = ((eshp->eshf & ESHORT_Fatal) ? (SCDEF_Fatal) : 0) + ((eshp->eshf & ESHORT_Victim) ? 0 : (SCDEF_Victim));

		/* Update the count of annulled events */
		if (ep->flags & EF_Annulled) ++sump->tab[P_Annulled];

		/* Update the summary */
		else ++sump->tab[summarytab[player_scdef[eshp->otherpg+i][eshp->thispg]]];
	}

	return 0;
}

int thook(event_p ep, int match, void* context, evolve__ctx_t* ectx)
{
    /* Don't call hook more than once */
    ectx->flags &= ~(EVOLVE_FuncEndCond|EVOLVE_FuncExhaust);

	/* Not called becase event matched? */
    if ((match != 2) && (match != 3)) return 0;

	/* Can authenticated player know about ref'd event? Assume not admin */
    if ((ep) && !(ep->flags & (EF_Ignore|EF_Invisible)) && (ep->flags & EF_Open)) return (match);

    /* Else pretend not to have found event */
    ectx->eptr = 0;
    ectx->endref = 0;

    /* Re-check termination conditions */
    return 0;
}

/***************************************************************************************************************************
** Output page content                                                                                                    **
***************************************************************************************************************************/
int printplinfo(player_p pptr, pstate_p pstp, player_p authp, summary_p sump,
				timeval endtime, char* fref, event_p feptr, int cond)
{
	player_p plp;
	ppseud_p pseudp;
	pexist_p exp;
	int i, admin;
	char* p;
	FILE *fp;

	/* Privs */
	admin = ((authp->flags & PF_Admin) ? 2 : 0) | ((authp->flags & PF_Observer) ? 1 : 0);

	/*fprintf(stderr, "%li:   wibble it printplinfo\n", uniqid);*/

	printf(LF "<HTML><HEAD><TITLE>" GAME " Statistics</TITLE></HEAD>"
		   LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
		   LF "<h1><hr>" GAME " - Player Statistics<hr></h1>");

	/* Information immediately before/after event (which was found) */
	if (feptr) printf(LF "<h3>Information taken immediately %s <a href=\""CGI_URL"event.cgi?ev_ref=%s\">event"
					  LF "&quot;%s&quot;</a> at "TDISP" on %s.</h3>", (cond & 4) ? "prior to" : "after",
					  feptr->ref, feptr->ref, Thm(endtime), fdayname(T_d(endtime), 0));
	else
	{
		/* Looking for a specific event, but not found? */
		if (fref) printf(LF "<h4><img src=\""CROSS_URL"\" align=middle "CROSS_SIZ" hspace=8>"
						 LF "<font color=#FFCCCC>Event &quot;%s&quot; was not be located within the time limit."
						 LF "</font></h4><hr>", fref);

		/* Information at end of game time */
		if (cond & 1) printf(LF "<h3>Information taken at end of all game time.</h3>");

		/* Information at a specific time */
		else printf(LF "<h3>Information taken %s all events at "TDISP" on %s.</h3>",
					(cond & 2) ? "before" : "after", Thm(endtime), fdayname(T_d(endtime), 0));
	}

	/* Open file, copy it over */
	if (fp = fopen(FILE_PlinfoGuide, "rb"))
	{
		i = ftofcpy(fp, stdout, -1, fpbuf, MAX_fpbuf, -1, -1);"";
		fclose(fp);
		if (i) printf(LF "<p><b><font color=#FFCCCC>Could not copy HTML file:</font> error code %i.</b></p>", i);
	}

	printf(LF "<form action=\"" CGI_URL "plinfo.cgi\" method=\"POST\">"
		   LF "<p><table rows=5 cols=2 border=0 cellpadding=0 cellspacing=2>"
		   LF "<tr><td align=right>Currently authenticated as:&nbsp;</td><td><b>%s</b>%s</td></tr>"
		   LF "<tr><td align=right>Change authentication to:&nbsp;</td><td><select size=1 name=\"authname\">",
		   authp->name, ((admin & 2) ? " (admin)" : ""));
	(void) makeidmenu(authp->id, PF_Login, PF_Login, i = (authp->flags & PF_Observer) ? 0 : 1);
	puts("</select></td></tr><tr><td align=right>Password:&nbsp;</td>"
		 LF "<td><input name=\"authpass\" type=\"password\" size=24 maxlength=64></td></tr>"
		 LF "<tr><td align=right>New player to query:&nbsp;</td><td><select size=1 name=\"id\">");
	(void) makeidmenu(pptr->id, ((authp) && (authp->flags & PF_Admin)) ? 0 : PF_Hidden, 0, i);
	puts("</select></td></tr>" LF "<tr><td></td><td>"
		 LF "<input value=\"(Re)authenticate and perform new query\" type=\"submit\" name=\"reauth\"></td></tr>"
		 LF "</table></p></form>");

	i = 0; p = "<p>Skip down to: ";
	if (sump) printf("%s<a href=\"#act\">Activity summary</a>", (i++) ? " | " : p);
	if ((pstp) && (pstp->exist)) printf("%s<a href=\"#hist\">Chronological history</a>", (i++) ? " | " : p);
	if (i) puts("</p>");

	puts(LF "<hr><h3>Player information for ");
	if (admin > 0) puts(pptr->name);
	else if (pptr->pseud) encapsulate(pptr->pseud, stdout);
	else printf("[%s]", pptr->id);
	puts(":</h3><p><table cols=4 cellspacing=4 cellpadding=4 border=0>");
	
	/* Static player information (from uinfo.txt file) */
	if (admin > 0)
	{
		printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">College:</font></td>"
			   LF "<td bgcolor="RED"><font color="WHITE">%s</font></td>"
			   LF "<td bgcolor="BLUE" align=right><font color="WHITE">Email address:</font></td>"
			   LF "<td bgcolor="RED"><font color="WHITE">", pptr->coll);
		printf(((pptr->flags & PF_Mailable) ? "<a href=\"mailto:%s\">%s</a>" : "%s<!--%s-->"),
			   pptr->email, pptr->email);
		printf("</font></td></tr>"
			   LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Address:</font></td>"
			   LF "<td bgcolor="RED" colspan=3><font color="WHITE">%s</font></td></tr>"
			   LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Comments:</font></td>"
			   LF "<td bgcolor="RED" colspan=3><font color="WHITE">%s</font></td></tr>",
			   pptr->addr, pptr->mafia);
	}

	/* Player information from evolution */
	if (pstp)
	{
		if (admin > 1)
		{
			target_p tarp;
			player_p qptr;
			int j;

			printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Assassins:</font></td>"
				   LF "<td bgcolor="RED"><font color="WHITE">");
	
			for (tarp = pstp->asses, j = 0; (tarp); tarp = tarp->next, j++)
			{
				if ((tarp->id) && (strcmp(tarp->id, NON_ID)))
				{
					for (qptr = plist; (qptr); qptr = qptr->link) if (!strcmp(qptr->id, tarp->id)) break;
					if (!qptr) printf("%s<b>[Unknown:%s]</b>", (j) ? "<br>" : "", tarp->id);
					else printf("%s<a href=\"plinfo.cgi?id=%s\">%s</a>", (j) ? "<br>" : "", qptr->id, qptr->name);
				}
				else puts("<b>Unallocated</b>");
			}
			if (!j) puts("<b>No assassins!</b>");
	
			printf("</font></td><td bgcolor="BLUE" align=right><font color="WHITE">"
				   "Targets:</font></td><td bgcolor="RED"><font color="WHITE">");

			for (tarp = pstp->targs, j = 0; (tarp); tarp = tarp->next, j++)
			{
				if ((tarp->id) && (strcmp(tarp->id, NON_ID)))
				{
					for (qptr = plist; (qptr); qptr = qptr->link) if (!strcmp(qptr->id, tarp->id)) break;
					if (!qptr) printf("%s<b>[Unknown:%s]</b>", (j) ? "<br>" : "", tarp->id);
					else printf("%s<a href=\"plinfo.cgi?id=%s\">%s</a>", (j) ? "<br>" : "", qptr->id, qptr->name);
				}
				else puts("<b>Unallocated</b>");
			}
			if (!j) puts("<b>No targets!</b>");
	
			printf("</font></td></tr>");
		}

		if (pstp->exist)
		{
			unsigned long ff = (pptr->flags & (~PF_MASK_Scoring)) | (pstp->exist->fl & PF_MASK_Scoring);

			printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Status:</font></td>"
				   LF "<td bgcolor="RED" colspan=3><font color="WHITE">");
			printpstatus(ff, pstp->exist->timer, 0);
			puts("</font></td></tr>");

	#if OPT_ShowScore

			i = 1;
			printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Score:</font></td>"
				   LF "<td bgcolor="RED"><font color="WHITE">%li points</font></td>",
				   ((pstp->score + SCORE_ROUND) >> SCORE_SHIFT));
	#else
			i = 3;
	#endif
			printf(LF "<td bgcolor="BLUE" align=right><font color="WHITE">Player kills:</font></td>"
				   LF "<td bgcolor="RED" colspan=%i><font color="WHITE">%d (%d failed attempts)</font></td></tr>",
				   i, pstp->scoretab[S_Kil], pstp->scoretab[S_Att]);

	#if OPT_ShowScore

			printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Dangerousness:</font></td>"
				   LF "<td bgcolor="RED"><font color="WHITE"><b>%0.2lf</b></font></td>",
				   (pstp->danger / ((double)(1L<<DANGER_SHIFT))));
	#endif
			printf(LF "<td bgcolor="BLUE" align=right><font color="WHITE">Total deaths:</font></td>"
				   LF "<td bgcolor="RED" colspan=%i><font color="WHITE">%d (survived %d attempts)</font></td></tr>",
				   i, pstp->scoretab[S_Dth], pstp->scoretab[S_Sur]);

	#if OPT_ShowScore

			printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Worth:</font></td>"
				   LF "<td bgcolor="RED"><font color="WHITE">%li points</font></td>",
				   ((((long)((dangerbonus(pstp->danger) * pstp->score) / 6.0)) + SCORE_ROUND) >> SCORE_SHIFT));
	#endif
			printf(LF "<td bgcolor="BLUE" align=right><font color="WHITE">Non-player kills:</font></td>"
				   LF "<td bgcolor="RED" colspan=%i><font color="WHITE">%d legal, %d illegal, %d failed</font></td></tr>",
				   i, pstp->scoretab[S_Leg], pstp->scoretab[S_Ill], pstp->scoretab[S_Non]);
	
			if (admin > 1)
			{
				printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Last active at:</font></td>"
					   LF "<td bgcolor="RED"><font color="WHITE">"TDISP" %s</font></td>"
					   LF "<td bgcolor="BLUE" align=right><font color="WHITE">Rank:</font></td>"
					   LF "<td bgcolor="RED"><font color="WHITE">",
					   Thm(pstp->exist->tlact), fdayname(T_d(pstp->exist->tlact), 0));

				if (!(pstp->exist->fl & PF_Police)) printf("Not on police force.</font></td></tr>");

				else printf("<b>(%li)</b>&nbsp;%s</font></td></tr>",
							pstp->exist->rank, rank_name[LIM_RANK(pstp->exist->rank)]);
			}

			if (pstp->exist->comment)
			{
				printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Case history:</font></td>"
					   LF "<td bgcolor="RED" colspan=3><font color="WHITE">%s</font></td></tr>", pstp->exist->comment);
			}
	
			if (pstp->exist->dps)
			{
				printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Pseudonym:</font></td>"
					   LF "<td bgcolor="RED" colspan=3><font color="WHITE">");
				encapsulate(pstp->exist->dps->text, stdout);
				printf("</font></td></tr>");
			}
		
			for (pseudp = pstp->allpseuds, i = 0; (pseudp); pseudp = pseudp->link_pl)
			{
	            if (pseudp == pstp->exist->dps) continue;
				if (i++) printf(LF "<tr><td></td><td bgcolor="RED" colspan=3><font color="WHITE">");
				else printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Previous pseudonyms:</font></td>"
							LF "<td bgcolor="RED" colspan=3><font color="WHITE">");
	            encapsulate(pseudp->text, stdout);
				printf("</font></td></tr>");
			}
	
			if ((!i) && (!pstp->exist->dps))
			{
				printf(LF "<tr><td bgcolor="BLUE" align=right><font color="WHITE">Pseudonym:</font></td>"
					   LF "<td bgcolor="RED" colspan=3><font color="WHITE"><b>(none yet)</b></font></td></tr>");
			}
		}
	}

	puts(LF "</table></p>");

	/* Player activity summary: table of players killed or killed by */
	if (sump)
	{
		player_p oplp;
		int i, j = 0;
		char *cc[2] = { BLUE, RED };

		/* Table header */
		puts(LF "<hr><h3><a name=\"act\">Player activity summary for ");

		if (admin > 0) puts(pptr->name);
		else if (pptr->pseud) encapsulate(pptr->pseud, stdout);
		else printf("[%s]", pptr->id);

		printf("</a></h3>" LF "<p><table cols=%d cellpadding=4 cellspacing=4 border=0 width=\"100%%\">"
			   LF "<thead><td bgcolor=#000000>As of "TDISP" %s:",
			   P_MAX + 1 - ((admin & 2) ? 0 : P_PUB), Thm(endtime), fdayname(T_d(endtime), 0));

		/* Print out each column title */
		for (i = (admin & 2) ? 0 : P_PUB; (i < P_MAX); i++) printf("</td><td bgcolor="GREEN">%s", summarytit[i]);
		
		puts("</td></thead>");

		/* Print out each row of summary */
		while (sump)
		{
			/* Alternate colours */
			j = 1 - j;

			/* Begin new row */
			printf(LF "<tr bgcolor=%s><td bgcolor="GREEN">", cc[j]);

			/* Print out player name */
			if (oplp = sump->pstp->plp)
			{
				p = (oplp->name) ? (oplp->name) : (oplp->id);
				i = (!strcmp(oplp->id, LEG_ID) || !strcmp(oplp->id, ILL_ID));
				if (!i) printf("<a href=\"plinfo.cgi?id=%s\">", oplp->id);
				if (admin > 0) puts(p);
				else if (oplp->pseud) encapsulate(oplp->pseud, stdout);
				else printf("[%s]", oplp->id);
				if (!i) puts("</a>");
			}
			else printf("<b>unknown player</b>");

			/* Print out each column */
			for (i = (admin & 2) ? 0 : P_PUB; (i < P_MAX); i++)
			{
				if (sump->tab[i]) printf("</td><td><b>%d</b>", sump->tab[i]);
				else fputs("</td><td>&mdash;", stdout);
			}
			puts("</td></tr>");

			/* Next per-player summary */
			sump = sump->next;
		}

		/* Table footer */
		puts(LF "</table></p>");
	}

	/*printf("<hr><h3><a name=\"targ\">Target history for %s</a></h3>"
		   "<p><b>(not implemented)</b></p>", pptr->name);*/

	/* Player history: show chronological details of events in which player was involved */
	if ((pstp) && (pstp->exist))
	{
		char *q;

		puts("<hr><h3><a name=\"hist\">Player history for");

		if (admin > 0) puts(pptr->name);
		else if (pptr->pseud) encapsulate(pptr->pseud, stdout);
		else printf("[%s]", pptr->id);

		puts("</a></h3><p>");

		for (exp = pstp->exist; (exp); exp = exp->link_pl)
		{
			p = !(exp->fl & PF_Wanted) ? "" : (exp->fl & PF_Incompetent) ? " incompetent" : " wanted";

			if (exp->fl & PF_Dead)
			{
				printf("<b>Died%s at "TDISP" %s.</b><br>",
					   p, Thm(exp->died), fdayname(T_d(exp->died), "%d %b"));
			}
			else if (exp == pstp->exist)
			{
				printf("<b>Currently alive (at "TDISP" on %s)%s%s.</b><br>",
					   Thm(endtime), fdayname(T_d(endtime), 0), (*p) ? (" and ") : (""), p);
			}

			if (exp->ext)
			{
				eshort_p eshp;
				player_p oplp;
				char *otherpname;

				puts("<ul>");

				/* Scan through list of relevent events for this existence */
				for (eshp = exp->ext; (eshp); eshp = eshp->next)
				{
					/* Short struct for other player */
					oplp = (eshp->other) ? (eshp->other->plp) : 0;

					/* Event title (XXX --- relies on these being bottom four bits, correctly ordered) */
					p = titletab[eshp->eshf & (ESHORT_Victim|ESHORT_AssDead|ESHORT_VictDead|ESHORT_Fatal)];

					printf("<li><a href=\"" NEWS_URL "news%s.html#%s\">[%s "TDISP"]</a>" LF
						   "<a href=\""CGI_URL"event.cgi?ev_ref=%s\">[Report]</a> %s " LF,
						   perid(eshp->ewhen), eshp->eref, fdayname(T_d(eshp->ewhen), "%d %b"),
						   Thm(eshp->ewhen), eshp->eref, p);

					if (oplp)
					{
						otherpname = (oplp->name) ? (oplp->name) : (oplp->id);
						i = (!strcmp(oplp->id, LEG_ID) || !strcmp(oplp->id, ILL_ID));
						if (!i) printf("<a href=\"plinfo.cgi?id=%s\">", oplp->id);
						if (admin > 0) puts(otherpname);
						else if (oplp->pseud) encapsulate(oplp->pseud, stdout);
						else printf("[%s]", oplp->id);
						if (!i) puts("</a>");
					}
					else printf("<b>unknown player</b>");

					if (eshp->eshf & ESHORT_NoDangerBonus) puts(" (dangerousness disabled)");

					puts(".</li>");
				}

				puts("</ul><br>");
			}

			q = (exp->fl & PF_Redeemed) ? "Redeemed" : "Incarnated ";

			p = (exp->fl & PF_Police) ? "policeman" : "assassin";

			printf("<b>%s as %s at "TDISP" %s.</b><br>", q, p, Thm(exp->born), fdayname(T_d(exp->born), "%d %b"));
		}

		puts("</p>");
	}

	printf(LF "<hr><form action=\"" CGI_URL "pled.cgi\" method=\"POST\">"
		   LF "<input type=\"hidden\" name=\"authname\" value=\"%s\">"
		   LF "<input type=\"hidden\" name=\"pl_1id\" value=\"%s\"><p>",
		   authp->id, pptr->id);

	if (authp->flags & PF_Admin)
	{
		puts("<input value=\"Act on player\" type=\"submit\" name=\"sub_act\">"
			 "&nbsp;&nbsp;to:&nbsp;&nbsp;<select name=\"act\" size=1>");
		for (i = 0; (i < MAX_act); i++) printf("<option value=\"%d\">%s</option>\n", i+1, acttab[i]);
		puts("</select>");
	}

	else
	{
		#if (OPT_ChangePseud)
		puts("<input value=\"Add/change players' pseudonym\" type=\"submit\" name=\"sub_pseud\">");
		#endif
		puts("<input value=\"Change players' password\" type=\"submit\" name=\"sub_passwd\">");
	}

	puts("</p></form>");

	return 0;
}

/***************************************************************************************************************************
** Main code. Doesn't make use of command line arguments.                                                                 **
***************************************************************************************************************************/
int main(int argc, char* argv[])
{
	char *p, *thetimestr,  *ipids[3] = { 0, 0, 0 }, *playerid = "", *authname = "", *authpass = "", *fref = 0;
	int i, rc = 0, lockfd = 0;
	pstate_p pslist, psp, *pspp;
	player_p plarr[3] = { 0, 0, 0 }, *plidx = plarr, authp = 0;
	event_p feptr = 0;
	evolve_flags ffl = 0;
	timeval rtime = 0, ftime = 0;
	bad_things badstuff = 0;
	u32b u;


	/* Output HTTP headers */
    printf("Content-type: text/html" LF LF);

	/* Initialse stuff: time, file creation umask, pid, make note in log and redirect stderr there */
	if ((rc = geninit(&thetimestr, 0117, 1, "Plinfo")) < 0) exit(send_doc(-rc, 0));

	/* Logged error handler now established */
	else if (rc) erf(rc, 1);

	/* Parse CGI input */
	if ((i = parsecgi()) > 0) erf(i, 1);

	/* No input given */
	/*else if (i == 0) erf(ERF_Void, 6);*/

	/* Check security */
	/*if (strcmp(findval("security", 1), SECURITY)) exit(insecure(1, "plinfo"));*/

	/* If no authentication name given, assume public user */
	if (!(authname = findval("authname", 0))) authname = PUB_ID;

	/* Authentication password */
	else authpass = findval("authpass", 1);

	/* Player id for query */ /* WAS: if (!(*playerid))) erf(ERF_NoIdGiven, 2); */
	playerid = findval("id", 1);

	/* Claim lockfile */
	if (!(lockfd = setflock("plInfo.c"))) erf(ERF_flockFailed, 3);

	/* Array of 'interesting' players - may wish to include 'accessing player' at some point */
	ipids[0] = playerid;
	ipids[1] = authname;
	ipids[2] = 0;
	
	/* Read full info for all players */
	if (rc = scanplinfo(SCANPL_AddFull|SCANPL_GetTargs, ipids, &plidx,
						0, &plist, statbuf, statbuf_end, &statbuf_pos)) erf(rc, 4);

	/* Invalid authentication? */
	if (!(authp = plidx[1]) || !authp->name || !(authp->flags & PF_Login)
											|| (!(authp->flags & PF_Public) && strcmp(authp->pass, authpass)))
	{
		authp = 0;
		erf(ERF_BadAuth, 5);
	}

	/* Present player info around a given event? */
	if ((fref = findval("fref", 0)) && !(authp->flags & PF_Admin)) ffl |= EVOLVE_FuncEndCond;
	
	/* Immediately prior to, or after, given event? Time limit is exclusive? */
	if (findval("prior", 0)) ffl |= (fref) ? (EVOLVE_Prior) : (EVOLVE_Exclusive);

	/* Read specified date and time */
	if (!(p = findval("fday", 0)) || !(ftime = strings2tim(p, findval("fhour", 0), findval("fmin", 0))))
	{
		/* Dodgy date */
		if (p) badstuff |= BAD_InvalidDate;

		/* if priv'd and searching for event, don't impose any time limit, otherwise, use current time */
		if ((authp->flags & PF_Admin) && (!fref)) ftime = thetime;
	}

	/* Sanity checks on date and time for non-priv'd players */
	if (!(authp->flags & PF_Admin))
	{
		/* Use current time if none given */
		if (!ftime) ftime = thetime;

		/* Otherwise, assume time invalid */
		else badstuff |= BAD_DateRange;

		/* Upper limits */
		if (ftime > TIME_plinfoMax) ftime = TIME_plinfoMax;

		/* Lower limits */
		else if (ftime < TIME_plinfoMin)
		{
			ftime = TIME_plinfoMin;
			ffl |= EVOLVE_Exclusive;
		}

		/* Time was valid, after all */
		else badstuff &= (~BAD_DateRange);
	}

	/*fprintf(stderr, "%li:   wibble the first: bad=%lu\n", uniqid, badstuff);*/

	/* Invalid or missing player */
	if ((!plidx[0]) || (!plidx[0]->name)) rc = ERF_UnknownPlayer;

	/* Specified player should not be hidden */
	else if ((!authp->flags & PF_Admin) && (plidx[0]->flags & PF_Hidden)) rc = ERF_NoPermission;

	/* Output player's stats */
	else
	{
		/* Context structure for event summarizer */
		struct recev__ctx rectx;

		/* Initialise context */
		rectx.psp = 0;
		rectx.sump = 0;
		rectx.pid = playerid;

		/* Only interested in selected player now */
		ipids[1] = 0;

		/* Scan through event list */
		if (rc = evolve(EVOLVE_TimersToLimit|EVOLVE_Comments|EVOLVE_AllExists|EVOLVE_AllPseuds|EVLadd_NoPseuds|ffl|
						EVOLVE_FuncAssass|EVOLVE_MatchAll|EVOLVE_InheritPseud|EVOLVE_ScorePlayer|EVOLVE_Targetting,
						ipids, &pspp, &pslist, 0, player_scdef, 0, 0, fref, &feptr, ftime, &rtime,
						statbuf_pos, statbuf_end, &statbuf_pos, recev, &rectx, plist, thook, 0, 0)) erf(rc, 8);

		/* Check for unknown players */
		for (psp = pslist; (psp); psp = psp->next) if (!psp->plp) erf(ERF_UnknownPlayer, 100);

		/* Some flags detailing the end condition */
		i = ((ftime) ? 0 : 1) | ((ffl & EVOLVE_Exclusive) ? 2 : 0) | ((ffl & EVOLVE_Prior) ? 4 : 0);

		/* Disgourge results */
		if (rc = printplinfo(plidx[0], pspp[0], authp, rectx.sump, rtime, fref, feptr, i)) erf(rc, 9);

		/* Finished */
		printf("<p align=right>Produced at %s.<!--(rc:%d)--></p></BODY></HTML>" LF, thetimestr, rc);
	}


	fprintf(stderr, "%li:   wibble the fourth\n", uniqid);

_erf: /***** Error handling *****/

	/* Debug */
	if (rc) fprintf(stderr, "%li:   main(%i):%i:%s authname=\"%s\" authpass=\"%s\" %sid=\"%s\"\n", uniqid,
					ERF_CONTEXT(rc), ERF_TYPE(rc), ERF_STR(rc), authname, authpass, ((authp) ? "good " : ""), playerid);

	/* Tidy up */
	if (lockfd) setfunlock(lockfd);

	/* Only retain low bits of normal rc's */
	if (rc) send_doc(ERF_LOWBITS(rc), authp);

	return 0;
}

/***************************************************************************************************************************
** Produce and output HTML document data for the various situations indicated by (w).                                     **
***************************************************************************************************************************/
int send_doc(int w, player_p authp)
{
	int rc = 0;
	int i = ((authp) && (authp->flags & PF_Observer)) ? 0 : OPT_IDMenuMode;
	char* p;

	switch (w)
	{
		case ERF_BadAuth:
		case ERF_NoIdGiven:
		case ERF_Void:
		case ERF_UnknownPlayer:
		case ERF_NoPermission:
		{
			puts(LF "<HTML><HEAD><TITLE>Mayweek 2001 Assassins' Game: Player information</TITLE></HEAD>"
				 LF "<BODY bgcolor=#000000 text=#FFFFDD link=#BBBBFF vlink=#FFCCFF alink=#FFFFFF>"
				 LF "<h1>Detailed Player Statistics</h1>"
				 LF "<form action=\"" CGI_URL "plinfo.cgi\" method=\"POST\">");

			if (w != ERF_Void) puts("<h3><img align=left src=\"" CROSS_URL "\" border=0 width=34 height=34 hspace=8>");
			if (w == ERF_BadAuth) puts("<font color=#FFCCCC>Authentication failed!</font></h3><p>");
			if (w == ERF_NoIdGiven) puts("<font color=#FFCCCC>No player id given for query.</font></h3>");
			if (w == ERF_UnknownPlayer) puts("<font color=#FFCCCC>Unknown player.</font></h3>");
			if (w == ERF_NoPermission) puts("<font color=#FFCCCC>Can't query that player.</font></h3>");

			puts(LF "<p><table rows=4 cols=2 border=0 cellpadding=0 cellspacing=4>"
				 LF "<tr><td align=right>Authenticate as:&nbsp;</td><td><select size=1 name=\"authname\">");

			if (rc = makeidmenu((authp) ? authp->id : PUB_ID, PF_Login, PF_Login, i)) erf(rc, 1);

			puts("</select></td></tr>" LF "<tr><td align=right>Password:&nbsp;</td>"
				 LF "<td><input name=\"authpass\" type=\"password\" size=24 maxlength=64></td></tr>");

			puts(LF "<tr><td align=right>Player to query:&nbsp;</td>");

			if ((!OPT_NoListPlayers) || ((authp) && (authp->flags & PF_Admin)))
			{
				puts(LF "<td><select size=1 name=\"id\">");
				rc = makeidmenu(findval("id", 1), ((authp) && (authp->flags & PF_Admin)) ? 0 : PF_Hidden, 0, i);
				if (rc) erf(rc, 1); 
				puts("</select></td></tr>");
			}
			else
			{
				puts(LF "<td><input name=\"id\" type=\"text\" value=\"");
				encapsulate(findval("id", 1), stdout);
				puts(LF "\" size=10 maxlength=10></td></tr>");
			}

			regurgitate(excl_list);

			puts(LF "<tr><td></td><td><input value=\"Submit\" type=\"submit\"></td></tr>"
				 LF "</table></p></form></BODY></HTML>");

			break;
		}

		default: error_doc(w);
	}

	return 0;


_erf: /***** Error handling *****/

	fprintf(stderr, "%li:   send_doc(%i):%i\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc));
	return 0;
}

