/*****************************************************
**  lib-cgi.c                            (library)  **
**  ----------------------------------------------  **
**  Decode CGI input (either on stdin or environ).  **
**  ----------------------------------------------  **
**  Converted to new time system?                   **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef NO_CGI
 #error "** lib-cgi #included when NO_CGI specified **"
#endif

#ifdef included__lib_cgi
 #error "** lib-cgi multiply #included **"
#else
 #define included__lib_cgi fromfile

/***************************************************************************************************************************
** Definitions and declarations                                                                                           **
***************************************************************************************************************************/

/* Global variables holding decoded cgi data */
int npairs = 0;
char* name[MAX_npairs], *val[MAX_npairs];
char* empty = "[[empty]]";

/* Convenient macro-interfaces for acquiring decoded data */
#define findval(x,y) findnameval((x),(y)&(~16),0)
#define findsuf(x,y) findnameval((x),(y)|(16),0)

/***************************************************************************************************************************
** Read the URL-encoded query string at (*fromptr), and update (*fromptr) accordingly for next invocation of function.    **
** Copy the data into buffer delimited by (*toptr) and (toend), and update (*toptr) to point to remaining free space.     **
** Place pointers to name/value parts of "name=value" pair in (*name) and (*val) [which may point to global (empty) var]  **
** Return (-1) if buffer exhausted, (1) if sucessful but no more pairs to read, (0) if sucessful and more pairs remain.   **
***************************************************************************************************************************/
int gextract(char** fromptr, char** toptr, char* toend, char** name, char** val)
{
	char c, pc = 0;
	char* to = *toptr;
	char* from = *fromptr; 

	/* No value yet found */
	*val = 0;

	/* Extract the entire encoded name=val pair and convert plusses to spaces */
	for (*name = to; (to < toend) && (c = *from) && (c != '&'); from++) *(to++) = (c == '+' ? ' ' : c);

	/* Return starting location of next pair */
	*fromptr = from + 1;

	/* Problem: out of buffer */
	if (to >= toend) return -1;

	/* Zero terminate the extracted string for convenience */
	*(to) = '\0';

	/* Expand escaped %xx characters and extract value string */
	for (from = to = *name; (*from); from++, to++)
	{
		if ((*from == '%') && (from[1]) && (from[2]))
		{
			/* Decode character sequence */
			c = ((from[1] > '9') ? ((from[1] & 0xDF) - 'A' + 10) : (from[1] - '0')) * 16
			  + ((from[2] > '9') ? ((from[2] & 0xDF) - 'A' + 10) : (from[2] - '0'));

			/* Skip the escape sequence */
			from += 2;
		}
		else c = *from;

		/* Turn CRLF or LFCR sequences into LF */
		if (((pc==13)&&(c==10)) || ((pc==10)&&(c==13))) { pc = 0; *(--to) = 10; }

		/* Disallow ctrl characters other than nl, lf */
		else *to = pc = ((c>31) || (c==10) || (c==13)) ? c : '?';

		/* Is this the beginning of the value data? */
		if ((*to == '=') && (!(*val)))
		{
			*val = to + 1;
			*to = '\0';
		}
	}

	/* Ensure termination of processed string */
	*to = '\0';

	/* Return exact extent of used buffer */
	*toptr = to + 1;

	/* Set 'value' content to point at the global variable 'empty' when necessary. */
	if (!(*val) || (**val == '\0')) *val = empty;

	/* More pairs to read? */
	return ((*fromptr)[-1]) ? 0 : 1;
}

/***************************************************************************************************************************
** Read up to (length) characters from the URL-encoded data stream on stdin, and leave things suitable for next call.     **
** Copy the data into buffer delimited by (*toptr) and (toend), and update (*toptr) to point to remaining free space.     **
** Place pointers to name/value parts of "name=value" pair in (*name) and (*val) [which may point to global (empty) var]  **
** Return (-1) if buffer exhausted, (1) if sucessful but no more pairs to read, (0) if sucessful and more pairs remain.   **
***************************************************************************************************************************/
int pextract(int* length, char** toptr, char* toend, char** name, char** val)
{
	int c, pc = 0;
	char* from, *to = *toptr;

	/* No value yet found */
	*val = 0; *name = to;

	/* Extract the entire encoded name=val pair and convert plusses to spaces */
	while ((to < toend) && ((*length)--) && ((c = getchar()) != EOF) && (c != '&'))
	{
		*(to++) = (c == '+' ? ' ' : c);
	}

	/* Problem: out of buffer */
	if (to >= toend) return -1;

	/* Zero terminate the extracted string for convenience */
	*(to) = '\0';

	/* Expand escaped %xx characters and extract value string */
	for (from = to = *name; (*from); from++, to++)
	{
		if ((*from == '%') && (from[1]) && (from[2]))
		{
			/* Decode character sequence */
			c = ((from[1] > '9') ? ((from[1] & 0xDF) - 'A' + 10) : (from[1] - '0')) * 16
			  + ((from[2] > '9') ? ((from[2] & 0xDF) - 'A' + 10) : (from[2] - '0'));

			/* Skip the escape sequence */
			from += 2;
		}
		else c = (*from);

		/* Turn CRLF or LFCR sequences into LF */
		if (((pc == 13) && (c == 10)) || ((pc == 10) && (c == 13))) { pc = 0; *(--to) = 10; }

		/* Disallow ctrl characters other than nl, lf */
		else *to = pc = ((c > 31) || (c == 10) || (c == 13)) ? c : '?';

		/* Is this the beginning of the value data? */
		if ((*to == '=') && (!(*val)))
		{
			*val = to + 1;
			*to = '\0';
		}
	}

	/* Ensure termination of processed string */
	*to = '\0';

	/* Return exact extent of used buffer */
	*toptr = to + 1;

	/* Set 'value' content to point at the global variable 'empty' when necessary. */
	if (!(*val) || (**val == '\0')) *val = empty;

	/* More pairs to read? */
	return ((*length) && (!feof(stdin))) ? 0 : 1;
}

/***************************************************************************************************************************
** Search decoded name/val pairs for name matches to (what). Sets 'v' to corresponding value. Flags as follows:           **
** b3 clear => demand exact match, set 's' to matched name; b3 set => initial chars must match, set 's' to name suffix.   **
** If b3 and b9 set (substring matching), only names with non-empty suffix are matched, ie. exact matches disallowed.     **
** If no match found: b5 set => set 'v' and 's' to null ptr; b5 clear => behave as given below with 'v' and 's' empty.    **
** If valid match, but 'v' empty: set 'v' to null ptr (b0 clear) or null string (b0 set).                                 **
** If valid match, but 's' empty (implies b3 set): set 's' to null ptr (b8 clear) or null string (b8 set).                **
** b2 set => strip all but CR/LF chars from 'v' string (if any), otherwise b1 unset => strip all ctrl chars.              **
** b7 set => strip all but CR/LF chars from 's' string (if any); b6 set => strip all ctrl chars from 's' string.          **
** b4 clear => store 's' as (*suffix) if (suffix) non-null; return 'v'; b4 set => swap: store 'v' (etc) and return 's'.   **
***************************************************************************************************************************/
char* findnameval(char* what, unsigned int flag, char** suffix)
{
	int i, n;
	char *p, *v, *s;

	/* Search for exact match */
	if (!(flag & 8))
	{
		for (i = n = 0; (i < npairs); i++) if (streq(what, name[i])) break;
	}

	/* Search for initial (proper) substring, no exact matches */
	else if (flag & 512)
	{
		for (i = 0, n = strlen(what); (i < npairs); i++) if (!strncmp(what, name[i], n) && (name[i][n] >= 32)) break;
	}

	/* Search for initial substring, exact matches allowed */
	else for (i = 0, n = strlen(what); (i < npairs); i++) if (!strncmp(what, name[i], n)) break;

	/* Match found? */
	if (i < npairs)
	{
		/* Get value */
		v = (val[i] != empty) ? val[i] : 0;
	
		/* Get name suffix (we cater for the null suffix case here, regardless) */
		if ((s = name[i] + n)[0] == '\0') s = 0;
	
		/* 'v' - Deal with CR/LF if flags bit 2 set */
		if ((v) && (flag & 4)) for (p = v; (*p); p++) if ((*p < 32) && (*p != 13) && (*p != 10)) *p = ' ';
	
		/* 'v' - else strip ctrl chars if flags bit 1 CLEAR */
		if ((v) && !(flag & 6)) for (p = v; (*p); p++) if (*p < 32) *p = ' ';
	
		/* 's' - Deal with CR/LF if flags bit 7 set */
		if ((s) && (flag & 128)) for (p = s; (*p); p++) if ((*p < 32) && (*p != 13) && (*p != 10)) *p = ' ';
	
		/* 's' - Strip ctrl chars if flags bit 6 SET */
		if ((s) && (flag & 64)) for (p = s; (*p); p++) if (*p < 32) *p = ' ';
	}

	/* No match */
	else v = s = 0;

	/* No match and flags bit 5 set => force return of null pointers */
	if ((i < npairs) || !(flag & 32))
	{
		/* Otherwise, use bits 0 and 8 to choose between null pointers and strings */
		if ((!v) && (flag & 1)) v = nullstr;
		if ((!s) && (flag & 256)) s = nullstr;
	}

	/* Return something in suffix? */
	if (suffix) *suffix = (flag & 16) ? v : s;

	/* Return value */
	return (flag & 16) ? s : v;
}

/***************************************************************************************************************************
** Parse CGI content                                                                                                      **
***************************************************************************************************************************/
int parsecgi(void)
{
	static char cgib[MAX_cgibuf];
	char *cgib_pos = cgib, *cgib_end = cgib + MAX_cgibuf;
	int len, i;
	char *p, *query;

	/* No input read yet */
	npairs = 0;

	/* If given a null PUT query, read POST'ed data on stdin */
	if (!(query = getenv("QUERY_STRING")) || (!query[0]))
	{
		/*fprintf(stderr, "{no-query}");*/

		/* Given a null query? */
		if (!(p = getenv("CONTENT_TYPE"))   || (strcasecmp(p, "application/x-www-form-urlencoded")) ||
			!(p = getenv("REQUEST_METHOD")) || (strcasecmp(p, "POST")) ||
			!(p = getenv("CONTENT_LENGTH")) || (0 > (len = atoi(p)))) return (0); /*exit(insecure(4+1, "plinfo"));*/
	
		/* Otherwise, extract name=val pairs */
		while ((npairs < MAX_npairs) && !(i = pextract(&len, &cgib_pos, cgib_end, &name[npairs], &val[npairs]))) ++npairs;

		/* When the dust has settled... */
		if     (i == -1) return ERF_CGIExhaustion;	/* Buffer exhaustion */
		else if (i == 0) return ERF_CGIManyPairs;	/* Too many name=value pairs */
		else if (i == 1) npairs++; 					/* Don't miss the last pair! */

		return (-2);
	}

	/* Otherwise parse query string for id and request security */
	else
	{
		/*fprintf(stderr, "{query-str=%s}", query);*/

		/* Extract name=val pairs from environment */
		for (p = query; ((npairs < MAX_npairs) &&
						 !(i = gextract(&p, &cgib_pos, cgib_end, &name[npairs], &val[npairs]))); npairs++);

		/* When the dust has settled... */
		if     (i == -1) return ERF_CGIExhaustion;	/* Buffer exhaustion */
		else if (i == 0) return ERF_CGIManyPairs;	/* Too many name=value pairs */
		else if (i == 1) npairs++; 					/* Don't miss the last pair! */
	
		return (-1); /*exit(insecure(1, "plinfo"));*/
	}
}

/***************************************************************************************************************************
** Old password protection method; not suited to multiuser use. Not used anymore.                                         **
** Enforce security - set mode b1 to include http headers, b2 to indicate that cgi has not been reliably parsed.          **
** b0 to include/fill-in the id field, b3 for which field, b4 for the ev_ref field.                                       **
***************************************************************************************************************************/
#if 0
int insecure(int mode, char* p)
{
	printf("%s</html><HTML><HEAD><TITLE>" GAME " Assassins: Authentication</TITLE></HEAD>"
		   "<BODY bgcolor=#DDDDFF text=#000000 link=#0000FF vlink=#FF0000 alink=#FF00FF>"
		   "<form action=\"" CGI_URL "%s.cgi\" method=\"POST\">"
		   "<h1>Unauthenticated access disallowed.</h1><p>Please supply the password: "
		   "&nbsp; <input name=\"security\" type=\"password\" size=64 maxlength=64><br>",
		   (mode & 2) ? "Content-type: text/html" LF LF : "", p);

	if (mode & 1) printf("Player id &nbsp; <input type=\"text\" name=\"id\" value=\"%s\" size=8 maxlength=8>",
						 (mode & 4) ? "" : findval("id", 1));

	if (mode & 8) printf("News page &nbsp; <input type=\"text\" name=\"which\" value=\"%s\" size=8 maxlength=8>",
						 (mode & 4) ? "" : findval("which", 1));

	if (mode & 16) printf("Event ref &nbsp; <input type=\"text\" name=\"ev_ref\" value=\"%s\" size=8 maxlength=8>",
						 (mode & 4) ? "" : findval("ev_ref", 1));

	if ((mode & (32+4)) == 32) printf("<input type=\"hidden\" name=\"vitality\" value=\"%s\">"
									  "<input type=\"hidden\" name=\"law\" value=\"%s\">"
									  "<input type=\"hidden\" name=\"rest\" value=\"%s\">"
  									  "<input type=\"hidden\" name=\"sort\" value=\"%s\">",
									  findval("vitality", 1), findval("law", 1), findval("rest", 1), findval("sort", 1));
	
	puts("<input value=\"Submit\" type=\"submit\"></p></form></BODY></HTML>");

	return 0;
}
#endif

/***************************************************************************************************************************
** Reprint all CGI content to stdout, ignoring pairs whose name matches an entry in the null terminated string array.     **
***************************************************************************************************************************/
void regurgitate(char **excl)
{
	int i;
	char **p;

	for (i = 0; (i < npairs); i++)
	{
		if ((!name[i]) || (name[i] == empty)) continue;

		if (excl) for (p = excl; (*p); p++) if (!strcasecmp(*p, name[i])) continue;

		fputs("<input type=\"hidden\" name=\"", stdout);

		encapsulate(name[i], stdout);

		fputs("\" value=\"", stdout);

		if ((val[i]) && (val[i] != empty)) encapsulate(val[i], stdout);

		fputs("\">" LF, stdout);
	}
}

#endif
/*** The End ***/
