/*****************************************************
**  lib-rep-edit.c                       (library)  **
**  ----------------------------------------------  **
**  converted to new time system?                   **
**                                                  **
************** BORING LEGAL NOTICE *******************
**                                                  **
**  Assassins' Game (Semi-)Automatic Umpire System  **
**  ----------------------------------------------  **
**  Prerelease: Copyright 2000-2002 Edward M Nokes  **
**  This code may not be distributed in source or   **
**  executable form without the express consent of  **
**  the author, Edward M Nokes (emn23@cam.ac.uk).   **
**  The author accepts no responsibility or         **
**  liability for anything occurring as a result    **
**  of using this software, etc. etc.               **
**                                                  **
*****************************************************/

#ifdef included__lib_rep_edit
 #error "** lib-rep-edit multiply #included **"
#else
 #define included__lib_rep_edit fromfile

/***************************************************************************************************************************
** Definitions and declarations                                                                                           **
***************************************************************************************************************************/

#define MAX_ReportLength	16000		/* Limit on allowed length of reports, in chars */

/***************************************************************************************************************************
** Add or edit reports.                                                                                                   **
** If (oldrid) is null, add a new news report (rptr) and place new report id in (rptr->rid)[32].                          **
** If (oldrid) is non-null, replace the report of id (oldrid) with report defined by (rptr) in the 'reports.txt' file.    **
** Optimisation when editing reports: if field (rptr->fpos) != -1L, take it as the file position of old/new report.       **
** If (fpp) and (*fpp) are non-null, use (*fpp) as the source file pointer, and close afterwards, setting (*fin) to zero. **
** Note that the file must be opened with both read and write privs for the case of adding a new report, but need only    **
** (mode) b2 set => if successful, and (fpp) non-null, maintain (add) or reopen (edit) new file and return as (*fpp).     **
** be readable when editing reports; for report editing, however, the global (fpbuf) is used for temporary storage.       **
** (mode) b0 set => update (rptr->fpos) with new report body location; clear => (new == old) report header location.      **
***************************************************************************************************************************/
int addreport(report_p rptr, char* oldrid, FILE** fpp, int mode, char* evref, char* buf, char* bufend)
{
	FILE *fp = 0, *fout = 0;
	char* p;
	int i, rc = 0;
	unsigned long idx;
	long len, dec, ltmp, lcom;


	/* Source file already open? Clear file pointer parameter in advance */
	if ((fpp) && (*fpp)) { fp = *fpp; *fpp = 0; }

	/* Paranoia */
	if ((!fpp) && (mode & 4)) mode &= (~4);

	/* Allow empty pseudonyms (ignore this now) */
	/*p = (rptr->pseud) ? (rptr->pseud) : "-";*/

	/* Check parameters */
	if ((!rptr) || (!rptr->rid) || (!rptr->pid) || (!rptr->text)) erf(ERF_BadParms, 1);

	/* Ensure minimum number of bytes free in buffer */
	if ((bufend - buf) < chk[CHK_reps].len) erf(ERF_BufferExhaust, 100);

	/* Paranoia: don't allow report ids to be edited */
	if ((oldrid) && (strcasecmp(oldrid, rptr->rid))) erf(ERF_BadParms, 2);

	/* Non-optimised mode searching unsupported at the moment */
	if ((oldrid) && (rptr->fpos <= 0L)) erf(ERF_BadParms, 3);

	/* Calculate and restrict size of report text */
	if ((rptr->length = strlen(rptr->text)) > MAX_ReportLength)
	{
		rptr->length = (MAX_ReportLength);
		rptr->text[rptr->length] = '\0';
	}

	/*fprintf(stderr, "%li:   woo!\n", uniqid);*/

	/* Adding new report: append to file */
	if (!oldrid)
	{
		/* Open player info file */
		if ((!fp) && !(fp = fopen(FILE_Reports, "r+b"))) erf(ERF_fopenReps, 4);

		/* Seek to start of file */
		if (fseek(fp, 0L, SEEK_SET)) erf(ERF_fseekFailed, 5);

		/* Check file header (ignoring initial comments) and get free index */
		if (i = chkheader(fp, CHK_reps, buf, 0, &ltmp, &idx, &lcom)) erf(i, 6);

		/* Seek to start of comments */
		if (fseek(fp, lcom, SEEK_SET)) erf(ERF_fseekFailed, 7);

		/* Form new report id, increment index */
		if (sprintf(rptr->rid, "r%lu", idx++) <= 2) erf(ERF_printfFailed, 8);

		/* Write new file header */
		if (i = prtheader(fp, CHK_reps, idx)) erf(i, 9);

		/* Paranoia: get file position */
		if ((lcom = ftell(fp)) == -1L) erf(ERF_ftellFailed, 10);

		/* Paranoia: check we are where we think we are */
		if (lcom != ltmp) erf(ERF_Paranoia, 11);

		/* Seek to end of file */
		if (fseek(fp, 0L, SEEK_END)) erf(ERF_fseekFailed, 12);

		/* Insert some blank lines */
		if (EOF == fputs(LF LF, fp)) erf(ERF_fputsFailed, 13);

		/* Note position of report header in file */
		if ((rptr->fpos = ftell(fp)) == -1L) erf(ERF_ftellFailed, 14);

		/* Appending => use same file for reading and writing; be careful to close only once */
		fout = fp;
		fp = 0;

		/*fprintf(stderr, "%li:   woo2!\n", uniqid);*/
	}

	/* Editing report */
	else
	{
		/* Open input reports file */
		if ((!fp) && !(fp = fopen(FILE_Reports, "rb"))) erf(ERF_fopenReps, 39);
	
		/* Check file header (ignoring initial comments) */
		if (i = chkheader(fp, CHK_reps, buf, 0, 0, 0, 0)) erf(i, 40);

		/* Open output reports file */
		if (!(fout = fopen(FILE_NewReports, "wb"))) erf(ERF_fopenReps, 41);

		/* Copy bytes from start of fin to fout, in a paranoid manner */
		if (i = ftofcpy(fp, fout, rptr->fpos, fpbuf, MAX_fpbuf, 0L, rptr->fpos)) erf(i, 42);

		/* Sanity check: next char in input file should be a '*' to indicate beginning of header */
		if (fgetc(fp) != '*') erf(ERF_InternalError, 46);

		/* Skip initial tabs, read 'repid' from input file */
		if (i = tread(fp, fpbuf, fpbuf + MAX_fpbuf, 0)) erf(ERF_BadTread+i, 47); 

		/* Sanity check: old report id in input file should be that of new/old report */
		if (strcasecmp(fpbuf, oldrid)) erf(ERF_InternalError, 48);

		/*fprintf(stderr, "%li:   woo3!\n", uniqid);*/

		/* Skip old report flags */
		if (i = thopp(fp, 0)) erf(ERF_BadTread+i, 49);

		/* Acquire old report length */
		if (i = tread(fp, fpbuf, fpbuf + MAX_fpbuf, 0)) erf(ERF_BadTread+i, 50);

		/* Convert it */
		len = atol(fpbuf);

		/* Ignore the rest of the line */
		toeol(fp);

		/* Seek past old report body in input file */
		if (fseek(fp, len, SEEK_CUR)) erf(ERF_fseekFailed, 51);

		/* Paranoia: check output file position */
		if ((ltmp = ftell(fout)) == -1L) erf(ERF_ftellFailed, 52);

		/* Paranoia: is it where it should be? */
		if (ltmp != rptr->fpos) erf(ERF_InternalError, 53);
	}

	/*fprintf(stderr, "%li:   woo33!\n", uniqid);*/

	/* Output new report header */
	i = fprintf(fout, "*%s\t%lX\t%lu\t%s\t%s%s", rptr->rid, rptr->flags, rptr->length, rptr->pid, ESC(rptr->pseud));

	/* Collect errors */
	if (i <= 0) erf(ERF_printfFailed, 54);

	/*fprintf(stderr, "%li:   woo4!\n", uniqid);*/

	/* Report edit time, if present */
	if (rptr->uptim) i = fprintf(fout, "\t"TFMT"", Tdhm(rptr->uptim));

	/* No edit time known */
	else if (fputs("\t-", fout)) erf(ERF_fputsFailed, 56);

	/* Output additional field: event ref */
	if (i > 0) i = fprintf(fout, "\t%s%s", ESC(evref));

	/* Terminate line */
	if ((i <= 0) || (EOF == fputc('\n', fout))) erf(ERF_printfFailed, 79);

	/* Option: Store file position of report body */
	if ((mode & 1) && ((rptr->fpos = ftell(fout)) == -1L)) erf(ERF_ftellFailed, 80);

	/* Store new report */
	if (EOF == fputs(rptr->text, fout)) erf(ERF_fputsFailed, 81);

	/* Guardword */	
	if (EOF == fputs(LF "####" LF, fout)) erf(ERF_fputsFailed, 82);

	/*fprintf(stderr, "%li:   woo7!\n", uniqid);*/

	/* Editing report: copy the rest of the input file piecewise */
	if (oldrid)
	{
		char *q;

		/* Read up to EOF on input file */
		for (len = 0; (!ferror(fp)) && (!feof(fp)); len += dec)
		{
			/* Read up to MAX_fpbuf bytes */
			dec = fread(q = fpbuf, 1, MAX_fpbuf, fp);

			/* Try not to copy too many guardwords */
			if ((len == 0) && (dec >= 6) && (!strncmp(q, LF "####" LF, 6))) { q += 6; dec -= 6; }

			/* Write them out */
			if ((dec) && ((ltmp = fwrite(q, 1, (size_t)dec, fout)) != dec))
			{
				fprintf(stderr, "%li:   addrep fwrite on \"%s\" failed after %li: dec=%li ltmp=%li ferror=%i feof=%i.\n",
						uniqid, oldrid, len, dec, ltmp, ferror(fout), feof(fout));
				erf(ERF_fwriteFailed, 83);
			}
		}

		/*fprintf(stderr, "%li:   Foom.\n", uniqid);*/

		/* Trap read errors */
		if (ferror(fp))
		{
			fprintf(stderr, "%li:   addrep fread on \"%s\" failed after %li: dec=%li ferror=%i feof=%i.\n",
					uniqid, oldrid, len, dec, ferror(fp), feof(fp));
			erf(ERF_freadFailed, 84);
		}

		/* Tidy up */
		fclose(fp); fp = 0;
		fclose(fout); fout = 0;

		/* Success - remove oldest file */
		remove(FILE_OldReports);

		/* Rename old file as oldest, new file as current */
		if (rename(FILE_Reports, FILE_OldReports) || rename(FILE_NewReports, FILE_Reports)) erf(ERF_FSwapFailed, 85);

		/* Attempt to reopen new file (correct mode here?) */
		if ((mode & 4) && !(*fpp = fopen(FILE_Reports, "rb"))) erf(ERF_fopenReps, 86);
	}

	/* Maintain open existing file */
	else if (mode & 4) { *fpp = fout; fout = 0; }

_erf: /***** Error handling *****/

	/* Tidy up */
	if (fp) fclose(fp);
	if (fout) fclose(fout);

	/* Debug */
	if (rc) fprintf(stderr, "%li:   addreport(%i):%i \"%s\"\n", uniqid, ERF_CONTEXT(rc), ERF_TYPE(rc), STR(oldrid));

	/* More debug information */
	if (rptr) fprintf(stderr, "%li:   --[%cr]--> pid=\"%s\" rid=\"%s\" fl=&%lx len=%lu eref=\"%s\"\n", uniqid,
					 (oldrid) ? '~' : '+', STR(rptr->pid), STR(rptr->rid), rptr->flags, rptr->length, STR(evref));

	/* Finished */
	return (ERF_LOWBITS(rc));
}

#endif
/*** The End ***/
